@@ -163,7 +163,7 @@ order for the change to take effect."
                (agda-input-drop-prefix "\\")
                (agda-input-or
                 (agda-input-compose
-                 (agda-input-drop '("^o"))
+                 (agda-input-drop '("^l" "^o" "^r" "^v"))
                  (agda-input-prefix "^"))
                 (agda-input-prefix "_")))))
     )
@@ -470,14 +470,63 @@ order for the change to take effect."
 
   ;; Blackboard bold letters.
 
-  ("bn"   . ("ℕ"))
-  ("bz"   . ("ℤ"))
-  ("bq"   . ("ℚ"))
-  ("br"   . ("ℝ"))
-  ("bc"   . ("ℂ"))
-  ("bp"   . ("ℙ"))
-  ("bb"   . ("𝔹"))
-  ("bsum" . ("⅀"))
+  ("bA"   . ("𝔸"))
+  ("bB"   . ("𝔹"))
+  ("bC"   . ("ℂ"))
+  ("bD"   . ("𝔻"))
+  ("bE"   . ("𝔼"))
+  ("bF"   . ("𝔽"))
+  ("bG"   . ("𝔾"))
+  ("bH"   . ("ℍ"))
+  ("bI"   . ("𝕀"))
+  ("bJ"   . ("𝕁"))
+  ("bK"   . ("𝕂"))
+  ("bL"   . ("𝕃"))
+  ("bM"   . ("𝕄"))
+  ("bN"   . ("ℕ"))
+  ("bO"   . ("𝕆"))
+  ("bP"   . ("ℙ"))
+  ("bQ"   . ("ℚ"))
+  ("bR"   . ("ℝ"))
+  ("bS"   . ("𝕊"))
+  ("bT"   . ("𝕋"))
+  ("bU"   . ("𝕌"))
+  ("bV"   . ("𝕍"))
+  ("bW"   . ("𝕎"))
+  ("bX"   . ("𝕏"))
+  ("bY"   . ("𝕐"))
+  ("bZ"   . ("ℤ"))
+  ("bGG"  . ("ℾ"))
+  ("bGP"  . ("ℿ"))
+  ("bGS"  . ("⅀"))
+  ("ba"   . ("𝕒"))
+  ("bb"   . ("𝕓"))
+  ("bc"   . ("𝕔"))
+  ("bd"   . ("𝕕"))
+  ("be"   . ("𝕖"))
+  ("bf"   . ("𝕗"))
+  ("bg"   . ("𝕘"))
+  ("bh"   . ("𝕙"))
+  ("bi"   . ("𝕚"))
+  ("bj"   . ("𝕛"))
+  ("bk"   . ("𝕜"))
+  ("bl"   . ("𝕝"))
+  ("bm"   . ("𝕞"))
+  ("bn"   . ("𝕟"))
+  ("bo"   . ("𝕠"))
+  ("bp"   . ("𝕡"))
+  ("bq"   . ("𝕢"))
+  ("br"   . ("𝕣"))
+  ("bs"   . ("𝕤"))
+  ("bt"   . ("𝕥"))
+  ("bu"   . ("𝕦"))
+  ("bv"   . ("𝕧"))
+  ("bw"   . ("𝕨"))
+  ("bx"   . ("𝕩"))
+  ("by"   . ("𝕪"))
+  ("bz"   . ("𝕫"))
+  ("bGg"  . ("ℽ"))
+  ("bGp"  . ("ℼ"))
 
   ;; Blackboard bold numbers.
 
@@ -512,6 +561,9 @@ order for the change to take effect."
   ("lbag" . ("⟅"))
   ("rbag" . ("⟆"))
 
+  ("(|" . ("⦇"))
+  ("|)" . ("⦈"))
+
   ;; Primes.
 
   ("'" . ,(agda-input-to-string-list "′″‴⁗"))
@@ -539,7 +591,6 @@ order for the change to take effect."
   ("\\"         . ("\\"))
   ("en"         . ("–"))
   ("em"         . ("—"))
-  ("^i"         . ("ⁱ"))
   ("!!"         . ("‼"))
   ("??"         . ("⁇"))
   ("?!"         . ("‽" "⁈"))
