@@ -10,7 +10,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.5.1.1"
+(defvar agda2-version "2.5.2"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
@@ -93,19 +93,27 @@ argument, and does not need to be listed here."
   :type '(repeat string)
   :group 'agda2)
 
+(defvar agda2-backends '("GHC" "GHCNoMain" "JS" "LaTeX")
+  "Compilation backends.")
+
 (defcustom agda2-backend
-  "MAlonzo"
-  "The backend which is used to compile Agda programs."
-  :type '(choice (const "MAlonzo")
-                 (const "MAlonzoNoMain")
-                 (const "Epic")
-                 (const "JS"))
+  'ask
+  "The backend used to compile Agda programs."
+  :type `(choice (const :tag "Ask every time" ask)
+                 ,@(mapcar (lambda (x) `(const ,x)) agda2-backends))
   :group 'agda2)
 
 (defcustom agda2-toplevel-module "Agda.Interaction.GhciTop"
   "The name of the Agda toplevel module."
   :type 'string :group 'agda2)
 
+(defcustom agda2-warning-window-max-height
+  0.35
+  "The maximum height of the warning window.
+A multiple of the frame height."
+  :type 'number
+  :group 'agda2)
+
 (defcustom agda2-information-window-max-height
   0.35
   "The maximum height of the information window.
@@ -197,7 +205,7 @@ to this variable to take effect."
                         (modify-syntax-entry keys "w" tbl)))
                     (standard-syntax-table))
     ;; Then override the remaining special cases.
-    (dolist (cs '((?- . "w 12b") (?\n . "> b")
+    (dolist (cs '((?{ . "(}1n") (?} . "){4n") (?- . "w 123b") (?\n . "> b")
                   (?. . ".") (?\; . ".") (?_ . ".") (?! . ".")))
       (modify-syntax-entry (car cs) (cdr cs) tbl))
     tbl)
@@ -223,7 +231,7 @@ constituents.")
     (agda2-remove-annotations                "\C-c\C-x\C-d"       (global)       "Remove goals and highlighting (\"deactivate\")")
     (agda2-display-implicit-arguments        "\C-c\C-x\C-h"       (global)       "Toggle display of hidden arguments")
     (agda2-show-constraints                  ,(kbd "C-c C-=")     (global)       "Show constraints")
-    (agda2-solveAll                          ,(kbd "C-c C-s")     (global)       "Solve constraints")
+    (agda2-solve-maybe-all                   ,(kbd "C-c C-s")     (local global) "Solve constraints")
     (agda2-show-goals                        ,(kbd "C-c C-?")     (global)       "Show goals")
     (agda2-next-goal                         "\C-c\C-f"           (global)       "Next goal") ; Forward.
     (agda2-previous-goal                     "\C-c\C-b"           (global)       "Previous goal") ; Back.
@@ -244,14 +252,14 @@ constituents.")
     (describe-char                           nil                  (global)       "Information about the character at point")
     (agda2-comment-dwim-rest-of-buffer       ,(kbd "C-c C-x M-;") (global)       "Comment/uncomment the rest of the buffer")
     (agda2-display-program-version           nil                  (global)       "Version")
-    (agda2-set-program-version               nil                  (global)       "Switch to another version of Agda")
+    (agda2-set-program-version               "\C-c\C-x\C-s"       (global)       "Switch to another version of Agda")
     (eri-indent                  ,(kbd "TAB"))
     (eri-indent-reverse          [S-iso-lefttab])
     (eri-indent-reverse          [S-lefttab])
     (eri-indent-reverse          [S-tab])
     (agda2-goto-definition-mouse [mouse-2])
     (agda2-goto-definition-keyboard "\M-.")
-    (agda2-go-back                  "\M-*")
+    (agda2-go-back                  ,(if (version< emacs-version "25.1") "\M-*" "\M-,"))
     )
   "Table of commands, used to build keymaps and menus.
 Each element has the form (CMD &optional KEYS WHERE DESC) where
@@ -288,6 +296,9 @@ menus.")
 (defvar agda2-info-buffer nil
   "Agda information buffer.")
 
+(defvar agda2-warning-buffer nil
+  "Agda warnings buffer.")
+
 (defvar agda2-process-buffer nil
   "Agda subprocess buffer.
 Set in `agda2-restart'.")
@@ -431,6 +442,8 @@ agda2-include-dirs is not bound." :warning))
  (force-mode-line-update)
  ;; Protect global value of default-input-method from set-input-method.
  (make-local-variable 'default-input-method)
+ ;; Don't take script into account when determining word boundaries
+ (set (make-local-variable 'word-combining-categories) (cons '(nil . nil) word-combining-categories))
  (set-input-method "Agda")
  ;; Highlighting etc. is removed when we switch from the Agda mode.
  ;; Use case: When a file M.lagda with a local variables list
@@ -453,7 +466,7 @@ agda2-include-dirs is not bound." :warning))
          (output (with-output-to-string
                    (call-process agda2-program-name
                                  nil standard-output nil "--version")))
-         (version (and (string-match "^Agda version \\([0-9.]+\\)$" output)
+         (version (and (string-match "^Agda version \\([0-9.]+\\)" output)
                        (match-string 1 output))))
     (unless (equal version agda2-version)
       (error "The Agda mode's version (%s) does not match that of %s (%s)."
@@ -746,7 +759,7 @@ An error is raised if no responses are received."
                   (or ask (string-match "\\`\\s *\\'" txt)))
              (setq txt (read-string (concat want ": ") nil nil txt t)))
             (t (setq input-from-goal t)))
-      (apply 'agda2-go t t t cmd
+      (apply 'agda2-go t input-from-goal t cmd
              (format "%d" g)
              (if input-from-goal (agda2-goal-Range o) (agda2-mkRange nil))
              (agda2-string-quote txt) args))))
@@ -774,22 +787,16 @@ An error is raised if no responses are received."
             ))
 
 (defun agda2-measure-load-time
-  (&optional highlighting-level dont-touch continuation)
+  (&optional highlighting-level continuation)
   "Load the current buffer and print how much time it takes.
 \(Wall-clock time.)
 
 The given HIGHLIGHTING-LEVEL is used (if non-nil).
 
-The file is first saved and \"touched\", unless DONT-TOUCH is
-non-nil.
-
 If CONTINUATION is non-nil, then CONTINUATION is applied to the
 resulting time (represented as a string)."
   (interactive)
   (agda2-abort-if-in-progress)
-  (unless dont-touch
-    (save-buffer)
-    (shell-command (concat "touch \"" (buffer-file-name) "\"")))
   (let* ((agda2-highlight-level
           (or highlighting-level agda2-highlight-level)))
     (setq agda2-measure-data (cons continuation (current-time)))
@@ -800,11 +807,22 @@ resulting time (represented as a string)."
 
 The variable `agda2-backend' determines which backend is used."
   (interactive)
-  (agda2-go t t t "Cmd_compile"
-            agda2-backend
-            (agda2-string-quote (buffer-file-name))
-            (agda2-list-quote agda2-program-args)
-            ))
+  (let ((backend (cond ((equal agda2-backend "MAlonzo")       "GHC")
+                       ((equal agda2-backend "MAlonzoNoMain") "GHCNoMain")
+                       ((equal agda2-backend 'ask)
+                        (completing-read "Backend: " agda2-backends
+                                         nil t nil nil nil
+                                         'inherit-input-method))
+                       (t agda2-backend))))
+    (unless (member backend agda2-backends)
+      (if (equal backend "")
+          (error "No backend chosen")
+        (error "Invalid backend: %s" backend)))
+    (agda2-go t t t "Cmd_compile"
+              backend
+              (agda2-string-quote (buffer-file-name))
+              (agda2-list-quote agda2-program-args)
+              )))
 
 (defun agda2-give()
   "Give to the goal at point the expression in it" (interactive)
@@ -867,21 +885,25 @@ Assumes that <clause> = {!<variables>!} is on one line."
 (defun agda2-make-case-action-extendlam (newcls)
   "Replace definition of extended lambda with new clauses NEWCLS and reload."
   (agda2-forget-all-goals);; we reload later anyway.
-  (let* ((pmax (- (re-search-forward "\\([^!]}\\)\\|;") 1))
+  (let* ((p0 (point))
+         (pmax (re-search-forward "!}"))
          (bracketCount 0)
+         (p1 (goto-char (+ (current-indentation) (line-beginning-position))))
+         (indent (current-column))
          cl)
+    (goto-char p0)
     (re-search-backward "{!")
-      (while (and (not (equal (preceding-char) ?\;)) (>= bracketCount 0))
-        (backward-char)
-        (if (equal (preceding-char) ?}) (incf bracketCount))
-        (if (equal (preceding-char) ?{) (decf bracketCount)))
-    (let ((p (point)))
+    (while (and (not (equal (preceding-char) ?\;)) (>= bracketCount 0) (> (point) p1))
+      (backward-char)
+      (if (equal (preceding-char) ?}) (incf bracketCount))
+      (if (equal (preceding-char) ?{) (decf bracketCount)))
+    (let* ((is-lambda-where (= (point) p1))
+           (p (point)))
       (delete-region (point) pmax)
-      (insert " ")
+      (if (not is-lambda-where) (insert " "))
       (while (setq cl (pop newcls))
         (insert cl)
-        (if newcls (insert " ; ")))
-      (insert " ")
+        (if newcls (if is-lambda-where (insert "\n" (make-string indent ? )) (insert " ; "))))
       (goto-char p)))
   (agda2-load))
 
@@ -891,6 +913,42 @@ Assumes that <clause> = {!<variables>!} is on one line."
 major mode)."
   (setq agda2-buffer-external-status status))
 
+(defun agda2-warning-buffer nil
+  "Creates the Agda warning buffer, if it does not already exist.
+The buffer is returned."
+  (unless (buffer-live-p agda2-warning-buffer)
+    (setq agda2-warning-buffer
+          (generate-new-buffer "*Agda warnings*"))
+
+    (with-current-buffer agda2-warning-buffer
+      (compilation-mode "AgdaInfo")
+      ;; Support for jumping to positions mentioned in the text.
+      (set (make-local-variable 'compilation-error-regexp-alist)
+           '(("\\([\\\\/][^[:space:]]*\\):\\([0-9]+\\),\\([0-9]+\\)-\\(\\([0-9]+\\),\\)?\\([0-9]+\\)"
+              1 (2 . 5) (3 . 6))))
+      ;; No support for recompilation. The key binding is removed, and
+      ;; attempts to run `recompile' will (hopefully) result in an
+      ;; error.
+      (let ((map (copy-keymap (current-local-map))))
+        (define-key map (kbd "g") 'undefined)
+        (use-local-map map))
+      (set (make-local-variable 'compile-command)
+           'agda2-does-not-support-compilation-via-the-compilation-mode)
+
+      (set-syntax-table agda2-mode-syntax-table)
+      (set (make-local-variable 'word-combining-categories) (cons '(nil . nil) word-combining-categories))
+      (set-input-method "Agda")))
+
+  agda2-warning-buffer)
+
+(defun agda2-font-syntactic-face (state)
+  (cond ((nth 4 state)
+         ( save-excursion
+           (goto-char (nth 8 state))
+           (cond ((looking-at "--[[:space:]\n]") 'font-lock-comment-face)
+                 ((looking-at "{-[^#]") 'font-lock-comment-face)
+          )))))
+
 (defun agda2-info-buffer nil
   "Creates the Agda info buffer, if it does not already exist.
 The buffer is returned."
@@ -914,6 +972,7 @@ The buffer is returned."
            'agda2-does-not-support-compilation-via-the-compilation-mode)
 
       (set-syntax-table agda2-mode-syntax-table)
+      (set (make-local-variable 'word-combining-categories) (cons '(nil . nil) word-combining-categories))
       (set-input-method "Agda")))
 
   agda2-info-buffer)
@@ -996,6 +1055,90 @@ is inserted, and point is placed before this text."
               (goto-char (point-max))
             (goto-char (point-min))))))))
 
+(defun agda2-close-warning nil
+  (interactive)
+  (when (buffer-live-p agda2-warning-buffer)
+        (delete-windows-on agda2-warning-buffer))
+)
+
+(defun agda2-warning-action (name text &optional append)
+  "Insert TEXT into the Agda warning buffer and display it.
+NAME is displayed in the buffer's mode line.
+
+If APPEND is non-nil, then TEXT is appended at the end of the
+buffer, and point placed after this text.
+
+If APPEND is nil, then any previous text is removed before TEXT
+is inserted, and point is placed before this text."
+  (interactive)
+  (let ((buf (agda2-warning-buffer)))
+    (with-current-buffer buf
+      ;; In some cases the jump-to-position-mentioned-in-text
+      ;; functionality (see compilation-error-regexp-alist above)
+      ;; didn't work: Emacs jumped to the wrong position. However, it
+      ;; seems to work if compilation-forget-errors is used. This
+      ;; problem may be related to Emacs bug #9679
+      ;; (http://debbugs.gnu.org/cgi/bugreport.cgi?bug=9679). The idea
+      ;; to use compilation-forget-errors comes from a comment due to
+      ;; Oleksandr Manzyuk
+      ;; (https://github.com/haskell/haskell-mode/issues/67).
+      (compilation-forget-errors)
+      (unless append (erase-buffer))
+      (save-excursion
+        (goto-char (point-max))
+        (insert text))
+      (put-text-property 0 (length name) 'face '(:weight bold) name)
+      (setq mode-line-buffer-identification name)
+      (force-mode-line-update))
+    ;; If the current window displays the information buffer, then the
+    ;; window configuration is left untouched.
+    (unless (equal (window-buffer) buf)
+      (let ((agda-info
+              (and agda2-info-buffer
+                   (car-safe
+                     ;; All windows, including minibuffers, on any
+                     ;; frame on the current terminal, displaying the
+                     ;; present Agda file buffer.
+                    (get-buffer-window-list agda2-info-buffer t 0)))))
+        (save-selected-window
+          ;; Select a window displaying the Agda file buffer (if such
+          ;; a window exists). With certain configurations of
+          ;; display-buffer this should increase the likelihood that
+          ;; the info buffer will be displayed on the same frame.
+          (when agda-info
+            (select-window agda-info 'no-record))
+          (let* (;; The warnings window should be displayed below the
+                 ;; Agda info one
+                 (split-width-threshold nil)
+                 (split-height-threshold 1)
+                 (window
+                   (display-buffer
+                     buf
+                     ;; Under Emacs 23 the effect of the following
+                     ;; argument is only that the current window
+                     ;; should not be used.
+                     '((display-buffer-below-selected)
+                       .
+                       (;; Do not use the same window
+                        (inhibit-same-window . t)
+                        ;; Do not raise or select another frame.
+                        (inhibit-switch-frame . t))))))
+            (if window
+                (fit-window-to-buffer window
+                  (truncate
+                    (* (frame-height)
+                       agda2-warning-window-max-height))))))))
+    ;; Move point in every window displaying the information buffer.
+    ;; Exception: If we are appending, don't move point in selected
+    ;; windows.
+    (dolist (window (get-buffer-window-list buf 'no-minibuffer t))
+      (unless (and append
+                   (equal window (selected-window)))
+        (with-selected-window window
+          (if append
+              (goto-char (point-max))
+            (goto-char (point-min))))))))
+
 (defun agda2-info-action-and-copy (name text &optional append)
   "Same as agda2-info-action but also puts TEXT in the kill ring."
   (kill-new text)
@@ -1108,6 +1251,36 @@ The form of the result depends on the prefix argument:
                    (concat ,cmd " " ,eval " "
                            (agda2-string-quote expr)))))))
 
+(defmacro agda2-maybe-normalised-global (name comment cmd)
+  "This macro constructs a function NAME which runs CMD.
+COMMENT is used to build the function's comments. The function
+NAME takes a prefix argument which tells whether it should
+normalise types or not when running CMD (through
+`agda2-go' t nil t;)."
+  (let ((eval (make-symbol "eval")))
+    `(defun ,name (prefix)
+       ,(concat comment ".
+
+The form of the result depends on the prefix argument:
+
+* If the prefix argument is `nil' (i.e., if no prefix argument is
+  given), then the result is simplified.
+
+* If the prefix argument is `(4)' (for instance if C-u is typed
+  exactly once right before the command is invoked), then the
+  result is neither explicitly normalised nor simplified.
+
+* If any other prefix argument is used (for instance if C-u is
+  typed twice right before the command is invoked), then the
+  result is normalised.")
+       (interactive "P")
+       (let ((,eval (cond ((equal prefix nil) "AsIs")
+                          ((equal prefix '(4)) "Simplified")
+                          ("Normalised"))))
+         (agda2-go t nil t
+                   (concat ,cmd " " ,eval " "
+                           ))))))
+
 (agda2-maybe-normalised
  agda2-goal-type
  "Show the type of the goal at point"
@@ -1213,10 +1386,27 @@ a goal, the top-level scope."
                           'agda2-module-contents
                         'agda2-module-contents-toplevel)))
 
-(defun agda2-solveAll ()
-  "Solves all goals that are already instantiated internally."
+(defun agda2-solve-maybe-all ()
+  "Solves goals that are already instantiated internally.
+Either only one if point is a goal, or all of them."
   (interactive)
-  (agda2-go t t t "Cmd_solveAll"))
+  (call-interactively (if (agda2-goal-at (point))
+                          'agda2-solveOne
+                          'agda2-solveAll))
+)
+
+(agda2-maybe-normalised-global
+  agda2-solveAll
+  "Solves all goals that are already instantiated internally."
+  "Cmd_solveAll"
+)
+
+(agda2-maybe-normalised
+  agda2-solveOne
+  "Solves the goal at point if it is already instantiated internally"
+  "Cmd_solveOne"
+  nil
+)
 
 (defun agda2-solveAll-action (iss)
   (while iss
@@ -1234,28 +1424,49 @@ a goal, the top-level scope."
 
 (defun agda2-compute-normalised (&optional arg)
   "Compute the normal form of the expression in the goal at point.
-With a prefix argument \"abstract\" is ignored during the computation."
+
+With a prefix argument distinct from `(4)' the normal form of
+\"show <expression>\" is computed, and then the resulting string
+is printed.
+
+With any prefix argument \"abstract\" is ignored during the
+computation."
   (interactive "P")
   (let ((cmd (concat "Cmd_compute"
-                     (if arg " True" " False"))))
+                      (cond ((equal arg nil) " DefaultCompute")
+                            ((equal arg '(4)) " IgnoreAbstract")
+                            (" UseShowInstance")))))
     (agda2-goal-cmd cmd "expression to normalise")))
 
 (defun agda2-compute-normalised-toplevel (expr &optional arg)
-  "Computes the normal form of the given expression.
-The scope used for the expression is that of the last point inside the current
-top-level module.
-With a prefix argument \"abstract\" is ignored during the computation."
+  "Compute the normal form of the given expression.
+The scope used for the expression is that of the last point
+inside the current top-level module.
+
+With a prefix argument distinct from `(4)' the normal form of
+\"show <expression>\" is computed, and then the resulting string
+is printed.
+
+With any prefix argument \"abstract\" is ignored during the
+computation."
   (interactive "MExpression: \nP")
   (let ((cmd (concat "Cmd_compute_toplevel"
-                     (if arg " True" " False")
-                     " ")))
+                     (cond ((equal arg nil) " DefaultCompute")
+                            ((equal arg '(4)) " IgnoreAbstract")
+                            (" UseShowInstance")) " ")))
     (agda2-go t nil t (concat cmd (agda2-string-quote expr)))))
 
 (defun agda2-compute-normalised-maybe-toplevel ()
-  "Computes the normal form of the given expression,
-using the scope of the current goal or, if point is not in a goal, the
-top-level scope.
-With a prefix argument \"abstract\" is ignored during the computation."
+  "Compute the normal form of the given expression.
+The scope used for the expression is that of the last point
+inside the current top-level module.
+
+With a prefix argument distinct from `(4)' the normal form of
+\"show <expression>\" is computed, and then the resulting string
+is printed.
+
+With any prefix argument \"abstract\" is ignored during the
+computation."
   (interactive)
   (if (agda2-goal-at (point))
       (call-interactively 'agda2-compute-normalised)
@@ -1578,7 +1789,7 @@ a file is loaded."
   ;; Enable highlighting of comments via Font Lock mode (which uses
   ;; the syntax table).
   (set (make-local-variable 'font-lock-defaults)
-       '(nil nil nil nil nil))
+       '(nil nil nil nil nil (font-lock-syntactic-face-function . agda2-font-syntactic-face)))
   ;; If the following s-expression is removed, then highlighting of
   ;; comments stops working.
   (when font-lock-mode
@@ -1700,6 +1911,24 @@ the argument is a positive number, otherwise turn it off."
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Switching to a different version of Agda
 
+(defun get-agda-program-versions ()
+  "Get \"version strings\" of executables starting with
+'agda-mode' in current path."
+  (delete-dups
+   (mapcar (lambda (path)
+             ;; strip 'agda-mode' prefix
+             (replace-regexp-in-string "^agda-mode-?" ""
+                                       (file-name-nondirectory path)))
+           (remove-if-not 'file-executable-p
+             ;; concatenate result
+             (reduce 'append
+                     ;; for each directory in exec-path, get list of
+                     ;; files whose name starts with 'agda-mode'
+                     (mapcar (lambda (path)
+                               (when (file-accessible-directory-p path)
+                                 (directory-files path 't "^agda-mode")))
+                             exec-path))))))
+
 ;; Note that other versions of Agda may use different protocols, so
 ;; this function unloads the Emacs mode.
 
@@ -1710,7 +1939,8 @@ This command assumes that the agda and agda-mode executables for
 Agda version VERSION are called agda-VERSION and
 agda-mode-VERSION, and that they are located on the PATH. (If
 VERSION is empty, then agda and agda-mode are used instead.)"
-  (interactive "M")
+  (interactive
+   (list (completing-read "Version: " (get-agda-program-versions))))
 
   (let*
       ((agda-buffers
