@@ -1,9 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE TupleSections #-}
-
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
 
 module Agda.Auto.Auto (auto) where
 
@@ -71,7 +66,7 @@ insertAbsurdPattern (c:s) = c : insertAbsurdPattern s
 getName :: A.Expr -> Maybe (Bool, I.QName)
 getName (A.ScopedExpr _ e) = getName e
 getName (A.Def qname) = Just (False, qname)
-getName (A.Proj qname) = Just (False, qname)
+getName (A.Proj _ qname) = Just (False, head $ I.unAmbQ qname)
 getName (A.Con qname) = Just (True, head $ I.unAmbQ qname)
 getName _ = Nothing
 
@@ -127,8 +122,7 @@ auto ii rng argstr = do
     -- Get the meta variable for the interaction point we are trying to fill.
     -- Add the @autohints@ for that meta to the hints collection.
     mi <- lookupInteractionId ii
-    --thisdefinfo <- catchError (liftM Just $ findClause mi) (\_ -> return Nothing)
-    thisdefinfo <- findClauseDeep mi
+    thisdefinfo <- findClauseDeep ii
     ehints <- (ehints ++) <$> do autohints hintmode mi $ fmap fst3 thisdefinfo
 
     -- If @thisdefinfo /= Nothing@ get the its type (normalized).
@@ -322,7 +316,7 @@ auto ii rng argstr = do
                         case lookup mi riis of
                          Nothing ->
                           -- catchError
-                           (giveExpr mi expr >> return (Nothing, Nothing))
+                           (giveExpr Nothing mi expr >> return (Nothing, Nothing))
                            -- (const retry)
                            -- (\_ -> return (Nothing, Just ("Failed to give expr for side solution of " ++ show mi)))
                          Just ii' -> do ae <- give ii' Nothing expr
@@ -375,10 +369,11 @@ auto ii rng argstr = do
               cls'' <- forM cls' $ \ (I.Clause _ tel ps body t catchall) -> do
                 withCurrentModule (AN.qnameModule def) $ do
                  -- Normalise the dot patterns
-                 ps <- addCtxTel tel $ normalise ps
-                 body <- etaContractBody body
-                 liftM modifyAbstractClause $ inContext [] $ reify $ AN.QNamed def $ I.Clause noRange tel ps body t catchall
-              pcs <- withInteractionId ii $ mapM prettyA cls''
+                 ps <- addContext tel $ normalise ps
+                 body <- etaContract body
+                 liftM modifyAbstractClause $ inTopContext $ reify $ AN.QNamed def $ I.Clause noRange tel ps body t catchall
+              moduleTel <- lookupSection (AN.qnameModule def)
+              pcs <- withInteractionId ii $ inTopContext $ addContext moduleTel $ mapM prettyA cls''
               ticks <- liftIO $ readIORef ticks
 
 
@@ -466,6 +461,7 @@ autohints AHMModule mi (Just def) = do
       c <- getConstInfo n
       case theDef c of
         Axiom{}    -> return True
+        AbstractDefn{} -> return True
         Function{} -> return True
         _          -> return False
 
