@@ -1,15 +1,13 @@
 {-# LANGUAGE CPP #-}
 
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
-
 module Agda.Auto.Convert where
 
 import Control.Applicative hiding (getConst, Const(..))
 import Data.IORef
 import Data.Map (Map)
 import qualified Data.Map as Map
+import Data.Maybe
+import Data.Traversable (traverse)
 import Control.Monad.State
 
 import Agda.Syntax.Concrete (exprFieldA)
@@ -20,16 +18,15 @@ import qualified Agda.Syntax.Abstract.Name as AN
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Position as SP
 import qualified Agda.TypeChecking.Monad.Base as MB
-import Agda.TypeChecking.Monad.State (getImportedSignature)
-import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars)
+import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars, ignoreAbstractMode)
 import Agda.Utils.Permutation (Permutation(Perm), permute, takeP, compactP)
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, ctxEntry, envContext, clEnv)
-import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo)
+import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo, lookupInteractionPoint)
 import Agda.TypeChecking.Monad.Context (getContextArgs)
 import Agda.TypeChecking.Monad.Constraints (getAllConstraints)
-import Agda.TypeChecking.Substitute (piApply, applySubst)
-import Agda.TypeChecking.Telescope (piApplyM, renamingR)
+import Agda.TypeChecking.Substitute (piApply, applySubst, renamingR)
+import Agda.TypeChecking.Telescope (piApplyM)
 import qualified Agda.TypeChecking.Substitute as I (absBody)
 import Agda.TypeChecking.Reduce (Normalise, normalise, instantiate)
 import Agda.TypeChecking.EtaContract (etaContract)
@@ -37,6 +34,8 @@ import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Free (freeIn)
 import qualified Agda.Utils.HashMap as HMap
 
+import Agda.Interaction.MakeCase (getClauseForIP)
+
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
 
@@ -109,6 +108,7 @@ tomy imi icns typs = do
            return (Def narg clauses' Nothing Nothing, [])
      (cont, projfcns2) <- case defn of
       MB.Axiom {} -> return (Postulate, [])
+      MB.AbstractDefn -> return (Postulate, [])
       MB.Function {MB.funClauses = clauses} -> clausesToDef clauses
       -- MB.Primitive {MB.primClauses = []} -> throwError $ strMsg "Auto: Primitive functions are not supported" -- Andreas, 2013-06-17 breaks interaction/AutoMisc
       MB.Primitive {MB.primClauses = clauses} -> clausesToDef clauses
@@ -298,18 +298,25 @@ tomyClauses (cl:cls) = do
   Nothing -> cls'
 
 tomyClause :: I.Clause -> TOM (Maybe ([Pat O], MExp O))
-tomyClause cl@(I.Clause {I.clauseBody = body}) = do
- let pats = I.clausePats cl
+tomyClause cl = do
+ let -- Jesper, 2016-07-28: I can't figure out if this should be the old or new
+     -- clause body (i.e. relative to the positions of pattern variables or
+     -- relative to the clauseTel). Both options pass the test suite, so I
+     -- have the impression it doesn't actually matter.
+     -- ALTERNATIVE CODE:
+     -- perm = fromMaybe __IMPOSSIBLE__ $ IP.clausePerm cl
+     -- body = applySubst (renamingR perm) $ I.clauseBody cl
+     body = I.clauseBody cl
+     pats = I.clausePats cl
  pats' <- mapM tomyPat $ IP.unnumberPatVars pats
- body' <- tomyBody body
+ body' <- traverse tomyExp =<< lift (norm body)
  return $ case body' of
-           Just (body', _) -> Just (pats', body')
-           Nothing -> Nothing
-
+           Just body' -> Just (pats', body')
+           Nothing    -> Nothing
 
 tomyPat :: Common.Arg I.Pattern -> TOM (Pat O)
 tomyPat p = case Common.unArg p of
- I.ProjP _ -> lift $ copatternsNotImplemented
+ I.ProjP{} -> lift $ copatternsNotImplemented
  I.VarP n -> return $ PatVar (show n)
  I.DotP _ -> return $ PatVar "_" -- because Agda includes these when referring to variables in the body
  I.ConP con _ pats -> do
@@ -322,19 +329,6 @@ tomyPat p = case Common.unArg p of
   return $ PatConApp c (replicate npar PatExp ++ pats')
  I.LitP _ -> throwError $ strMsg "Auto: Literals in patterns are not supported"
 
-tomyBody :: I.ClauseBodyF I.Term -> TOM (Maybe (MExp O, Int))
-tomyBody (I.Body t) = do
- t <- lift $ norm t
- t' <- tomyExp t
- return $ Just (t', 0)
-tomyBody (I.Bind (I.Abs _ b)) = do
- res <- tomyBody b
- return $ case res of
-  Nothing -> Nothing
-  Just (b', i) -> Just (b', i + 1)
-tomyBody (I.Bind (I.NoAbs _ b)) = tomyBody b
-tomyBody I.NoBody = return Nothing
-
 weaken :: Int -> MExp O -> MExp O
 weaken _ e@(Meta m) = e
 weaken i (NotM e) =
@@ -398,7 +392,7 @@ tomyExp v0 =
       c   <- getConst False name TMAll
       as' <- tomyExps as
       return $ NotM $ App Nothing (NotM OKVal) (Const c) as'
-    I.Con con as -> do
+    I.Con con ci as -> do
       let name = I.conName con
       c   <- getConst True name TMAll
       as' <- tomyExps as
@@ -453,7 +447,7 @@ fmExp m (I.Lam _ b) = fmExp m (I.unAbs b)
 fmExp m (I.Lit _) = False
 fmExp m (I.Level (I.Max as)) = any (fmLevel m) as
 fmExp m (I.Def _ as) = fmExps m $ I.argsFromElims as
-fmExp m (I.Con _ as) = fmExps m as
+fmExp m (I.Con _ ci as) = fmExps m as
 fmExp m (I.Pi x y)  = fmType m (Common.unDom x) || fmType m (I.unAbs y)
 fmExp m (I.Sort _) = False
 fmExp m (I.MetaV mid _) = mid == m
@@ -481,7 +475,9 @@ cnvh info = case Common.getHiding info of
     Common.Hidden    -> Hidden
 
 icnvh :: FMode -> Common.ArgInfo
-icnvh h = (Common.setHiding h' Common.defaultArgInfo)
+icnvh h = Common.setHiding h' $
+          Common.setOrigin Common.Inserted $
+          Common.defaultArgInfo
     where
     h' = case h of
         NotHidden -> Common.NotHidden
@@ -518,7 +514,7 @@ frommyExp (NotM e) =
         frommyExps n as v
 -}
        (ndrop, h) = case iscon of
-                      Just n -> (n, \ q -> I.Con (I.ConHead q Common.Inductive [])) -- TODO: restore fields
+                      Just n -> (n, \ q -> I.Con (I.ConHead q Common.Inductive []) Common.ConOSystem) -- TODO: restore fields
                       Nothing -> (0, \ f vs -> I.Def f $ map I.Apply vs)
    frommyExps ndrop as (h name [])
   Lam hid (Abs mid t) -> do
@@ -567,7 +563,7 @@ frommyExps ndrop (NotM as) trm =
   ALConPar _ -> __IMPOSSIBLE__
  where
   addend x (I.Var h xs) = I.Var h (xs ++ [I.Apply x])
-  addend x (I.Con h xs) = I.Con h (xs ++ [x])
+  addend x (I.Con h ci xs) = I.Con h ci (xs ++ [x])
   addend x (I.Def h xs) = I.Def h (xs ++ [I.Apply x])
   addend x (I.Shared p) = addend x (I.derefPtr p)
   addend _ _ = __IMPOSSIBLE__
@@ -591,7 +587,8 @@ modifyAbstractExpr = f
   f e = e
 
 modifyAbstractClause :: A.Clause -> A.Clause
-modifyAbstractClause (A.Clause lhs (A.RHS e) decls catchall) = A.Clause lhs (A.RHS (modifyAbstractExpr e)) decls catchall
+modifyAbstractClause (A.Clause lhs dots (A.RHS e mc) decls catchall) =
+  A.Clause lhs dots (A.RHS (modifyAbstractExpr e) mc) decls catchall
 modifyAbstractClause cl = cl
 
 -- ---------------------------------
@@ -667,7 +664,7 @@ frommyClause (ids, pats, mrhs) = do
      cnvps n (_ : _) = __IMPOSSIBLE__
      cnvp (HI hid p) = do
       p' <- case p of
-       CSPatVar v -> return (I.VarP $ let HI _ (Id n, _) = ids !! v in n)
+       CSPatVar v -> return (I.varP $ let HI _ (Id n, _) = ids !! v in n)
        CSPatConApp c ps -> do
         cdef <- lift $ readIORef c
         let (Just ndrop, name) = cdorigin cdef
@@ -682,19 +679,14 @@ frommyClause (ids, pats, mrhs) = do
       return $ Common.Arg (icnvh hid) $ Common.unnamed p'   -- TODO: recover names
  ps <- cnvps 0 pats
  body <- case mrhs of
-          Nothing -> return $ I.NoBody
-          Just e -> do
-           e' <- frommyExp e {- renm e -} -- renaming before adding to clause below
-           let r 0 = I.Body e'
-               r n = I.Bind $ I.Abs "h" $ r (n - 1)
-               e'' = r nv
-           return e''
+          Nothing -> return $ Nothing
+          Just e -> Just <$> frommyExp e
  let cperm =  Perm nv perm
  return $ I.Clause
    { I.clauseRange = SP.noRange
    , I.clauseTel   = tel
-   , I.namedClausePats = IP.numberPatVars cperm $ applySubst (renamingR $ compactP cperm) ps
-   , I.clauseBody  = applySubst (renamingR cperm) <$> body
+   , I.namedClausePats = IP.numberPatVars __IMPOSSIBLE__ cperm $ applySubst (renamingR $ compactP cperm) ps
+   , I.clauseBody  = body
    , I.clauseType  = Nothing -- TODO: compute clause type
    , I.clauseCatchall = False
    }
@@ -706,14 +698,6 @@ contains_constructor = any f
          CSPatConApp{} -> True
          _ -> False
 
-
-etaContractBody :: I.ClauseBody -> MB.TCM I.ClauseBody
-etaContractBody (I.NoBody) = return I.NoBody
-etaContractBody (I.Body b) = etaContract b >>= \b -> return (I.Body b)
-etaContractBody (I.Bind (I.Abs id b)) = etaContractBody b >>= \b -> return (I.Bind (I.Abs id b))
-etaContractBody (I.Bind (I.NoAbs x b))  = I.Bind . I.NoAbs x <$> etaContractBody b
-
-
 -- ---------------------------------
 
 freeIn :: Nat -> MExp o -> Bool
@@ -751,39 +735,16 @@ negtype ee = f (0 :: Int)
 
 -- ---------------------------------------
 
-findClauseDeep :: I.MetaId -> MB.TCM (Maybe (AN.QName, I.Clause, Bool))
-findClauseDeep m = do
-  sig <- getImportedSignature
-  let res = do
-        def <- HMap.elems $ sig ^. MB.sigDefinitions
-        MB.Function{MB.funClauses = cs} <- [MB.theDef def]
-        c <- cs
-        unless (peelbinds False findMeta $ I.clauseBody c) []
-        return (MB.defName def, c, peelbinds __IMPOSSIBLE__ toplevel $ I.clauseBody c)
-  return $ case res of
-    [] -> Nothing
-    r:_ -> Just r   -- TODO: with pattern matching lambdas we might get more than one hit, which to choose?
+findClauseDeep :: Common.InteractionId -> MB.TCM (Maybe (AN.QName, I.Clause, Bool))
+findClauseDeep ii = ignoreAbstractMode $ do  -- Andreas, 2016-09-04, issue #2162
+  MB.InteractionPoint { MB.ipClause = ipCl} <- lookupInteractionPoint ii
+  (f, clauseNo) <- case ipCl of
+    MB.IPClause f clauseNo _ -> return (f, clauseNo)
+    MB.IPNoClause -> MB.typeError $ MB.GenericError $
+      "Cannot apply the auto tactic here, we are not in a function clause"
+  (_, c, _) <- getClauseForIP f clauseNo
+  return $ Just (f, c, maybe __IMPOSSIBLE__ toplevel $ I.clauseBody c)
   where
-    peelbinds d f = r
-     where r b = case b of
-                  I.Bind b -> r $ I.absBody b
-                  I.NoBody -> d
-                  I.Body e -> f e
-    findMeta e =
-     case I.ignoreSharing e of
-      I.Var _ es -> findMetas $ I.argsFromElims es
-      I.Lam _ b -> findMeta (I.absBody b)
-      I.Lit{} -> False
-      I.Level (I.Max as) -> any (fmLevel m) as
-      I.Def _ es -> findMetas $ I.argsFromElims es
-      I.Con _ as -> findMetas as
-      I.Pi it ot -> findMetat (Common.unDom it) || findMetat (I.unAbs ot)
-      I.Sort{} -> False
-      I.MetaV m' _  -> m == m'
-      I.DontCare _ -> False
-      I.Shared{} -> __IMPOSSIBLE__
-    findMetas = any (findMeta . Common.unArg)
-    findMetat (I.El _ e) = findMeta e
     toplevel e =
      case I.ignoreSharing e of
       I.MetaV{} -> True
@@ -821,7 +782,7 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
       (I.Lam hid1 b1, I.Lam hid2 b2) | hid1 == hid2 -> f (nl + 1) n c (I.absBody b1) (I.absBody b2)
       (I.Lit lit1, I.Lit lit2) | lit1 == lit2 -> c (n + 1)
       (I.Def n1 as1, I.Def n2 as2) | n1 == n2 -> fes nl (n + 1) c as1 as2
-      (I.Con n1 as1, I.Con n2 as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
+      (I.Con n1 _ as1, I.Con n2 _ as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
       (I.Pi (Common.Dom info1 it1) ot1, I.Pi (Common.Dom info2 it2) ot2) | Common.argInfoHiding info1 == Common.argInfoHiding info2 -> ft nl n (\n -> ft (nl + 1) n c (I.absBody ot1) (I.absBody ot2)) it1 it2
       (I.Sort{}, I.Sort{}) -> c n -- sloppy
       _ -> Nothing
@@ -831,6 +792,6 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
      _ -> Nothing
     fes nl n c es1 es2 = case (es1, es2) of
      ([], []) -> c n
-     (I.Proj f : es1, I.Proj f' : es2) | f == f' -> fes nl n c es1 es2
+     (I.Proj _ f : es1, I.Proj _ f' : es2) | f == f' -> fes nl n c es1 es2
      (I.Apply (Common.Arg info1 e1) : es1, I.Apply (Common.Arg info2 e2) : es2) | Common.argInfoHiding info1 == Common.argInfoHiding info2 -> f nl n (\n -> fes nl n c es1 es2) e1 e2
      _ -> Nothing
