@@ -1,7 +1,5 @@
 {-# LANGUAGE CPP                       #-}
-{-# LANGUAGE FunctionalDependencies    #-}
 {-# LANGUAGE IncoherentInstances       #-}
-{-# LANGUAGE MultiParamTypeClasses     #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
 
 #if __GLASGOW_HASKELL__ <= 708
@@ -67,12 +65,14 @@ data Phase
     -- ^ Subphase for 'Termination'.
   | ModuleName
     -- ^ Subphase for 'Import'.
+  | BuildInterface
+    -- ^ Subphase for 'Serialization'.
   | Sort
-    -- ^ Subphase for 'Serialize'.
+    -- ^ Subphase for 'Serialization'.
   | BinaryEncode
-    -- ^ Subphase for 'Serialize'.
+    -- ^ Subphase for 'Serialization'.
   | Compress
-    -- ^ Subphase for 'Serialize'.
+    -- ^ Subphase for 'Serialization'.
   | Operators
     -- ^ Subphase for 'Parsing'.
   | Free
@@ -81,6 +81,10 @@ data Phase
     -- ^ Subphase for 'Typing': occurs check for solving metas.
   | CheckLHS
     -- ^ Subphase for 'Typing': checking the LHS
+  | CheckRHS
+    -- ^ Subphase for 'Typing': checking the RHS
+  | TypeSig
+    -- ^ Subphase for 'Typing': checking a type signature
   | UnifyIndices
     -- ^ Subphase for 'CheckLHS': unification of the indices
   | InverseScopeLookup
