@@ -125,13 +125,13 @@ compileModule i = do
             uptodate <- liftIO $ isNewerThan eifFile ifile
             (eif, imps') <- case uptodate of
                 True  -> do
-                    lift $ reportSLn "" 1 $
+                    lift $ reportSLn "compile.epic" 2 $
                       (prettyShow . iModuleName) i ++ " : no compilation is needed."
                     eif <- readEInterface eifFile
                     modify $ \s -> s { curModule = eif }
                     return (eif, Set.insert file imps)
                 False -> do
-                    lift $ reportSLn "" 1 $
+                    lift $ reportSLn "compile.epic" 1 $
                         "Compiling: " ++ (prettyShow . iModuleName) i
                     resetNameSupply
                     initialAnalysis i
@@ -182,6 +182,7 @@ initialAnalysis inter = do
         case defEpicDef def of
           Nothing -> putDelayed q True
           _       -> return ()
+      AbstractDefn -> __IMPOSSIBLE__
       _ -> return ()
 
 idPrint :: String -> (a -> Compile TCM b) -> a -> Compile TCM b
