@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
 
 module Agda.Compiler.Epic.Injection where
 
@@ -98,13 +96,14 @@ patternToTerm :: Nat -> Pattern -> Term
 patternToTerm n p = case p of
     VarP v          -> var n
     DotP t          -> t
-    ConP c typ args -> Con c $ zipWith (\ arg t -> arg {unArg = t}) args
+    ConP c typ args -> Con c ci $ zipWith (\ arg t -> arg {unArg = t}) args
                              $ snd
                              $ foldr (\ arg (n, ts) -> (n + nrBinds arg, patternToTerm n arg : ts))
                                      (n , [])
                              $ map namedArg args
+      where ci = fromConPatternInfo typ
     LitP l          -> Lit l
-    ProjP d         -> Def d [] -- Andreas, 2012-10-31 that might not be enought to get a term from list of patterns (TODO)
+    ProjP _ d       -> Def d [] -- Andreas, 2012-10-31 that might not be enought to get a term from list of patterns (TODO)
 
 nrBinds :: Num i => Pattern -> i
 nrBinds p = case p of
@@ -140,19 +139,20 @@ isInjectiveHere nam idx clause = do
    , text ("argumentNo=" ++ show idx)
    -- , prettyTCM (clausePats clause)
    ]
- case getBody clause of
+ case compiledClauseBody clause of
   Nothing -> return emptyC
   Just body -> do
-    let t    = patternToTerm idxR $ unArg $ fromMaybe __IMPOSSIBLE__ $
-                 unnumberPatVars (clausePats clause) !!! idx
+    let vars = unnumberPatVars $ namedClausePats clause
+    let t    = patternToTerm idxR $ namedArg $ fromMaybe __IMPOSSIBLE__ $
+                 vars !!! idx
         t'   = applySubst (substForDot $ namedClausePats clause) t
-        idxR = sum . map (nrBinds . unArg) . genericDrop (idx + 1) $ unnumberPatVars $ clausePats clause
+        idxR = sum . map (nrBinds . namedArg) . drop (idx + 1) $ vars
     body' <- lift $ reduce body
     lift $ reportSLn "epic.injection" 40 "reduced body"
     injFs <- gets (injectiveFuns . importedModules)
     lift $ reportSLn "epic.injection" 40 "calculated injFs"
     res <- (t' <: body') `runReaderT` (Map.insert nam (InjectiveFun idx
-                                                      (length (clausePats clause))) injFs)
+                                                      (length (namedClausePats clause))) injFs)
     lift $ reportSDoc "epic.injection" 20 $ vcat
       [ text "isInjective:" <+> text (show nam)
       , text "at Index   :" <+> text (show idx)
@@ -258,7 +258,7 @@ instance Injectible a => Injectible [a] where
 instance Injectible a => Injectible (Elim' a) where
   e1 <: e2 =
     case (e1, e2) of
-      (Proj f1 , Proj f2 ) | f1 == f2 -> return $ Just []
+      (Proj _ f1, Proj _ f2) | f1==f2 -> return $ Just []
       (Apply a1, Apply a2)            -> a1 <: a2
       _                               -> return Nothing
 
@@ -290,7 +290,7 @@ instance Injectible Term where
               Just (Apply a) -> t1 <: unArg a
       (Var i1 es1, Var i2 es2) | i1 == i2 -> es1 <: es2
       (Def q1 es1, Def q2 es2) | q1 == q2 -> es1 <: es2
-      (Con con1 args1, Con con2 args2) -> do
+      (Con con1 _ args1, Con con2 _ args2) -> do
         let c1 = conName con1
             c2 = conName con2
         args1' <- flip notForced args1 <$> do lift . getForcedArgs $ c1
@@ -313,7 +313,7 @@ instance Injectible Term where
         args1' <- map unArg <$> mapM (lift . lift . reduce) args1
         args2' <- map unArg <$> mapM (lift . lift . reduce) args2
         unionConstraints <$> zipWithM (\a b -> (a <: b)) args1' args2'
-      (Con con1 args1, Con con2 args2) -> do
+      (Con con1 _ args1, Con con2 _ args2) -> do
         let c1 = conName con1
             c2 = conName con2
         args1' <- map unArg <$> flip notForced args1 <$> getForcedArgs c1
