@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleInstances #-}
 
 -- | Find the places where the builtin static is used and do some normalisation
 --   there.
@@ -57,11 +56,12 @@ etaExpand :: Term -> Compile TCM Term
 etaExpand def@(Def n ts) = do
     defs <- lift $ use $ stImports . sigDefinitions
     let f   = maybe __IMPOSSIBLE__ theDef (HM.lookup n defs)
-        len = length . clausePats . head .  funClauses $ f
+        len = length . namedClausePats . head .  funClauses $ f
         toEta :: Num a => a
         toEta = fromIntegral $ len - length ts
         term  = raise toEta def `applys` map var (downFrom toEta)
-    return $ foldr (\ v t -> Lam defaultArgInfo (Abs v t)) term $ replicate toEta "staticVar"
+        info  = setOrigin Inserted defaultArgInfo
+    return $ foldr (\ v t -> Lam info (Abs v t)) term $ replicate toEta "staticVar"
 etaExpand x = return x
 
 class Evaluate a where
@@ -99,7 +99,7 @@ instance Evaluate Term where
             ]
           ]
         return f
-    Con c args   -> Con c <$> evaluate args
+    Con c ci args   -> Con c ci <$> evaluate args
     Pi  arg abs  -> return term
     Sort s       -> return term
     MetaV i args -> return term
@@ -116,7 +116,4 @@ instance Evaluate Term where
     isStatic :: QName -> Compile TCM Bool
     isStatic q = do
       defs <- lift $ use $ stImports . sigDefinitions
-      return $ case fmap theDef $ HM.lookup q defs of
-          Nothing -> False
-          Just (f@Function{}) -> funStatic f
-          Just _              -> False
+      return $ maybe False (^. theDefLens . funStatic) $ HM.lookup q defs
