@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP                 #-}
-{-# LANGUAGE FlexibleContexts    #-}
-{-# LANGUAGE PatternGuards       #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Compiler.MAlonzo.Compiler where
@@ -14,7 +12,7 @@ import Control.Monad.Reader hiding (mapM_, forM_, mapM, forM, sequence)
 import Control.Monad.State  hiding (mapM_, forM_, mapM, forM, sequence)
 
 import Data.Generics.Geniplate
-import Data.Foldable hiding (any, foldr, sequence_)
+import Data.Foldable hiding (any, all, foldr, sequence_)
 import Data.Function
 import qualified Data.List as List
 import Data.Map (Map)
@@ -24,10 +22,9 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable hiding (for)
 
-import qualified Language.Haskell.Exts.Extension as HS
-import qualified Language.Haskell.Exts.Parser as HS
-import qualified Language.Haskell.Exts.Pretty as HS
-import qualified Language.Haskell.Exts.Syntax as HS
+import Numeric.IEEE
+
+import qualified Agda.Utils.Haskell.Syntax as HS
 
 import System.Directory (createDirectoryIfMissing)
 import System.FilePath hiding (normalise)
@@ -39,12 +36,14 @@ import Agda.Compiler.MAlonzo.Pretty
 import Agda.Compiler.MAlonzo.Primitives
 import Agda.Compiler.ToTreeless
 import Agda.Compiler.Treeless.Unused
+import Agda.Compiler.Treeless.Erase
 
 import Agda.Interaction.FindFile
 import Agda.Interaction.Imports
 import Agda.Interaction.Options
 
 import Agda.Syntax.Common
+import Agda.Syntax.Fixity
 import qualified Agda.Syntax.Abstract.Name as A
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Internal as I
@@ -71,7 +70,7 @@ import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Utils.Pretty (prettyShow, Pretty)
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Singleton
@@ -96,12 +95,12 @@ compile i = do
     yesComp
     writeModule =<< decl <$> curHsMod <*> (definitions =<< curDefs) <*> imports
   where
-  decl mn ds imp = HS.Module dummy mn [] Nothing Nothing imp (map fakeDecl (reverse $ iHaskellCode i) ++ ds)
+  decl mn ds imp = HS.Module mn [] imp (map fakeDecl (reverse $ iHaskellCode i) ++ ds)
   uptodate = liftIO =<< (isNewerThan <$> outFile_ <*> ifile)
   ifile    = maybe __IMPOSSIBLE__ filePath <$>
                (findInterfaceFile . toTopLevelModuleName =<< curMName)
-  noComp   = reportSLn "" 1 . (++ " : no compilation is needed.") . show . A.mnameToConcrete =<< curMName
-  yesComp  = reportSLn "" 1 . (`repl` "Compiling <<0>> in <<1>> to <<2>>") =<<
+  noComp   = reportSLn "compile.ghc" 2 . (++ " : no compilation is needed.") . show . A.mnameToConcrete =<< curMName
+  yesComp  = reportSLn "compile.ghc" 1 . (`repl` "Compiling <<0>> in <<1>> to <<2>>") =<<
              sequence [show . A.mnameToConcrete <$> curMName, ifile, outFile_] :: TCM ()
 
 --------------------------------------------------
@@ -118,23 +117,20 @@ imports = (++) <$> hsImps <*> imps where
              getHaskellImports
 
   unqualRTE :: HS.ImportDecl
-  unqualRTE = HS.ImportDecl dummy mazRTE False False False Nothing Nothing $ Just $
-#if MIN_VERSION_haskell_src_exts(1,17,0)
-              (False, [HS.IVar $ HS.Ident x | x <- [mazCoerceName, mazErasedName]])
-#else
-              (False, [HS.IVar HS.NoNamespace $ HS.Ident x | x <- [mazCoerceName, mazErasedName]])
-#endif
+  unqualRTE = HS.ImportDecl mazRTE False $ Just $
+              (False, [ HS.IVar $ HS.Ident x
+                      | x <- [mazCoerceName, mazErasedName] ++
+                             map treelessPrimName [T.PAdd, T.PSub, T.PMul, T.PQuot, T.PRem, T.PGeq, T.PLt, T.PEqI, T.PEqF] ])
 
   imps :: TCM [HS.ImportDecl]
   imps = List.map decl . uniq <$>
            ((++) <$> importsForPrim <*> (List.map mazMod <$> mnames))
 
   decl :: HS.ModuleName -> HS.ImportDecl
-  decl m = HS.ImportDecl dummy m True False False Nothing Nothing Nothing
+  decl m = HS.ImportDecl m True Nothing
 
   mnames :: TCM [ModuleName]
-  mnames = (++) <$> (Set.elems <$> use stImportedModules)
-                <*> (List.map fst . iImportedModules <$> curIF)
+  mnames = Set.elems <$> use stImportedModules
 
   uniq :: [HS.ModuleName] -> [HS.ModuleName]
   uniq = List.map head . List.group . List.sort
@@ -173,11 +169,11 @@ definition kit (Defn UnusedArg _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 definition kit (Defn NonStrict _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 -}
 definition kit Defn{defArgInfo = info, defName = q} | isIrrelevant info = do
-  reportSDoc "malonzo.definition" 10 $
+  reportSDoc "compile.ghc.definition" 10 $
     text "Not compiling" <+> prettyTCM q <> text "."
   return []
 definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef = d} = do
-  reportSDoc "malonzo.definition" 10 $ vcat
+  reportSDoc "compile.ghc.definition" 10 $ vcat
     [ text "Compiling" <+> prettyTCM q <> text ":"
     , nest 2 $ text (show d)
     ]
@@ -197,11 +193,11 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
             a    = ihname "a" 0
             b    = ihname "a" 1
             vars = [a, b]
-        return [ HS.TypeDecl dummy infT
+        return [ HS.TypeDecl infT
                              (List.map HS.UnkindedVar vars)
                              (HS.TyVar b)
-               , HS.FunBind [HS.Match dummy infV
-                                      (List.map HS.PVar vars) Nothing
+               , HS.FunBind [HS.Match infV
+                                      (List.map HS.PVar vars)
                                       (HS.UnGuardedRhs HS.unit_con)
                                       emptyBinds]
                ]
@@ -209,11 +205,10 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
         let sharp = unqhname "d" q
             x     = ihname "x" 0
         return $
-          [ HS.TypeSig dummy [sharp] $ fakeType $
+          [ HS.TypeSig [sharp] $ fakeType $
               "forall a. a -> a"
-          , HS.FunBind [HS.Match dummy sharp
+          , HS.FunBind [HS.Match sharp
                                  [HS.PVar x]
-                                 Nothing
                                  (HS.UnGuardedRhs (HS.Var (HS.UnQual x)))
                                  emptyBinds]
           ]
@@ -221,11 +216,10 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
         let flat = unqhname "d" q
             x    = ihname "x" 0
         return $
-          [ HS.TypeSig dummy [flat] $ fakeType $
+          [ HS.TypeSig [flat] $ fakeType $
               "forall a. a -> a"
-          , HS.FunBind [HS.Match dummy flat
+          , HS.FunBind [HS.Match flat
                                  [HS.PVar x]
-                                 Nothing
                                  (HS.UnGuardedRhs (HS.Var (HS.UnQual x)))
                                  emptyBinds]
           ]
@@ -239,14 +233,16 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
         | Just (HsType ty) <- compiledHaskell compiled -> do
         ccscov <- ifM (noCheckCover q) (return []) $ do
             ccs <- List.concat <$> mapM checkConstructorType cs
-            cov <- checkCover q ty np cs
+            cov <- checkCover q ty (np + ni) cs
             return $ ccs ++ cov
         return $ tvaldecl q (dataInduction d) 0 (np + ni) [] (Just __IMPOSSIBLE__) ++ ccscov
       Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs } -> do
+        computeErasedConstructorArgs q
         (ars, cds) <- unzip <$> mapM condecl cs
         return $ tvaldecl q (dataInduction d) (List.maximum (np:ars) - np) (np + ni) cds cl
       Constructor{} -> return []
       Record{ recClause = cl, recConHead = con, recFields = flds } -> do
+        computeErasedConstructorArgs q
         let c = conName con
         let noFields = length flds
         let ar = I.arity ty
@@ -255,6 +251,7 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
   --         Nothing -> return $ cdecl q noFields
   --         Just c  -> snd <$> condecl c
         return $ tvaldecl q Inductive noFields ar [cd] cl
+      AbstractDefn -> __IMPOSSIBLE__
   where
   function :: Maybe HaskellExport -> TCM [HS.Decl] -> TCM [HS.Decl]
   function mhe fun = do
@@ -263,10 +260,10 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
       Nothing -> return ccls
       Just (HsExport t name) -> do
         let tsig :: HS.Decl
-            tsig = HS.TypeSig dummy [HS.Ident name] (fakeType t)
+            tsig = HS.TypeSig [HS.Ident name] (fakeType t)
 
             def :: HS.Decl
-            def = HS.FunBind [HS.Match dummy (HS.Ident name) [] Nothing (HS.UnGuardedRhs (hsVarUQ $ dname q)) emptyBinds]
+            def = HS.FunBind [HS.Match (HS.Ident name) [] (HS.UnGuardedRhs (hsVarUQ $ dname q)) emptyBinds]
         return ([tsig,def] ++ ccls)
 
   functionViaTreeless :: QName -> TCM [HS.Decl]
@@ -280,15 +277,15 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
     let (ps, b) = lamView e
         lamView e =
           case stripTopCoerce e of
-            HS.Lambda _ ps b -> (ps, b)
+            HS.Lambda ps b -> (ps, b)
             b                -> ([], b)
-        stripTopCoerce (HS.Lambda i ps b) = HS.Lambda i ps $ stripTopCoerce b
+        stripTopCoerce (HS.Lambda ps b) = HS.Lambda ps $ stripTopCoerce b
         stripTopCoerce e =
           case hsAppView e of
             [c,  e] | c == mazCoerce -> e
             _                        -> e
 
-        funbind f ps b = HS.FunBind [HS.Match dummy f ps Nothing (HS.UnGuardedRhs b) emptyBinds]
+        funbind f ps b = HS.FunBind [HS.Match f ps (HS.UnGuardedRhs b) emptyBinds]
 
     -- The definition of the non-stripped function
     (ps0, _) <- lamView <$> closedTerm (foldr ($) T.TErased $ replicate (length used) T.TLam)
@@ -300,25 +297,20 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
       else [ funbind (dname q) ps b ]
 
   mkwhere :: [HS.Decl] -> [HS.Decl]
-  mkwhere (HS.FunBind [m0, HS.Match _     dn ps mt rhs emptyBinds] :
-           fbs@(_:_)) =
-          [HS.FunBind [m0, HS.Match dummy dn ps mt rhs bindsAux]]
+  mkwhere (HS.FunBind [m0, HS.Match dn ps rhs emptyBinds] : fbs@(_:_)) =
+          [HS.FunBind [m0, HS.Match dn ps rhs bindsAux]]
     where
-#if MIN_VERSION_haskell_src_exts(1,17,0)
     bindsAux :: Maybe HS.Binds
     bindsAux = Just $ HS.BDecls fbs
-#else
-    bindsAux :: HS.Binds
-    bindsAux = HS.BDecls fbs
-#endif
+
   mkwhere fbs = fbs
 
   fbWithType :: HaskellType -> HS.Exp -> [HS.Decl]
   fbWithType ty e =
-    [ HS.TypeSig dummy [unqhname "d" q] $ fakeType ty ] ++ fb e
+    [ HS.TypeSig [unqhname "d" q] $ fakeType ty ] ++ fb e
 
   fb :: HS.Exp -> [HS.Decl]
-  fb e  = [HS.FunBind [HS.Match dummy (unqhname "d" q) [] Nothing
+  fb e  = [HS.FunBind [HS.Match (unqhname "d" q) []
                                 (HS.UnGuardedRhs $ e) emptyBinds]]
 
   axiomErr :: HS.Exp
@@ -367,8 +359,8 @@ intros n cont = freshNames n $ \xs ->
 checkConstructorType :: QName -> TCM [HS.Decl]
 checkConstructorType q = do
   Just (HsDefn ty hs) <- compiledHaskell . defCompiledRep <$> getConstInfo q
-  return [ HS.TypeSig dummy [unqhname "check" q] $ fakeType ty
-         , HS.FunBind [HS.Match dummy (unqhname "check" q) [] Nothing
+  return [ HS.TypeSig [unqhname "check" q] $ fakeType ty
+         , HS.FunBind [HS.Match (unqhname "check" q) []
                                 (HS.UnGuardedRhs $ fakeExp hs) emptyBinds]
          ]
 
@@ -376,19 +368,19 @@ checkCover :: QName -> HaskellType -> Nat -> [QName] -> TCM [HS.Decl]
 checkCover q ty n cs = do
   let tvs = [ "a" ++ show i | i <- [1..n] ]
       makeClause c = do
-        (a, _) <- conArityAndPars c
+        a <- erasedArity c
         Just (HsDefn _ hsc) <- compiledHaskell . defCompiledRep <$> getConstInfo c
         let pat = HS.PApp (HS.UnQual $ HS.Ident hsc) $ replicate a HS.PWildCard
-        return $ HS.Alt dummy pat (HS.UnGuardedRhs $ HS.unit_con) emptyBinds
+        return $ HS.Alt pat (HS.UnGuardedRhs $ HS.unit_con) emptyBinds
 
   cs <- mapM makeClause cs
   let rhs = case cs of
               [] -> fakeExp "()" -- There is no empty case statement in Haskell
               _  -> HS.Case (HS.Var $ HS.UnQual $ HS.Ident "x") cs
 
-  return [ HS.TypeSig dummy [unqhname "cover" q] $ fakeType $ unwords (ty : tvs) ++ " -> ()"
-         , HS.FunBind [HS.Match dummy (unqhname "cover" q) [HS.PVar $ HS.Ident "x"]
-                                Nothing (HS.UnGuardedRhs rhs) emptyBinds]
+  return [ HS.TypeSig [unqhname "cover" q] $ fakeType $ unwords (ty : tvs) ++ " -> ()"
+         , HS.FunBind [HS.Match (unqhname "cover" q) [HS.PVar $ HS.Ident "x"]
+                                (HS.UnGuardedRhs rhs) emptyBinds]
          ]
 
 closedTerm :: T.TTerm -> TCM HS.Exp
@@ -404,13 +396,32 @@ term tm0 = case tm0 of
     return $ hsVarUQ x
   T.TApp (T.TDef f) ts -> do
     used <- lift $ getCompiledArgUse f
-    if any not used && length ts >= length used
-      then do
+    let given   = length ts
+        needed  = length used
+        missing = drop given used
+    if any not used
+      then if any not missing then term (etaExpand (needed - given) tm0) else do
         f <- lift $ HS.Var <$> xhqn "du" f  -- used stripped function
         f `apps` [ t | (t, True) <- zip ts $ used ++ repeat True ]
       else do
         t' <- term (T.TDef f)
         t' `apps` ts
+  T.TApp (T.TCon c) ts -> do
+    kit <- lift coinductionKit
+    if Just c == (nameOfSharp <$> kit)
+      then do
+        t' <- HS.Var <$> lift (xhqn "d" c)
+        apps t' ts
+      else do
+        (ar, _) <- lift $ conArityAndPars c
+        erased  <- lift $ getErasedConArgs c
+        let missing = drop (length ts) erased
+            notErased = not
+        case all notErased missing of
+          False -> term $ etaExpand (length missing) tm0
+          True  -> do
+            f <- lift $ HS.Con <$> conhqn c
+            f `apps` [ t | (t, False) <- zip ts erased ]
   T.TApp t ts -> do
     t' <- term t
     t' `apps` ts
@@ -428,18 +439,14 @@ term tm0 = case tm0 of
     sc' <- term (T.TVar sc)
     alts' <- traverse (alt sc) alts
     def' <- term def
-    let defAlt = HS.Alt dummy HS.PWildCard (HS.UnGuardedRhs def') emptyBinds
+    let defAlt = HS.Alt HS.PWildCard (HS.UnGuardedRhs def') emptyBinds
 
     return $ HS.Case (hsCast sc') (alts' ++ [defAlt])
 
-  T.TLit l -> lift $ literal l
+  T.TLit l -> return $ literal l
   T.TDef q -> do
     HS.Var <$> (lift $ xhqn "d" q)
-  T.TCon q -> do
-    kit <- lift coinductionKit
-    if Just q == (nameOfSharp <$> kit)
-      then HS.Var <$> lift (xhqn "d" q)
-      else hsCast' . HS.Con <$> lift (conhqn q)
+  T.TCon q   -> term (T.TApp (T.TCon q) [])
   T.TPrim p  -> return $ compilePrim p
   T.TUnit    -> return HS.unit_con
   T.TSort    -> return HS.unit_con
@@ -447,62 +454,75 @@ term tm0 = case tm0 of
   T.TError e -> return $ case e of
     T.TUnreachable ->  rtmUnreachableError
   where apps =  foldM (\ h a -> HS.App h <$> term a)
+        etaExpand n t =
+          foldr (const T.TLam)
+                (T.mkTApp (raise n t) [T.TVar i | i <- [n - 1, n - 2..0]])
+                (replicate n ())
+
 
 compilePrim :: T.TPrim -> HS.Exp
-compilePrim s =
-  case s of
-    T.PQuot -> fakeExp "(Prelude.quot :: Integer -> Integer -> Integer)"
-    T.PRem -> fakeExp "(Prelude.rem :: Integer -> Integer -> Integer)"
-    T.PSub -> fakeExp "((Prelude.-) :: Integer -> Integer -> Integer)"
-    T.PAdd -> fakeExp "((Prelude.+) :: Integer -> Integer -> Integer)"
-    T.PMul -> fakeExp "((Prelude.*) :: Integer -> Integer -> Integer)"
-    T.PGeq -> fakeExp "((Prelude.>=) :: Integer -> Integer -> Bool)"
-    T.PLt  -> fakeExp "((Prelude.<) :: Integer -> Integer -> Bool)"
-    T.PEq  -> fakeExp "((Prelude.==) :: Integer -> Integer -> Bool)"
-    T.PSeq -> HS.Var (hsName "seq")
-    -- primitives only used by GuardsToPrims transformation, which MAlonzo doesn't use
-    T.PIf  -> __IMPOSSIBLE__
+compilePrim s = HS.Var $ hsName $ treelessPrimName s
 
 alt :: Int -> T.TAlt -> CC HS.Alt
 alt sc a = do
   case a of
-    T.TACon {} -> do
-      intros (T.aArity a) $ \xs -> do
-        hConNm <- lift $ conhqn $ T.aCon a
-        mkAlt (HS.PApp hConNm $ map HS.PVar xs)
+    T.TACon {T.aCon = c} -> do
+      intros (T.aArity a) $ \ xs -> do
+        erased <- lift $ getErasedConArgs c
+        hConNm <- lift $ conhqn c
+        mkAlt (HS.PApp hConNm $ map HS.PVar [ x | (x, False) <- zip xs erased ])
     T.TAGuard g b -> do
       g <- term g
       b <- term b
-      return $ HS.Alt dummy HS.PWildCard
-                      (HS.GuardedRhss [HS.GuardedRhs dummy [HS.Qualifier g] b])
+      return $ HS.Alt HS.PWildCard
+                      (HS.GuardedRhss [HS.GuardedRhs [HS.Qualifier g] b])
                       emptyBinds
-    T.TALit { T.aLit = (LitQName _ q) } -> mkAlt (litqnamepat q)
-    T.TALit { T.aLit = (LitString _ s) , T.aBody = b } -> do
-      b <- term b
+    T.TALit { T.aLit = LitQName _ q } -> mkAlt (litqnamepat q)
+    T.TALit { T.aLit = l@LitFloat{},   T.aBody = b } -> mkGuarded (treelessPrimName T.PEqF) (literal l) b
+    T.TALit { T.aLit = LitString _ s , T.aBody = b } -> mkGuarded "(==)" (litString s) b
+    T.TALit {} -> mkAlt (HS.PLit $ hslit $ T.aLit a)
+  where
+    mkGuarded eq lit b = do
+      b  <- term b
       sc <- term (T.TVar sc)
       let guard =
-            HS.Var (HS.UnQual (HS.Ident "(==)")) `HS.App`
-              sc`HS.App`
-              litString s
-      return $ HS.Alt dummy HS.PWildCard
-                      (HS.GuardedRhss [HS.GuardedRhs dummy [HS.Qualifier guard] b])
+            HS.Var (HS.UnQual (HS.Ident eq)) `HS.App`
+              sc `HS.App` lit
+      return $ HS.Alt HS.PWildCard
+                      (HS.GuardedRhss [HS.GuardedRhs [HS.Qualifier guard] b])
                       emptyBinds
-    T.TALit {} -> mkAlt (HS.PLit HS.Signless $ hslit $ T.aLit a)
-  where
+
     mkAlt :: HS.Pat -> CC HS.Alt
     mkAlt pat = do
         body' <- term $ T.aBody a
-        return $ HS.Alt dummy pat (HS.UnGuardedRhs $ hsCast body') emptyBinds
+        return $ HS.Alt pat (HS.UnGuardedRhs $ hsCast body') emptyBinds
 
-literal :: Literal -> TCM HS.Exp
+literal :: Literal -> HS.Exp
 literal l = case l of
-  LitNat    _ _   -> return $ typed "Integer"
-  LitFloat  _ _   -> return $ typed "Double"
-  LitQName  _ x   -> return $ litqname x
-  LitString _ s   -> return $ litString s
-  _               -> return $ l'
-  where l'    = HS.Lit $ hslit l
-        typed = HS.ExpTypeSig dummy l' . HS.TyCon . rtmQual
+  LitNat    _ _   -> typed "Integer"
+  LitFloat  _ x   -> floatExp x "Double"
+  LitQName  _ x   -> litqname x
+  LitString _ s   -> litString s
+  _               -> l'
+  where
+    l'    = HS.Lit $ hslit l
+    typed = HS.ExpTypeSig l' . HS.TyCon . rtmQual
+
+    -- ASR (2016-09-14): See Issue #2169.
+    -- Ulf, 2016-09-28: and #2218.
+    floatExp :: Double -> String -> HS.Exp
+    floatExp x s
+      | isNegativeZero x = rte "negativeZero"
+      | isNegativeInf  x = rte "negativeInfinity"
+      | isInfinite x     = rte "positiveInfinity"
+      | isNegativeNaN x  = rte "negativeNaN"
+      | isNaN x          = rte "positiveNaN"
+      | otherwise        = typed s
+
+    rte = HS.Var . HS.Qual mazRTE . HS.Ident
+
+    isNegativeInf x = isInfinite x && x < 0.0
+    isNegativeNaN x = isNaN x && not (identicalIEEE x (0.0 / 0.0))
 
 hslit :: Literal -> HS.Literal
 hslit l = case l of LitNat    _ x -> HS.Int    x
@@ -519,26 +539,47 @@ litString s =
 
 litqname :: QName -> HS.Exp
 litqname x =
-  HS.Con (HS.Qual mazRTE $ HS.Ident "QName") `HS.App`
-  hsTypedInt n `HS.App`
-  hsTypedInt m `HS.App`
-  HS.Lit (HS.String $ show x )
+  rteCon "QName" `apps`
+  [ hsTypedInt n
+  , hsTypedInt m
+  , HS.Lit $ HS.String $ show x
+  , rteCon "Fixity" `apps`
+    [ litAssoc (fixityAssoc fx)
+    , litPrec  (fixityLevel fx) ] ]
   where
+    apps = foldl HS.App
+    rteCon name = HS.Con $ HS.Qual mazRTE $ HS.Ident name
     NameId n m = nameId $ qnameName x
+    fx = theFixity $ nameFixity $ qnameName x
+
+    litAssoc NonAssoc   = rteCon "NonAssoc"
+    litAssoc LeftAssoc  = rteCon "LeftAssoc"
+    litAssoc RightAssoc = rteCon "RightAssoc"
+
+    litPrec Unrelated   = rteCon "Unrelated"
+    litPrec (Related l) = rteCon "Related" `HS.App` hsTypedInt l
 
 litqnamepat :: QName -> HS.Pat
 litqnamepat x =
   HS.PApp (HS.Qual mazRTE $ HS.Ident "QName")
-          [ HS.PLit HS.Signless (HS.Int n)
-          , HS.PLit HS.Signless (HS.Int m)
-          , HS.PWildCard]
+          [ HS.PLit (HS.Int $ fromIntegral n)
+          , HS.PLit (HS.Int $ fromIntegral m)
+          , HS.PWildCard, HS.PWildCard ]
   where
     NameId n m = nameId $ qnameName x
 
+erasedArity :: QName -> TCM Nat
+erasedArity q = do
+  (ar, _) <- conArityAndPars q
+  erased  <- length . filter id <$> getErasedConArgs q
+  return (ar - erased)
+
 condecl :: QName -> TCM (Nat, HS.ConDecl)
 condecl q = do
   (ar, np) <- conArityAndPars q
-  return $ (ar + np, cdecl q ar)
+  erased   <- length . filter id <$> getErasedConArgs q
+  let ar' = ar - erased
+  return $ (ar' + np, cdecl q ar')
 
 cdecl :: QName -> Nat -> HS.ConDecl
 cdecl q n = HS.ConDecl (unqhname "C" q)
@@ -549,10 +590,8 @@ tvaldecl :: QName
             -- ^ Is the type inductive or coinductive?
          -> Nat -> Nat -> [HS.ConDecl] -> Maybe Clause -> [HS.Decl]
 tvaldecl q ind ntv npar cds cl =
-  HS.FunBind [HS.Match dummy vn pvs Nothing
-                       (HS.UnGuardedRhs HS.unit_con) emptyBinds] :
-  maybe [HS.DataDecl dummy kind [] tn tvs
-                     (List.map (HS.QualConDecl dummy [] []) cds) []]
+  HS.FunBind [HS.Match vn pvs (HS.UnGuardedRhs HS.unit_con) emptyBinds] :
+  maybe [HS.DataDecl kind tn tvs cds []]
         (const []) cl
   where
   (tn, vn) = (unqhname "T" q, unqhname "d" q)
@@ -563,7 +602,6 @@ tvaldecl q ind ntv npar cds cl =
   -- single argument are translated into newtypes.
   kind = case (ind, cds) of
     (Inductive, [HS.ConDecl _ [_]]) -> HS.NewType
-    (Inductive, [HS.RecDecl _ [_]]) -> HS.NewType
     _                               -> HS.DataType
 
 infodecl :: QName -> [HS.Decl] -> [HS.Decl]
@@ -581,7 +619,7 @@ hsCast = addcast . go where
   addcast es = foldl HS.App mazCoerce es
   -- this need to be extended if you generate other kinds of exps.
   go (HS.App e1 e2    ) = go e1 ++ [hsCast e2]
-  go (HS.Lambda _ ps e) = [ HS.Lambda dummy ps (hsCast e) ]
+  go (HS.Lambda _ ps e) = [ HS.Lambda ps (hsCast e) ]
   go e = [e]
 -}
 
@@ -590,7 +628,7 @@ hsCast e = hsCoerce (hsCast' e)
 
 hsCast' :: HS.Exp -> HS.Exp
 hsCast' (HS.InfixApp e1 op e2) = hsCoerce $ HS.InfixApp (hsCast' e1) op (hsCast' e2)
-hsCast' (HS.Lambda _ ps e)     = HS.Lambda dummy ps $ hsCast' e
+hsCast' (HS.Lambda ps e)       = HS.Lambda ps $ hsCast' e
 hsCast' (HS.Let bs e)          = HS.Let bs $ hsCast' e
 hsCast' (HS.Case sc alts)      = HS.Case (hsCast' sc) (map (hsMapAlt hsCast') alts)
 hsCast' e =
@@ -601,7 +639,7 @@ hsCast' e =
 -- We still have to coerce function applications in arguments to coerced
 -- functions.
 hsCastApp :: HS.Exp -> HS.Exp
-hsCastApp (HS.Lambda i ps b) = HS.Lambda i ps (hsCastApp b)
+hsCastApp (HS.Lambda ps b) = HS.Lambda ps (hsCastApp b)
 hsCastApp (HS.Case sc bs) = HS.Case (hsCastApp sc) (map (hsMapAlt hsCastApp) bs)
 hsCastApp (HS.InfixApp e1 op e2) = HS.InfixApp (hsCastApp e1) op (hsCastApp e2)
 hsCastApp e =
@@ -611,7 +649,7 @@ hsCastApp e =
 
 -- No coercion for literal integers
 hsCoerce :: HS.Exp -> HS.Exp
-hsCoerce e@(HS.ExpTypeSig _ (HS.Lit (HS.Int{})) _) = e
+hsCoerce e@(HS.ExpTypeSig (HS.Lit (HS.Int{})) _) = e
 hsCoerce (HS.Case sc alts) = HS.Case sc (map (hsMapAlt hsCoerce) alts)
 hsCoerce (HS.Let bs e) = HS.Let bs $ hsCoerce e
 hsCoerce e =
@@ -630,13 +668,13 @@ copyRTEModules = do
   (lift . copyDirContent srcDir) =<< compileDir
 
 writeModule :: HS.Module -> TCM ()
-writeModule (HS.Module l m ps w ex imp ds) = do
+writeModule (HS.Module m ps imp ds) = do
   -- Note that GHC assumes that sources use ASCII or UTF-8.
   out <- outFile m
   liftIO $ UTF8.writeFile out $ prettyPrint $
-    HS.Module l m (p : ps) w ex imp ds
+    HS.Module m (p : ps) imp ds
   where
-  p = HS.LanguagePragma dummy $ List.map HS.Ident $
+  p = HS.LanguagePragma $ List.map HS.Ident $
         [ "EmptyDataDecls"
         , "ExistentialQuantification"
         , "ScopedTypeVariables"
@@ -645,8 +683,7 @@ writeModule (HS.Module l m ps w ex imp ds) = do
         ]
 
 
-outFile' :: (HS.Pretty a, TransformBi HS.ModuleName (Wrap a)) =>
-            a -> TCM (FilePath, FilePath)
+outFile' :: Pretty a => a -> TCM (FilePath, FilePath)
 outFile' m = do
   mdir <- compileDir
   let (fdir, fn) = splitFileName $ repldot pathSeparator $
