@@ -6,7 +6,7 @@ import Control.Monad.State
 import Data.Char
 import Data.List as L
 import Data.Map as M
-import qualified Language.Haskell.Exts.Syntax as HS
+import qualified Agda.Utils.Haskell.Syntax as HS
 
 import Agda.Compiler.Common
 import Agda.Compiler.ToTreeless
@@ -15,6 +15,7 @@ import Agda.Compiler.MAlonzo.Misc
 import Agda.Compiler.MAlonzo.Pretty
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Treeless
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Primitive
@@ -59,28 +60,47 @@ checkTypeOfMain q ty ret
           [prettyTCM io] ++ pwords " A, for some A. The given type is" ++ [prettyTCM ty]
         typeError $ GenericError $ show err
   where
-    mainAlias = HS.FunBind [HS.Match dummy mainLHS [] Nothing mainRHS emptyBinds ]
+    mainAlias = HS.FunBind [HS.Match mainLHS [] mainRHS emptyBinds ]
     mainLHS   = HS.Ident "main"
     mainRHS   = HS.UnGuardedRhs $ HS.Var $ HS.UnQual $ unqhname "d" q
 
+treelessPrimName :: TPrim -> String
+treelessPrimName p =
+  case p of
+    PQuot -> "quotInt"
+    PRem  -> "remInt"
+    PSub  -> "subInt"
+    PAdd  -> "addInt"
+    PMul  -> "mulInt"
+    PGeq  -> "geqInt"
+    PLt   -> "ltInt"
+    PEqI  -> "eqInt"
+    PEqF  -> "eqFloat"
+    -- MAlonzo uses literal patterns, so we don't need equality for the other primitive types
+    PEqC  -> __IMPOSSIBLE__
+    PEqS  -> __IMPOSSIBLE__
+    PEqQ  -> __IMPOSSIBLE__
+    PSeq  -> "seq"
+    -- primitives only used by GuardsToPrims transformation, which MAlonzo doesn't use
+    PIf   -> __IMPOSSIBLE__
+
 -- Haskell modules to be imported for BUILT-INs
 importsForPrim :: TCM [HS.ModuleName]
 importsForPrim =
   fmap (++ [HS.ModuleName "Data.Text"]) $
   xForPrim $
   L.map (\(s, ms) -> (s, return (L.map HS.ModuleName ms))) $
-  [ "CHAR"           |-> ["Data.Char"]
-  , "primIsDigit"    |-> ["Data.Char"]
-  , "primIsLower"    |-> ["Data.Char"]
-  , "primIsDigit"    |-> ["Data.Char"]
-  , "primIsAlpha"    |-> ["Data.Char"]
-  , "primIsSpace"    |-> ["Data.Char"]
-  , "primIsAscii"    |-> ["Data.Char"]
-  , "primIsLatin1"   |-> ["Data.Char"]
-  , "primIsPrint"    |-> ["Data.Char"]
-  , "primIsHexDigit" |-> ["Data.Char"]
-  , "primToUpper"    |-> ["Data.Char"]
-  , "primToLower"    |-> ["Data.Char"]
+  [ "CHAR"              |-> ["Data.Char"]
+  , "primIsAlpha"       |-> ["Data.Char"]
+  , "primIsAscii"       |-> ["Data.Char"]
+  , "primIsDigit"       |-> ["Data.Char"]
+  , "primIsHexDigit"    |-> ["Data.Char"]
+  , "primIsLatin1"      |-> ["Data.Char"]
+  , "primIsLower"       |-> ["Data.Char"]
+  , "primIsPrint"       |-> ["Data.Char"]
+  , "primIsSpace"       |-> ["Data.Char"]
+  , "primToLower"       |-> ["Data.Char"]
+  , "primToUpper"       |-> ["Data.Char"]
   ]
   where (|->) = (,)
 
@@ -91,8 +111,9 @@ xForPrim table = do
   qs <- HMap.keys <$> curDefs
   bs <- toList <$> gets stBuiltinThings
   let getName (Builtin (Def q _))    = q
-      getName (Builtin (Con q _))    = conName q
+      getName (Builtin (Con q _ _))  = conName q
       getName (Builtin (Shared p))   = getName (Builtin $ derefPtr p)
+      getName (Builtin (Lam _ b))    = getName (Builtin $ unAbs b)
       getName (Builtin _)            = __IMPOSSIBLE__
       getName (Prim (PrimFun q _ _)) = q
   concat <$> sequence [ maybe (return []) id $ L.lookup s table
@@ -132,18 +153,17 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
 
   -- Floating point functions
   , "primNatToFloat"        |-> return "(fromIntegral :: Integer -> Double)"
-  , "primFloatPlus"         |-> return "((+) :: Double -> Double -> Double)"
-  , "primFloatMinus"        |-> return "((-) :: Double -> Double -> Double)"
-  , "primFloatTimes"        |-> return "((*) :: Double -> Double -> Double)"
-  , "primFloatDiv"          |-> return "((/) :: Double -> Double -> Double)"
-  , "primFloatEquality"     |-> return "((\\ x y -> if isNaN x && isNaN y then True else x == y) :: Double -> Double -> Bool)"
-  , "primFloatLess"         |-> return (unwords
-                                  [ "((\\ x y ->"
-                                  , "let isNegInf z = z < 0 && isInfinite z in"
-                                  , "if isNegInf y then False else"
-                                  , "if isNegInf x then True  else"
-                                  , "if isNaN x    then True  else"
-                                  , "x < y) :: Double -> Double -> Bool)" ])
+  , "primFloatPlus"         |-> return "((+)          :: Double -> Double -> Double)"
+  , "primFloatMinus"        |-> return "((-)          :: Double -> Double -> Double)"
+  , "primFloatTimes"        |-> return "((*)          :: Double -> Double -> Double)"
+  , "primFloatNegate"       |-> return "(negate       :: Double -> Double)"
+  , "primFloatDiv"          |-> return "((/)          :: Double -> Double -> Double)"
+  -- ASR (2016-09-14). We use bitwise equality for comparing Double
+  -- because Haskell's Eq, which equates 0.0 and -0.0, allows to prove
+  -- a contradiction (see Issue #2169).
+  , "primFloatEquality"     |-> return "MAlonzo.RTE.eqFloat"
+  , "primFloatNumericalEquality" |-> return "MAlonzo.RTE.eqNumFloat"
+  , "primFloatNumericalLess" |-> return "MAlonzo.RTE.ltNumFloat"
   , "primFloatSqrt"         |-> return "(sqrt :: Double -> Double)"
   , "primRound"             |-> return "(round :: Double -> Integer)"
   , "primFloor"             |-> return "(floor :: Double -> Integer)"
@@ -151,7 +171,13 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   , "primExp"               |-> return "(exp :: Double -> Double)"
   , "primLog"               |-> return "(log :: Double -> Double)"
   , "primSin"               |-> return "(sin :: Double -> Double)"
-  , "primShowFloat"         |-> return "(Data.Text.pack . (\\ x -> if isNegativeZero x then \"0.0\" else show x) :: Double -> Data.Text.Text)"
+  , "primCos"               |-> return "(cos :: Double -> Double)"
+  , "primTan"               |-> return "(tan :: Double -> Double)"
+  , "primASin"              |-> return "(asin :: Double -> Double)"
+  , "primACos"              |-> return "(acos :: Double -> Double)"
+  , "primATan"              |-> return "(atan :: Double -> Double)"
+  , "primATan2"             |-> return "(atan2 :: Double -> Double -> Double)"
+  , "primShowFloat"         |-> return "(Data.Text.pack . show :: Double -> Data.Text.Text)"
 
   -- Character functions
   , "primCharEquality"   |-> rel "(==)" "Char"
@@ -174,12 +200,13 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   , "primStringFromList" |-> return "Data.Text.pack"
   , "primStringAppend"   |-> binAsis "Data.Text.append" "Data.Text.Text"
   , "primStringEquality" |-> rel "(==)" "Data.Text.Text"
-  , "primShowString"     |-> return "Data.Text.unpack"
+  , "primShowString"     |-> return "id"
 
   -- Reflection
   , "primQNameEquality"   |-> rel "(==)" "MAlonzo.RTE.QName"
   , "primQNameLess"       |-> rel "(<)" "MAlonzo.RTE.QName"
   , "primShowQName"       |-> return "Data.Text.pack . MAlonzo.RTE.qnameString"
+  , "primQNameFixity"     |-> return "MAlonzo.RTE.qnameFixity"
   , "primMetaEquality"    |-> rel "(==)" "Integer"
   , "primMetaLess"        |-> rel "(<)" "Integer"
   , "primShowMeta"        |-> return "\\ x -> Data.Text.pack (\"_\" ++ show (x :: Integer))"
@@ -223,7 +250,7 @@ noCheckCover q = (||) <$> isBuiltin q builtinNat <*> isBuiltin q builtinInteger
 
 pconName :: String -> TCM String
 pconName s = toS . ignoreSharing =<< getBuiltin s where
-  toS (Con q _) = prettyPrint <$> conhqn (conName q)
+  toS (Con q _ _) = prettyPrint <$> conhqn (conName q)
   toS (Lam _ t) = toS (unAbs t)
   toS _ = mazerror $ "pconName" ++ s
 
