@@ -1,10 +1,5 @@
 {-# LANGUAGE CPP                   #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE FlexibleContexts      #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TupleSections         #-}
 {-# LANGUAGE UndecidableInstances  #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
@@ -29,10 +24,12 @@ import qualified Agda.Syntax.Concrete as C -- ToDo: Remove with instance of ToCo
 import Agda.Syntax.Position
 import Agda.Syntax.Abstract as A hiding (Open, Apply, Assign)
 import Agda.Syntax.Abstract.Views as A
+import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Common
 import Agda.Syntax.Info (ExprInfo(..),MetaInfo(..),emptyMetaInfo,exprNoRange)
 import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Internal as I
+import Agda.Syntax.Literal
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.Syntax.Translation.ConcreteToAbstract
@@ -42,6 +39,7 @@ import Agda.Syntax.Fixity(Precedence(..))
 import Agda.Syntax.Parser
 
 import Agda.TheTypeChecker
+import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Monad as M hiding (MetaInfo)
 import Agda.TypeChecking.MetaVars
@@ -55,8 +53,11 @@ import Agda.TypeChecking.Irrelevance (wakeIrrelevantVars)
 import Agda.TypeChecking.Pretty (prettyTCM)
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.CheckInternal
+import Agda.TypeChecking.SizedTypes.Solve
 import qualified Agda.TypeChecking.Pretty as TP
 
+import Agda.Termination.TermCheck (termMutual)
+
 import Agda.Utils.Except ( Error(strMsg), MonadError(catchError, throwError) )
 import Agda.Utils.Functor
 import Agda.Utils.Lens
@@ -75,7 +76,7 @@ import Agda.Utils.Impossible
 
 parseExpr :: Range -> String -> TCM C.Expr
 parseExpr rng s = do
-  C.ExprWhere e wh <- liftIO $ parsePosString exprWhereParser pos s
+  C.ExprWhere e wh <- runPM $ parsePosString exprWhereParser pos s
   unless (null wh) $ typeError $ GenericError $
     "where clauses are not supported in holes"
   return e
@@ -89,10 +90,10 @@ parseExprIn ii rng s = do
     e   <- parseExpr rng s
     concreteToAbstract (clScope mi) e
 
-giveExpr :: MetaId -> Expr -> TCM Expr
+giveExpr :: Maybe InteractionId -> MetaId -> Expr -> TCM ()
 -- When translator from internal to abstract is given, this function might return
 -- the expression returned by the type checker.
-giveExpr mi e = do
+giveExpr mii mi e = do
     mv <- lookupMeta mi
     -- In the context (incl. signature) of the meta variable,
     -- type check expression and assign meta
@@ -112,6 +113,7 @@ giveExpr mi e = do
           TP.text "give: instantiated meta type =" TP.<+> prettyTCM t'
         v <- checkExpr e t'
         case mvInstantiation mv of
+
           InstV xs v' -> unlessM ((Irrelevant ==) <$> asks envRelevance) $ do
             reportSDoc "interaction.give" 20 $ TP.sep
               [ TP.text "meta was already set to value v' = " TP.<+> prettyTCM v'
@@ -130,8 +132,34 @@ giveExpr mi e = do
               [ TP.text "in meta context, v' = " TP.<+> prettyTCM v'
               ]
             equalTerm t' v v'  -- Note: v' now lives in context of meta
-          _ -> updateMeta mi v
-        reify v
+
+          _ -> do -- updateMeta mi v
+            reportSLn "interaction.give" 20 "give: meta unassigned, assigning..."
+            args <- getContextArgs
+            nowSolvingConstraints $ assign DirEq mi args v
+
+        reportSDoc "interaction.give" 20 $ TP.text "give: meta variable updated!"
+        redoChecks mii
+        wakeupConstraints mi
+        solveSizeConstraints DontDefaultToInfty
+        -- Double check.
+        vfull <- instantiateFull v
+        checkInternal vfull t'
+
+-- | After a give, redo termination etc. checks for function which was complemented.
+redoChecks :: Maybe InteractionId -> TCM ()
+redoChecks Nothing = return ()
+redoChecks (Just ii) = do
+  reportSLn "interaction.give" 20 $
+    "give: redoing termination check for function surrounding " ++ show ii
+  ip <- lookupInteractionPoint ii
+  case ipClause ip of
+    IPNoClause -> return ()
+    IPClause f _ _ -> do
+      mb <- mutualBlockOf f
+      terErrs <- local (\ e -> e { envMutualBlock = Just mb }) $ termMutual []
+      unless (null terErrs) $ typeError $ TerminationCheckFailed terErrs
+  -- TODO redo positivity check!
 
 -- | Try to fill hole by expression.
 --
@@ -149,11 +177,10 @@ give ii mr e = liftTCM $ do
   reportSDoc "interaction.give" 10 $ TP.text "giving expression" TP.<+> prettyTCM e
   reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
   -- Try to give mi := e
-  _ <- catchError (giveExpr mi e) $ \ err -> case err of
+  catchError (giveExpr (Just ii) mi e) $ \ err -> case err of
     -- Turn PatternErr into proper error:
-    PatternErr{} -> do
-      err <- withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
-      typeError $ GenericError $ show err
+    PatternErr{} -> typeError . GenericDocError =<< do
+      withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
     _ -> throwError err
   removeInteractionPoint ii
   return e
@@ -192,7 +219,7 @@ refine ii mr e = do
         appMeta e = do
           let rng = rightMargin r -- Andreas, 2013-05-01 conflate range to its right margin to ensure that appended metas are last in numbering.  This fixes issue 841.
           -- Make new interaction point
-          ii <- registerInteractionPoint rng Nothing
+          ii <- registerInteractionPoint False rng Nothing
           let info = Info.MetaInfo
                 { Info.metaRange = rng
                 , Info.metaScope = scope { scopePrecedence = ArgumentCtx }
@@ -241,15 +268,35 @@ evalInMeta ii e =
 data Rewrite =  AsIs | Instantiated | HeadNormal | Simplified | Normalised
     deriving (Read)
 
-normalForm :: Rewrite -> Type -> TCM Type
+normalForm :: (Reduce t, Simplify t, Normalise t) => Rewrite -> t -> TCM t
 normalForm AsIs         t = return t
 normalForm Instantiated t = return t   -- reify does instantiation
 normalForm HeadNormal   t = {- etaContract =<< -} reduce t
 normalForm Simplified   t = {- etaContract =<< -} simplify t
 normalForm Normalised   t = {- etaContract =<< -} normalise t
 
+data ComputeMode = DefaultCompute | IgnoreAbstract | UseShowInstance
+  deriving (Read, Eq)
 
-data OutputForm a b = OutputForm Range ProblemId (OutputConstraint a b)
+computeIgnoreAbstract :: ComputeMode -> Bool
+computeIgnoreAbstract DefaultCompute  = False
+computeIgnoreAbstract IgnoreAbstract  = True
+computeIgnoreAbstract UseShowInstance = True
+  -- UseShowInstance requires the result to be a string literal so respecting
+  -- abstract can only ever break things.
+
+computeWrapInput :: ComputeMode -> String -> String
+computeWrapInput UseShowInstance s = "show (" ++ s ++ ")"
+computeWrapInput _               s = s
+
+showComputed :: ComputeMode -> Expr -> TCM Doc
+showComputed UseShowInstance e =
+  case e of
+    A.Lit (LitString _ s) -> pure (text s)
+    _                     -> (text "Not a string:" $$) <$> prettyATop e
+showComputed _ e = prettyATop e
+
+data OutputForm a b = OutputForm Range [ProblemId] (OutputConstraint a b)
   deriving (Functor)
 
 data OutputConstraint a b
@@ -295,13 +342,21 @@ outputFormId (OutputForm _ _ o) = out o
       FindInScopeOF _ _ _        -> __IMPOSSIBLE__
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
-  reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) (last pids) <$> reify c)
+  reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) pids <$> reify c)
+
+reifyElimToExpr :: I.Elim -> TCM Expr
+reifyElimToExpr e = case e of
+    I.Apply v -> appl "apply" <$> reify v
+    I.Proj _o f -> appl "proj" <$> reify ((defaultArg $ I.Def f []) :: Arg Term)
+  where
+    appl :: String -> Arg Expr -> Expr
+    appl s v = A.App exprNoRange (A.Lit (LitString noRange s)) $ fmap unnamed v
 
 instance Reify Constraint (OutputConstraint Expr Expr) where
     reify (ValueCmp cmp t u v)   = CmpInType cmp <$> reify t <*> reify u <*> reify v
     reify (ElimCmp cmp t v es1 es2) =
-      CmpElim cmp <$> reify t <*> reify es1
-                              <*> reify es2
+      CmpElim cmp <$> reify t <*> mapM reifyElimToExpr es1
+                              <*> mapM reifyElimToExpr es2
     reify (LevelCmp cmp t t')    = CmpLevels cmp <$> reify t <*> reify t'
     reify (TypeCmp cmp t t')     = CmpTypes cmp <$> reify t <*> reify t'
     reify (TelCmp a b cmp t t')  = CmpTeles cmp <$> (ETel <$> reify t) <*> (ETel <$> reify t')
@@ -336,12 +391,11 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
               OfType tac <$> reify goal
           Open{}  -> __IMPOSSIBLE__
           OpenIFS{}  -> __IMPOSSIBLE__
-          InstS{} -> __IMPOSSIBLE__
           InstV{} -> __IMPOSSIBLE__
     reify (FindInScope m _b mcands) = FindInScopeOF
       <$> (reify $ MetaV m [])
       <*> (reify =<< getMetaType m)
-      <*> (forM (fromMaybe [] mcands) $ \ (Candidate tm ty eti) -> do
+      <*> (forM (fromMaybe [] mcands) $ \ (Candidate tm ty eti _) -> do
             (,) <$> reify tm <*> reify ty)
     reify (IsEmpty r a) = IsEmptyType <$> reify a
     reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
@@ -355,8 +409,8 @@ showComparison cmp = " " ++ prettyShow cmp ++ " "
 instance (Show a,Show b) => Show (OutputForm a b) where
   show o =
     case o of
-      OutputForm r 0   c -> show c ++ range r
-      OutputForm r pid c -> "[" ++ prettyShow pid ++ "] " ++ show c ++ range r
+      OutputForm r []   c -> show c ++ range r
+      OutputForm r pids c -> show pids ++ " " ++ show c ++ range r
     where
       range r | null s    = ""
               | otherwise = " [ at " ++ s ++ " ]"
@@ -433,14 +487,14 @@ getConstraints = liftTCM $ do
     cs <- forM cs $ \c -> do
             cl <- reify c
             enterClosure cl abstractToConcrete_
-    ss <- mapM toOutputForm =<< getSolvedInteractionPoints True -- get all
+    ss <- mapM toOutputForm =<< getSolvedInteractionPoints True AsIs -- get all
     return $ ss ++ cs
   where
     toOutputForm (ii, mi, e) = do
       mv <- getMetaInfo <$> lookupMeta mi
       withMetaInfo mv $ do
         let m = QuestionMark emptyMetaInfo{ metaNumber = Just $ fromIntegral ii } ii
-        abstractToConcrete_ $ OutputForm noRange 0 $ Assign m e
+        abstractToConcrete_ $ OutputForm noRange [] $ Assign m e
 
 -- | @getSolvedInteractionPoints True@ returns all solutions,
 --   even if just solved by another, non-interaction meta.
@@ -448,8 +502,8 @@ getConstraints = liftTCM $ do
 --   @getSolvedInteractionPoints False@ only returns metas that
 --   are solved by a non-meta.
 
-getSolvedInteractionPoints :: Bool -> TCM [(InteractionId, MetaId, Expr)]
-getSolvedInteractionPoints all = concat <$> do
+getSolvedInteractionPoints :: Bool -> Rewrite -> TCM [(InteractionId, MetaId, Expr)]
+getSolvedInteractionPoints all norm = concat <$> do
   mapM solution =<< getInteractionIdsAndMetas
   where
     solution (i, m) = do
@@ -462,12 +516,11 @@ getSolvedInteractionPoints all = concat <$> do
               v <- ignoreSharing <$> instantiate v
               let isMeta = case v of MetaV{} -> True; _ -> False
               if isMeta && not all then return [] else do
-                e <- reify v
+                e <- reify =<< normalForm norm v
                 return [(i, m, ScopedExpr scope e)]
             unsol = return []
         case mvInstantiation mv of
           InstV{}                        -> sol (MetaV m $ map Apply args)
-          InstS{}                        -> sol (Level $ Max [Plus 0 $ MetaLevel m $ map Apply args])
           Open{}                         -> unsol
           OpenIFS{}                      -> unsol
           BlockedConst{}                 -> unsol
@@ -522,7 +575,6 @@ typesOfHiddenMetas norm = liftTCM $ do
   openAndImplicit is x m =
     case mvInstantiation m of
       M.InstV{} -> False
-      M.InstS{} -> False
       M.Open    -> x `notElem` is
       M.OpenIFS -> x `notElem` is  -- OR: True !?
       M.BlockedConst{} -> True
@@ -530,8 +582,9 @@ typesOfHiddenMetas norm = liftTCM $ do
 
 metaHelperType :: Rewrite -> InteractionId -> Range -> String -> TCM (OutputConstraint' Expr Expr)
 metaHelperType norm ii rng s = case words s of
-  []    -> fail "C-c C-h expects an argument of the form f e1 e2 .. en"
+  []    -> failure
   f : _ -> do
+    ensureName f
     A.Application h args <- A.appView . getBody . deepUnscope <$> parseExprIn ii rng ("let " ++ f ++ " = _ in " ++ s)
     withInteractionId ii $ do
       cxtArgs  <- getContextArgs
@@ -546,8 +599,29 @@ metaHelperType norm ii rng s = case words s of
           (delta1, delta2, _, a', as', vs') = splitTelForWith tel a (map OtherType as) vs
       a <- local (\e -> e { envPrintDomainFreePi = True }) $ do
         reify =<< cleanupType arity args =<< normalForm norm =<< fst <$> withFunctionType delta1 vs' as' delta2 a'
+      reportSDoc "interaction.helper" 10 $ TP.vcat
+        [ TP.text "generating helper function"
+        , TP.nest 2 $ TP.text "tel    = " TP.<+> inTopContext (prettyTCM tel)
+        , TP.nest 2 $ TP.text "a      = " TP.<+> prettyTCM a
+        , TP.nest 2 $ TP.text "vs     = " TP.<+> prettyTCM vs
+        , TP.nest 2 $ TP.text "as     = " TP.<+> prettyTCM as
+        , TP.nest 2 $ TP.text "delta1 = " TP.<+> inTopContext (prettyTCM delta1)
+        , TP.nest 2 $ TP.text "delta2 = " TP.<+> inTopContext (addContext delta1 $ prettyTCM delta2)
+        , TP.nest 2 $ TP.text "a'     = " TP.<+> inTopContext (addContext delta1 $ addContext delta2 $ prettyTCM a')
+        , TP.nest 2 $ TP.text "as'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM as')
+        , TP.nest 2 $ TP.text "vs'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM vs')
+        ]
       return (OfType' h a)
   where
+    failure = typeError $ GenericError $ "Expected an argument of the form f e1 e2 .. en"
+    ensureName f = do
+      ce <- parseExpr rng f
+      case ce of
+        C.Ident{} -> return ()
+        C.RawApp _ [C.Ident{}] -> return ()
+        _ -> do
+         reportSLn "interaction.helper" 10 $ "ce = " ++ show ce
+         failure
     cleanupType arity args t = do
       -- Get the arity of t
       TelV ttel _ <- telView t
@@ -593,7 +667,7 @@ metaHelperType norm ii rng s = case words s of
     onNamesTm f v = case v of
       I.Var x es   -> I.Var x <$> onNamesElims f es
       I.Def q es   -> I.Def q <$> onNamesElims f es
-      I.Con c args -> I.Con c <$> onNamesArgs f args
+      I.Con c ci args -> I.Con c ci <$> onNamesArgs f args
       I.Lam i b    -> I.Lam i <$> onNamesAbs f onNamesTm b
       I.Pi a b     -> I.Pi <$> traverse (onNames f) a <*> onNamesAbs f onNames b
       I.DontCare v -> I.DontCare <$> onNamesTm f v
@@ -706,10 +780,10 @@ introTactic pmLambda ii = do
           I.Def d _ -> do
             def <- getConstInfo d
             case theDef def of
-              Datatype{}    -> addCtxTel tel' $ introData t
+              Datatype{}    -> addContext tel' $ introData t
               Record{ recNamedCon = name }
-                | name      -> addCtxTel tel' $ introData t
-                | otherwise -> addCtxTel tel' $ introRec d
+                | name      -> addContext tel' $ introData t
+                | otherwise -> addContext tel' $ introRec d
               _ -> fallback
           _ -> fallback
      `catchError` \_ -> return []
@@ -720,7 +794,7 @@ introTactic pmLambda ii = do
 
     showTCM v = show <$> prettyTCM v
 
-    introFun tel = addCtxTel tel' $ do
+    introFun tel = addContext tel' $ do
         reportSDoc "interaction.intro" 10 $ do TP.text "introFun" TP.<+> prettyTCM (telFromList tel)
         imp <- showImplicitArguments
         let okHiding0 h = imp || h == NotHidden
@@ -746,7 +820,7 @@ introTactic pmLambda ii = do
 
     introData t = do
       let tel  = telFromList [defaultDom ("_", t)]
-          pat  = [defaultArg $ unnamed $ I.VarP (0,"c")]
+          pat  = [defaultArg $ unnamed $ debruijnNamedVar "c" 0]
       r <- splitLast CoInductive tel pat
       case r of
         Left err -> return []
