@@ -1,5 +1,4 @@
 -- {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleInstances #-}
 
 module Agda.Interaction.EmacsTop
     ( mimicGHCi
@@ -28,7 +27,7 @@ import Agda.Interaction.EmacsCommand
 import Agda.Interaction.Highlighting.Emacs
 import Agda.Interaction.Options
 
-import Agda.Version
+import Agda.VersionCommit
 
 ----------------------------------
 
@@ -88,9 +87,34 @@ lispifyResponse (Resp_HighlightingInfo info modFile) =
 lispifyResponse (Resp_DisplayInfo info) = return $ case info of
     Info_CompilationOk -> f "The module was successfully compiled." "*Compilation result*"
     Info_Constraints s -> f s "*Constraints*"
-    Info_AllGoals s -> f s "*All Goals*"
+    Info_AllGoalsWarnings g  w e -> f body ("*All" ++ title ++ "*")
+      where
+        isG = not $ null g
+        isW = not $ null w
+        isE = not $ null e
+        title = intercalate "," $ catMaybes
+                  [ " Goals"    <$ guard isG
+                  , " Warnings" <$ guard isW
+                  , " Errors"   <$ guard isE
+                  , " Done"     <$ guard (not (isG || isW || isE))
+                  ]
+        delimiter s = concat [ replicate 4 '\x2014'
+                             , " ", s, " "
+                             , replicate (54 - length s) '\x2014'
+                             ]
+
+        body = intercalate "\n" $ catMaybes
+                 [ g                    <$ guard isG
+                 , delimiter "Warnings" <$ guard (isW && (isG || isE))
+                 , w                    <$ guard isW
+                 , delimiter "Errors"   <$ guard (isE && (isG || isW))
+                 , e                    <$ guard isE
+                 ]
     Info_Auto s -> f s "*Auto*"
     Info_Error s -> f s "*Error*"
+    -- FNF: if Info_Warning comes back into use, the above should be
+    -- clearWarning : f s "*Error*"
+    --Info_Warning s -> [ display_warning "*Errors*" s ] -- FNF: currently unused
     Info_Time s -> f (render s) "*Time*"
     Info_NormalForm s -> f (render s) "*Normal Form*"   -- show?
     Info_InferredType s -> f (render s) "*Inferred Type*"
@@ -107,10 +131,12 @@ lispifyResponse (Resp_DisplayInfo info) = return $ case info of
                                  ]
                              ]
     Info_Intro s -> f (render s) "*Intro*"
-    Info_Version -> f ("Agda version " ++ version) "*Agda Version*"
+    Info_Version -> f ("Agda version " ++ versionWithCommitInfo) "*Agda Version*"
   where f content bufname = [ display_info' False bufname content ]
 lispifyResponse Resp_ClearHighlighting = return [ L [ A "agda2-highlight-clear" ] ]
 lispifyResponse Resp_ClearRunningInfo = return [ clearRunningInfo ]
+-- FNF: if Info_Warning comes back into use, the above should be
+-- return [ clearRunningInfo, clearWarning ]
 lispifyResponse (Resp_RunningInfo n s)
   | n <= 1    = return [ displayRunningInfo s ]
   | otherwise = return [ L [A "agda2-verbose", A (quote s)] ]
