@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP              #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE RelaxedPolyRec   #-}
 
 -- | Generates data used for precise syntax highlighting.
 
@@ -16,7 +14,6 @@ module Agda.Interaction.Highlighting.Generate
   , computeUnsolvedMetaWarnings
   , computeUnsolvedConstraints
   , storeDisambiguatedName
-  , Agda.Interaction.Highlighting.Generate.tests
   ) where
 
 import Prelude hiding (null)
@@ -37,19 +34,21 @@ import qualified Data.IntMap as IntMap
 import Data.Void
 
 import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
-import Agda.Interaction.Highlighting.Precise hiding (tests)
-import Agda.Interaction.Highlighting.Range   hiding (tests)
+import Agda.Interaction.Highlighting.Precise
+import Agda.Interaction.Highlighting.Range
 
 import qualified Agda.TypeChecking.Errors as E
 import Agda.TypeChecking.MetaVars (isBlockedTerm)
 import Agda.TypeChecking.Monad
   hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import qualified Agda.TypeChecking.Monad as M
+import Agda.TypeChecking.Positivity.Occurrence
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment'(..))
 import qualified Agda.Syntax.Common as Common
 import qualified Agda.Syntax.Concrete as C
+import Agda.Syntax.Fixity
 import qualified Agda.Syntax.Info as SI
 import qualified Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Literal as L
@@ -58,13 +57,13 @@ import qualified Agda.Syntax.Parser.Tokens as T
 import qualified Agda.Syntax.Position as P
 
 import Agda.Utils.FileName
+import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
-import Agda.Utils.TestHelpers
 import Agda.Utils.HashMap (HashMap)
 import qualified Agda.Utils.HashMap as HMap
 
@@ -167,6 +166,8 @@ generateAndPrintSyntaxInfo decl hlLevel = do
       Full{} -> generateConstructorInfo modMap file kinds decl
       _      -> return mempty
 
+    warnInfo <- Fold.fold . map (warningHighlighting . tcWarning) <$> use stTCWarnings
+
     let (from, to) = case P.rangeToInterval (P.getRange decl) of
           Nothing -> __IMPOSSIBLE__
           Just i  -> ( fromIntegral $ P.posPos $ P.iStart i
@@ -183,6 +184,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     let syntaxInfo = compress (mconcat [ constructorInfo
                                        , theRest modMap file
                                        , nameInfo
+                                       , warnInfo
                                        ])
                        `mappend`
                      curTokens
@@ -223,7 +225,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     , Fold.foldMap getNamedArg    $ universeBi decl
     ]
     where
-    bound n = nameToFile modMap file [] (A.nameConcrete n)
+    bound n = nameToFile modMap file [] (A.nameConcrete n) P.noRange
                          (\isOp -> mempty { aspect = Just $ Name (Just Bound) isOp })
                          (Just $ A.nameBindingSite n)
 
@@ -233,17 +235,17 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     macro n = nameToFileA modMap file n True $ \isOp ->
                   mempty { aspect = Just $ Name (Just Macro) isOp }
 
-    field m n = nameToFile modMap file m n
+    field m n = nameToFile modMap file m n P.noRange
                            (\isOp -> mempty { aspect = Just $ Name (Just Field) isOp })
                            Nothing
     asName n = nameToFile modMap file []
-                          n
+                          n P.noRange
                           (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
                           Nothing
 
     mod isTopLevelModule n =
       nameToFile modMap file []
-                 (A.nameConcrete n)
+                 (A.nameConcrete n) P.noRange
                  (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
                  (Just $ (if isTopLevelModule then P.beginningOfFile else id)
                            (A.nameBindingSite n))
@@ -281,7 +283,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     getPattern' :: A.Pattern' e -> File
     getPattern' (A.VarP x)    = bound x
     getPattern' (A.AsP _ x _) = bound x
-    getPattern' (A.DotP pi _) =
+    getPattern' (A.DotP pi _ _) =
       singleton (rToR $ P.getRange pi)
                 (mempty { otherAspects = [DottedPattern] })
     getPattern' (A.PatternSynP _ q _) = patsyn q
@@ -336,13 +338,13 @@ generateTokenInfo
   :: AbsolutePath          -- ^ The module to highlight.
   -> TCM CompressedFile
 generateTokenInfo file =
-  liftIO $ tokenHighlighting <$> Pa.parseFile' Pa.tokensParser file
+  runPM $ tokenHighlighting <$> Pa.parseFile' Pa.tokensParser file
 
 -- | Same as 'generateTokenInfo' but takes a string instead of a filename.
 generateTokenInfoFromString :: P.Range -> String -> TCM CompressedFile
 generateTokenInfoFromString r _ | r == P.noRange = return mempty
 generateTokenInfoFromString r s = do
-  liftIO $ tokenHighlighting <$> Pa.parsePosString Pa.tokensParser p s
+  runPM $ tokenHighlighting <$> Pa.parsePosString Pa.tokensParser p s
   where
     Just p = P.rStart r
 
@@ -422,10 +424,11 @@ nameKinds hlLevel decl = do
   defnToKind   M.Record{}                          = Record
   defnToKind   M.Constructor{ M.conInd = i }       = Constructor i
   defnToKind   M.Primitive{}                       = Primitive
+  defnToKind   M.AbstractDefn                      = __IMPOSSIBLE__
 
   declToKind :: A.Declaration ->
                 HashMap A.QName NameKind -> HashMap A.QName NameKind
-  declToKind (A.Axiom _ i _ q _)
+  declToKind (A.Axiom _ i _ _ q _)
     | SI.defMacro i == Common.MacroDef = insert q Macro
     | otherwise                        = insert q Postulate
   declToKind (A.Field _ q _)        = insert q Field -- Function
@@ -506,7 +509,7 @@ printErrorInfo e = printHighlightingInfo . compress =<< errorHighlighting e
 
 errorHighlighting :: TCErr -> TCM File
 
-errorHighlighting (TypeError s cl@(Closure sig env scope (TerminationCheckFailed termErrs))) =
+errorHighlighting (TypeError s cl@Closure{ clValue = TerminationCheckFailed termErrs }) =
   -- For termination errors, we keep the previous highlighting,
   -- just additionally mark the bad calls.
   return $ terminationErrorHighlighting termErrs
@@ -525,6 +528,15 @@ errorHighlighting e = do
                   }
   return $ mconcat [ erase, error ]
 
+-- | Generate syntax highlighting for warnings.
+
+warningHighlighting :: Warning -> File
+warningHighlighting w = case w of
+  TerminationIssue terrs    -> terminationErrorHighlighting terrs
+  NotStrictlyPositive d ocs -> positivityErrorHighlighting d ocs
+  _ -> mempty
+
+
 -- | Generate syntax highlighting for termination errors.
 
 terminationErrorHighlighting :: [TerminationError] -> File
@@ -536,6 +548,17 @@ terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
     callSites    = Fold.foldMap (\r -> singleton (rToR r) m) $
                    concatMap (map M.callInfoRange . M.termErrCalls) termErrs
 
+-- | Generate syntax highlighting for not-strictly-positive inductive
+-- definitions.
+
+-- TODO: highlight also the problematic occurrences
+positivityErrorHighlighting :: I.QName -> OccursWhere -> File
+positivityErrorHighlighting q o = several (rToR <$> P.getRange q : rs) m
+  where
+    rs = case o of Unknown -> []; Known r _ -> [r]
+    m  = mempty { otherAspects = [PositivityProblem] }
+
+
 -- | Generates and prints syntax highlighting information for unsolved
 -- meta-variables and certain unsolved constraints.
 
@@ -614,6 +637,8 @@ nameToFile :: SourceToModule
               -- ^ The name qualifier (may be empty).
            -> C.Name
               -- ^ The base name.
+           -> P.Range
+              -- ^ The 'Range' of the name in its fixity declaration (if any).
            -> (Bool -> Aspects)
               -- ^ Meta information to be associated with the name.
               -- The argument is 'True' iff the name is an operator.
@@ -622,7 +647,7 @@ nameToFile :: SourceToModule
               -- meta information is extended with this information,
               -- if possible.
            -> File
-nameToFile modMap file xs x m mR =
+nameToFile modMap file xs x fr m mR =
   -- We don't care if we get any funny ranges.
   if all (== Strict.Just file) fileNames then
     several (map rToR rs)
@@ -631,7 +656,7 @@ nameToFile modMap file xs x m mR =
     mempty
   where
   fileNames  = catMaybes $ map (fmap P.srcFile . P.rStart . P.getRange) (x : xs)
-  rs         = map P.getRange (x : xs)
+  rs         = applyWhen (not $ null fr) (fr :) $ map P.getRange (x : xs)
   mFilePos   = do
     r <- mR
     P.Pn { P.srcFile = Strict.Just f, P.posPos = p } <- P.rStart r
@@ -659,8 +684,16 @@ nameToFileA modMap file x include m =
              file
              (concreteQualifier x)
              (concreteBase x)
+             r
              m
              (if include then Just $ bindingSite x else Nothing)
+  where
+    -- Andreas, 2016-09-08, for issue #2140:
+    -- Range of name from fixity declaration:
+    fr = theNameRange $ A.nameFixity $ A.qnameName x
+    -- Somehow we import fixity ranges from other files, we should ignore them.
+    -- (I do not understand how we get them as they should not be serialized...)
+    r = if P.rangeFile fr == Strict.Just file then fr else P.noRange
 
 concreteBase :: I.QName -> C.Name
 concreteBase = A.nameConcrete . A.qnameName
@@ -678,11 +711,3 @@ storeDisambiguatedName q = whenJust (start $ P.getRange q) $ \ i ->
   stDisambiguatedNames %= IntMap.insert i q
   where
   start r = fromIntegral . P.posPos <$> P.rStart' r
-
-------------------------------------------------------------------------
--- All tests
-
--- | All the properties.
-
-tests :: IO Bool
-tests = runTests "Agda.Interaction.Highlighting.Generate" []
