@@ -1,12 +1,5 @@
 {-# LANGUAGE CPP                   #-}
-{-# LANGUAGE DeriveFoldable        #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE DeriveTraversable     #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE Rank2Types            #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TupleSections         #-}
 
 {-# OPTIONS_GHC -fno-cse #-}
 
@@ -49,6 +42,7 @@ import Agda.Syntax.Fixity
 import Agda.Syntax.Position
 import Agda.Syntax.Parser
 import Agda.Syntax.Common
+import Agda.Syntax.Literal
 import Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Generic as C
 import Agda.Syntax.Concrete.Pretty ()
@@ -67,13 +61,14 @@ import Agda.Interaction.SearchAbout
 import Agda.Interaction.Response hiding (Function, ExtendedLambda)
 import qualified Agda.Interaction.Response as R
 import qualified Agda.Interaction.BasicOps as B
+import Agda.Interaction.BasicOps hiding (whyInScope)
 import Agda.Interaction.Highlighting.Precise hiding (Postulate)
 import qualified Agda.Interaction.Imports as Imp
 import Agda.Interaction.Highlighting.Generate
+import qualified Agda.Interaction.Highlighting.LaTeX as LaTeX
 import qualified Agda.Interaction.Highlighting.Range as H
 
 import Agda.Compiler.Common (IsMain (..))
-import qualified Agda.Compiler.Epic.Compiler as Epic
 import qualified Agda.Compiler.MAlonzo.Compiler as MAlonzo
 import qualified Agda.Compiler.JS.Compiler as JS
 
@@ -87,6 +82,7 @@ import Agda.Utils.Except
   )
 
 import Agda.Utils.FileName
+import Agda.Utils.Function
 import Agda.Utils.Hash
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Lens
@@ -144,6 +140,22 @@ initCommandState = CommandState
 
 type CommandM = StateT CommandState TCM
 
+-- | Restore both 'TCState' and 'CommandState'.
+
+localStateCommandM :: CommandM a -> CommandM a
+localStateCommandM m = do
+  cSt <- get
+  tcSt <- lift $ get
+  x <- m
+  lift $ put tcSt
+  put cSt
+  return x
+
+-- | Restore 'TCState', do not touch 'CommandState'.
+
+liftLocalState :: TCM a -> CommandM a
+liftLocalState = lift . localState
+
 -- | Build an opposite action to 'lift' for state monads.
 
 revLift
@@ -168,6 +180,11 @@ commandMToIO ci_i = revLift runStateT lift $ \ct -> revLift runSafeTCM liftIO $
 liftCommandMT :: (forall a . TCM a -> TCM a) -> CommandM a -> CommandM a
 liftCommandMT f m = revLift runStateT lift $ f . ($ m)
 
+-- | Ditto, but restore state.
+
+liftCommandMTLocalState :: (forall a . TCM a -> TCM a) -> CommandM a -> CommandM a
+liftCommandMTLocalState f = liftCommandMT f . localStateCommandM
+
 -- | Put a response by the callback function given by 'stInteractionOutputCallback'.
 
 putResponse :: Response -> CommandM ()
@@ -212,8 +229,24 @@ handleCommand_ = handleCommand id (return ())
 
 handleCommand :: (forall a. CommandM a -> CommandM a) -> CommandM () -> CommandM () -> CommandM ()
 handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
-    res <- (`catchErr` (return . Just)) $ Nothing <$ cmd
-    maybe (return ()) (\ e -> onFail >> handleErr e) res
+    tcSt <- lift get
+
+    -- -- Andreas, 2016-11-18 OLD CODE:
+    -- -- onFail and handleErr are executed in "new" command state (not TCState).
+    -- -- But it seems that if an exception is raised, it is identical to the old state,
+    -- -- see code for catchErr.
+    -- res <- (`catchErr` (return . Just)) $ Nothing <$ cmd
+    -- maybe (return ()) (\ e -> onFail >> handleErr e) res
+
+    -- Andreas, 2016-11-18 NEW CODE: execute onFail and handleErr in handler
+    -- which means (looking at catchErr) they run in state s rathern than s'.
+    -- Yet, it looks like s == s' in case the command failed.
+    cmd `catchErr` \ e -> do
+      onFail
+      handleErr e
+      -- Andreas, 2016-11-18, issue #2174
+      -- Reset TCState after error is handled, to get rid of metas created during failed command
+      lift $ put tcSt
 
   where
     -- Preserves state so we can do unsolved meta highlighting
@@ -237,16 +270,17 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
     -- error. Because this function may switch the focus to another file
     -- the status information is also updated.
     handleErr e = do
+        unsolvedNotOK <- lift $ not . optAllowUnsolved <$> pragmaOptions
         meta    <- lift $ computeUnsolvedMetaWarnings
         constr  <- lift $ computeUnsolvedConstraints
         err     <- lift $ errorHighlighting e
         modFile <- lift $ use stModuleToSource
-        let info = compress $ mconcat
+        let info = compress $ mconcat $
                      -- Errors take precedence over unsolved things.
-                     [err, meta, constr]
+                     err : if unsolvedNotOK then [meta, constr] else []
         s <- lift $ prettyError e
         x <- lift $ optShowImplicit <$> use stPragmaOptions
-        mapM_ putResponse $
+        unless (null s) $ mapM_ putResponse $
             [ Resp_DisplayInfo $ Info_Error s ] ++
             tellEmacsToJumpToError (getRange e) ++
             [ Resp_HighlightingInfo info modFile ] ++
@@ -306,6 +340,9 @@ data Interaction' range
     -- show those instead.
   | Cmd_metas
 
+    -- | Display all warnings.
+  | Cmd_warnings
+
     -- | Shows all the top-level names in the given module, along with
     -- their types. Uses the top-level scope.
   | Cmd_show_module_contents_toplevel
@@ -316,7 +353,10 @@ data Interaction' range
     -- identifiers in their type.
   | Cmd_search_about_toplevel B.Rewrite String
 
-  | Cmd_solveAll
+    -- | Solve (all goals / the goal at point) whose values are determined by
+    -- the constraints.
+  | Cmd_solveAll B.Rewrite
+  | Cmd_solveOne B.Rewrite InteractionId range String
 
     -- | Parse the given expression (as if it were defined at the
     -- top-level of the current module) and infer its type.
@@ -326,7 +366,7 @@ data Interaction' range
 
     -- | Parse and type check the given expression (as if it were defined
     -- at the top-level of the current module) and normalise it.
-  | Cmd_compute_toplevel Bool -- Ignore abstract?
+  | Cmd_compute_toplevel B.ComputeMode
                          String
 
     ------------------------------------------------------------------------
@@ -403,7 +443,7 @@ data Interaction' range
 
   | Cmd_make_case       InteractionId range String
 
-  | Cmd_compute         Bool -- Ignore abstract?
+  | Cmd_compute         B.ComputeMode
                         InteractionId range String
 
   | Cmd_why_in_scope    InteractionId range String
@@ -519,14 +559,15 @@ interpret (Cmd_load m argv) =
   cmd_load' m argv True $ \_ -> interpret Cmd_metas
 
 interpret (Cmd_compile b file argv) =
-  cmd_load' file argv False $ \(i, mw) -> do
+  cmd_load' file argv (b == LaTeX) $ \(i, mw) -> do
+    mw <- lift $ Imp.applyFlagsToMaybeWarnings RespectFlags mw
     case mw of
       Imp.NoWarnings -> do
         lift $ case b of
-          MAlonzo -> MAlonzo.compilerMain IsMain i
-          MAlonzoNoMain -> MAlonzo.compilerMain NotMain i
-          Epic    -> Epic.compilerMain i
-          JS      -> JS.compilerMain i
+          GHC       -> MAlonzo.compilerMain IsMain i
+          GHCNoMain -> MAlonzo.compilerMain NotMain i
+          JS        -> JS.compilerMain i
+          LaTeX     -> LaTeX.generateLaTeX i
         display_info $ Info_CompilationOk
       Imp.SomeWarnings w ->
         display_info $ Info_Error $ unlines
@@ -538,11 +579,17 @@ interpret Cmd_constraints =
     display_info . Info_Constraints . unlines . map show =<< lift B.getConstraints
 
 interpret Cmd_metas = do -- CL.showMetas []
-  ms <- lift $ showOpenMetas
-  -- If we do not have open metas, but open constaints, display those.
-  ifM (return (null ms) `and2M` do not . null <$> lift B.getConstraints)
-    {-then-} (interpret Cmd_constraints)
-    {-else-} (display_info $ Info_AllGoals $ unlines ms)
+  unsolvedNotOK <- lift $ not . optAllowUnsolved <$> pragmaOptions
+  ms <- lift showOpenMetas
+  (pwe, pwa) <- interpretWarnings
+  display_info $ Info_AllGoalsWarnings (unlines ms) pwa pwe
+
+interpret Cmd_warnings = do
+  -- Ulf, 2016-08-09: Warnings are now printed in the info buffer by Cmd_metas.
+  -- pws <- interpretWarnings
+  -- unless (null pwd) $ display_info $ Info_Warning pws
+  return ()
+
 
 interpret (Cmd_show_module_contents_toplevel norm s) =
   liftCommandMT B.atTopLevel $ showModuleContents norm noRange s
@@ -550,25 +597,27 @@ interpret (Cmd_show_module_contents_toplevel norm s) =
 interpret (Cmd_search_about_toplevel norm s) =
   liftCommandMT B.atTopLevel $ searchAbout norm noRange s
 
-interpret Cmd_solveAll = do
-  out <- lift $ mapM lowr =<< B.getSolvedInteractionPoints False -- only solve metas which have a proper instantiation, i.e., not another meta
-  putResponse $ Resp_SolveAll out
-  where
-      lowr (i, m, e) = do
-        mi <- getMetaInfo <$> lookupMeta m
-        e <- withMetaInfo mi $ lowerMeta <$> abstractToConcreteCtx TopCtx e
-        return (i, e)
+interpret (Cmd_solveAll norm)        = solveInstantiatedGoals norm Nothing
+interpret (Cmd_solveOne norm ii _ _) = solveInstantiatedGoals norm' (Just ii)
+  -- `solveOne` is called via `agda2-maybe-normalised` which does not use
+  -- AsIs < Simplified < Normalised but rather Simplified < Instantiated < Normalised
+  -- So we remap the Rewrite modifiers to match solveAll's behaviour.
+  -- NB: instantiate is called in getSolvedInteractionPoints no matter what.
+  where norm' = case norm of
+                  Simplified   -> AsIs
+                  Instantiated -> Simplified
+                  _            -> norm
 
 interpret (Cmd_infer_toplevel norm s) =
   parseAndDoAtToplevel (B.typeInCurrent norm) Info_InferredType s
 
-interpret (Cmd_compute_toplevel ignore s) =
-  parseAndDoAtToplevel (allowNonTerminatingReductions .
-                        if ignore then ignoreAbstractMode . c
-                                  else inConcreteMode . c)
-                       Info_NormalForm
-                       s
-  where c = B.evalInCurrent
+interpret (Cmd_compute_toplevel cmode s) =
+  parseAndDoAtToplevel' action Info_NormalForm $ computeWrapInput cmode s
+  where
+  action = allowNonTerminatingReductions
+         . (if computeIgnoreAbstract cmode then ignoreAbstractMode else inConcreteMode)
+         . (B.showComputed cmode <=< B.evalInCurrent)
+
 
 interpret (ShowImplicitArgs showImpl) = do
   opts <- lift commandLineOptions
@@ -667,7 +716,7 @@ interpret (Cmd_auto ii rng s) = do
       -- For highlighting, Resp_GiveAction needs to access
       -- the @oldInteractionScope@s of the interaction points solved by Auto.
       -- We dig them out from the state before Auto was invoked.
-      insertOldInteractionScope ii =<< lift (localState (put st >> getInteractionScope ii))
+      insertOldInteractionScope ii =<< liftLocalState (put st >> getInteractionScope ii)
       -- Andreas, 2014-07-07: NOT TRUE:
       -- -- Andreas, 2014-07-05: The following should be obsolete,
       -- -- as Auto has removed the interaction points already:
@@ -686,19 +735,19 @@ interpret (Cmd_auto ii rng s) = do
    Right (Right s) -> give_gen ii rng s Refine
 
 interpret (Cmd_context norm ii _ _) =
-  display_info . Info_Context =<< lift (prettyContext norm False ii)
+  display_info . Info_Context =<< liftLocalState (prettyContext norm False ii)
 
 interpret (Cmd_helper_function norm ii rng s) =
-  display_info . Info_HelperFunction =<< lift (cmd_helper_function norm ii rng s)
+  display_info . Info_HelperFunction =<< liftLocalState (cmd_helper_function norm ii rng s)
 
 interpret (Cmd_infer norm ii rng s) =
   display_info . Info_InferredType
-    =<< lift (B.withInteractionId ii
+    =<< liftLocalState (B.withInteractionId ii
           (prettyATop =<< B.typeInMeta ii norm =<< B.parseExprIn ii rng s))
 
 interpret (Cmd_goal_type norm ii _ _) =
   display_info . Info_CurrentGoal
-    =<< lift (B.withInteractionId ii $ prettyTypeOfMeta norm ii)
+    =<< liftLocalState (B.withInteractionId ii $ prettyTypeOfMeta norm ii)
 
 interpret (Cmd_goal_type_context norm ii rng s) =
   cmd_goal_type_context_and empty norm ii rng s
@@ -707,8 +756,8 @@ interpret (Cmd_goal_type_context_infer norm ii rng s) = do
   -- In case of the empty expression to type, don't fail with
   -- a stupid parse error, but just fall back to
   -- Cmd_goal_type_context.
-  have <- if all Char.isSpace s then return empty else do
-    typ <- lift $ B.withInteractionId ii $
+  have <- if all Char.isSpace s then return empty else liftLocalState $ do
+    typ <- B.withInteractionId ii $
       prettyATop =<< B.typeInMeta ii norm =<< B.parseExprIn ii rng s
     return $ text "Have:" <+> typ
   cmd_goal_type_context_and have norm ii rng s
@@ -723,11 +772,23 @@ interpret (Cmd_why_in_scope ii rng s) =
   liftCommandMT (B.withInteractionId ii) $ whyInScope s
 
 interpret (Cmd_make_case ii rng s) = do
-  (casectxt , cs) <- lift $ makeCase ii rng s
+  (f, casectxt, cs) <- lift $ makeCase ii rng s
   liftCommandMT (B.withInteractionId ii) $ do
     hidden <- lift $ showImplicitArguments
-    pcs <- lift $ mapM prettyA $ List.map (extlam_dropLLifted casectxt hidden) cs
-    putResponse $ Resp_MakeCase (makeCaseVariant casectxt) (List.map (extlam_dropName casectxt . render) pcs)
+    tel <- lift $ lookupSection (qnameModule f) -- don't shadow the names in this telescope
+    let cs'  :: [A.Clause] = List.map (extlam_dropLLifted casectxt hidden) cs
+    pcs      :: [Doc]     <- lift $ inTopContext $ addContext tel $ mapM prettyA cs'
+    let pcs' :: [String]   = List.map (extlam_dropName casectxt . render) pcs
+    lift $ reportSDoc "interaction.case" 60 $ TCP.vcat
+      [ TCP.text "InteractionTop.Cmd_make_case"
+      , TCP.nest 2 $ TCP.vcat
+        [ TCP.text "cs   = " TCP.<+> TCP.vcat (map prettyA cs)
+        , TCP.text "cs'  = " TCP.<+> TCP.vcat (map prettyA cs')
+        , TCP.text "pcs  = " TCP.<+> TCP.vcat (map return pcs)
+        , TCP.text "pcs' = " TCP.<+> TCP.vcat (map TCP.text pcs')
+        ]
+      ]
+    putResponse $ Resp_MakeCase (makeCaseVariant casectxt) pcs'
   where
     render = renderStyle (style { mode = OneLineMode })
 
@@ -750,31 +811,66 @@ interpret (Cmd_make_case ii rng s) = do
     -- Drops pattern added to extended lambda functions when lambda lifting them
     extlam_dropLLifted :: CaseContext -> Bool -> A.Clause -> A.Clause
     extlam_dropLLifted Nothing _ x = x
-    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info A.LHSProj{} ps) rhs decl catchall) = __IMPOSSIBLE__
-    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info (A.LHSHead name nps) ps) rhs decl catchall)
+    extlam_dropLLifted _ _ (A.Clause (A.LHS _ A.LHSProj{} _) _ _ _ _) = __IMPOSSIBLE__
+    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info (A.LHSHead name nps) ps) dots rhs decl catchall)
       = let n = if hidden then h + nh else nh
         in
-         (A.Clause (A.LHS info (A.LHSHead name (drop n nps)) ps) rhs decl catchall)
+         (A.Clause (A.LHS info (A.LHSHead name (drop n nps)) ps) dots rhs decl catchall)
+
+interpret (Cmd_compute cmode ii rng s) = display_info . Info_NormalForm =<< do
+  liftLocalState $ do
+    e <- B.parseExprIn ii rng $ computeWrapInput cmode s
+    B.withInteractionId ii $ do
+      showComputed cmode =<< do applyWhen (computeIgnoreAbstract cmode) ignoreAbstractMode $ B.evalInCurrent e
 
-interpret (Cmd_compute ignore ii rng s) = do
-  e <- lift $ B.parseExprIn ii rng s
-  d <- lift $ B.withInteractionId ii $ do
-         let c = B.evalInCurrent e
-         v <- if ignore then ignoreAbstractMode c else c
-         prettyATop v
-  display_info $ Info_NormalForm d
 
 interpret Cmd_show_version = display_info Info_Version
 
+-- | Show warnings
+interpretWarnings :: CommandM (String, String)
+interpretWarnings = do
+  mws <- lift $ Imp.getAllWarnings Imp.AllWarnings RespectFlags
+  case filter isNotMeta <$> mws of
+    Imp.SomeWarnings ws@(_:_) -> do
+      let (we, wa) = Imp.classifyWarnings ws
+      pwe <- lift $ prettyTCWarnings we
+      pwa <- lift $ prettyTCWarnings wa
+      return (pwe, pwa)
+    _ -> return ("", "")
+   where isNotMeta w = case tcWarning w of
+                         UnsolvedInteractionMetas{} -> False
+                         UnsolvedMetaVariables{}    -> False
+                         _                          -> True
+
+
+-- | Solved goals already instantiated internally
+-- The second argument potentially limits it to one specific goal.
+solveInstantiatedGoals :: B.Rewrite -> Maybe InteractionId -> CommandM ()
+solveInstantiatedGoals norm mii = do
+  -- Andreas, 2016-10-23 issue #2280: throw away meta elims.
+  out <- lift $ local (\ e -> e { envPrintMetasBare = True }) $ do
+    sip <- B.getSolvedInteractionPoints False norm
+           -- only solve metas which have a proper instantiation, i.e., not another meta
+    maybe id (\ ii -> filter ((ii ==) . fst)) mii <$> mapM prt sip
+  putResponse $ Resp_SolveAll out
+  where
+      prt (i, m, e) = do
+        mi <- getMetaInfo <$> lookupMeta m
+        e <- withMetaInfo mi $ abstractToConcreteCtx TopCtx e
+        return (i, e)
+
+
 -- | Print open metas nicely.
 showOpenMetas :: TCM [String]
 showOpenMetas = do
   ims <- B.typesOfVisibleMetas B.AsIs
   di <- forM ims $ \ i ->
-    B.withInteractionId (B.outputFormId $ B.OutputForm noRange 0 i) $
+    B.withInteractionId (B.outputFormId $ B.OutputForm noRange [] i) $
       showATop i
   -- Show unsolved implicit arguments simplified.
-  dh <- mapM showA' =<< B.typesOfHiddenMetas B.Simplified
+  unsolvedNotOK <- not . optAllowUnsolved <$> pragmaOptions
+  hms <- (guard unsolvedNotOK >>) <$> B.typesOfHiddenMetas B.Simplified
+  dh <- mapM showA' hms
   return $ di ++ dh
   where
     metaId (B.OfType i _) = i
@@ -822,8 +918,8 @@ cmd_load' file argv unsolvedOK cmd = do
     case z of
       Left err   -> lift $ typeError $ GenericError err
       Right opts -> do
-        lift $ TM.setCommandLineOptions' relativeTo $
-          mapPragmaOptions (\ o -> o { optAllowUnsolved = unsolvedOK }) opts
+        let update o = o { optAllowUnsolved = unsolvedOK && optAllowUnsolved o}
+        lift $ TM.setCommandLineOptions' relativeTo $ mapPragmaOptions update opts
         displayStatus
 
     -- Reset the state, preserving options and decoded modules. Note
@@ -863,10 +959,11 @@ withCurrentFile m = do
 
 -- | Available backends.
 
-data Backend = MAlonzo
-             | MAlonzoNoMain
-             | Epic | JS
-    deriving (Show, Read)
+data Backend = GHC
+             | GHCNoMain
+             | JS
+             | LaTeX
+    deriving (Show, Read, Eq)
 
 data GiveRefine = Give | Refine | Intro
   deriving (Eq, Show)
@@ -916,7 +1013,10 @@ give_gen ii rng s0 giveRefine = do
     modifyTheInteractionPoints $ replace ii iis
     -- print abstract expr
     ce        <- lift $ abstractToConcreteEnv (makeEnv scope) ae
-    lift $ reportSLn "interaction.give" 30 $ "ce = " ++ show ce
+    lift $ reportSLn "interaction.give" 30 $ unlines
+      [ "ce = " ++ show ce
+      , "scopePrecedence = " ++ show (scopePrecedence scope)
+      ]
     -- if the command was @Give@, use the literal user input;
     -- Andreas, 2014-01-15, see issue 1020:
     -- Refine could solve a goal by introducing the sole constructor
@@ -957,7 +1057,7 @@ highlightExpr e =
   where
     dummy = mkName_ (NameId 0 0) "dummy"
     info  = mkDefInfo (nameConcrete dummy) noFixity' PublicAccess ConcreteDef (getRange e)
-    decl  = A.Axiom NoFunSig info defaultArgInfo (qnameFromList [dummy]) e
+    decl  = A.Axiom NoFunSig info defaultArgInfo Nothing (qnameFromList [dummy]) e
 
 -- | Sorts interaction points based on their ranges.
 
@@ -987,8 +1087,8 @@ prettyContext norm rev ii = B.withInteractionId ii $ do
   ctx <- B.contextOfMeta ii norm
   es  <- mapM (prettyATop . B.ofExpr) ctx
   ns  <- mapM (showATop   . B.ofName) ctx
-  let shuffle = if rev then reverse else id
-  return $ align 10 $ filter (not . null. fst) $ shuffle $ zip ns (map (text ":" <+>) es)
+  return $ align 10 $ applyWhen rev reverse $
+    filter (not . null . fst) $ zip ns $ map (text ":" <+>) es
 
 -- | Create type of application of new helper function that would solve the goal.
 
@@ -997,33 +1097,39 @@ cmd_helper_function norm ii r s = B.withInteractionId ii $ inTopContext $
   prettyATop =<< B.metaHelperType norm ii r s
 
 -- | Displays the current goal, the given document, and the current
--- context.
+--   context.
+--
+--   Should not modify the state.
 
 cmd_goal_type_context_and :: Doc -> B.Rewrite -> InteractionId -> Range ->
                              String -> StateT CommandState (TCMT IO) ()
-cmd_goal_type_context_and doc norm ii _ _ = do
-  goal <- lift $ B.withInteractionId ii $ prettyTypeOfMeta norm ii
-  ctx  <- lift $ prettyContext norm True ii
-  display_info $ Info_GoalType
-                (text "Goal:" <+> goal $+$
-                 doc $+$
-                 text (replicate 60 '\x2014') $+$
-                 ctx)
+cmd_goal_type_context_and doc norm ii _ _ = display_info . Info_GoalType =<< do
+  lift $ do
+    goal <- B.withInteractionId ii $ prettyTypeOfMeta norm ii
+    ctx  <- prettyContext norm True ii
+    return $ vcat
+      [ text "Goal:" <+> goal
+      , doc
+      , text (replicate 60 '\x2014')
+      , ctx
+      ]
 
 -- | Shows all the top-level names in the given module, along with
 -- their types.
 
 showModuleContents :: B.Rewrite -> Range -> String -> CommandM ()
-showModuleContents norm rng s = do
-  (modules, types) <- lift $ B.moduleContents norm rng s
-  types' <- lift $ forM types $ \ (x, t) -> do
-     t <- TCP.prettyTCM t
-     return (show x, text ":" <+> t)
-  display_info $ Info_ModuleContents $
-    text "Modules" $$
-    nest 2 (vcat $ map (text . show) modules) $$
-    text "Names" $$
-    nest 2 (align 10 types')
+showModuleContents norm rng s = display_info . Info_ModuleContents =<< do
+  liftLocalState $ do
+    (modules, types) <- B.moduleContents norm rng s
+    types' <- forM types $ \ (x, t) -> do
+      t <- TCP.prettyTCM t
+      return (show x, text ":" <+> t)
+    return $ vcat
+      [ text "Modules"
+      , nest 2 $ vcat $ map (text . show) modules
+      , text "Names"
+      , nest 2 $ align 10 types'
+      ]
 
 -- | Shows all the top-level names in scope which mention all the given
 -- identifiers in their type.
@@ -1044,12 +1150,12 @@ searchAbout norm rg nm = do
 -- | Explain why something is in scope.
 
 whyInScope :: String -> CommandM ()
-whyInScope s = do
-  (v, xs, ms) <- lift $ B.whyInScope s
-  cwd <- do
-    Just (file, _) <- gets $ theCurrentFile
-    return $ takeDirectory $ filePath file
-  display_info . Info_WhyInScope =<< do lift $ explanation cwd v xs ms
+whyInScope s = display_info . Info_WhyInScope =<< do
+  Just (file, _) <- gets theCurrentFile
+  let cwd = takeDirectory $ filePath file
+  liftLocalState $ do
+    (v, xs, ms) <- B.whyInScope s
+    explanation cwd v xs ms
   where
     explanation _ Nothing [] [] = TCP.text (s ++ " is not in scope.")
     explanation cwd v xs ms = TCP.vcat
@@ -1129,6 +1235,8 @@ setCommandLineOpts opts = do
 
 
 -- | Computes some status information.
+--
+--   Does not change the state.
 
 status :: CommandM Status
 status = do
@@ -1145,7 +1253,7 @@ status = do
       case t == t' of
         False -> return False
         True  -> do
-          mm <- Map.lookup f <$> sourceToModule
+          mm <- lookupModuleFromSource f
           case mm of
             Nothing -> return False -- work-around for Issue1007
             Just m  -> not . miWarnings . fromMaybe __IMPOSSIBLE__ <$> getVisitedModule m
@@ -1154,15 +1262,17 @@ status = do
                   , sChecked               = checked
                   }
 
--- | Displays\/updates status information.
+-- | Displays or updates status information.
+--
+--   Does not change the state.
 
 displayStatus :: CommandM ()
 displayStatus =
   putResponse . Resp_Status  =<< status
 
 -- | @display_info@ does what @'display_info'' False@ does, but
--- additionally displays some status information (see 'status' and
--- 'displayStatus').
+--   additionally displays some status information (see 'status' and
+--   'displayStatus').
 
 display_info :: DisplayInfo -> CommandM ()
 display_info info = do
@@ -1179,43 +1289,28 @@ nameModifiers :: [String]
 nameModifiers = "" : "'" : "''" : [show i | i <-[3..]]
 
 
--- | Kill meta numbers and ranges from all metas (@?@ and @_@).
-lowerMeta :: (C.ExprLike a) => a -> a
-lowerMeta = C.mapExpr kill where
-  kill e =
-    case e of
-      C.QuestionMark{} -> preMeta
-      C.Underscore{}   -> preUscore
-      C.App{}          -> case appView e of
-        C.AppView (C.QuestionMark _ _) _ -> preMeta
-        C.AppView (C.Underscore   _ _) _ -> preUscore
-        _ -> e
-      C.Paren r q@(C.QuestionMark _ Nothing) -> q
-      _ -> e
-
-  preMeta   = C.QuestionMark noRange Nothing
-  preUscore = C.Underscore   noRange Nothing
-
-
 -- | Parses and scope checks an expression (using the \"inside scope\"
 -- as the scope), performs the given command with the expression as
 -- input, and displays the result.
 
-parseAndDoAtToplevel
-  :: (A.Expr -> TCM A.Expr)
+parseAndDoAtToplevel'
+  :: (A.Expr -> TCM Doc)
      -- ^ The command to perform.
   -> (Doc -> DisplayInfo)
      -- ^ The name to use for the buffer displaying the output.
   -> String
      -- ^ The expression to parse.
   -> CommandM ()
-parseAndDoAtToplevel cmd title s = do
-  e   <- liftIO $ parse exprParser s
-  (time, res) <-
+parseAndDoAtToplevel' cmd title s = do
+  (time, res) <- localStateCommandM $ do
+    e <- lift $ runPM $ parse exprParser s
     maybeTimed (lift $ B.atTopLevel $
-                prettyA =<< cmd =<< concreteToAbstract_ e)
+                cmd =<< concreteToAbstract_ e)
   display_info (title $ fromMaybe empty time $$ res)
 
+parseAndDoAtToplevel :: (A.Expr -> TCM A.Expr) -> (Doc -> DisplayInfo) -> String -> CommandM ()
+parseAndDoAtToplevel cmd = parseAndDoAtToplevel' (prettyA <=< cmd)
+
 maybeTimed :: CommandM a -> CommandM (Maybe Doc, a)
 maybeTimed work = do
   doTime <- lift $ hasVerbosity "profile.interactive" 10
