@@ -2,6 +2,7 @@
 
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
+    , IgnoreFlags(..)
     , PragmaOptions(..)
     , OptionsPragma
     , Flag, OptM, runOptM
@@ -14,12 +15,12 @@ module Agda.Interaction.Options
     , defaultInteractionOptions
     , defaultVerbosity
     , defaultCutOff
+    , defaultPragmaOptions
     , standardOptions_
     , unsafePragmaOptions
     , isLiterate
     , mapFlag
     , usage
-    , tests
     , defaultLibDir
     -- Reused by PandocAgda
     , inputFlag
@@ -42,7 +43,7 @@ import Data.List                ( isSuffixOf , intercalate )
 import System.Console.GetOpt    ( getOpt', usageInfo, ArgOrder(ReturnInOrder)
                                 , OptDescr(..), ArgDescr(..)
                                 )
-import System.Directory         ( doesDirectoryExist )
+import System.Directory         ( doesFileExist, doesDirectoryExist )
 
 import Text.EditDistance
 
@@ -56,13 +57,12 @@ import Agda.Utils.Except
   , runExceptT
   )
 
-import Agda.Utils.TestHelpers   ( runTests )
-import Agda.Utils.QuickCheck    ( quickCheck' )
 import Agda.Utils.FileName      ( absolute, AbsolutePath, filePath )
 import Agda.Utils.Monad         ( ifM, readM )
 import Agda.Utils.List          ( groupOn, wordsBy )
 import Agda.Utils.String        ( indent )
 import Agda.Utils.Trie          ( Trie )
+import Agda.Syntax.Parser.Literate ( literateExts )
 import qualified Agda.Utils.Trie as Trie
 
 import Agda.Version
@@ -71,12 +71,17 @@ import Paths_Agda ( getDataFileName )
 
 -- | This should probably go somewhere else.
 isLiterate :: FilePath -> Bool
-isLiterate file = ".lagda" `isSuffixOf` file
+isLiterate file = any (`isSuffixOf` file) literateExts
 
 -- OptDescr is a Functor --------------------------------------------------
 
 type Verbosity = Trie String Int
 
+-- ignore or respect the flags --allow-unsolved-metas,
+-- --no-termination-check, --no-positivity-check?
+data IgnoreFlags = IgnoreFlags | RespectFlags
+  deriving Eq
+
 data CommandLineOptions = Options
   { optProgramName      :: String
   , optInputFile        :: Maybe FilePath
@@ -86,11 +91,12 @@ data CommandLineOptions = Options
   , optOverrideLibrariesFile :: Maybe FilePath
   -- ^ Use this (if Just) instead of .agda/libraries
   , optDefaultLibs      :: Bool
-  -- ^ Use ~/.agda/defaults or look for .agda-lib file.
+  -- ^ Use ~/.agda/defaults
+  , optUseLibs          :: Bool
+  -- ^ look for .agda-lib files
   , optShowVersion      :: Bool
   , optShowHelp         :: Bool
   , optInteractive      :: Bool
-  , optRunTests         :: Bool
   , optGHCiInteraction  :: Bool
   , optCompileNoMain    :: Bool
   , optGhcCompile       :: Bool
@@ -150,6 +156,10 @@ data PragmaOptions = PragmaOptions
   , optExactSplit                :: Bool
   , optEta                       :: Bool
   , optRewriting                 :: Bool  -- ^ Can rewrite rules be added and used?
+  , optPostfixProjections        :: Bool
+      -- ^ Should system generated projections 'ProjSystem' be printed
+      --   postfix (True) or prefix (False).
+  , optInstanceSearchDepth       :: Int
   }
   deriving (Show,Eq)
 
@@ -180,10 +190,10 @@ defaultOptions = Options
   , optLibraries        = []
   , optOverrideLibrariesFile = Nothing
   , optDefaultLibs      = True
+  , optUseLibs          = True
   , optShowVersion      = False
   , optShowHelp         = False
   , optInteractive      = False
-  , optRunTests         = False
   , optGHCiInteraction  = False
   , optCompileNoMain    = False
   , optGhcCompile       = False
@@ -239,6 +249,8 @@ defaultPragmaOptions = PragmaOptions
   , optExactSplit                = False
   , optEta                       = True
   , optRewriting                 = False
+  , optPostfixProjections        = False
+  , optInstanceSearchDepth       = 500
   }
 
 -- | The default termination depth.
@@ -256,10 +268,6 @@ defaultLaTeXDir = "latex"
 defaultHTMLDir :: String
 defaultHTMLDir = "html"
 
-prop_defaultOptions :: IO Bool
-prop_defaultOptions =
-  either (const False) (const True) <$> runOptM (checkOpts defaultOptions)
-
 type OptM = ExceptT String IO
 
 runOptM :: OptM a -> IO (Either String a)
@@ -321,16 +329,6 @@ unsafePragmaOptions opts =
   [ "--rewriting"                                | optRewriting opts                 ] ++
   []
 
--- | The default pragma options should be considered safe.
-
-defaultPragmaOptionsSafe :: IO Bool
-defaultPragmaOptionsSafe
-    | null unsafe = return True
-    | otherwise   = do putStrLn $ "Following pragmas are default but not safe: "
-                                        ++ intercalate ", " unsafe
-                       return False
-  where unsafe = unsafePragmaOptions defaultPragmaOptions
-
 inputFlag :: FilePath -> Flag CommandLineOptions
 inputFlag f o =
     case optInputFile o of
@@ -373,9 +371,6 @@ showImplicitFlag o = return $ o { optShowImplicit = True }
 showIrrelevantFlag :: Flag PragmaOptions
 showIrrelevantFlag o = return $ o { optShowIrrelevant = True }
 
-runTestsFlag :: Flag CommandLineOptions
-runTestsFlag o = return $ o { optRunTests = True }
-
 ghciInteractionFlag :: Flag CommandLineOptions
 ghciInteractionFlag o = return $ o { optGHCiInteraction = True }
 
@@ -453,6 +448,14 @@ noExactSplitFlag o = return $ o { optExactSplit = False }
 rewritingFlag :: Flag PragmaOptions
 rewritingFlag o = return $ o { optRewriting = True }
 
+postfixProjectionsFlag :: Flag PragmaOptions
+postfixProjectionsFlag o = return $ o { optPostfixProjections = True }
+
+instanceDepthFlag :: String -> Flag PragmaOptions
+instanceDepthFlag s o = do
+  d <- integerArgument "--instance-search-depth" s
+  return $ o { optInstanceSearchDepth = d }
+
 interactiveFlag :: Flag CommandLineOptions
 interactiveFlag  o = return $ o { optInteractive    = True
                                 , optPragmaOptions  = (optPragmaOptions o)
@@ -508,9 +511,6 @@ uhcTraceLevelFlag i o = return $ o { optUHCTraceLevel = read i }
 uhcFlagsFlag :: String -> Flag CommandLineOptions
 uhcFlagsFlag s o = return $ o { optUHCFlags = optUHCFlags o ++ [s] }
 
-optimNoSmashing :: Flag CommandLineOptions
-optimNoSmashing o = return $ o {optOptimSmashing = False }
-
 htmlFlag :: Flag CommandLineOptions
 htmlFlag o = return $ o { optGenerateHTML = True }
 
@@ -530,11 +530,17 @@ libraryFlag :: String -> Flag CommandLineOptions
 libraryFlag s o = return $ o { optLibraries = optLibraries o ++ [s] }
 
 overrideLibrariesFileFlag :: String -> Flag CommandLineOptions
-overrideLibrariesFileFlag s o = return $ o { optOverrideLibrariesFile = Just s }
+overrideLibrariesFileFlag s o = do
+  ifM (liftIO $ doesFileExist s)
+    {-then-} (return $ o { optOverrideLibrariesFile = Just s })
+    {-else-} (throwError $ "Libraries file not found: " ++ s)
 
 noDefaultLibsFlag :: Flag CommandLineOptions
 noDefaultLibsFlag o = return $ o { optDefaultLibs = False }
 
+noLibsFlag :: Flag CommandLineOptions
+noLibsFlag o = return $ o { optUseLibs = False }
+
 verboseFlag :: String -> Flag PragmaOptions
 verboseFlag s o =
     do  (k,n) <- parseVerbose s
@@ -568,7 +574,7 @@ standardOptions =
     , Option []     ["interaction"] (NoArg ghciInteractionFlag)
                     "for use with the Emacs mode"
     , Option ['c']  ["compile", "ghc"] (NoArg compileGhcFlag)
-                    "compile program using the GHC backend (experimental)"
+                    "compile program using the GHC backend"
     , Option []     ["ghc-dont-call-ghc"] (NoArg ghcDontCallGhcFlag) "Don't call ghc, just write the GHC Haskell files."
     , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
                     "give the flag GHC-FLAG to GHC when compiling using the GHC backend"
@@ -588,7 +594,6 @@ standardOptions =
     , Option []     ["uhc-gen-trace"] (ReqArg uhcTraceLevelFlag "TRACE") "Add tracing code to generated executable."
     , Option []     ["uhc-flag"] (ReqArg uhcFlagsFlag "UHC-FLAG")
                     "give the flag UHC-FLAG to UHC when compiling using the UHC backend"
-    , Option []     ["no-smashing"] (NoArg optimNoSmashing) "Don't apply the smashing optimization."
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
                     ("directory for compiler output (default: the project root)")
 
@@ -597,8 +602,6 @@ standardOptions =
     --                "give the flag EPIC-FLAG to Epic when compiling using Epic"
                     "the Epic backend has been removed"
 
-    , Option []     ["test"] (NoArg runTestsFlag)
-                    "run internal test suite"
     , Option []     ["vim"] (NoArg vimFlag)
                     "generate Vim highlighting files"
     , Option []     ["latex"] (NoArg latexFlag)
@@ -623,6 +626,8 @@ standardOptions =
                     "use library LIB"
     , Option []     ["library-file"] (ReqArg overrideLibrariesFileFlag "FILE")
                     "use FILE instead of the standard libraries file"
+    , Option []     ["no-libraries"] (NoArg noLibsFlag)
+                    "don't use any library files"
     , Option []     ["no-default-libraries"] (NoArg noDefaultLibsFlag)
                     "don't use default libraries"
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
@@ -700,6 +705,10 @@ pragmaOptions =
                     "disable eta rules for records"
     , Option []     ["rewriting"] (NoArg rewritingFlag)
                     "enable declaration and use of REWRITE rules"
+    , Option []     ["postfix-projections"] (NoArg postfixProjectionsFlag)
+                    "make postfix projection notation the default"
+    , Option []     ["instance-search-depth"] (ReqArg instanceDepthFlag "N")
+                    "set instance search depth to N (default: 500)"
     ]
 
 -- | Used for printing usage info.
@@ -813,12 +822,3 @@ defaultLibDir = do
   ifM (doesDirectoryExist libdir)
       (return libdir)
       (error $ "The lib directory " ++ libdir ++ " does not exist")
-
-------------------------------------------------------------------------
--- All tests
-
-tests :: IO Bool
-tests = runTests "Agda.Interaction.Options"
-  [ prop_defaultOptions
-  , defaultPragmaOptionsSafe
-  ]
