@@ -1,10 +1,5 @@
 {-# LANGUAGE CPP                   #-}
 {-# LANGUAGE DeriveDataTypeable    #-}
-{-# LANGUAGE DeriveFoldable        #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE DeriveTraversable     #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE TemplateHaskell       #-}
 
 {-| The abstract syntax. This is what you get after desugaring and scope
@@ -16,11 +11,12 @@ module Agda.Syntax.Abstract
     , module Agda.Syntax.Abstract.Name
     ) where
 
-import Prelude hiding (foldl, foldr)
+import Prelude
 import Control.Arrow (first)
 import Control.Applicative
 
-import Data.Foldable as Fold
+import Data.Foldable (Foldable)
+import qualified Data.Foldable as Fold
 import Data.Map (Map)
 import Data.Maybe
 import Data.Sequence (Seq, (<|), (><))
@@ -29,6 +25,7 @@ import Data.Traversable
 import Data.Typeable (Typeable)
 import Data.Void
 
+import Agda.Syntax.Concrete.Name (NumHoles(..))
 import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Pretty ()
@@ -39,9 +36,14 @@ import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Abstract.Name as A (QNamed)
 import Agda.Syntax.Literal
 import Agda.Syntax.Scope.Base
+import qualified Agda.Syntax.Internal as I
 
+import Agda.TypeChecking.Positivity.Occurrence
+
+import Agda.Utils.Functor
 import Agda.Utils.Geniplate
 import Agda.Utils.Lens
+import Agda.Utils.Pretty
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -52,8 +54,8 @@ type Args = [NamedArg Expr]
 data Expr
   = Var  Name                          -- ^ Bound variable.
   | Def  QName                         -- ^ Constant: axiom, function, data or record type.
-  | Proj QName                         -- ^ Projection.
-  | Con  AmbiguousQName                -- ^ Constructor.
+  | Proj ProjOrigin AmbiguousQName     -- ^ Projection (overloaded).
+  | Con  AmbiguousQName                -- ^ Constructor (overloaded).
   | PatternSyn QName                   -- ^ Pattern synonym.
   | Macro QName                        -- ^ Macro.
   | Lit Literal                        -- ^ Literal.
@@ -66,6 +68,7 @@ data Expr
     --   'metaNumber' to 'Nothing' while keeping the 'InteractionId'.
   | Underscore   MetaInfo
     -- ^ Meta variable for hidden argument (must be inferred locally).
+  | Dot ExprInfo Expr                  -- ^ @.e@, for postfix projection.
   | App  ExprInfo Expr (NamedArg Expr) -- ^ Ordinary (binary) application.
   | WithApp ExprInfo Expr [Expr]       -- ^ With application.
   | Lam  ExprInfo LamBinding Expr      -- ^ @λ bs → e@.
@@ -106,13 +109,35 @@ data Axiom
 -- | Renaming (generic).
 type Ren a = [(a, a)]
 
+data ScopeCopyInfo = ScopeCopyInfo
+  { renModules :: Ren ModuleName
+  , renNames   :: Ren QName }
+  deriving (Eq, Show)
+
+initCopyInfo :: ScopeCopyInfo
+initCopyInfo = ScopeCopyInfo
+  { renModules = []
+  , renNames   = []
+  }
+
+instance Pretty ScopeCopyInfo where
+  pretty i = vcat [ prRen "renModules =" (renModules i)
+                  , prRen "renNames   =" (renNames i) ]
+    where
+      prRen s r = sep [ text s, nest 2 $ vcat (map pr r) ]
+      pr (x, y) = pretty x <+> text "->" <+> pretty y
+
 data Declaration
-  = Axiom      Axiom DefInfo ArgInfo QName Expr      -- ^ type signature (can be irrelevant and colored, but not hidden)
+  = Axiom      Axiom DefInfo ArgInfo (Maybe [Occurrence]) QName Expr
+    -- ^ Type signature (can be irrelevant, but not hidden).
+    --
+    -- The fourth argument contains an optional assignment of
+    -- polarities to arguments.
   | Field      DefInfo QName (Arg Expr)              -- ^ record field
   | Primitive  DefInfo QName Expr                    -- ^ primitive function
   | Mutual     MutualInfo [Declaration]              -- ^ a bunch of mutually recursive definitions
   | Section    ModuleInfo ModuleName [TypedBindings] [Declaration]
-  | Apply      ModuleInfo ModuleName ModuleApplication (Ren QName) (Ren ModuleName) ImportDirective
+  | Apply      ModuleInfo ModuleName ModuleApplication ScopeCopyInfo ImportDirective
     -- ^ The @ImportDirective@ is for highlighting purposes.
   | Import     ModuleInfo ModuleName ImportDirective
     -- ^ The @ImportDirective@ is for highlighting purposes.
@@ -139,7 +164,7 @@ class GetDefInfo a where
   getDefInfo :: a -> Maybe DefInfo
 
 instance GetDefInfo Declaration where
-  getDefInfo (Axiom _ i _ _ _)      = Just i
+  getDefInfo (Axiom _ i _ _ _ _)    = Just i
   getDefInfo (Field i _ _)          = Just i
   getDefInfo (Primitive i _ _)      = Just i
   getDefInfo (ScopedDecl _ (d:_))   = getDefInfo d
@@ -177,7 +202,6 @@ data Pragma
   | CompiledJSPragma QName String
   | CompiledUHCPragma QName String
   | CompiledDataUHCPragma QName String [String]
-  | NoSmashingPragma QName
   | StaticPragma QName
   | InlinePragma QName
   | DisplayPragma QName [NamedArg Pattern] Expr
@@ -189,8 +213,8 @@ data LetBinding
     -- ^ @LetBind info rel name type defn@
   | LetPatBind LetInfo Pattern Expr
     -- ^ Irrefutable pattern binding.
-  | LetApply ModuleInfo ModuleName ModuleApplication (Ren QName) (Ren ModuleName) ImportDirective
-    -- ^ @LetApply mi newM (oldM args) renaming moduleRenaming dir@.
+  | LetApply ModuleInfo ModuleName ModuleApplication ScopeCopyInfo ImportDirective
+    -- ^ @LetApply mi newM (oldM args) renamings dir@.
     -- The @ImportDirective@ is for highlighting purposes.
   | LetOpen ModuleInfo ModuleName ImportDirective
     -- ^ only for highlighting and abstractToConcrete
@@ -238,11 +262,20 @@ data TypedBinding
 
 type Telescope  = [TypedBindings]
 
+data NamedDotPattern = NamedDot Name I.Term I.Type
+  deriving (Typeable, Show)
+
+instance Eq NamedDotPattern where
+  _ == _ = True -- These are not relevant for caching purposes
+
 -- | We could throw away @where@ clauses at this point and translate them to
 --   @let@. It's not obvious how to remember that the @let@ was really a
 --   @where@ clause though, so for the time being we keep it here.
 data Clause' lhs = Clause
   { clauseLHS        :: lhs
+  , clauseNamedDots  :: [NamedDotPattern]
+      -- ^ Only in with-clauses where we inherit some already checked dot patterns from the parent.
+      --   These live in the context of the parent clause left-hand side.
   , clauseRHS        :: RHS
   , clauseWhereDecls :: [Declaration]
   , clauseCatchall   :: Bool
@@ -252,19 +285,38 @@ type Clause = Clause' LHS
 type SpineClause = Clause' SpineLHS
 
 data RHS
-  = RHS Expr
+  = RHS
+    { rhsExpr     :: Expr
+    , rhsConcrete :: Maybe C.Expr
+      -- ^ We store the original concrete expression in case
+      --   we have to reproduce it during interactive case splitting.
+      --   'Nothing' for internally generated rhss.
+    }
   | AbsurdRHS
   | WithRHS QName [Expr] [Clause]
       -- ^ The 'QName' is the name of the with function.
-  | RewriteRHS [(QName, Expr)] RHS [Declaration]
-      -- ^ The 'QName's are the names of the generated with functions.
-      --   One for each 'Expr'.
-      --   The RHS shouldn't be another @RewriteRHS@.
-  deriving (Typeable, Show, Eq)
+  | RewriteRHS
+    { rewriteExprs      :: [(QName, Expr)]
+      -- ^ The 'QName's are the names of the generated with functions,
+      --   one for each 'Expr'.
+    , rewriteRHS        :: RHS
+      -- ^ The RHS should not be another @RewriteRHS@.
+    , rewriteWhereDecls :: [Declaration]
+      -- ^ The where clauses are attached to the @RewriteRHS@ by
+      ---  the scope checker (instead of to the clause).
+    }
+  deriving (Typeable, Show)
+
+instance Eq RHS where
+  RHS e _          == RHS e' _            = e == e'
+  AbsurdRHS        == AbsurdRHS           = True
+  WithRHS a b c    == WithRHS a' b' c'    = and [ a == a', b == b', c == c' ]
+  RewriteRHS a b c == RewriteRHS a' b' c' = and [ a == a', b == b', c == c' ]
+  _                == _                   = False
 
 -- | The lhs of a clause in spine view (inside-out).
 --   Projection patterns are contained in @spLhsPats@,
---   represented as @DefP d []@.
+--   represented as @ProjP d@.
 data SpineLHS = SpineLHS
   { spLhsInfo     :: LHSInfo             -- ^ Range.
   , spLhsDefName  :: QName               -- ^ Name of function we are defining.
@@ -294,11 +346,8 @@ data LHSCore' e
              , lhsPats     :: [NamedArg (Pattern' e)]  -- ^ Applied to patterns @ps@.
              }
     -- | Projection
-  | LHSProj  { lhsDestructor :: QName
+  | LHSProj  { lhsDestructor :: AmbiguousQName
                -- ^ Record projection identifier.
-             , lhsPatsLeft   :: [NamedArg (Pattern' e)]
-               -- ^ Indices of the projection.
-               --   Currently none @[]@, since we do not have indexed records.
              , lhsFocus      :: NamedArg (LHSCore' e)
                -- ^ Main branch.
              , lhsPatsRight  :: [NamedArg (Pattern' e)]
@@ -331,29 +380,33 @@ instance LHSToSpine LHS SpineLHS where
 
 lhsCoreToSpine :: LHSCore' e -> A.QNamed [NamedArg (Pattern' e)]
 lhsCoreToSpine (LHSHead f ps) = QNamed f ps
-lhsCoreToSpine (LHSProj d ps1 h ps2) = (++ (p : ps2)) <$> lhsCoreToSpine (namedArg h)
-  where p = updateNamedArg (const $ DefP patNoRange d ps1) h
+lhsCoreToSpine (LHSProj d h ps) = (++ (p : ps)) <$> lhsCoreToSpine (namedArg h)
+  where p = updateNamedArg (const $ ProjP patNoRange ProjPrefix d) h
 
-spineToLhsCore :: QNamed [NamedArg (Pattern' e)] -> LHSCore' e
+spineToLhsCore :: IsProjP e => QNamed [NamedArg (Pattern' e)] -> LHSCore' e
 spineToLhsCore (QNamed f ps) = lhsCoreAddSpine (LHSHead f []) ps
 
 -- | Add applicative patterns (non-projection patterns) to the right.
-lhsCoreApp :: LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
-lhsCoreApp (LHSHead f ps)        ps' = LHSHead f $ ps ++ ps'
-lhsCoreApp (LHSProj d ps1 h ps2) ps' = LHSProj d ps1 h $ ps2 ++ ps'
+lhsCoreApp :: IsProjP e => LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
+lhsCoreApp (LHSHead f ps)   ps' = LHSHead f   $ ps ++ ps'
+lhsCoreApp (LHSProj d h ps) ps' = LHSProj d h $ ps ++ ps'
 
 -- | Add projection and applicative patterns to the right.
-lhsCoreAddSpine :: LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
+lhsCoreAddSpine :: IsProjP e => LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
 lhsCoreAddSpine core ps = case ps2 of
-    (Arg info (Named n (DefP i d ps0)) : ps2') ->
-       LHSProj d ps0 (Arg info $ Named n $ lhsCoreApp core ps1) []
-         `lhsCoreAddSpine` ps2'
     [] -> lhsCoreApp core ps
+    p@(Arg info (Named n (ProjP i o d))) : ps2' | let nh = numHoles d->
+      -- Andreas, 2016-06-13
+      -- If the projection was written prefix by the user
+      -- or it is fully applied an operator
+      -- we turn it to prefix projection form.
+      (if o == ProjPrefix || nh > 0 && nh <= 1 + length ps2' then
+        LHSProj d (Arg info $ Named n $ lhsCoreApp core ps1) []
+      else lhsCoreApp core $ ps1 ++ [p])
+        `lhsCoreAddSpine` ps2'
     _ -> __IMPOSSIBLE__
   where
-    (ps1, ps2) = break (isDefP . namedArg) ps
-    isDefP DefP{} = True
-    isDefP _      = False
+    (ps1, ps2) = break (isJust . isProjP) ps
 
 -- | Used for checking pattern linearity.
 lhsCoreAllPatterns :: LHSCore' e -> [Pattern' e]
@@ -363,15 +416,14 @@ lhsCoreAllPatterns = map namedArg . qnamed . lhsCoreToSpine
 lhsCoreToPattern :: LHSCore -> Pattern
 lhsCoreToPattern lc =
   case lc of
-    LHSHead f aps -> DefP noInfo f aps
-    LHSProj d aps1 lhscore aps2 -> DefP noInfo d $
-      aps1 ++ fmap (fmap lhsCoreToPattern) lhscore : aps2
+    LHSHead f aps -> DefP noInfo (AmbQ [f]) aps
+    LHSProj d lhscore aps -> DefP noInfo d $
+      fmap (fmap lhsCoreToPattern) lhscore : aps
   where noInfo = patNoRange -- TODO, preserve range!
 
 mapLHSHead :: (QName -> [NamedArg Pattern] -> LHSCore) -> LHSCore -> LHSCore
-mapLHSHead f (LHSHead x ps)        = f x ps
-mapLHSHead f (LHSProj d ps1 l ps2) =
-  LHSProj d ps1 (fmap (fmap (mapLHSHead f)) l) ps2
+mapLHSHead f (LHSHead x ps)   = f x ps
+mapLHSHead f (LHSProj d l ps) = LHSProj d (fmap (fmap (mapLHSHead f)) l) ps
 
 ---------------------------------------------------------------------------
 -- * Patterns
@@ -381,13 +433,17 @@ mapLHSHead f (LHSProj d ps1 l ps2) =
 data Pattern' e
   = VarP Name
   | ConP ConPatInfo AmbiguousQName [NamedArg (Pattern' e)]
-  | DefP PatInfo QName          [NamedArg (Pattern' e)]
-    -- ^ Defined pattern: function definition @f ps@ or destructor pattern @d p ps@.
+  | ProjP PatInfo ProjOrigin AmbiguousQName
+    -- ^ Destructor pattern @d@.
+  | DefP PatInfo AmbiguousQName [NamedArg (Pattern' e)]
+    -- ^ Defined pattern: function definition @f ps@.
+    --   It is also abused to convert destructor patterns into concrete syntax
+    --   thus, we put AmbiguousQName here as well.
   | WildP PatInfo
     -- ^ Underscore pattern entered by user.
     --   Or generated at type checking for implicit arguments.
   | AsP PatInfo Name (Pattern' e)
-  | DotP PatInfo e
+  | DotP PatInfo Origin e
   | AbsurdP PatInfo
   | LitP Literal
   | PatternSynP PatInfo QName [NamedArg (Pattern' e)]
@@ -397,38 +453,49 @@ data Pattern' e
 type Pattern  = Pattern' Expr
 type Patterns = [NamedArg Pattern]
 
--- | Check whether we are a projection pattern.
-class IsProjP a where
-  isProjP :: a -> Maybe QName
-
 instance IsProjP (Pattern' e) where
-  isProjP (DefP _ d []) = Just d
-  isProjP _             = Nothing
+  isProjP (ProjP _ o d) = Just (o, d)
+  isProjP _ = Nothing
 
-instance IsProjP a => IsProjP (Arg a) where
-  isProjP = isProjP . unArg
+instance IsProjP Expr where
+  isProjP (Proj o ds)      = Just (o, ds)
+  isProjP (ScopedExpr _ e) = isProjP e
+  isProjP _ = Nothing
 
-instance IsProjP a => IsProjP (Named n a) where
-  isProjP = isProjP . namedThing
+class MaybePostfixProjP a where
+  maybePostfixProjP :: a -> Maybe (ProjOrigin, AmbiguousQName)
+
+instance IsProjP e => MaybePostfixProjP (Pattern' e) where
+  maybePostfixProjP (DotP _ _ e)  = isProjP e <&> \ (_o, d) -> (ProjPostfix, d)
+  maybePostfixProjP (ProjP _ o d) = Just (o, d)
+  maybePostfixProjP _ = Nothing
+
+instance MaybePostfixProjP a => MaybePostfixProjP (Arg a) where
+  maybePostfixProjP = maybePostfixProjP . unArg
+
+instance MaybePostfixProjP a => MaybePostfixProjP (Named n a) where
+  maybePostfixProjP = maybePostfixProjP . namedThing
 
 {--------------------------------------------------------------------------
     Instances
  --------------------------------------------------------------------------}
 
 -- | Does not compare 'ScopeInfo' fields.
+--   Does not distinguish between prefix and postfix projections.
 
 instance Eq Expr where
   ScopedExpr _ a1         == ScopedExpr _ a2         = a1 == a2
 
   Var a1                  == Var a2                  = a1 == a2
   Def a1                  == Def a2                  = a1 == a2
-  Proj a1                 == Proj a2                 = a1 == a2
+  Proj _ a1               == Proj _ a2               = a1 == a2
   Con a1                  == Con a2                  = a1 == a2
   PatternSyn a1           == PatternSyn a2           = a1 == a2
   Macro a1                == Macro a2                = a1 == a2
   Lit a1                  == Lit a2                  = a1 == a2
   QuestionMark a1 b1      == QuestionMark a2 b2      = (a1, b1) == (a2, b2)
   Underscore a1           == Underscore a2           = a1 == a2
+  Dot r1 e1               == Dot r2 e2               = (r1, e1) == (r2, e2)
   App a1 b1 c1            == App a2 b2 c2            = (a1, b1, c1) == (a2, b2, c2)
   WithApp a1 b1 c1        == WithApp a2 b2 c2        = (a1, b1, c1) == (a2, b2, c2)
   Lam a1 b1 c1            == Lam a2 b2 c2            = (a1, b1, c1) == (a2, b2, c2)
@@ -457,12 +524,12 @@ instance Eq Expr where
 instance Eq Declaration where
   ScopedDecl _ a1                == ScopedDecl _ a2                = a1 == a2
 
-  Axiom a1 b1 c1 d1 e1           == Axiom a2 b2 c2 d2 e2           = (a1, b1, c1, d1, e1) == (a2, b2, c2, d2, e2)
+  Axiom a1 b1 c1 d1 e1 f1        == Axiom a2 b2 c2 d2 e2 f2        = (a1, b1, c1, d1, e1, f1) == (a2, b2, c2, d2, e2, f2)
   Field a1 b1 c1                 == Field a2 b2 c2                 = (a1, b1, c1) == (a2, b2, c2)
   Primitive a1 b1 c1             == Primitive a2 b2 c2             = (a1, b1, c1) == (a2, b2, c2)
   Mutual a1 b1                   == Mutual a2 b2                   = (a1, b1) == (a2, b2)
   Section a1 b1 c1 d1            == Section a2 b2 c2 d2            = (a1, b1, c1, d1) == (a2, b2, c2, d2)
-  Apply a1 b1 c1 d1 e1 f1        == Apply a2 b2 c2 d2 e2 f2        = (a1, b1, c1, d1, e1, f1) == (a2, b2, c2, d2, e2, f2)
+  Apply a1 b1 c1 d1 e1           == Apply a2 b2 c2 d2 e2           = (a1, b1, c1, d1, e1) == (a2, b2, c2, d2, e2)
   Import a1 b1 c1                == Import a2 b2 c2                = (a1, b1, c1) == (a2, b2, c2)
   Pragma a1 b1                   == Pragma a2 b2                   = (a1, b1) == (a2, b2)
   Open a1 b1 c1                  == Open a2 b2 c2                  = (a1, b1, c1) == (a2, b2, c2)
@@ -505,11 +572,12 @@ instance HasRange TypedBinding where
 instance HasRange Expr where
     getRange (Var x)               = getRange x
     getRange (Def x)               = getRange x
-    getRange (Proj x)              = getRange x
+    getRange (Proj _ x)            = getRange x
     getRange (Con x)               = getRange x
     getRange (Lit l)               = getRange l
     getRange (QuestionMark i _)    = getRange i
     getRange (Underscore  i)       = getRange i
+    getRange (Dot i _)             = getRange i
     getRange (App i _ _)           = getRange i
     getRange (WithApp i _ _)       = getRange i
     getRange (Lam i _ _)           = getRange i
@@ -535,11 +603,11 @@ instance HasRange Expr where
     getRange (Macro x)             = getRange x
 
 instance HasRange Declaration where
-    getRange (Axiom    _ i _ _ _    ) = getRange i
+    getRange (Axiom    _ i _ _ _ _  ) = getRange i
     getRange (Field      i _ _      ) = getRange i
     getRange (Mutual     i _        ) = getRange i
     getRange (Section    i _ _ _    ) = getRange i
-    getRange (Apply      i _ _ _ _ _) = getRange i
+    getRange (Apply      i _ _ _ _)   = getRange i
     getRange (Import     i _ _      ) = getRange i
     getRange (Primitive  i _ _      ) = getRange i
     getRange (Pragma     i _        ) = getRange i
@@ -557,10 +625,11 @@ instance HasRange Declaration where
 instance HasRange (Pattern' e) where
     getRange (VarP x)            = getRange x
     getRange (ConP i _ _)        = getRange i
+    getRange (ProjP i _ _)       = getRange i
     getRange (DefP i _ _)        = getRange i
     getRange (WildP i)           = getRange i
     getRange (AsP i _ _)         = getRange i
-    getRange (DotP i _)          = getRange i
+    getRange (DotP i _ _)        = getRange i
     getRange (AbsurdP i)         = getRange i
     getRange (LitP l)            = getRange l
     getRange (PatternSynP i _ _) = getRange i
@@ -574,21 +643,21 @@ instance HasRange LHS where
 
 instance HasRange (LHSCore' e) where
     getRange (LHSHead f ps) = fuseRange f ps
-    getRange (LHSProj d ps1 lhscore ps2) = d `fuseRange` ps1 `fuseRange` lhscore `fuseRange` ps2
+    getRange (LHSProj d lhscore ps) = d `fuseRange` lhscore `fuseRange` ps
 
 instance HasRange a => HasRange (Clause' a) where
-    getRange (Clause lhs rhs ds catchall) = getRange (lhs,rhs,ds)
+    getRange (Clause lhs _ rhs ds catchall) = getRange (lhs,rhs,ds)
 
 instance HasRange RHS where
     getRange AbsurdRHS                = noRange
-    getRange (RHS e)                  = getRange e
+    getRange (RHS e _)                = getRange e
     getRange (WithRHS _ e cs)         = fuseRange e cs
     getRange (RewriteRHS xes rhs wh)  = getRange (map snd xes, rhs, wh)
 
 instance HasRange LetBinding where
     getRange (LetBind  i _ _ _ _     ) = getRange i
     getRange (LetPatBind  i _ _      ) = getRange i
-    getRange (LetApply i _ _ _ _ _   ) = getRange i
+    getRange (LetApply i _ _ _ _     ) = getRange i
     getRange (LetOpen  i _ _         ) = getRange i
     getRange (LetDeclaredVariable x)   = getRange x
 
@@ -596,10 +665,11 @@ instance HasRange LetBinding where
 instance SetRange (Pattern' a) where
     setRange r (VarP x)             = VarP (setRange r x)
     setRange r (ConP i ns as)       = ConP (setRange r i) ns as
-    setRange r (DefP _ n as)        = DefP (PatRange r) (setRange r n) as
+    setRange r (ProjP _ o ns)       = ProjP (PatRange r) o ns
+    setRange r (DefP _ ns as)       = DefP (PatRange r) ns as -- (setRange r n) as
     setRange r (WildP _)            = WildP (PatRange r)
     setRange r (AsP _ n p)          = AsP (PatRange r) (setRange r n) p
-    setRange r (DotP _ e)           = DotP (PatRange r) e
+    setRange r (DotP _ o e)         = DotP (PatRange r) o e
     setRange r (AbsurdP _)          = AbsurdP (PatRange r)
     setRange r (LitP l)             = LitP (setRange r l)
     setRange r (PatternSynP _ n as) = PatternSynP (PatRange r) (setRange r n) as
@@ -619,11 +689,12 @@ instance KillRange TypedBinding where
 instance KillRange Expr where
   killRange (Var x)                = killRange1 Var x
   killRange (Def x)                = killRange1 Def x
-  killRange (Proj x)               = killRange1 Proj x
+  killRange (Proj o x)             = killRange1 (Proj o) x
   killRange (Con x)                = killRange1 Con x
   killRange (Lit l)                = killRange1 Lit l
   killRange (QuestionMark i ii)    = killRange2 QuestionMark i ii
   killRange (Underscore  i)        = killRange1 Underscore i
+  killRange (Dot i e)              = killRange2 Dot i e
   killRange (App i e1 e2)          = killRange3 App i e1 e2
   killRange (WithApp i e es)       = killRange3 WithApp i e es
   killRange (Lam i b e)            = killRange3 Lam i b e
@@ -649,13 +720,11 @@ instance KillRange Expr where
   killRange (Macro x)              = killRange1 Macro x
 
 instance KillRange Declaration where
-  killRange (Axiom    p i rel a b     ) = killRange4 (Axiom p)  i rel a b
+  killRange (Axiom    p i a b c d     ) = killRange4 (\i a c d -> Axiom p i a b c d) i a c d
   killRange (Field      i a b         ) = killRange3 Field      i a b
   killRange (Mutual     i a           ) = killRange2 Mutual     i a
   killRange (Section    i a b c       ) = killRange4 Section    i a b c
-  killRange (Apply      i a b c d e   ) = killRange3 Apply      i a b c d (killRange e)
-   -- the arguments c and d of Apply are name maps, so nothing to kill
-   -- Andreas, 2016-01-24 really?
+  killRange (Apply      i a b c d     ) = killRange5 Apply      i a b c d
   killRange (Import     i a b         ) = killRange3 Import     i a b
   killRange (Primitive  i a b         ) = killRange3 Primitive  i a b
   killRange (Pragma     i a           ) = Pragma (killRange i) a
@@ -674,13 +743,17 @@ instance KillRange ModuleApplication where
   killRange (SectionApp a b c  ) = killRange3 SectionApp a b c
   killRange (RecordModuleIFS a ) = killRange1 RecordModuleIFS a
 
+instance KillRange ScopeCopyInfo where
+  killRange (ScopeCopyInfo a b) = killRange2 ScopeCopyInfo a b
+
 instance KillRange e => KillRange (Pattern' e) where
   killRange (VarP x)            = killRange1 VarP x
   killRange (ConP i a b)        = killRange3 ConP i a b
+  killRange (ProjP i o a)       = killRange3 ProjP i o a
   killRange (DefP i a b)        = killRange3 DefP i a b
   killRange (WildP i)           = killRange1 WildP i
   killRange (AsP i a b)         = killRange3 AsP i a b
-  killRange (DotP i a)          = killRange2 DotP i a
+  killRange (DotP i o a)        = killRange3 DotP i o a
   killRange (AbsurdP i)         = killRange1 AbsurdP i
   killRange (LitP l)            = killRange1 LitP l
   killRange (PatternSynP i a p) = killRange3 PatternSynP i a p
@@ -693,22 +766,25 @@ instance KillRange LHS where
   killRange (LHS i a b)   = killRange3 LHS i a b
 
 instance KillRange e => KillRange (LHSCore' e) where
-  killRange (LHSHead a b)     = killRange2 LHSHead a b
-  killRange (LHSProj a b c d) = killRange4 LHSProj a b c d
+  killRange (LHSHead a b)   = killRange2 LHSHead a b
+  killRange (LHSProj a b c) = killRange3 LHSProj a b c
 
 instance KillRange a => KillRange (Clause' a) where
-  killRange (Clause lhs rhs ds catchall) = killRange4 Clause lhs rhs ds catchall
+  killRange (Clause lhs dots rhs ds catchall) = killRange5 Clause lhs dots rhs ds catchall
+
+instance KillRange NamedDotPattern where
+  killRange (NamedDot a b c) = killRange3 NamedDot a b c
 
 instance KillRange RHS where
   killRange AbsurdRHS                = AbsurdRHS
-  killRange (RHS e)                  = killRange1 RHS e
+  killRange (RHS e c)                = killRange2 RHS e c
   killRange (WithRHS q e cs)         = killRange3 WithRHS q e cs
   killRange (RewriteRHS xes rhs wh)  = killRange3 RewriteRHS xes rhs wh
 
 instance KillRange LetBinding where
   killRange (LetBind    i info a b c) = killRange5 LetBind  i info a b c
   killRange (LetPatBind i a b       ) = killRange3 LetPatBind i a b
-  killRange (LetApply   i a b c d e ) = killRange3 LetApply i a b c d (killRange e)
+  killRange (LetApply   i a b c d   ) = killRange5 LetApply i a b c d
   killRange (LetOpen    i x dir     ) = killRange3 LetOpen  i x dir
   killRange (LetDeclaredVariable x)   = killRange1 LetDeclaredVariable x
 
@@ -760,7 +836,7 @@ instance AllNames QName where
   allNames q = Seq.singleton q
 
 instance AllNames Declaration where
-  allNames (Axiom   _ _ _ q _)        = Seq.singleton q
+  allNames (Axiom   _ _ _ _ q _)      = Seq.singleton q
   allNames (Field     _   q _)        = Seq.singleton q
   allNames (Primitive _   q _)        = Seq.singleton q
   allNames (Mutual     _ defs)        = allNames defs
@@ -780,10 +856,10 @@ instance AllNames Declaration where
   allNames (ScopedDecl _ decls)       = allNames decls
 
 instance AllNames Clause where
-  allNames (Clause _ rhs decls _) = allNames rhs >< allNames decls
+  allNames (Clause _ _ rhs decls _) = allNames rhs >< allNames decls
 
 instance AllNames RHS where
-  allNames (RHS e)                   = allNames e
+  allNames (RHS e _)                 = allNames e
   allNames AbsurdRHS{}               = Seq.empty
   allNames (WithRHS q _ cls)         = q <| allNames cls
   allNames (RewriteRHS qes rhs cls) = Seq.fromList (map fst qes) >< allNames rhs >< allNames cls
@@ -796,6 +872,7 @@ instance AllNames Expr where
   allNames Lit{}                   = Seq.empty
   allNames QuestionMark{}          = Seq.empty
   allNames Underscore{}            = Seq.empty
+  allNames (Dot _ e)               = allNames e
   allNames (App _ e1 e2)           = allNames e1 >< allNames e2
   allNames (WithApp _ e es)        = allNames e >< allNames es
   allNames (Lam _ b e)             = allNames b >< allNames e
@@ -832,11 +909,11 @@ instance AllNames TypedBinding where
   allNames (TLet _ lbs)  = allNames lbs
 
 instance AllNames LetBinding where
-  allNames (LetBind _ _ _ e1 e2)    = allNames e1 >< allNames e2
-  allNames (LetPatBind _ _ e)       = allNames e
-  allNames (LetApply _ _ app _ _ _) = allNames app
-  allNames LetOpen{}                = Seq.empty
-  allNames (LetDeclaredVariable _)  = Seq.empty
+  allNames (LetBind _ _ _ e1 e2)   = allNames e1 >< allNames e2
+  allNames (LetPatBind _ _ e)      = allNames e
+  allNames (LetApply _ _ app _ _)  = allNames app
+  allNames LetOpen{}               = Seq.empty
+  allNames (LetDeclaredVariable _) = Seq.empty
 
 instance AllNames ModuleApplication where
   allNames (SectionApp bindss _ es) = allNames bindss >< allNames es
@@ -847,7 +924,7 @@ instance AllNames ModuleApplication where
 -- Precondition: The declaration has to be a (scoped) 'Axiom'.
 
 axiomName :: Declaration -> QName
-axiomName (Axiom _ _ _ q _)    = q
+axiomName (Axiom _ _ _ _ q _)  = q
 axiomName (ScopedDecl _ (d:_)) = axiomName d
 axiomName _                    = __IMPOSSIBLE__
 
@@ -861,7 +938,7 @@ instance AnyAbstract a => AnyAbstract [a] where
   anyAbstract = Fold.any anyAbstract
 
 instance AnyAbstract Declaration where
-  anyAbstract (Axiom _ i _ _ _)      = defAbstract i == AbstractDef
+  anyAbstract (Axiom _ i _ _ _ _)    = defAbstract i == AbstractDef
   anyAbstract (Field i _ _)          = defAbstract i == AbstractDef
   anyAbstract (Mutual     _ ds)      = anyAbstract ds
   anyAbstract (ScopedDecl _ ds)      = anyAbstract ds
@@ -878,7 +955,7 @@ nameExpr :: AbstractName -> Expr
 nameExpr d = mk (anameKind d) $ anameName d
   where
     mk DefName        x = Def x
-    mk FldName        x = Proj x
+    mk FldName        x = Proj ProjSystem $ AmbQ [x]
     mk ConName        x = Con $ AmbQ [x]
     mk PatternSynName x = PatternSyn x
     mk MacroName      x = Macro x
@@ -896,11 +973,13 @@ patternToExpr :: Pattern -> Expr
 patternToExpr (VarP x)            = Var x
 patternToExpr (ConP _ c ps)       =
   Con c `app` map (fmap (fmap patternToExpr)) ps
-patternToExpr (DefP _ f ps)       =
+patternToExpr (ProjP _ o ds)      = Proj o ds
+patternToExpr (DefP _ (AmbQ [f]) ps) =
   Def f `app` map (fmap (fmap patternToExpr)) ps
+patternToExpr (DefP _ (AmbQ _) ps) = __IMPOSSIBLE__
 patternToExpr (WildP _)           = Underscore emptyMetaInfo
 patternToExpr (AsP _ _ p)         = patternToExpr p
-patternToExpr (DotP _ e)          = e
+patternToExpr (DotP _ _ e)        = e
 patternToExpr (AbsurdP _)         = Underscore emptyMetaInfo  -- TODO: could this happen?
 patternToExpr (LitP l)            = Lit l
 patternToExpr (PatternSynP _ _ _) = __IMPOSSIBLE__
@@ -919,12 +998,13 @@ substPattern s p = case p of
   VarP z        -> fromMaybe p (lookup z s)
   ConP i q ps   -> ConP i q (map (fmap (fmap (substPattern s))) ps)
   RecP i ps     -> RecP i (map (fmap (substPattern s)) ps)
+  ProjP{}       -> p
   WildP i       -> p
-  DotP i e      -> DotP i (substExpr (map (fmap patternToExpr) s) e)
+  DotP i o e    -> DotP i o (substExpr (map (fmap patternToExpr) s) e)
   AbsurdP i     -> p
   LitP l        -> p
   DefP{}        -> p              -- destructor pattern
-  AsP{}         -> __IMPOSSIBLE__ -- @-patterns (not supported anyways)
+  AsP i x p     -> AsP i x (substPattern s p) -- Note: cannot substitute into as-variable
   PatternSynP{} -> __IMPOSSIBLE__ -- pattern synonyms (already gone)
 
 class SubstExpr a where
@@ -964,6 +1044,7 @@ instance SubstExpr Expr where
     Lit _                 -> e
     QuestionMark{}        -> e
     Underscore   _        -> e
+    Dot i e               -> Dot i (substExpr s e)
     App  i e e'           -> App i (substExpr s e) (substExpr s e')
     WithApp i e es        -> WithApp i (substExpr s e) (substExpr s es)
     Lam  i lb e           -> Lam i lb (substExpr s e)
