@@ -1,12 +1,6 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE DeriveFoldable #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE DeriveTraversable #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE CPP                        #-}
+{-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE StandaloneDeriving #-}
 
 {-| Abstract names carry unique identifiers and stuff.
 -}
@@ -24,8 +18,6 @@ import Data.List
 import Data.Function
 import Data.Hashable
 
-import Test.QuickCheck
-
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import {-# SOURCE #-} Agda.Syntax.Fixity
@@ -33,6 +25,7 @@ import Agda.Syntax.Concrete.Name (IsNoName(..), NumHoles(..))
 import qualified Agda.Syntax.Concrete.Name as C
 
 import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Pretty
 import Agda.Utils.Size
@@ -81,7 +74,17 @@ newtype ModuleName = MName { mnameToList :: [Name] }
 -- Invariant: All the names in the list must have the same concrete,
 -- unqualified name.  (This implies that they all have the same 'Range').
 newtype AmbiguousQName = AmbQ { unAmbQ :: [QName] }
-  deriving (Eq, Typeable)
+  deriving (Eq, Ord, Typeable)
+
+-- | Check whether we are a projection pattern.
+class IsProjP a where
+  isProjP :: a -> Maybe (ProjOrigin, AmbiguousQName)
+
+instance IsProjP a => IsProjP (Arg a) where
+  isProjP = isProjP . unArg
+
+instance IsProjP a => IsProjP (Named n a) where
+  isProjP = isProjP . namedThing
 
 -- | A module is anonymous if the qualification path ends in an underscore.
 isAnonymousModuleName :: ModuleName -> Bool
@@ -262,6 +265,10 @@ instance NumHoles Name where
 instance NumHoles QName where
   numHoles = numHoles . qnameName
 
+-- | We can have an instance for ambiguous names as all share a common concrete name.
+instance NumHoles AmbiguousQName where
+  numHoles (AmbQ qs) = numHoles $ fromMaybe __IMPOSSIBLE__ $ headMaybe qs
+
 ------------------------------------------------------------------------
 -- * Show instances
 ------------------------------------------------------------------------
@@ -381,29 +388,6 @@ instance Sized QName where
 instance Sized ModuleName where
   size = size . mnameToList
 
-------------------------------------------------------------------------
--- * Arbitrary instances
-------------------------------------------------------------------------
-
--- | The generated names all have the same 'Fixity'': 'noFixity''.
-
-instance Arbitrary Name where
-  arbitrary =
-    Name <$> arbitrary <*> arbitrary <*> arbitrary
-         <*> return noFixity'
-
-instance CoArbitrary Name where
-  coarbitrary = coarbitrary . nameId
-
-instance Arbitrary QName where
-  arbitrary = do
-    ms <- arbitrary
-    n  <- arbitrary
-    return (QName (MName ms) n)
-
-instance CoArbitrary QName where
-  coarbitrary = coarbitrary . qnameName
-
 ------------------------------------------------------------------------
 -- * NFData instances
 ------------------------------------------------------------------------
