@@ -1,12 +1,7 @@
 {-# LANGUAGE CPP                       #-}
-{-# LANGUAGE DefaultSignatures         #-}
+{-# LANGUAGE GADTs                     #-}
+{-# LANGUAGE NoMonoLocalBinds          #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
-{-# LANGUAGE PatternGuards             #-}
-{-# LANGUAGE TupleSections             #-}
-
-#if __GLASGOW_HASKELL__ <= 706
-{-# LANGUAGE FlexibleContexts #-}
-#endif
 
 module Agda.Syntax.Abstract.Views where
 
@@ -37,11 +32,23 @@ data AppView = Application Expr [NamedArg Expr]
 appView :: Expr -> AppView
 appView e =
   case e of
-    App i e1 arg | Application hd es <- appView e1
+    App _ e1 e2
+      | Dot _ e2' <- unScope $ namedArg e2
+      , Just f <- maybeProjTurnPostfix e2'
+                   -> Application f [defaultNamedArg e1]
+    App i e1 arg
+      | Application hd es <- appView e1
                    -> Application hd $ es ++ [arg]
     ScopedExpr _ e -> appView e
     _              -> Application e []
 
+maybeProjTurnPostfix :: Expr -> Maybe Expr
+maybeProjTurnPostfix e =
+  case e of
+    ScopedExpr i e' -> ScopedExpr i <$> maybeProjTurnPostfix e'
+    Proj _ x        -> return $ Proj ProjPostfix x
+    _               -> Nothing
+
 unAppView :: AppView -> Expr
 unAppView (Application h es) =
   foldl (App (ExprRange noRange)) h es
@@ -90,8 +97,8 @@ deepUnscopeDecl d                                = [deepUnscope d]
 class ExprLike a where
   -- | The first expression is pre-traversal, the second one post-traversal.
   recurseExpr :: (Applicative m) => (Expr -> m Expr -> m Expr) -> a -> m a
-  default recurseExpr :: (Traversable f, Applicative m)
-                      => (Expr -> m Expr -> m Expr) -> f a -> m (f a)
+  default recurseExpr :: (Traversable f, ExprLike a', a ~ f a', Applicative m)
+                      => (Expr -> m Expr -> m Expr) -> a -> m a
   recurseExpr = traverse . recurseExpr
 
   foldExpr :: Monoid m => (Expr -> m) -> a -> m
@@ -120,6 +127,7 @@ instance ExprLike Expr where
       Lit{}                   -> pure e0
       QuestionMark{}          -> pure e0
       Underscore{}            -> pure e0
+      Dot ei e                -> Dot ei <$> recurse e
       App ei e arg            -> App ei <$> recurse e <*> recurse arg
       WithApp ei e es         -> WithApp ei <$> recurse e <*> recurse es
       Lam ei b e              -> Lam ei <$> recurse b <*> recurse e
@@ -155,6 +163,7 @@ instance ExprLike Expr where
       Lit{}                -> m
       QuestionMark{}       -> m
       Underscore{}         -> m
+      Dot _ e              -> m `mappend` fold e
       App _ e e'           -> m `mappend` fold e `mappend` fold e'
       WithApp _ e es       -> m `mappend` fold e `mappend` fold es
       Lam _ b e            -> m `mappend` fold b `mappend` fold e
@@ -190,6 +199,7 @@ instance ExprLike Expr where
       Lit{}                   -> f e
       QuestionMark{}          -> f e
       Underscore{}            -> f e
+      Dot ei e                -> f =<< Dot ei <$> trav e
       App ei e arg            -> f =<< App ei <$> trav e <*> trav arg
       WithApp ei e es         -> f =<< WithApp ei <$> trav e <*> trav es
       Lam ei b e              -> f =<< Lam ei <$> trav b <*> trav e
@@ -301,13 +311,13 @@ instance ExprLike LetBinding where
 instance ExprLike a => ExprLike (Pattern' a) where
 
 instance ExprLike a => ExprLike (Clause' a) where
-  recurseExpr f (Clause lhs rhs ds ca) = Clause <$> rec lhs <*> rec rhs <*> rec ds <*> pure ca
+  recurseExpr f (Clause lhs dots rhs ds ca) = Clause <$> rec lhs <*> pure dots <*> rec rhs <*> rec ds <*> pure ca
     where rec = recurseExpr f
 
 instance ExprLike RHS where
   recurseExpr f rhs =
     case rhs of
-      RHS e                   -> RHS <$> rec e
+      RHS e c                 -> RHS <$> rec e <*> pure c
       AbsurdRHS{}             -> pure rhs
       WithRHS x es cs         -> WithRHS x <$> rec es <*> rec cs
       RewriteRHS xes rhs ds   -> RewriteRHS <$> rec xes <*> rec rhs <*> rec ds
@@ -336,7 +346,6 @@ instance ExprLike Pragma where
       CompiledJSPragma{}          -> pure p
       CompiledUHCPragma{}         -> pure p
       CompiledDataUHCPragma{}     -> pure p
-      NoSmashingPragma{}          -> pure p
       StaticPragma{}              -> pure p
       InlinePragma{}              -> pure p
       DisplayPragma f xs e        -> DisplayPragma f <$> rec xs <*> rec e
@@ -353,12 +362,12 @@ instance ExprLike SpineLHS where
 instance ExprLike Declaration where
   recurseExpr f d =
     case d of
-      Axiom a d i x e           -> Axiom a d i x <$> rec e
+      Axiom a d i mp x e        -> Axiom a d i mp x <$> rec e
       Field i x e               -> Field i x <$> rec e
       Primitive i x e           -> Primitive i x <$> rec e
       Mutual i ds               -> Mutual i <$> rec ds
       Section i m tel ds        -> Section i m <$> rec tel <*> rec ds
-      Apply i m a rd rm d       -> (\ a -> Apply i m a rd rm d) <$> rec a
+      Apply i m a ci d          -> (\ a -> Apply i m a ci d) <$> rec a
       Import{}                  -> pure d
       Pragma i p                -> Pragma i <$> rec p
       Open{}                    -> pure d
