@@ -1,10 +1,6 @@
 {-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
-{-# LANGUAGE DeriveFoldable             #-}
-{-# LANGUAGE DeriveFunctor              #-}
 {-# LANGUAGE DeriveGeneric              #-}
-{-# LANGUAGE DeriveTraversable          #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 {-| Some common syntactic entities are defined in this module.
@@ -19,18 +15,17 @@ import qualified Data.ByteString.Char8 as ByteString
 import Data.Foldable
 import Data.Hashable
 import qualified Data.Strict.Maybe as Strict
-import Data.Monoid
+import Data.Semigroup hiding (Arg)
 import Data.Traversable
 import Data.Typeable (Typeable)
+import Data.Word
 
 import GHC.Generics (Generic)
 
-import Test.QuickCheck hiding (Small)
-
 import Agda.Syntax.Position
 
 import Agda.Utils.Functor
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty hiding ((<>))
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -63,13 +58,6 @@ instance HasRange Induction where
 instance KillRange Induction where
   killRange = id
 
-instance Arbitrary Induction where
-  arbitrary = elements [Inductive, CoInductive]
-
-instance CoArbitrary Induction where
-  coarbitrary Inductive   = variant 0
-  coarbitrary CoInductive = variant 1
-
 instance NFData Induction where
   rnf Inductive   = ()
   rnf CoInductive = ()
@@ -83,13 +71,16 @@ data Hiding  = Hidden | Instance | NotHidden
 
 -- | 'Hiding' is an idempotent partial monoid, with unit 'NotHidden'.
 --   'Instance' and 'NotHidden' are incompatible.
+instance Semigroup Hiding where
+  NotHidden <> h         = h
+  h         <> NotHidden = h
+  Hidden    <> Hidden    = Hidden
+  Instance  <> Instance  = Instance
+  _         <> _         = __IMPOSSIBLE__
+
 instance Monoid Hiding where
   mempty = NotHidden
-  mappend NotHidden h         = h
-  mappend h         NotHidden = h
-  mappend Hidden    Hidden    = Hidden
-  mappend Instance  Instance  = Instance
-  mappend _         _         = __IMPOSSIBLE__
+  mappend = (<>)
 
 instance KillRange Hiding where
   killRange = id
@@ -148,10 +139,6 @@ instance LensHiding (WithHiding a) where
 mergeHiding :: LensHiding a => WithHiding a -> a
 mergeHiding (WithHiding h a) = mapHiding (mappend h) a
 
--- | @isHidden@ does not apply to 'Instance', only to 'Hidden'.
-isHidden :: LensHiding a => a -> Bool
-isHidden a = getHiding a == Hidden
-
 -- | Visible ('NotHidden') arguments are @notHidden@. (DEPRECATED, use 'visible'.)
 notHidden :: LensHiding a => a -> Bool
 notHidden a = getHiding a == NotHidden
@@ -232,9 +219,6 @@ allRelevances =
 instance KillRange Relevance where
   killRange rel = rel -- no range to kill
 
-instance Arbitrary Relevance where
-  arbitrary = elements allRelevances
-
 instance Ord Relevance where
   (<=) = moreRelevant
 
@@ -350,13 +334,48 @@ ignoreForced Irrelevant = Irrelevant
 -- | Irrelevant function arguments may appear non-strictly in the codomain type.
 irrToNonStrict :: Relevance -> Relevance
 irrToNonStrict Irrelevant = NonStrict
--- irrToNonStrict NonStrict  = Relevant -- TODO: is that what we want (OR: NonStrict)  -- better be more conservative
+-- irrToNonStrict NonStrict  = Relevant -- TODO: this is bad if we apply irrToNonStrict several times!
 irrToNonStrict rel        = rel
 
+-- | Applied when working on types (unless --experimental-irrelevance).
+nonStrictToRel :: Relevance -> Relevance
+nonStrictToRel NonStrict = Relevant
+nonStrictToRel rel       = rel
+
 nonStrictToIrr :: Relevance -> Relevance
 nonStrictToIrr NonStrict = Irrelevant
 nonStrictToIrr rel       = rel
 
+---------------------------------------------------------------------------
+-- * Origin of arguments (user-written, inserted or reflected)
+---------------------------------------------------------------------------
+
+data Origin = UserWritten | Inserted | Reflected
+  deriving (Typeable, Show, Eq, Ord)
+
+instance KillRange Origin where
+  killRange = id
+
+instance NFData Origin where
+  rnf UserWritten = ()
+  rnf Inserted = ()
+  rnf Reflected = ()
+
+class LensOrigin a where
+
+  getOrigin :: a -> Origin
+
+  setOrigin :: Origin -> a -> a
+  setOrigin o = mapOrigin (const o)
+
+  mapOrigin :: (Origin -> Origin) -> a -> a
+  mapOrigin f a = setOrigin (f $ getOrigin a) a
+
+instance LensOrigin Origin where
+  getOrigin = id
+  setOrigin = const
+  mapOrigin = id
+
 ---------------------------------------------------------------------------
 -- * Argument decoration
 ---------------------------------------------------------------------------
@@ -364,12 +383,14 @@ nonStrictToIrr rel       = rel
 -- | A function argument can be hidden and/or irrelevant.
 
 data ArgInfo = ArgInfo
-  { argInfoHiding    :: Hiding
-  , argInfoRelevance :: Relevance
+  { argInfoHiding       :: Hiding
+  , argInfoRelevance    :: Relevance
+  , argInfoOrigin       :: Origin
+  , argInfoOverlappable :: Bool
   } deriving (Typeable, Eq, Ord, Show)
 
 instance KillRange ArgInfo where
-  killRange (ArgInfo h r) = killRange2 ArgInfo h r
+  killRange (ArgInfo h r o v) = killRange3 ArgInfo h r o v
 
 class LensArgInfo a where
   getArgInfo :: a -> ArgInfo
@@ -384,7 +405,7 @@ instance LensArgInfo ArgInfo where
   mapArgInfo = id
 
 instance NFData ArgInfo where
-  rnf (ArgInfo a b) = rnf a `seq` rnf b
+  rnf (ArgInfo a b c d) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
 
 instance LensHiding ArgInfo where
   getHiding = argInfoHiding
@@ -396,9 +417,16 @@ instance LensRelevance ArgInfo where
   setRelevance h ai = ai { argInfoRelevance = h }
   mapRelevance f ai = ai { argInfoRelevance = f (argInfoRelevance ai) }
 
+instance LensOrigin ArgInfo where
+  getOrigin = argInfoOrigin
+  setOrigin o ai = ai { argInfoOrigin = o }
+  mapOrigin f ai = ai { argInfoOrigin = f (argInfoOrigin ai) }
+
 defaultArgInfo :: ArgInfo
-defaultArgInfo =  ArgInfo { argInfoHiding    = NotHidden
-                          , argInfoRelevance = Relevant }
+defaultArgInfo =  ArgInfo { argInfoHiding       = NotHidden
+                          , argInfoRelevance    = Relevant
+                          , argInfoOrigin       = UserWritten
+                          , argInfoOverlappable = False }
 
 
 ---------------------------------------------------------------------------
@@ -423,14 +451,14 @@ instance KillRange a => KillRange (Arg a) where
   killRange (Arg info a) = killRange2 Arg info a
 
 instance Eq a => Eq (Arg a) where
-  Arg (ArgInfo h1 _) x1 == Arg (ArgInfo h2 _) x2 = (h1, x1) == (h2, x2)
+  Arg (ArgInfo h1 _ _ _) x1 == Arg (ArgInfo h2 _ _ _) x2 = (h1, x1) == (h2, x2)
 
 instance Show a => Show (Arg a) where
-    show (Arg (ArgInfo h r) x) = showR r $ showH h $ show x
+    show (Arg (ArgInfo h r o v) x) = showR r $ showO o $ showH h $ show x
       where
         showH Hidden     s = "{" ++ s ++ "}"
         showH NotHidden  s = "(" ++ s ++ ")"
-        showH Instance   s = "{{" ++ s ++ "}}"
+        showH Instance   s = (if v then "overlap " else "") ++ "{{" ++ s ++ "}}"
         showR r s = case r of
           Irrelevant   -> "." ++ s
           NonStrict    -> "?" ++ s
@@ -438,6 +466,10 @@ instance Show a => Show (Arg a) where
           Forced Small -> "!" ++ s
           UnusedArg    -> "k" ++ s -- constant
           Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
+        showO o s = case o of
+          UserWritten -> "u" ++ s
+          Inserted    -> "i" ++ s
+          Reflected   -> "g" ++ s -- generated by reflection
 
 instance NFData e => NFData (Arg e) where
   rnf (Arg a b) = rnf a `seq` rnf b
@@ -450,16 +482,9 @@ instance LensRelevance (Arg e) where
   getRelevance = getRelevance . argInfo
   mapRelevance = mapArgInfo . mapRelevance
 
-{- RETIRED
-hide :: Arg a -> Arg a
-hide = setArgHiding Hidden
-
-makeInstance :: Arg a -> Arg a
-makeInstance = setHiding Instance
-
-isHiddenArg :: Arg a -> Bool
-isHiddenArg arg = argHiding arg /= NotHidden
--}
+instance LensOrigin (Arg e) where
+  getOrigin = getOrigin . argInfo
+  mapOrigin = mapArgInfo . mapOrigin
 
 instance LensArgInfo (Arg a) where
   getArgInfo        = argInfo
@@ -514,7 +539,7 @@ instance Underscore Doc where
 data Dom e = Dom
   { domInfo   :: ArgInfo
   , unDom     :: e
-  } deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
+  } deriving (Typeable, Ord, Functor, Foldable, Traversable)
 
 instance Decoration Dom where
   traverseF f (Dom ai a) = Dom ai <$> f a
@@ -525,6 +550,10 @@ instance HasRange a => HasRange (Dom a) where
 instance KillRange a => KillRange (Dom a) where
   killRange (Dom info a) = killRange2 Dom info a
 
+instance Eq a => Eq (Dom a) where
+  Dom (ArgInfo h1 r1 _ _) x1 == Dom (ArgInfo h2 r2 _ _) x2 =
+    (h1, ignoreForced r1, x1) == (h2, ignoreForced r2, x2)
+
 instance Show a => Show (Dom a) where
   show = show . argFromDom
 
@@ -658,16 +687,34 @@ stringToRawName = id
 type RString = Ranged RawName
 
 ---------------------------------------------------------------------------
--- * Constructor pattern info
+-- * Further constructor and projection info
 ---------------------------------------------------------------------------
 
--- | Where does the 'ConP' of come from?
-data ConPOrigin
-  = ConPImplicit  -- ^ Expanded from an implicit pattern.
-  | ConPCon       -- ^ User wrote a constructor pattern.
-  | ConPRec       -- ^ User wrote a record pattern.
+-- | Where does the 'ConP' or 'Con' come from?
+data ConOrigin
+  = ConOSystem  -- ^ Inserted by system or expanded from an implicit pattern.
+  | ConOCon     -- ^ User wrote a constructor (pattern).
+  | ConORec     -- ^ User wrote a record (pattern).
   deriving (Typeable, Show, Eq, Ord, Enum, Bounded)
 
+instance KillRange ConOrigin where
+  killRange = id
+
+-- | Prefer user-written over system-inserted.
+bestConInfo :: ConOrigin -> ConOrigin -> ConOrigin
+bestConInfo ConOSystem o = o
+bestConInfo o _ = o
+
+-- | Where does a projection come from?
+data ProjOrigin
+  = ProjPrefix    -- ^ User wrote a prefix projection.
+  | ProjPostfix   -- ^ User wrote a postfix projection.
+  | ProjSystem    -- ^ Projection was generated by the system.
+  deriving (Typeable, Show, Eq, Ord, Enum, Bounded)
+
+instance KillRange ProjOrigin where
+  killRange = id
+
 ---------------------------------------------------------------------------
 -- * Infixity, access, abstract, etc.
 ---------------------------------------------------------------------------
@@ -678,11 +725,24 @@ data IsInfix = InfixDef | PrefixDef
     deriving (Typeable, Show, Eq, Ord)
 
 -- | Access modifier.
-data Access = PrivateAccess | PublicAccess
-            | OnlyQualified  -- ^ Visible from outside, but not exported when opening the module
+data Access
+  = PrivateAccess Origin
+      -- ^ Store the 'Origin' of the private block that lead to this qualifier.
+      --   This is needed for more faithful printing of declarations.
+  | PublicAccess
+  | OnlyQualified  -- ^ Visible from outside, but not exported when opening the module
                              --   Used for qualified constructors.
     deriving (Typeable, Show, Eq, Ord)
 
+instance NFData Access where
+  rnf _ = ()
+
+instance HasRange Access where
+  getRange _ = noRange
+
+instance KillRange Access where
+  killRange = id
+
 -- | Abstract or concrete
 data IsAbstract = AbstractDef | ConcreteDef
     deriving (Typeable, Show, Eq, Ord)
@@ -720,7 +780,7 @@ type Arity  = Nat
 
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
-data NameId = NameId !Integer !Integer
+data NameId = NameId {-# UNPACK #-} !Word64 {-# UNPACK #-} !Word64
     deriving (Eq, Ord, Typeable, Generic)
 
 instance KillRange NameId where
@@ -742,11 +802,6 @@ instance Hashable NameId where
   {-# INLINE hashWithSalt #-}
   hashWithSalt salt (NameId n m) = hashWithSalt salt (n, m)
 
-instance Arbitrary NameId where
-  arbitrary = elements [ NameId x y | x <- [-1, 1], y <- [-1, 1] ]
-
-instance CoArbitrary NameId
-
 ---------------------------------------------------------------------------
 -- * Meta variables
 ---------------------------------------------------------------------------
@@ -843,11 +898,14 @@ data ImportDirective' a b = ImportDirective
 data Using' a b = UseEverything | Using [ImportedName' a b]
   deriving (Typeable, Eq)
 
+instance Semigroup (Using' a b) where
+  UseEverything <> u = u
+  u <> UseEverything = u
+  Using xs <> Using ys = Using (xs ++ ys)
+
 instance Monoid (Using' a b) where
   mempty = UseEverything
-  mappend UseEverything u = u
-  mappend u UseEverything = u
-  mappend (Using xs) (Using ys) = Using (xs ++ ys)
+  mappend = (<>)
 
 -- | Default is directive is @private@ (use everything, but do not export).
 defaultImportDir :: ImportDirective' a b
