@@ -1,9 +1,7 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 {-| Names in the concrete syntax are just strings (or lists of strings for
     qualified names).
@@ -21,12 +19,12 @@ import GHC.Generics (Generic)
 
 import System.FilePath
 
-import Test.QuickCheck
-
 import Agda.Syntax.Common
 import Agda.Syntax.Position
+
 import Agda.Utils.FileName
 import Agda.Utils.Pretty
+import Agda.Utils.Size
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -111,7 +109,7 @@ instance Underscore QName where
 
 newtype TopLevelModuleName
   = TopLevelModuleName { moduleNameParts :: [String] }
-  deriving (Show, Eq, Ord, Typeable)
+  deriving (Show, Eq, Ord, Typeable, Sized)
 
 ------------------------------------------------------------------------
 -- * Operations on 'Name' and 'NamePart'
@@ -202,6 +200,17 @@ qnameParts (QName x)  = [x]
 toTopLevelModuleName :: QName -> TopLevelModuleName
 toTopLevelModuleName = TopLevelModuleName . map prettyShow . qnameParts
 
+-- | Turns a top level module into a qualified name with 'noRange'.
+
+fromTopLevelModuleName :: TopLevelModuleName -> QName
+fromTopLevelModuleName (TopLevelModuleName [])     = __IMPOSSIBLE__
+fromTopLevelModuleName (TopLevelModuleName (x:xs)) = loop x xs
+  where
+  loop x []       = QName (mk x)
+  loop x (y : ys) = Qual  (mk x) $ loop y ys
+  mk :: String -> Name
+  mk x = Name noRange [Id x]
+
 -- | Turns a top-level module name into a file name with the given
 -- suffix.
 
@@ -250,6 +259,7 @@ instance IsNoName Name where
   isNoName (NoName _ _)    = True
   isNoName (Name _ [Hole]) = True   -- TODO: Track down where these come from
   isNoName (Name _ [])     = True
+  isNoName (Name _ [Id x]) = isNoName x
   isNoName _               = False
 
 instance IsNoName QName where
@@ -292,43 +302,14 @@ instance Pretty NamePart where
   pretty (Id s) = text $ rawNameToString s
 
 instance Pretty QName where
-  pretty (Qual m x) = pretty m <> pretty "." <> pretty x
+  pretty (Qual m x)
+    | isUnderscore m = pretty x -- don't print anonymous modules
+    | otherwise      = pretty m <> pretty "." <> pretty x
   pretty (QName x)  = pretty x
 
 instance Pretty TopLevelModuleName where
   pretty (TopLevelModuleName ms) = text $ intercalate "." ms
 
-------------------------------------------------------------------------
--- * QuickCheck instances
-------------------------------------------------------------------------
-
-instance Arbitrary TopLevelModuleName where
-  arbitrary = TopLevelModuleName <$> listOf1 (listOf1 $ elements "AB")
-
-instance CoArbitrary TopLevelModuleName where
-  coarbitrary (TopLevelModuleName m) = coarbitrary m
-
-instance Arbitrary Name where
-  arbitrary = oneof
-    [ Name   <$> arbitrary <*> parts
-    , NoName <$> arbitrary <*> arbitrary
-    ]
-    where
-    parts = do
-      parts         <- listOf1 (elements ["x", "y", "z"])
-      startWithHole <- arbitrary
-      endWithHole   <- arbitrary
-      return $
-        (if startWithHole then (Hole :)    else id) $
-        (if endWithHole   then (++ [Hole]) else id) $
-        intersperse Hole (map Id parts)
-
-instance CoArbitrary NamePart
-
-instance CoArbitrary Name where
-  coarbitrary (Name _ ns)  = variant 0 . coarbitrary ns
-  coarbitrary (NoName _ i) = variant 1 . coarbitrary i
-
 ------------------------------------------------------------------------
 -- * Range instances
 ------------------------------------------------------------------------
