@@ -1,9 +1,5 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE DeriveFoldable #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE DeriveTraversable #-}
-{-# LANGUAGE PatternGuards #-}
 
 {-| Definitions for fixity, precedence levels, and declared syntax.
 -}
@@ -41,8 +37,14 @@ import Agda.Utils.Impossible
 data Fixity' = Fixity'
     { theFixity   :: !Fixity
     , theNotation :: Notation
+    , theNameRange :: Range
+      -- ^ Range of the name in the fixity declaration
+      --   (used for correct highlighting, see issue #2140).
     }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Show)
+
+instance Eq Fixity' where
+  Fixity' f n _ == Fixity' f' n' _ = f == f' && n == n'
 
 -- | Decorating something with @Fixity'@.
 data ThingWithFixity x = ThingWithFixity x Fixity'
@@ -75,7 +77,7 @@ namesToNotation q n = NewNotation
   , notation       = if null syn then syntaxOf (unqualify q) else syn
   , notaIsOperator = null syn
   }
-  where Fixity' f syn = A.nameFixity n
+  where Fixity' f syn _ = A.nameFixity n
 
 -- | Replace 'noFixity' by 'defaultFixity'.
 useDefaultFixity :: NewNotation -> NewNotation
@@ -114,7 +116,7 @@ syntaxOf (Name _ xs)  = mkSyn 0 xs
     mkSyn n (Id x : xs) = IdPart x : mkSyn n xs
 
 noFixity' :: Fixity'
-noFixity' = Fixity' noFixity noNotation
+noFixity' = Fixity' noFixity noNotation noRange
 
 -- | Merges 'NewNotation's that have the same precedence level and
 -- notation, with two exceptions:
@@ -208,7 +210,11 @@ noSection n = NotationSection
 
 -- | Precedence levels for operators.
 
-data PrecedenceLevel = Unrelated | Related !Integer
+data PrecedenceLevel
+  = Unrelated
+    -- ^ No fixity declared.
+  | Related !Integer
+    -- ^ Fixity level declared as the @Integer@.
   deriving (Eq, Ord, Show, Typeable)
 
 -- | Associativity.
@@ -218,11 +224,12 @@ data Associativity = NonAssoc | LeftAssoc | RightAssoc
 
 -- | Fixity of operators.
 
-data Fixity =
-  Fixity { fixityRange :: Range
-         , fixityLevel :: !PrecedenceLevel
-         , fixityAssoc :: !Associativity
-         }
+data Fixity = Fixity
+  { fixityRange :: Range
+    -- ^ Range of the whole fixity declaration.
+  , fixityLevel :: !PrecedenceLevel
+  , fixityAssoc :: !Associativity
+  }
   deriving (Typeable, Show)
 
 instance Eq Fixity where
@@ -318,7 +325,7 @@ instance KillRange Fixity where
   killRange f = f { fixityRange = noRange }
 
 instance KillRange Fixity' where
-  killRange (Fixity' f n) = killRange2 Fixity' f n
+  killRange (Fixity' f n r) = killRange3 Fixity' f n r
 
 instance KillRange x => KillRange (ThingWithFixity x) where
   killRange (ThingWithFixity c f) = ThingWithFixity (killRange c) f
@@ -345,7 +352,7 @@ _fixityLevel f r = f (fixityLevel r) <&> \x -> r { fixityLevel = x }
 ------------------------------------------------------------------------
 
 instance NFData Fixity' where
-  rnf (Fixity' _ a) = rnf a
+  rnf (Fixity' _ a _) = rnf a
 
 -- | Ranges are not forced.
 
