@@ -1,4 +1,3 @@
-{-# LANGUAGE FlexibleInstances #-}
 
 -- | Extract used definitions from terms.
 module Agda.Syntax.Internal.Defs where
@@ -45,16 +44,10 @@ class GetDefs a where
 instance GetDefs Clause where
   getDefs = getDefs . clauseBody
 
-instance GetDefs ClauseBody where
-  getDefs b = case b of
-    Body v -> getDefs v
-    Bind b -> getDefs b
-    NoBody -> return ()
-
 instance GetDefs Term where
   getDefs v = case v of
     Def d vs   -> doDef d >> getDefs vs
-    Con c vs   -> getDefs vs
+    Con _ _ vs -> getDefs vs
     Lit l      -> return ()
     Var i vs   -> getDefs vs
     Lam _ v    -> getDefs v
