@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
 
 module Agda.Syntax.Internal.Generic where
 
@@ -100,13 +99,18 @@ instance TermLike a => TermLike (Ptr a) where
   traverseTermM f = traverse (traverseTermM f)
   foldTerm f = foldMap (foldTerm f)
 
+instance TermLike a => TermLike (Blocked a) where
+  traverseTerm f = fmap (traverseTerm f)
+  traverseTermM f = traverse (traverseTermM f)
+  foldTerm f = foldMap (foldTerm f)
+
 -- * Real terms
 
 instance TermLike Term where
   traverseTerm f t = case t of
     Var i xs    -> f $ Var i $ traverseTerm f xs
     Def c xs    -> f $ Def c $ traverseTerm f xs
-    Con c xs    -> f $ Con c $ traverseTerm f xs
+    Con c ci xs -> f $ Con c ci $ traverseTerm f xs
     Lam h b     -> f $ Lam h $ traverseTerm f b
     Pi a b      -> f $ uncurry Pi $ traverseTerm f (a, b)
     MetaV m xs  -> f $ MetaV m $ traverseTerm f xs
@@ -119,7 +123,7 @@ instance TermLike Term where
   traverseTermM f t = case t of
     Var i xs    -> f =<< Var i <$> traverseTermM f xs
     Def c xs    -> f =<< Def c <$> traverseTermM f xs
-    Con c xs    -> f =<< Con c <$> traverseTermM f xs
+    Con c ci xs -> f =<< Con c ci <$> traverseTermM f xs
     Lam h b     -> f =<< Lam h <$> traverseTermM f b
     Pi a b      -> f =<< uncurry Pi <$> traverseTermM f (a, b)
     MetaV m xs  -> f =<< MetaV m <$> traverseTermM f xs
@@ -132,7 +136,7 @@ instance TermLike Term where
   foldTerm f t = f t `mappend` case t of
     Var i xs    -> foldTerm f xs
     Def c xs    -> foldTerm f xs
-    Con c xs    -> foldTerm f xs
+    Con c ci xs -> foldTerm f xs
     Lam h b     -> foldTerm f b
     Pi a b      -> foldTerm f (a, b)
     MetaV m xs  -> foldTerm f xs
