@@ -19,6 +19,7 @@ import qualified Agda.Syntax.Abstract as A
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.CompiledClause
 
+import Agda.Utils.Functor
 import Agda.Utils.Impossible
 #include "undefined.h"
 
@@ -36,7 +37,6 @@ instance NamesIn a => NamesIn (Named n a)     where namesIn = namesInFoldable
 instance NamesIn a => NamesIn (Abs a)         where namesIn = namesInFoldable
 instance NamesIn a => NamesIn (WithArity a)   where namesIn = namesInFoldable
 instance NamesIn a => NamesIn (Tele a)        where namesIn = namesInFoldable
-instance NamesIn a => NamesIn (ClauseBodyF a) where namesIn = namesInFoldable
 
 instance NamesIn a => NamesIn (C.FieldAssignment' a) where namesIn = namesInFoldable
 
@@ -55,9 +55,10 @@ instance NamesIn Defn where
     Function    { funClauses = cl, funCompiled = cc }              -> namesIn (cl, cc)
     Datatype    { dataClause = cl, dataCons = cs, dataSort = s }   -> namesIn (cl, cs, s)
     Record      { recClause = cl, recConHead = c, recFields = fs } -> namesIn (cl, c, fs)
-      -- Don't need recTel or recConType since those will be reachable from the constructor
+      -- Don't need recTel since those will be reachable from the constructor
     Constructor { conSrcCon = c, conData = d }                     -> namesIn (c, d)
     Primitive   { primClauses = cl, primCompiled = cc }            -> namesIn (cl, cc)
+    AbstractDefn -> __IMPOSSIBLE__
 
 instance NamesIn Clause where
   namesIn Clause{ clauseTel = tel, namedClausePats = ps, clauseBody = b, clauseType = t } =
@@ -78,7 +79,7 @@ instance NamesIn (Pattern' a) where
     LitP l        -> namesIn l
     DotP v        -> namesIn v
     ConP c _ args -> namesIn (c, args)
-    ProjP f       -> namesIn f
+    ProjP _ f     -> namesIn f
 
 instance NamesIn a => NamesIn (Type' a) where
   namesIn (El s t) = namesIn (s, t)
@@ -97,7 +98,7 @@ instance NamesIn Term where
     Lam _ b      -> namesIn b
     Lit l        -> namesIn l
     Def f args   -> namesIn (f, args)
-    Con c args   -> namesIn (c, args)
+    Con c _ args -> namesIn (c, args)
     Pi a b       -> namesIn (a, b)
     Sort s       -> namesIn s
     Level l      -> namesIn l
@@ -131,7 +132,7 @@ instance NamesIn Literal where
 
 instance NamesIn a => NamesIn (Elim' a) where
   namesIn (Apply arg) = namesIn arg
-  namesIn (Proj f)    = namesIn f
+  namesIn (Proj _ f)  = namesIn f
 
 instance NamesIn QName   where namesIn x = Set.singleton x
 instance NamesIn ConHead where namesIn h = namesIn (conName h)
@@ -139,13 +140,15 @@ instance NamesIn ConHead where namesIn h = namesIn (conName h)
 instance NamesIn a => NamesIn (Open a) where
   namesIn = namesIn . openThing
 
+instance NamesIn a => NamesIn (Local a) where namesIn = namesIn . dget
+
 instance NamesIn DisplayForm where
   namesIn (Display _ ps v) = namesIn (ps, v)
 
 instance NamesIn DisplayTerm where
   namesIn v = case v of
-    DWithApp v us vs -> namesIn (v, us, vs)
-    DCon c vs        -> namesIn (c, vs)
+    DWithApp v us es -> namesIn (v, us, es)
+    DCon c _ vs      -> namesIn (c, vs)
     DDef f es        -> namesIn (f, es)
     DDot v           -> namesIn v
     DTerm v          -> namesIn v
@@ -160,6 +163,7 @@ instance NamesIn (A.Pattern' a) where
   namesIn p = case p of
     A.VarP{}               -> Set.empty
     A.ConP _ c args        -> namesIn (c, args)
+    A.ProjP _ _ d          -> namesIn d
     A.DefP _ f args        -> namesIn (f, args)
     A.WildP{}              -> Set.empty
     A.AsP _ _ p            -> namesIn p
