@@ -5,6 +5,9 @@ module Agda.Syntax.Literal where
 import Control.DeepSeq
 import Data.Char
 import Data.Typeable (Typeable)
+
+import Numeric.IEEE ( IEEE(identicalIEEE) )
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract.Name
@@ -53,7 +56,10 @@ showChar' c
 
 instance Eq Literal where
   LitNat _ n    == LitNat _ m    = n == m
-  LitFloat _ x  == LitFloat _ y  = x == y
+  -- ASR (2016-09-29). We use bitwise equality for comparing Double
+  -- because Haskell's Eq, which equates 0.0 and -0.0, allows to prove
+  -- a contradiction (see Issue #2169).
+  LitFloat _ x  == LitFloat _ y  = identicalIEEE x y || (isNaN x && isNaN y)
   LitString _ s == LitString _ t = s == t
   LitChar _ c   == LitChar _ d   = c == d
   LitQName _ x  == LitQName _ y  = x == y
@@ -62,7 +68,7 @@ instance Eq Literal where
 
 instance Ord Literal where
   LitNat _ n    `compare` LitNat _ m    = n `compare` m
-  LitFloat _ x  `compare` LitFloat _ y  = x `compare` y
+  LitFloat _ x  `compare` LitFloat _ y  = compareFloat x y
   LitString _ s `compare` LitString _ t = s `compare` t
   LitChar _ c   `compare` LitChar _ d   = c `compare` d
   LitQName _ x  `compare` LitQName _ y  = x `compare` y
@@ -80,6 +86,23 @@ instance Ord Literal where
   -- compare LitMeta{} _   = LT
   -- compare _ LitMeta{}   = GT
 
+-- NOTE: This is not the same ordering as primFloatNumericalEquality!
+-- This ordering must be a total order of all allowed float values,
+-- while primFloatNumericalEquality is only a preorder
+compareFloat :: Double -> Double -> Ordering
+compareFloat x y
+  | identicalIEEE x y          = EQ
+  | isNegInf x                 = LT
+  | isNegInf y                 = GT
+  | isNaN x && isNaN y         = EQ
+  | isNaN x                    = LT
+  | isNaN y                    = GT
+  | isNegativeZero x && x == y = LT
+  | isNegativeZero y && x == y = GT
+  | otherwise                  = compare x y
+  where
+    isNegInf z = z < 0 && isInfinite z
+
 instance HasRange Literal where
   getRange (LitNat    r _) = r
   getRange (LitFloat  r _) = r
