@@ -1,12 +1,12 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable    #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
 
 module Agda.Syntax.Parser.Monad
     ( -- * The parser monad
       Parser
     , ParseResult(..)
     , ParseState(..)
-    , ParseError(..)
+    , ParseError(..), ParseWarning(..)
     , LexState
     , LayoutContext(..)
     , ParseFlags (..)
@@ -14,8 +14,9 @@ module Agda.Syntax.Parser.Monad
     , initState
     , defaultParseFlags
     , parse
-    , parsePosString
     , parseFile
+    , parsePosString
+    , parseFromSrc
       -- * Manipulating the state
     , setParsePos, setLastPos, getParseInterval
     , setPrevToken
@@ -30,7 +31,7 @@ module Agda.Syntax.Parser.Monad
     )
     where
 
-import Control.Exception
+import Control.Exception (catch)
 import Data.Int
 import Data.Typeable ( Typeable )
 
@@ -40,12 +41,19 @@ import Control.Applicative
 import Agda.Syntax.Position
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
+#if !(MIN_VERSION_mtl(2,2,1))
+import Agda.Utils.Except ( Error(noMsg) )
+#endif
 
 import Agda.Utils.FileName
 import qualified Agda.Utils.IO.UTF8 as UTF8
+import qualified Agda.Utils.Maybe.Strict as Strict
 
 import Agda.Utils.Pretty
 
+#include "undefined.h"
+import Agda.Utils.Impossible
+
 {--------------------------------------------------------------------------
     The parse monad
  --------------------------------------------------------------------------}
@@ -92,7 +100,9 @@ data ParseFlags = ParseFlags
   deriving Show
 
 -- | What you get if parsing fails.
-data ParseError = ParseError
+data ParseError =
+  -- | Errors that arise at a specific position in the file
+  ParseError
   { errSrcFile   :: !SrcFile
                     -- ^ The file in which the error occurred.
   , errPos       :: !PositionWithoutFile
@@ -103,10 +113,31 @@ data ParseError = ParseError
                     -- ^ The previous token.
   , errMsg       :: String
                     -- ^ Hopefully an explanation of what happened.
+  } |
+  -- | Parse errors that concern a range in a file.
+  OverlappingTokensError
+  { errRange     :: !(Range' SrcFile)
+                    -- ^ The range of the bigger overlapping token
+  } |
+  -- | Parse errors that concern a whole file.
+  InvalidExtensionError
+  { errPath      :: !AbsolutePath
+                    -- ^ The file which the error concerns.
+  , errValidExts :: [String]
+  } |
+  ReadFileError
+  { errPath      :: !AbsolutePath
+  , errIOError   :: IOError
   }
-    deriving (Typeable)
+  deriving (Typeable)
 
-instance Exception ParseError
+-- | Warnings for parsing
+data ParseWarning =
+  -- | Parse errors that concern a range in a file.
+  OverlappingTokensWarning
+  { warnRange    :: !(Range' SrcFile)
+                    -- ^ The range of the bigger overlapping token
+  }
 
 -- | The result of parsing something.
 data ParseResult a  = ParseOk ParseState a
@@ -153,16 +184,52 @@ instance Show ParseError where
   show = prettyShow
 
 instance Pretty ParseError where
-  pretty err = vcat
-      [ pretty ((errPos err) { srcFile = errSrcFile err }) <> colon <+>
-        text (errMsg err)
-      , text $ errPrevToken err ++ "<ERROR>"
-      , text $ take 30 (errInput err) ++ "..."
+  pretty ParseError{errPos,errSrcFile,errMsg,errPrevToken,errInput} = vcat
+      [ pretty (errPos { srcFile = errSrcFile }) <> colon <+>
+        text errMsg
+      , text $ errPrevToken ++ "<ERROR>"
+      , text $ take 30 errInput ++ "..."
+      ]
+  pretty OverlappingTokensError{errRange} = vcat
+      [ pretty errRange <> colon <+>
+        text "Multi-line comment spans one or more literate text blocks."
+      ]
+  pretty InvalidExtensionError{errPath,errValidExts} = vcat
+      [ pretty errPath <> colon <+>
+        text "Unsupported extension."
+      , text "Supported extensions are:" <+> prettyList errValidExts
+      ]
+  pretty ReadFileError{errPath,errIOError} = vcat
+      [ text "Cannot read file" <+> pretty errPath
+        -- TODO: `show` should be replaced by `displayException` once we
+        -- cease to support versions of GHC under 7.10.
+      , text "Error:" <+> text (show errIOError)
       ]
 
 instance HasRange ParseError where
-  getRange err = posToRange' (errSrcFile err) p p
-    where p = errPos err
+  getRange ParseError{errSrcFile,errPos=p} = posToRange' errSrcFile p p
+  getRange OverlappingTokensError{errRange} = errRange
+  getRange InvalidExtensionError{errPath} = posToRange p p
+    where p = startPos (Just errPath)
+  getRange ReadFileError{errPath} = posToRange p p
+    where p = startPos (Just errPath)
+
+#if !(MIN_VERSION_mtl(2,2,1))
+-- Stupid ErrorT!
+instance Error ParseError where
+  noMsg = __IMPOSSIBLE__
+#endif
+
+instance Show ParseWarning where
+  show = prettyShow
+
+instance Pretty ParseWarning where
+  pretty OverlappingTokensWarning{warnRange} = vcat
+      [ pretty warnRange <> colon <+>
+        text "Multi-line comment spans one or more literate text blocks."
+      ]
+instance HasRange ParseWarning where
+  getRange OverlappingTokensWarning{warnRange} = warnRange
 
 {--------------------------------------------------------------------------
     Running the parser
@@ -198,7 +265,7 @@ defaultParseFlags = ParseFlags { parseKeepComments = False }
 --   more specialised functions that supply the 'ParseFlags' and the
 --   'LexState'.
 parse :: ParseFlags -> [LexState] -> Parser a -> String -> ParseResult a
-parse flags st p input = unP p (initState Nothing flags input st)
+parse flags st p input = parseFromSrc flags st p Strict.Nothing input
 
 -- | The even more general way of parsing a string.
 parsePosString :: Position -> ParseFlags -> [LexState] -> Parser a -> String ->
@@ -214,8 +281,18 @@ parsePosString pos flags st p input = unP p (initStatePos pos flags input st)
 parseFile :: ParseFlags -> [LexState] -> Parser a -> AbsolutePath
           -> IO (ParseResult a)
 parseFile flags st p file =
-    do  input <- liftIO $ UTF8.readTextFile $ filePath file
-        return $ unP p (initState (Just file) flags input st)
+    do  res <- (Right <$> (UTF8.readTextFile (filePath file))) `catch`
+          (return . Left . ReadFileError file)
+        case res of
+          Left  error -> return$ ParseFailed error
+          Right input -> return$ parseFromSrc flags st p (Strict.Just file) input
+
+-- | Parses a string as if it were the contents of the given file
+--   Useful for integrating preprocessors.
+parseFromSrc :: ParseFlags -> [LexState] -> Parser a -> SrcFile -> String
+              -> ParseResult a
+parseFromSrc flags st p src input = unP p (initState (Strict.toLazy src) flags input st)
+
 
 {--------------------------------------------------------------------------
     Manipulating the state
