@@ -11,6 +11,7 @@ import Agda.Syntax.Position
 data Keyword
         = KwLet | KwIn | KwWhere | KwData | KwCoData
         | KwPostulate | KwMutual | KwAbstract | KwPrivate | KwInstance
+        | KwOverlap
         | KwOpen | KwImport | KwModule | KwPrimitive | KwMacro
         | KwInfix | KwInfixL | KwInfixR | KwWith | KwRewrite
         | KwSet | KwProp | KwForall | KwRecord | KwConstructor | KwField
@@ -21,7 +22,7 @@ data Keyword
         | KwCOMPILED_DATA | KwCOMPILED_DECLARE_DATA | KwCOMPILED_TYPE | KwCOMPILED | KwCOMPILED_EXPORT
         | KwHASKELL
         | KwCOMPILED_EPIC | KwCOMPILED_JS | KwCOMPILED_UHC | KwCOMPILED_DATA_UHC
-        | KwIMPORT | KwIMPORT_UHC | KwIMPOSSIBLE | KwSTATIC | KwINLINE | KwNO_SMASHING
+        | KwIMPORT | KwIMPORT_UHC | KwIMPOSSIBLE | KwSTATIC | KwINLINE
         | KwNO_TERMINATION_CHECK | KwTERMINATING | KwNON_TERMINATING
         | KwMEASURE | KwDISPLAY
         | KwREWRITE
@@ -29,7 +30,7 @@ data Keyword
         | KwUnquote | KwUnquoteDecl | KwUnquoteDef
         | KwSyntax
         | KwPatternSyn | KwTactic | KwCATCHALL
-        | KwNO_POSITIVITY_CHECK
+        | KwNO_POSITIVITY_CHECK | KwPOLARITY
     deriving (Eq, Show)
 
 layoutKeywords :: [Keyword]
@@ -41,6 +42,7 @@ data Symbol
         | SymColon | SymArrow | SymEqual | SymLambda
         | SymUnderscore | SymQuestionMark   | SymAs
         | SymOpenParen        | SymCloseParen
+        | SymOpenIdiomBracket | SymCloseIdiomBracket
         | SymDoubleOpenBrace  | SymDoubleCloseBrace
         | SymOpenBrace        | SymCloseBrace
         | SymOpenVirtualBrace | SymCloseVirtualBrace
