@@ -1,14 +1,9 @@
 {-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
-{-# LANGUAGE DeriveFoldable             #-}
-{-# LANGUAGE DeriveFunctor              #-}
 {-# LANGUAGE DeriveGeneric              #-}
-{-# LANGUAGE DeriveTraversable          #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
 {-# LANGUAGE ScopedTypeVariables        #-}
-{-# LANGUAGE TemplateHaskell            #-}
 
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
@@ -28,6 +23,7 @@ module Agda.Syntax.Position
   , movePos
   , movePosByString
   , backupPos
+  , startPos'
 
     -- * Intervals
   , Interval
@@ -37,13 +33,18 @@ module Agda.Syntax.Position
   , posToInterval
   , takeI
   , dropI
+  , getIntervalFile
+  , iLength
+  , fuseIntervals
+  , setIntervalFile
 
     -- * Ranges
   , Range
-  , Range'
+  , Range'(..)
   , rangeInvariant
   , consecutiveAndSeparated
   , intervalsToRange
+  , intervalToRange
   , rangeIntervals
   , rangeFile
   , rightMargin
@@ -52,6 +53,7 @@ module Agda.Syntax.Position
   , rStart, rStart'
   , rEnd, rEnd'
   , rangeToInterval
+  , rangeToIntervalWithFile
   , continuous
   , continuousPerLine
   , PrintRange(..)
@@ -68,15 +70,14 @@ module Agda.Syntax.Position
   , fuseRanges
   , beginningOf
   , beginningOfFile
-
-    -- * Tests
-  , tests
+  , interleaveRanges
   ) where
 
 import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
 import Control.Monad
+import Control.Monad.Writer (runWriter, Writer, tell)
 
 import Data.Foldable (Foldable)
 import qualified Data.Foldable as Fold
@@ -95,14 +96,11 @@ import Data.Void
 
 import GHC.Generics (Generic)
 
-import Test.QuickCheck.All
-
-import Agda.Utils.FileName hiding (tests)
-import Agda.Utils.List hiding (tests)
+import Agda.Utils.FileName
+import Agda.Utils.List
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
 import Agda.Utils.Pretty
-import Agda.Utils.QuickCheck
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -173,6 +171,11 @@ setIntervalFile :: a -> Interval' b -> Interval' a
 setIntervalFile f (Interval p1 p2) =
   Interval (p1 { srcFile = f }) (p2 { srcFile = f })
 
+-- | Gets the 'srcFile' component of the interval. Because of the invariant,
+--   they are both the same.
+getIntervalFile :: Interval' a -> a
+getIntervalFile = srcFile . iStart
+
 -- | Converts a file name and two positions to an interval.
 posToInterval ::
   a -> PositionWithoutFile -> PositionWithoutFile -> Interval' a
@@ -615,7 +618,7 @@ instance (Pretty a, HasRange a) => Pretty (PrintRange a) where
   pretty (PrintRange a) = pretty a <+> parens (text "at" <+> pretty (getRange a))
 
 {--------------------------------------------------------------------------
-    Functions on postitions and ranges
+    Functions on positions and ranges
  --------------------------------------------------------------------------}
 
 -- | The first position in a file: position 1, line 1, column 1.
@@ -825,216 +828,36 @@ beginningOfFile (Range f _) = posToRange' f p p
 withRangeOf :: (SetRange t, HasRange u) => t -> u -> t
 x `withRangeOf` y = setRange (getRange y) x
 
-------------------------------------------------------------------------
--- Test suite
-
--- | The positions corresponding to the interval. The positions do not
--- refer to characters, but to the positions between characters, with
--- zero pointing to the position before the first character.
-iPositions :: Interval' a -> Set Int32
-iPositions i = Set.fromList [posPos (iStart i) .. posPos (iEnd i)]
-
--- | The positions corresponding to the range, including the
--- end-points.
-rPositions :: Range' a -> Set Int32
-rPositions r = Set.unions (map iPositions $ rangeIntervals r)
-
--- | Constructs the least interval containing all the elements in the
--- set.
-makeInterval :: Set Int32 -> Set Int32
-makeInterval s
-  | Set.null s = Set.empty
-  | otherwise  = Set.fromList [Set.findMin s .. Set.findMax s]
-
-prop_iLength :: Interval' Integer -> Bool
-prop_iLength i = iLength i >= 0
-
-prop_startPos' :: Bool
-prop_startPos' = positionInvariant (startPos' ())
-
-prop_startPos :: Maybe AbsolutePath -> Bool
-prop_startPos = positionInvariant . startPos
-
-prop_noRange :: Bool
-prop_noRange = rangeInvariant (noRange :: Range)
-
-prop_takeI_dropI :: Interval' Integer -> Property
-prop_takeI_dropI i =
-  forAll (choose (0, toInteger $ iLength i)) $ \n ->
-    let s = genericReplicate n ' '
-        t = takeI s i
-        d = dropI s i
-    in
-    intervalInvariant t &&
-    intervalInvariant d &&
-    fuseIntervals t d == i
-
-prop_posToRange' ::
-  Integer -> PositionWithoutFile -> PositionWithoutFile -> Bool
-prop_posToRange' f p1 p2 =
-  rangeInvariant (posToRange' f p1 p2)
-
-prop_posToRange :: Position' Integer -> Position' Integer -> Bool
-prop_posToRange p1 p2 =
-  rangeInvariant (posToRange p1 (p2 { srcFile = srcFile p1 }))
-
-prop_intervalToRange :: Integer -> IntervalWithoutFile -> Bool
-prop_intervalToRange f i = rangeInvariant (intervalToRange f i)
-
-rangeToIntervalPropertyTemplate ::
-  Ord a =>
-  (Range' Integer -> Maybe (Interval' a)) ->
-  Range' Integer -> Bool
-rangeToIntervalPropertyTemplate r2i r = case r2i r of
-  Nothing -> r == noRange
-  Just i  ->
-    r /= noRange
-      &&
-    intervalInvariant i
-      &&
-    iPositions i == makeInterval (rPositions r)
-
-prop_rangeToIntervalWithFile :: Range' Integer -> Bool
-prop_rangeToIntervalWithFile =
-  rangeToIntervalPropertyTemplate rangeToIntervalWithFile
-
-prop_rangeToInterval :: Range' Integer -> Bool
-prop_rangeToInterval =
-  rangeToIntervalPropertyTemplate rangeToInterval
-
-prop_continuous :: Range -> Bool
-prop_continuous r =
-  rangeInvariant cr &&
-  rPositions cr == makeInterval (rPositions r)
-  where cr = continuous r
-
-prop_continuousPerLine :: Range -> Bool
-prop_continuousPerLine r =
-  rangeInvariant r'
-    &&
-  distinct lineNumbers
-    &&
-  rangeFile r' == rangeFile r
+-- | Interleaves two streams of ranged elements
+--
+--   It will report the conflicts as a list of conflicting pairs.
+--   In case of conflict, the element with the earliest start position
+--   is placed first. In case of a tie, the element with the earliest
+--   ending position is placed first. If both tie, the element from the
+--   first list is placed first.
+interleaveRanges :: (HasRange a) => [a] -> [a] -> ([a], [(a,a)])
+interleaveRanges as bs = runWriter$ go as bs
   where
-  r' = continuousPerLine r
+    go []         as = return as
+    go as         [] = return as
+    go as@(a:as') bs@(b:bs') =
+      let ra = getRange a
+          rb = getRange b
 
-  lineNumbers = concatMap lines (rangeIntervals r')
-    where
-    lines i | s == e    = [s]
-            | otherwise = [s, e]
-      where
-      s = posLine (iStart i)
-      e = posLine (iEnd   i)
+          ra0 = rStart ra
+          rb0 = rStart rb
 
-prop_fuseIntervals :: Interval' Integer -> Property
-prop_fuseIntervals i1 =
-  forAll (intervalInSameFileAs i1) $ \i2 ->
-    let i = fuseIntervals i1 i2 in
-    intervalInvariant i &&
-    iPositions i ==
-      makeInterval (Set.union (iPositions i1) (iPositions i2))
+          ra1 = rEnd ra
+          rb1 = rEnd rb
+      in
+      if ra1 <= rb0 then
+        (a:) <$> go as' bs
+      else if rb1 <= ra0 then
+        (b:) <$> go as bs'
+      else do
+        tell [(a,b)]
+        if ra0 < rb0 || (ra0 == rb0 && ra1 <= rb1) then
+          (a:) <$> go as' bs
+        else
+          (b:) <$> go as bs'
 
-prop_fuseRanges :: Range -> Property
-prop_fuseRanges r1 =
-  forAll (rangeInSameFileAs r1) $ \r2 ->
-    let r = fuseRanges r1 r2 in
-    rangeInvariant r
-      &&
-    rPositions r == Set.union (rPositions r1) (rPositions r2)
-
-prop_beginningOf :: Range -> Bool
-prop_beginningOf r = rangeInvariant (beginningOf r)
-
-prop_beginningOfFile :: Range -> Bool
-prop_beginningOfFile r = rangeInvariant (beginningOfFile r)
-
-instance Arbitrary a => Arbitrary (Position' a) where
-  arbitrary = do
-    srcFile       <- arbitrary
-    Positive pos' <- arbitrary
-    let pos  = fromInteger pos'
-        line = pred pos `div` 10 + 1
-        col  = pred pos `mod` 10 + 1
-    return (Pn {srcFile = srcFile, posPos = pos,
-                posLine = line, posCol = col })
-
--- | Generates an interval located in the same file as the given
--- interval.
-
-intervalInSameFileAs ::
-  (Arbitrary a, Ord a) => Interval' a -> Gen (Interval' a)
-intervalInSameFileAs i =
-  setIntervalFile (srcFile $ iStart i) <$>
-    (arbitrary :: Gen IntervalWithoutFile)
-
-prop_intervalInSameFileAs :: Interval' Integer -> Property
-prop_intervalInSameFileAs i =
-  forAll (intervalInSameFileAs i) $ \i' ->
-    intervalInvariant i' &&
-    srcFile (iStart i) == srcFile (iStart i')
-
--- | Generates a range located in the same file as the given
--- range (if possible).
-
-rangeInSameFileAs :: (Arbitrary a, Ord a) => Range' a -> Gen (Range' a)
-rangeInSameFileAs NoRange      = arbitrary
-rangeInSameFileAs (Range f is) = do
-  Range _f is <- arbitrary `suchThat` (not . null)
-  return $ Range (f `asTypeOf` _f) is
-
-prop_rangeInSameFileAs :: Range' Integer -> Property
-prop_rangeInSameFileAs r =
-  forAll (rangeInSameFileAs r) $ \r' ->
-    rangeInvariant r'
-      &&
-    case (r, r') of
-      (NoRange, _)            -> True
-      (Range f _, Range f' _) -> f == f'
-      (Range _ _, NoRange)    -> False
-
-instance (Arbitrary a, Ord a) => Arbitrary (Interval' a) where
-  arbitrary = do
-    (p1, p2 :: Position' a) <- liftM2 (,) arbitrary arbitrary
-    let [p1', p2'] = sort [p1, p2 { srcFile = srcFile p1 }]
-    return (Interval p1' p2')
-
-instance (Ord a, Arbitrary a) => Arbitrary (Range' a) where
-  arbitrary = do
-    f <- arbitrary
-    intervalsToRange f . fuse . sort <$> arbitrary
-    where
-    fuse (i1 : i2 : is)
-      | iEnd i1 >= iStart i2 = fuse (fuseIntervals i1 i2 : is)
-      | otherwise            = i1 : fuse (i2 : is)
-    fuse is = is
-
-instance CoArbitrary a => CoArbitrary (Position' a)
-instance CoArbitrary a => CoArbitrary (Interval' a)
-instance CoArbitrary a => CoArbitrary (Range' a)
-
-prop_positionInvariant :: Position' Integer -> Bool
-prop_positionInvariant = positionInvariant
-
-prop_intervalInvariant :: Interval' Integer -> Bool
-prop_intervalInvariant = intervalInvariant
-
-prop_rangeInvariant :: Range -> Bool
-prop_rangeInvariant = rangeInvariant
-
-instance Show (Position' Integer) where show = show . fmap Strict.Just
-instance Show (Interval' Integer) where show = show . fmap Strict.Just
-instance Show (Range'    Integer) where show = show . fmap Strict.Just
-
-------------------------------------------------------------------------
--- * All tests
-------------------------------------------------------------------------
-
--- Template Haskell hack to make the following $quickCheckAll work
--- under ghc-7.8.
-return [] -- KEEP!
-
--- | Test suite.
-tests :: IO Bool
-tests = do
-  putStrLn "Agda.Syntax.Position"
-  $quickCheckAll
