@@ -1,10 +1,7 @@
 {-# LANGUAGE CPP                 #-}
 {-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE GADTs               #-}
-{-# LANGUAGE LambdaCase          #-}
-{-# LANGUAGE RankNTypes          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections       #-}
 
 {-| This module defines the notion of a scope and operations on scopes.
 -}
@@ -73,12 +70,12 @@ data NameSpaceId
 type ScopeNameSpaces = [(NameSpaceId, NameSpace)]
 
 localNameSpace :: Access -> NameSpaceId
-localNameSpace PublicAccess  = PublicNS
-localNameSpace PrivateAccess = PrivateNS
-localNameSpace OnlyQualified = OnlyQualifiedNS
+localNameSpace PublicAccess    = PublicNS
+localNameSpace PrivateAccess{} = PrivateNS
+localNameSpace OnlyQualified   = OnlyQualifiedNS
 
 nameSpaceAccess :: NameSpaceId -> Access
-nameSpaceAccess PrivateNS = PrivateAccess
+nameSpaceAccess PrivateNS = PrivateAccess Inserted
 nameSpaceAccess _         = PublicAccess
 
 -- | Get a 'NameSpace' from 'Scope'.
@@ -662,7 +659,7 @@ flattenScope ms scope =
                [ qual c (build ms' exportedNamesInScope $ moduleScope a)
                | (c, a) <- Map.toList $ scopeImports root
                , let -- get the suffixes of c in ms
-                     ms' = mapMaybe (maybePrefixMatch $ C.qnameParts c) ms
+                     ms' = mapMaybe (List.stripPrefix $ C.qnameParts c) ms
                , not $ null ms' ]
     qual c = Map.mapKeys (q c)
       where
@@ -791,7 +788,13 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ maybeToList to
 
 -- * Inverse look-up
 
-data AllowAmbiguousNames = AmbiguousAnything | AmbiguousConstructors | AmbiguousNothing
+data AllowAmbiguousNames
+  = AmbiguousAnything
+      -- ^ Used for instance arguments to check whether a name is in scope,
+      --   but we do not care whether is is ambiguous
+  | AmbiguousConProjs
+      -- ^ Ambiguous constructors or projections.
+  | AmbiguousNothing
   deriving (Eq)
 
 isNameInScope :: A.QName -> ScopeInfo -> Bool
@@ -803,7 +806,7 @@ isNameInScope q scope =
 --   Sort by length, shortest first.
 
 inverseScopeLookup :: Either A.ModuleName A.QName -> ScopeInfo -> [C.QName]
-inverseScopeLookup = inverseScopeLookup' AmbiguousConstructors
+inverseScopeLookup = inverseScopeLookup' AmbiguousConProjs
 
 inverseScopeLookup' :: AllowAmbiguousNames -> Either A.ModuleName A.QName -> ScopeInfo -> [C.QName]
 inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ] $
@@ -829,7 +832,11 @@ inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ]
     unique (_:_:_) = False
 
     unambiguousModule q = amb == AmbiguousAnything || unique (scopeLookup q scope :: [AbstractModule])
-    unambiguousName   q = amb == AmbiguousAnything || unique xs || amb == AmbiguousConstructors && all ((ConName ==) . anameKind) xs
+    unambiguousName   q = amb == AmbiguousAnything
+        || unique xs
+        || amb == AmbiguousConProjs
+           && (all ((ConName ==) . anameKind) xs
+            || all ((FldName ==) . anameKind) xs)
       where xs = scopeLookup q scope
 
 recomputeInverseScopeMaps :: ScopeInfo -> ScopeInfo
