@@ -1,16 +1,6 @@
 {-# LANGUAGE CPP                    #-}
-{-# LANGUAGE FlexibleInstances      #-}
-{-# LANGUAGE FunctionalDependencies #-}
-{-# LANGUAGE LambdaCase             #-}
-{-# LANGUAGE MultiParamTypeClasses  #-}
-{-# LANGUAGE PatternGuards          #-}
-{-# LANGUAGE TupleSections          #-}
 {-# LANGUAGE UndecidableInstances   #-}
 
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
-
 -- {-# OPTIONS -fwarn-unused-binds #-}
 
 {-| The translation of abstract syntax to concrete syntax has two purposes.
@@ -67,6 +57,7 @@ import Agda.Utils.Functor
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
+import Agda.Utils.Singleton
 import Agda.Utils.Tuple
 import Agda.Utils.Pretty (prettyShow)
 
@@ -155,11 +146,18 @@ lookupQName ambCon x = do
     [] -> do
       let y = qnameToConcrete x
       if isUnderscore y
+        -- -- || any (isUnderscore . A.nameConcrete) (A.mnameToList $ A.qnameModule x)
         then return y
         else return $ C.Qual (C.Name noRange [Id empty]) y
         -- this is what happens for names that are not in scope (private names)
 
 lookupModule :: A.ModuleName -> AbsToCon C.QName
+lookupModule (A.MName []) = return $ C.QName $ C.Name noRange [Id "-1"]
+  -- Andreas, 2016-10-10 it can happen that we have an empty module name
+  -- for instance when we query the current module inside the
+  -- frontmatter or module telescope of the top level module.
+  -- In this case, we print it as an invalid module name.
+  -- (Should only affect debug printing.)
 lookupModule x =
     do  scope <- asks currentScope
         case inverseScopeLookupModule x scope of
@@ -233,18 +231,23 @@ withInfixDecls = foldr (.) id . map (uncurry withInfixDecl)
 
 -- Dealing with private definitions ---------------------------------------
 
+-- | Add @abstract@, @private@, @instance@ modifiers.
 withAbstractPrivate :: DefInfo -> AbsToCon [C.Declaration] -> AbsToCon [C.Declaration]
 withAbstractPrivate i m =
-    case (defAccess i, defAbstract i) of
-        (PublicAccess, ConcreteDef) -> m
-        (p,a)                       ->
-            do  ds <- m
-                return $ abst a $ priv p $ ds
+    priv (defAccess i)
+      . abst (defAbstract i)
+      . addInstanceB (defInstance i == InstanceDef)
+      <$> m
     where
-        priv PrivateAccess ds = [ C.Private (getRange ds) ds ]
-        priv _ ds             = ds
-        abst AbstractDef ds   = [ C.Abstract (getRange ds) ds ]
-        abst _ ds             = ds
+        priv (PrivateAccess UserWritten)
+                         ds = [ C.Private  (getRange ds) UserWritten ds ]
+        priv _           ds = ds
+        abst AbstractDef ds = [ C.Abstract (getRange ds) ds ]
+        abst ConcreteDef ds = ds
+
+addInstanceB :: Bool -> [C.Declaration] -> [C.Declaration]
+addInstanceB True  ds = [ C.InstanceB (getRange ds) ds ]
+addInstanceB False ds = ds
 
 -- The To Concrete Class --------------------------------------------------
 
@@ -347,7 +350,7 @@ instance ToConcrete A.Name C.Name where
   bindToConcrete x = bindName x
 
 instance ToConcrete A.QName C.QName where
-  toConcrete = lookupQName AmbiguousConstructors
+  toConcrete = lookupQName AmbiguousConProjs
 
 instance ToConcrete A.ModuleName C.QName where
   toConcrete = lookupModule
@@ -357,7 +360,10 @@ instance ToConcrete A.ModuleName C.QName where
 instance ToConcrete A.Expr C.Expr where
     toConcrete (Var x)            = Ident . C.QName <$> toConcrete x
     toConcrete (Def x)            = Ident <$> toConcrete x
-    toConcrete (Proj x)           = Ident <$> toConcrete x
+    toConcrete (Proj ProjPrefix (AmbQ (x:_))) = Ident <$> toConcrete x
+    toConcrete (Proj _          (AmbQ (x:_))) =
+      C.Dot (getRange x) . Ident <$> toConcrete x
+    toConcrete Proj{}             = __IMPOSSIBLE__
     toConcrete (A.Macro x)        = Ident <$> toConcrete x
     toConcrete (Con (AmbQ (x:_))) = Ident <$> toConcrete x
     toConcrete (Con (AmbQ []))    = __IMPOSSIBLE__
@@ -380,6 +386,9 @@ instance ToConcrete A.Expr C.Expr where
       C.Underscore (getRange i) $
         prettyShow . NamedMeta (metaNameSuggestion i) . MetaId . metaId <$> metaNumber i
 
+    toConcrete (A.Dot i e) =
+      C.Dot (getRange i) <$> toConcrete e
+
     toConcrete e@(A.App i e1 e2)    =
         tryToRecoverOpApp e
         -- or fallback to App
@@ -462,7 +471,7 @@ instance ToConcrete A.Expr C.Expr where
                            Irrelevant -> addDot a e
                            NonStrict  -> addDot a (addDot a e)
                            _          -> e
-            addDot a e = Dot (getRange a) e
+            addDot a e = C.Dot (getRange a) e
             mkArg (Arg info e) = case getHiding info of
                                           Hidden    -> HiddenArg   (getRange e) (unnamed e)
                                           Instance  -> InstanceArg (getRange e) (unnamed e)
@@ -567,8 +576,9 @@ instance ToConcrete A.TypedBinding C.TypedBinding where
 instance ToConcrete LetBinding [C.Declaration] where
     bindToConcrete (LetBind i info x t e) ret =
         bindToConcrete x $ \x ->
-        do (t,(e, [], [], [])) <- toConcrete (t, A.RHS e)
-           ret [ C.TypeSig info x t
+        do (t,(e, [], [], [])) <- toConcrete (t, A.RHS e Nothing)
+           ret $ addInstanceB (getHiding info == Instance) $
+               [ C.TypeSig info x t
                , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
                              e C.NoWhere False
                ]
@@ -577,7 +587,7 @@ instance ToConcrete LetBinding [C.Declaration] where
         p <- toConcrete p
         e <- toConcrete e
         ret [ C.FunClause (C.LHS p [] [] []) (C.RHS e) NoWhere False ]
-    bindToConcrete (LetApply i x modapp _ _ _) ret = do
+    bindToConcrete (LetApply i x modapp _ _) ret = do
       x' <- unqualify <$> toConcrete x
       modapp <- toConcrete modapp
       let r = getRange modapp
@@ -602,7 +612,9 @@ instance ToConcrete AsWhereDecls WhereClause where
   bindToConcrete (AsWhereDecls ds@[Section _ am _ _]) ret = do
     ds' <- declsToConcrete ds
     cm  <- unqualify <$> lookupModule am
-    let wh' = (if isNoName cm then AnyWhere else SomeWhere cm) $ ds'
+    -- Andreas, 2016-07-08 I put PublicAccess in the following SomeWhere
+    -- Should not really matter for printing...
+    let wh' = (if isNoName cm then AnyWhere else SomeWhere cm PublicAccess) $ ds'
     local (openModule' am defaultImportDir id) $ ret wh'
   bindToConcrete (AsWhereDecls ds) ret =
     ret . AnyWhere =<< declsToConcrete ds
@@ -633,7 +645,8 @@ declsToConcrete :: [A.Declaration] -> AbsToCon [C.Declaration]
 declsToConcrete ds = mergeSigAndDef . concat <$> toConcrete ds
 
 instance ToConcrete A.RHS (C.RHS, [C.Expr], [C.Expr], [C.Declaration]) where
-    toConcrete (A.RHS e) = do
+    toConcrete (A.RHS e (Just c)) = return (C.RHS c, [], [], [])
+    toConcrete (A.RHS e Nothing) = do
       e <- toConcrete e
       return (C.RHS e, [], [], [])
     toConcrete A.AbsurdRHS = return (C.AbsurdRHS, [], [], [])
@@ -658,14 +671,15 @@ instance ToConcrete (Maybe A.QName) (Maybe C.Name) where
 instance ToConcrete (Constr A.Constructor) C.Declaration where
   toConcrete (Constr (A.ScopedDecl scope [d])) =
     withScope scope $ toConcrete (Constr d)
-  toConcrete (Constr (A.Axiom _ i info x t)) = do
+  toConcrete (Constr (A.Axiom _ i info Nothing x t)) = do
     x' <- unsafeQNameToName <$> toConcrete x
     t' <- toConcreteTop t
     return $ C.TypeSig info x' t'
+  toConcrete (Constr (A.Axiom _ _ _ (Just _) _ _)) = __IMPOSSIBLE__
   toConcrete (Constr d) = head <$> toConcrete d
 
 instance ToConcrete a C.LHS => ToConcrete (A.Clause' a) [C.Declaration] where
-  toConcrete (A.Clause lhs rhs wh catchall) =
+  toConcrete (A.Clause lhs _ rhs wh catchall) =
       bindToConcrete lhs $ \lhs ->
         case lhs of
           C.LHS p wps _ _ -> do
@@ -693,12 +707,16 @@ instance ToConcrete A.Declaration [C.Declaration] where
   toConcrete (ScopedDecl scope ds) =
     withScope scope (declsToConcrete ds)
 
-  toConcrete (Axiom _ i info x t) = do
+  toConcrete (Axiom _ i info mp x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteTop t
-      return [C.Postulate (getRange i) [C.TypeSig info x' t']]
+      return $
+        (case mp of
+           Nothing   -> []
+           Just occs -> [C.Pragma (PolarityPragma noRange x' occs)]) ++
+        [C.Postulate (getRange i) [C.TypeSig info x' t']]
 
   toConcrete (A.Field i x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
@@ -752,7 +770,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
       ds <- declsToConcrete ds
       return [ C.Module (getRange i) x (concat tel) ds ]
 
-  toConcrete (A.Apply i x modapp _ _ _) = do
+  toConcrete (A.Apply i x modapp _ _) = do
     x  <- unsafeQNameToName <$> toConcrete x
     modapp <- toConcrete modapp
     let r = getRange modapp
@@ -799,7 +817,7 @@ instance ToConcrete RangeAndPragma C.Pragma where
     A.BuiltinPragma b e      -> C.BuiltinPragma r b <$> toConcrete e
     A.BuiltinNoDefPragma b x -> C.BuiltinPragma r b . C.Ident <$>
       toConcrete x
-    A.RewritePragma x        -> C.RewritePragma r <$> toConcrete x
+    A.RewritePragma x        -> C.RewritePragma r . singleton <$> toConcrete x
     A.CompiledTypePragma x hs -> do
       x <- toConcrete x
       return $ C.CompiledTypePragma r x hs
@@ -827,11 +845,10 @@ instance ToConcrete RangeAndPragma C.Pragma where
     A.CompiledDataUHCPragma x crd crcs -> do
       x <- toConcrete x
       return $ C.CompiledDataUHCPragma r x crd crcs
-    A.NoSmashingPragma x -> C.NoSmashingPragma r <$> toConcrete x
     A.StaticPragma x -> C.StaticPragma r <$> toConcrete x
     A.InlinePragma x -> C.InlinePragma r <$> toConcrete x
     A.DisplayPragma f ps rhs ->
-      C.DisplayPragma r <$> toConcrete (A.DefP (PatRange noRange) f ps) <*> toConcrete rhs
+      C.DisplayPragma r <$> toConcrete (A.DefP (PatRange noRange) (AmbQ [f]) ps) <*> toConcrete rhs
 
 -- Left hand sides --------------------------------------------------------
 
@@ -845,14 +862,41 @@ instance ToConcrete A.LHS C.LHS where
           ret $ C.LHS lhs wps [] []
 
 instance ToConcrete A.LHSCore C.Pattern where
-    bindToConcrete = bindToConcrete . lhsCoreToPattern
+  bindToConcrete = bindToConcrete . lhsCoreToPattern
 
 appBrackets' :: [arg] -> Precedence -> Bool
 appBrackets' []    _   = False
 appBrackets' (_:_) ctx = appBrackets ctx
 
--- TODO: bind variables properly
+newtype BindingPattern = BindingPat A.Pattern
+newtype FreshName = FreshenName A.Name
+
+instance ToConcrete FreshName A.Name where
+  bindToConcrete (FreshenName x) ret = bindToConcrete x $ \ y -> ret x{ nameConcrete = y }
+
+-- Takes care of freshening and binding pattern variables, but doesn't actually
+-- translate anything to Concrete.
+instance ToConcrete BindingPattern A.Pattern where
+  bindToConcrete (BindingPat p) ret =
+    case p of
+      A.VarP x               -> bindToConcrete (FreshenName x) $ ret . A.VarP
+      A.WildP{}              -> ret p
+      A.ProjP{}              -> ret p
+      A.AbsurdP{}            -> ret p
+      A.LitP{}               -> ret p
+      A.DotP{}               -> ret p
+      A.ConP i c args        -> bindToConcrete ((map . fmap . fmap) BindingPat args) $ ret . A.ConP i c
+      A.DefP i f args        -> bindToConcrete ((map . fmap . fmap) BindingPat args) $ ret . A.DefP i f
+      A.PatternSynP i f args -> bindToConcrete ((map . fmap . fmap) BindingPat args) $ ret . A.PatternSynP i f
+      A.RecP i args          -> bindToConcrete ((map . fmap)        BindingPat args) $ ret . A.RecP i
+      A.AsP i x p            -> bindToConcrete (FreshenName x) $ \ x ->
+                                bindToConcrete (BindingPat p)  $ \ p ->
+                                ret (A.AsP i x p)
+
 instance ToConcrete A.Pattern C.Pattern where
+  bindToConcrete p ret = do
+    prec <- currentPrecedence
+    bindToConcrete (BindingPat p) (ret <=< withPrecedence prec . toConcrete)
   toConcrete p =
     case p of
       A.VarP x ->
@@ -862,8 +906,14 @@ instance ToConcrete A.Pattern C.Pattern where
         return $ C.WildP (getRange i)
 
       A.ConP i (AmbQ []) args        -> __IMPOSSIBLE__
-      p@(A.ConP i xs@(AmbQ (x:_)) args) -> tryOp x (A.ConP i xs) args
-      p@(A.DefP i x args)               -> tryOp x (A.DefP i x)  args
+      A.ConP i xs@(AmbQ (x:_)) args  -> tryOp x (A.ConP i xs) args
+
+      A.ProjP _ _ (AmbQ []) -> __IMPOSSIBLE__
+      A.ProjP i ProjPrefix xs@(AmbQ (x:_)) -> C.IdentP <$> toConcrete x
+      A.ProjP i _ xs@(AmbQ (x:_)) -> C.DotP (getRange x) UserWritten . C.Ident <$> toConcrete x
+
+      A.DefP i (AmbQ []) _ -> __IMPOSSIBLE__
+      A.DefP i xs@(AmbQ (x:_)) args -> tryOp x (A.DefP i xs)  args
 
       A.AsP i x p -> do
         (x, p) <- toConcreteCtx ArgumentCtx (x,p)
@@ -878,20 +928,24 @@ instance ToConcrete A.Pattern C.Pattern where
       A.LitP l ->
         return $ C.LitP l
 
-      A.DotP i e  -> do
+      A.DotP i o e -> do
         c <- toConcreteCtx DotPatternCtx e
         case c of
           -- Andreas, 2016-02-04 print ._ pattern as _ pattern,
           -- following the fusing of WildP and ImplicitP.
           C.Underscore{} -> return $ C.WildP $ getRange i
-          _ -> return $ C.DotP (getRange i) c
+          _ -> return $ C.DotP (getRange i) o c
 
       A.PatternSynP i n _ ->
+        -- Ulf, 2016-11-29: This doesn't seem right. The underscore is a list
+        -- of arguments, which we shouldn't really throw away! I guess this
+        -- case is __IMPOSSIBLE__?
         C.IdentP <$> toConcrete n
 
       A.RecP i as ->
         C.RecP (getRange i) <$> mapM (traverse toConcrete) as
     where
+    tryOp :: A.QName -> (A.Patterns -> A.Pattern) -> A.Patterns -> AbsToCon C.Pattern
     tryOp x f args = do
       -- Andreas, 2016-02-04, Issue #1792
       -- To prevent failing of tryToRecoverOpAppP for overapplied operators,
@@ -936,8 +990,10 @@ tryToRecoverOpAppP = recoverOpApp bracketP_ opApp view
 
     view p = case p of
       ConP _ (AmbQ (c:_)) ps -> Just (HdCon c, ps)
-      DefP _ f            ps -> Just (HdDef f, ps)
-      _                      -> Nothing
+      DefP _ (AmbQ (f:_)) ps -> Just (HdDef f, ps)
+      _ -> __IMPOSSIBLE__
+      -- ProjP _ _ (AmbQ (d:_))   -> Just (HdDef d, [])   -- ? Andreas, 2016-04-21
+      -- _                      -> Nothing
 
 recoverOpApp :: (ToConcrete a c, HasRange c)
   => ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c)
