@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE ImplicitParams    #-}
 
 -- | An Abstract domain of relative sizes, i.e., differences
@@ -8,7 +7,7 @@
 
 module Agda.Termination.Order
   ( -- * Structural orderings
-    Order(Mat), decr
+    Order(..), decr
   , increase, decrease
   , (.*.)
   , supremum, infimum
@@ -16,21 +15,19 @@ module Agda.Termination.Order
   , le, lt, unknown, orderMat, collapseO
   , nonIncreasing, decreasing, isDecr
   , NotWorse(..)
-  , tests
+  , isOrder
   ) where
 
 import qualified Data.Foldable as Fold
-import Data.List as List hiding (union, insert)
+import qualified Data.List as List
 
 import Agda.Termination.CutOff
-import Agda.Termination.SparseMatrix as Matrix hiding (tests)
+import Agda.Termination.SparseMatrix as Matrix
 import Agda.Termination.Semiring (HasZero(..), Semiring)
 import qualified Agda.Termination.Semiring as Semiring
 
-import Agda.Utils.PartialOrd hiding (tests)
+import Agda.Utils.PartialOrd
 import Agda.Utils.Pretty
-import Agda.Utils.QuickCheck
-import Agda.Utils.TestHelpers
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -113,19 +110,17 @@ instance NotWorse Order where
 -}
 
 -- | We assume the matrices have the same dimension.
-instance (Ord i) => NotWorse (Matrix i Order) where
+instance (Ord i, HasZero o, NotWorse o) => NotWorse (Matrix i o) where
   m `notWorse` n
     | size m /= size n = __IMPOSSIBLE__
-    | otherwise        = Fold.all isTrue $
-                           zipMatrices onlym onlyn both trivial m n
+    | otherwise        = Fold.and $ zipMatrices onlym onlyn both trivial m n
     where
-      -- If an element is only in @m@, then its 'Unknown' in @n@
-      -- so it gotten better at best, in any case, not worse.
-      onlym o = True     -- @== o `notWorse` Unknown@
-      onlyn o = Unknown `notWorse` o
-      both    = notWorse
-      isTrue  = id
-      trivial = id
+    -- If an element is only in @m@, then its 'Unknown' in @n@
+    -- so it gotten better at best, in any case, not worse.
+    onlym o = True     -- @== o `notWorse` Unknown@
+    onlyn o = zeroElement `notWorse` o
+    both    = notWorse
+    trivial = id  -- @True@ counts as zero as it is neutral for @and@
 
 -- | Raw increase which does not cut off.
 increase :: Int -> Order -> Order
@@ -141,7 +136,7 @@ decrease i o = increase (-i) o
 -- | Smart constructor for @Decr k :: Order@ which cuts off too big values.
 --
 -- Possible values for @k@: @- ?cutoff '<=' k '<=' ?cutoff + 1@.
---
+
 decr :: (?cutoff :: CutOff) => Int -> Order
 decr k = case ?cutoff of
   CutOff c | k < -c -> Unknown
@@ -150,10 +145,10 @@ decr k = case ?cutoff of
 
 -- | Smart constructor for matrix shaped orders, avoiding empty and singleton matrices.
 orderMat :: Matrix Int Order -> Order
-orderMat m | Matrix.isEmpty m  = Decr 0                -- 0x0 Matrix = neutral element
-           | otherwise         = case isSingleton m of
-                                   Just o -> o         -- 1x1 Matrix
-                                   Nothing -> Mat m    -- nxn Matrix
+orderMat m
+ | Matrix.isEmpty m        = Decr 0 -- 0x0 Matrix = neutral element
+ | Just o <- isSingleton m = o      -- 1x1 Matrix
+ | otherwise               = Mat m  -- nxn Matrix
 
 withinCutOff :: (?cutoff :: CutOff) => Int -> Bool
 withinCutOff k = case ?cutoff of
@@ -197,88 +192,47 @@ instance Pretty Order where
   pretty Unknown  = text "?"
   pretty (Mat m)  = text "Mat" <+> pretty m
 
---instance Ord Order where
---    max = maxO
 
-{- instances cannot have implicit arguments?! GHC manual says:
-
-7.8.3.1. Implicit-parameter type constraints
-
-You can't have an implicit parameter in the context of a class or instance declaration. For example, both these declarations are illegal:
-
-  class (?x::Int) => C a where ...
-  instance (?x::a) => Foo [a] where ...
-
-Reason: exactly which implicit parameter you pick up depends on
-exactly where you invoke a function. But the ``invocation'' of
-instance declarations is done behind the scenes by the compiler, so
-it's hard to figure out exactly where it is done. Easiest thing is to
-outlaw the offending types.
-
-instance (?cutoff :: CutOff) => Arbitrary Order where
-  arbitrary = frequency
-    [(20, return Unknown)
-    ,(80, elements [- ?cutoff .. ?cutoff + 1] >>= Decr)
-    ] -- no embedded matrices generated for now.
--}
-instance Arbitrary Order where
-  arbitrary = frequency
-    [(30, return Unknown)
-    ,(70, elements [0,1] >>= return . Decr)
-    ] -- no embedded matrices generated for now.
-
-instance CoArbitrary Order where
-  coarbitrary (Decr k) = variant 0
-  coarbitrary Unknown  = variant 1
-  coarbitrary (Mat m)  = variant 2
-
--- | Multiplication of 'Order's. (Corresponds to sequential
--- composition.)
+-- | Multiplication of 'Order's.
+--   (Corresponds to sequential composition.)
 
 -- I think this funny pattern matching is because overlapping patterns
 -- are producing a warning and thus an error (strict compilation settings)
 (.*.) :: (?cutoff :: CutOff) => Order -> Order -> Order
-Unknown  .*. _         = Unknown
-(Mat m)  .*. Unknown   = Unknown
-(Decr k) .*. Unknown   = Unknown
-(Decr k) .*. (Decr l)  = decr (k + l)
-(Decr 0) .*. (Mat m)   = Mat m
-(Decr k) .*. (Mat m)   = (Decr k) .*. (collapse m)
-(Mat m1) .*. (Mat m2) = if (okM m1 m2) then
-                            Mat $ mul orderSemiring m1 m2
-                        else
-                            (collapse m1) .*. (collapse m2)
+Unknown  .*. _        = Unknown
+(Mat m)  .*. Unknown  = Unknown
+(Decr k) .*. Unknown  = Unknown
+(Decr k) .*. (Decr l) = decr (k + l)
+(Decr 0) .*. (Mat m)  = Mat m
+(Decr k) .*. (Mat m)  = (Decr k) .*. (collapse m)
+(Mat m1) .*. (Mat m2)
+  | okM m1 m2         = Mat $ mul orderSemiring m1 m2
+  | otherwise         = (collapse m1) .*. (collapse m2)
 (Mat m) .*. (Decr 0)  = Mat m
 (Mat m) .*. (Decr k)  = (collapse m) .*. (Decr k)
 
-{- collapse m
+-- | collapse @m@
+--
+-- We assume that @m@ codes a permutation:  each row has at most one column
+-- that is not @Unknown@.
+--
+-- To collapse a matrix into a single value, we take the best value of
+-- each column and multiply them.  That means if one column is all @Unknown@,
+-- i.e., no argument relates to that parameter, then the collapsed value
+-- is also @Unknown@.
+--
+-- This makes order multiplication associative.
 
-We assume that m codes a permutation:  each row has at most one column
-that is not Un.
-
-To collapse a matrix into a single value, we take the best value of
-each column and multiply them.  That means if one column is all Un,
-i.e., no argument relates to that parameter, than the collapsed value
-is also Un.
-
-This makes order multiplication associative.
-
--}
 collapse :: (?cutoff :: CutOff) => Matrix Int Order -> Order
-collapse m = -- if not $ Matrix.matrixInvariant m then __IMPOSSIBLE__ else
-  case toLists $ Matrix.transpose m of
-   [] -> __IMPOSSIBLE__   -- This can never happen if order matrices are generated by the smart constructor
-   m' -> foldl1 (.*.) $ map (foldl1 maxO) m'
-
-{- OLD CODE, does not give associative matrix multiplication:
-collapse :: (?cutoff :: CutOff) => Matrix Int Order -> Order
-collapse m = foldl (.*.) le (Data.Array.elems $ diagonal m)
--}
+collapse m = case toLists $ Matrix.transpose m of
+  [] -> __IMPOSSIBLE__   -- This can never happen if order matrices are generated by the smart constructor
+  m' -> foldl1 (.*.) $ map (foldl1 maxO) m'
 
 collapseO :: (?cutoff :: CutOff) => Order -> Order
 collapseO (Mat m) = collapse m
 collapseO o       = o
 
+-- | Can two matrices be multplied together?
 okM :: Matrix Int Order -> Matrix Int Order -> Bool
 okM m1 m2 = (rows $ size m2) == (cols $ size m1)
 
@@ -288,64 +242,46 @@ okM m1 m2 = (rows $ size m2) == (cols $ size m1)
 supremum :: (?cutoff :: CutOff) => [Order] -> Order
 supremum = foldr maxO Unknown
 
--- | @('Order', 'maxO', '.*.')@ forms a semiring, with 'Unknown' as
--- zero and 'Le' as one.
+-- | @('Order', 'maxO', '.*.')@ forms a semiring,
+--   with 'Unknown' as zero and 'Le' as one.
 
 maxO :: (?cutoff :: CutOff) => Order -> Order -> Order
 maxO o1 o2 = case (o1,o2) of
-               (Decr k, Decr l) -> Decr (max k l) -- cut off not needed, within borders
-               (Unknown,_) -> o2
-               (_,Unknown) -> o1
-               (Mat m1, Mat m2) -> Mat (Matrix.add maxO m1 m2)
-               (Mat m,_) -> maxO (collapse m) o2
-               (_,Mat m) ->  maxO o1 (collapse m)
+  (Decr k, Decr l) -> Decr (max k l) -- cut off not needed, within borders
+  (Unknown, _)     -> o2
+  (_, Unknown)     -> o1
+  (Mat m1, Mat m2) -> Mat (Matrix.add maxO m1 m2)
+  (Mat m, _)       -> maxO (collapse m) o2
+  (_, Mat m)       -> maxO o1 (collapse m)
 
 -- | The infimum of a (non empty) list of 'Order's.
+--   Gets the worst information.
 --  'Unknown' is the least element, thus, dominant.
 infimum :: (?cutoff :: CutOff) => [Order] -> Order
-infimum (o:l) = foldl' minO o l
+infimum (o:l) = List.foldl' minO o l
 infimum []    = __IMPOSSIBLE__
 
+-- | Pick the worst information.
 minO :: (?cutoff :: CutOff) => Order -> Order -> Order
 minO o1 o2 = case (o1,o2) of
-               (Unknown,_) -> Unknown
-               (_,Unknown) -> Unknown
-               (Decr k, Decr l) -> Decr (min k l) -- cut off not needed, within borders
-               (Mat m1, Mat m2) -> if (size m1 == size m2) then
-                                       Mat $ Matrix.intersectWith minO m1 m2
-                                   else
-                                       minO (collapse m1) (collapse m2)
-               (Mat m1,_) -> minO (collapse m1) o2
-               (_,Mat m2) -> minO o1 (collapse m2)
+  (Unknown, _)           -> Unknown
+  (_, Unknown)           -> Unknown
+  (Decr k, Decr l)       -> Decr (min k l) -- cut off not needed, within borders
+  (Mat m1, Mat m2)
+    | size m1 == size m2 -> Mat $ Matrix.intersectWith minO m1 m2
+    | otherwise          -> minO (collapse m1) (collapse m2)
+  (Mat m1, _)            -> minO (collapse m1) o2
+  (_, Mat m2)            -> minO o1 (collapse m2)
 
 
-{- Cannot have implicit arguments in instances.  Too bad!
-
-instance Monoid Order where
-  mempty = Unknown
-  mappend = maxO
-
-instance (cutoff :: Int) => SemiRing Order where
-  multiply = (.*.)
--}
+-- | We use a record for semiring instead of a type class
+--   since implicit arguments cannot occur in instance constraints,
+--   like @instance (?cutoff :: Int) => SemiRing Order@.
 
 orderSemiring :: (?cutoff :: CutOff) => Semiring Order
-orderSemiring =
-  Semiring.Semiring { Semiring.add = maxO
-                    , Semiring.mul = (.*.)
-                    , Semiring.zero = Unknown
---                    , Semiring.one = Le
-                    }
-
-prop_orderSemiring :: (?cutoff :: CutOff) => Order -> Order -> Order -> Bool
-prop_orderSemiring = Semiring.semiringInvariant orderSemiring
-
-------------------------------------------------------------------------
--- All tests
-
-tests :: IO Bool
-tests = runTests "Agda.Termination.Order"
-  [ quickCheck' prop_decr
-  , quickCheck' prop_orderSemiring
-  ]
-  where ?cutoff = DontCutOff -- CutOff 2  -- don't cut off in tests!
+orderSemiring = Semiring.Semiring
+  { Semiring.add  = maxO
+  , Semiring.mul  = (.*.)
+  , Semiring.zero = Unknown
+  -- , Semiring.one  = Le
+  }
