@@ -1,11 +1,7 @@
 {-# LANGUAGE CPP                        #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ImplicitParams             #-}
-{-# LANGUAGE NamedFieldPuns             #-}
-{-# LANGUAGE PatternGuards              #-}
-{-# LANGUAGE StandaloneDeriving         #-}
-{-# LANGUAGE TupleSections              #-}
+{-# LANGUAGE NondecreasingIndentation   #-}
 
 {- Checking for Structural recursion
    Authors: Andreas Abel, Nils Anders Danielsson, Ulf Norell,
@@ -16,12 +12,14 @@
 
 module Agda.Termination.TermCheck
     ( termDecl
+    , termMutual
     , Result, DeBruijnPat
     ) where
 
 import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
+import Control.Monad.Reader
 import Control.Monad.State
 
 import Data.Foldable (toList)
@@ -39,6 +37,7 @@ import Agda.Syntax.Internal.Generic
 import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Position
 import Agda.Syntax.Common
+import Agda.Syntax.Translation.InternalToAbstract ( reifyPatterns )
 
 import Agda.Termination.CutOff
 import Agda.Termination.Monad
@@ -59,6 +58,7 @@ import Agda.TypeChecking.Records -- (isRecordConstructor, isInductiveRecord)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Records
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Datatypes
@@ -78,6 +78,7 @@ import Agda.Utils.Maybe
 import Agda.Utils.Monad -- (mapM', forM', ifM, or2M, and2M)
 import Agda.Utils.Null
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Singleton
 import qualified Agda.Utils.VarSet as VarSet
 
@@ -94,30 +95,26 @@ type Calls = CallGraph CallPath
 type Result = [TerminationError]
 
 -- | Entry point: Termination check a single declaration.
+--
+--   Precondition: 'envMutualBlock' must be set correctly.
 
-termDecl :: MutualId -> A.Declaration -> TCM Result
-termDecl mid d = inTopContext $ ignoreAbstractMode $ termDecl' mid d
-
-
--- | Termination check a sequence of declarations.
-
-termDecls :: MutualId -> [A.Declaration] -> TCM Result
-termDecls mid ds = concat <$> mapM (termDecl' mid) ds
+termDecl :: A.Declaration -> TCM Result
+termDecl d = inTopContext $ termDecl' d
 
 
 -- | Termination check a single declaration
 --   (without necessarily ignoring @abstract@).
 
-termDecl' :: MutualId -> A.Declaration -> TCM Result
-termDecl' mid d = case d of
+termDecl' :: A.Declaration -> TCM Result
+termDecl' d = case d of
     A.Axiom {}            -> return mempty
     A.Field {}            -> return mempty
     A.Primitive {}        -> return mempty
     A.Mutual _ ds
       | [A.RecSig{}, A.RecDef _ _ _ _ _ _ _ rds] <- unscopeDefs ds
-                          -> termDecls mid rds
-    A.Mutual i ds         -> termMutual mid i ds
-    A.Section _ _ _ ds    -> termDecls mid ds
+                          -> termDecls rds
+    A.Mutual i ds         -> termMutual $ getNames ds
+    A.Section _ _ _ ds    -> termDecls ds
         -- section structure can be ignored as we are termination checking
         -- definitions lifted to the top-level
     A.Apply {}            -> return mempty
@@ -126,10 +123,10 @@ termDecl' mid d = case d of
     A.Open {}             -> return mempty
     A.PatternSynDef {}    -> return mempty
         -- open and pattern synonym defs are just artifacts from the concrete syntax
-    A.ScopedDecl _ ds     -> termDecls mid ds
+    A.ScopedDecl scope ds -> {- withScope_ scope $ -} termDecls ds
         -- scope is irrelevant as we are termination checking Syntax.Internal
     A.RecSig{}            -> return mempty
-    A.RecDef _ r _ _ _ _ _ ds -> termDecls mid ds
+    A.RecDef _ r _ _ _ _ _ ds -> termDecls ds
     -- These should all be wrapped in mutual blocks
     A.FunDef{}      -> __IMPOSSIBLE__
     A.DataSig{}     -> __IMPOSSIBLE__
@@ -137,33 +134,58 @@ termDecl' mid d = case d of
     A.UnquoteDecl{} -> __IMPOSSIBLE__
     A.UnquoteDef{}  -> __IMPOSSIBLE__
   where
+    termDecls ds = concat <$> mapM termDecl' ds
+
     unscopeDefs = concatMap unscopeDef
 
     unscopeDef (A.ScopedDecl _ ds) = unscopeDefs ds
     unscopeDef d = [d]
 
+    -- The mutual names mentioned in the abstract syntax
+    -- for symbols that need to be termination-checked.
+    getNames = concatMap getName
+    getName (A.FunDef i x delayed cs)   = [x]
+    getName (A.RecDef _ _ _ _ _ _ _ ds) = getNames ds
+    getName (A.Mutual _ ds)             = getNames ds
+    getName (A.Section _ _ _ ds)        = getNames ds
+    getName (A.ScopedDecl _ ds)         = getNames ds
+    getName (A.UnquoteDecl _ _ xs _)    = xs
+    getName (A.UnquoteDef _ xs _)       = xs
+    getName _                           = []
 
--- | Termination check a bunch of mutually inductive recursive definitions.
 
-termMutual :: MutualId -> Info.MutualInfo -> [A.Declaration] -> TCM Result
-termMutual mid i ds =
+-- | Entry point: Termination check the current mutual block.
 
-  -- We set the range to avoid panics when printing error messages.
-  setCurrentRange i $ do
+termMutual
+  :: [QName]
+     -- ^ The function names defined in this block on top-level.
+     --   (For error-reporting only.)
+  -> TCM Result
+termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempty) $ {-else-}
+ disableDestructiveUpdate $ do
 
   -- Get set of mutually defined names from the TCM.
   -- This includes local and auxiliary functions introduced
   -- during type-checking.
+  mid <- fromMaybe __IMPOSSIBLE__ <$> asks envMutualBlock
   mutualBlock <- lookupMutualBlock mid
-  let allNames = Set.elems mutualBlock
+  let allNames = Set.elems $ mutualNames mutualBlock
+      names    = if null names0 then allNames else names0
+      i        = mutualInfo mutualBlock
       -- Andreas, 2014-03-26
       -- Keeping recursion check after experiments on the standard lib.
       -- Seems still to save 1s.
       -- skip = return False
       -- No need to term-check if the declarations are acyclic!
       skip = not <$> do
+        -- Andreas, 2016-10-01 issue #2231
+        -- Recursivity checker has to see through abstract definitions!
+        ignoreAbstractMode $ do
         billTo [Benchmark.Termination, Benchmark.RecCheck] $ recursive allNames
 
+  -- We set the range to avoid panics when printing error messages.
+  setCurrentRange i $ do
+
   reportSLn "term.mutual" 10 $ "Termination checking " ++ show allNames
 
   -- NO_TERMINATION_CHECK
@@ -203,25 +225,14 @@ termMutual mid i ds =
          -- Else: Old check, all at once.
          (runTerm $ termMutual')
 
-     -- record result of termination check in signature
+     -- Record result of termination check in signature.
+     -- If there are some termination errors, we collect them in
+     -- the state and mark the definition as non-terminating so
+     -- that it does not get unfolded
      let terminates = null res
      forM_ allNames $ \ q -> setTerminates q terminates
      return res
 
-  where
-    getName (A.FunDef i x delayed cs) = [x]
-    getName (A.RecDef _ _ _ _ _ _ _ ds) = concatMap getName ds
-    getName (A.Mutual _ ds)       = concatMap getName ds
-    getName (A.Section _ _ _ ds)  = concatMap getName ds
-    getName (A.ScopedDecl _ ds)   = concatMap getName ds
-    getName (A.UnquoteDecl _ _ xs _) = xs
-    getName (A.UnquoteDef _ xs _)    = xs
-    getName _                     = []
-
-    -- the mutual names mentioned in the abstract syntax
-    names = concatMap getName ds
-
-
 -- | @termMutual'@ checks all names of the current mutual block,
 --   henceforth called @allNames@, for termination.
 --
@@ -411,10 +422,9 @@ typeEndsInDef t = liftTCM $ do
 --   Only the latter depends on the choice whether we
 --   consider dot patterns or not.
 termDef :: QName -> TerM Calls
-termDef name = terSetCurrent name $ do
+termDef name = terSetCurrent name $ inConcreteOrAbstractMode name $ \ def -> do
 
   -- Retrieve definition
-  def <- liftTCM $ getConstInfo name
   let t = defType def
 
   liftTCM $ reportSDoc "term.def.fun" 5 $
@@ -516,7 +526,7 @@ termToDBP t = ifNotM terGetUseDotPatterns (return unusedVar) $ {- else -} do
         t <- constructorForm t
         case ignoreSharing t of
           -- Constructors.
-          Con c args  -> ConDBP (conName c) <$> mapM (loop . unArg) args
+          Con c ci args -> ConDBP (conName c) <$> mapM (loop . unArg) args
           Def s [Apply arg] | Just s == suc
                       -> ConDBP s . (:[]) <$> loop (unArg arg)
           DontCare t  -> __IMPOSSIBLE__  -- removed by stripAllProjections
@@ -544,37 +554,24 @@ stripCoConstructors p = do
     ProjDBP{} -> return p
 
 -- | Masks all non-data/record type patterns if --without-K.
+--   See issue #1023.
 maskNonDataArgs :: [DeBruijnPat] -> TerM [Masked DeBruijnPat]
 maskNonDataArgs ps = zipWith mask ps <$> terGetMaskArgs
   where
     mask p@ProjDBP{} _ = Masked False p
     mask p           d = Masked d     p
 
--- | cf. 'TypeChecking.Coverage.Match.buildMPatterns'
-openClause :: Permutation -> [Pattern] -> ClauseBody -> TerM ([DeBruijnPat], Maybe Term)
-openClause perm ps body = do
-  -- invariant: xs has enough variables for the body
-  unless (permRange perm == genericLength xs) __IMPOSSIBLE__
-  dbps <- evalStateT (mapM build ps) xs
-  return . (dbps,) $ case body `applys` map var xs of
-    NoBody -> Nothing
-    Body v -> Just v
-    _      -> __IMPOSSIBLE__
+-- | Convert patterns to the format used by the termination checker
+--   TODO: refactor termination checking to use regular patterns
+convertPatterns :: [DeBruijnPattern] -> TerM [DeBruijnPat]
+convertPatterns ps = mapM build ps
   where
-    -- TODO: express build using numberPatVars
-    -- length of the telescope
-    n    = size perm
-    -- the variables as a map from the body variables to the clause telescope
-    xs   = permPicks $ flipP $ invertP __IMPOSSIBLE__ perm
-
-    tick = do x : xs <- get; put xs; return x
-
-    build :: Pattern -> StateT [Nat] TerM DeBruijnPat
-    build (VarP _)        = VarDBP <$> tick
+    build :: DeBruijnPattern -> TerM DeBruijnPat
+    build (VarP x)        = return $ VarDBP $ dbPatVarIndex x
     build (ConP con _ ps) = ConDBP (conName con) <$> mapM (build . namedArg) ps
-    build (DotP t)        = tick *> do lift $ termToDBP t
+    build (DotP t)        = termToDBP t
     build (LitP l)        = return $ LitDBP l
-    build (ProjP d)       = return $ ProjDBP d
+    build (ProjP o d)     = return $ ProjDBP o d
 
 -- | Extract recursive calls from one clause.
 termClause :: Clause -> TerM Calls
@@ -586,21 +583,21 @@ termClause clause = do
 
 termClause' :: Clause -> TerM Calls
 termClause' clause = do
-  cl @ Clause { clauseTel  = tel
-              , clauseBody = body } <- introHiddenLambdas clause
-  let argPats' = clausePats cl
-      perm     = clausePerm cl
+  cl <- introHiddenLambdas clause
+  let tel      = clauseTel cl
+      argPats' = namedClausePats cl
+      body     = clauseBody cl
   liftTCM $ reportSDoc "term.check.clause" 25 $ vcat
     [ text "termClause"
     , nest 2 $ text "tel      =" <+> prettyTCM tel
-    , nest 2 $ text ("perm     = " ++ show perm)
-    -- how to get the following right?
-    -- , nest 2 $ text "argPats' =" <+> do prettyA =<< reifyPatterns tel perm argPats'
+    , nest 2 $ text "argPats' =" <+> do
+      addContext tel $ do
+       aps <- reifyPatterns argPats'
+       fsep $ map prettyA aps
     ]
-  addCtxTel tel $ do
-    ps <- liftTCM $ normalise $ map unArg argPats'
-    (dbpats, res) <- openClause perm (unnumberPatVars ps) body
-    case res of
+  addContext tel $ do
+    dbpats <- convertPatterns $ map namedArg argPats'
+    case body of
       Nothing -> return empty
       Just v -> do
         dbpats <- mapM stripCoConstructors dbpats
@@ -609,13 +606,6 @@ termClause' clause = do
           terSetSizeDepth tel $ do
             reportBody v
             extract v
-  {-
-  -- if we are checking a delayed definition, we treat it as if there were
-  -- a guarding coconstructor (sharp)
-  terModifyGuarded (const $ case delayed of
-        Delayed    -> Order.lt
-        NotDelayed -> Order.le) $ do
-  -}
 
   where
     reportBody :: Term -> TerM ()
@@ -639,8 +629,10 @@ termClause' clause = do
 introHiddenLambdas :: MonadTCM tcm => Clause -> tcm Clause
 introHiddenLambdas clause = liftTCM $ do
   case clause of
-    Clause range ctel ps body Nothing catchall  -> return clause
-    Clause range ctel ps body (Just t) catchall -> do
+    Clause range ctel ps body        Nothing  catchall -> return clause
+
+    Clause range ctel ps Nothing     (Just t) catchall -> return clause
+    Clause range ctel ps (Just body) (Just t) catchall -> do
       case removeHiddenLambdas body of
         -- nobody or no hidden lambdas
         ([], _) -> return clause
@@ -654,26 +646,17 @@ introHiddenLambdas clause = liftTCM $ do
           -- join with lhs telescope
           let ctel' = telFromList $ telToList ctel ++ telToList ttel
               ps'   = raise n ps ++ zipWith toPat (downFrom $ size axs) axs
-          return $ Clause range ctel' ps' body' (Just (t $> t')) catchall
+          return $ Clause range ctel' ps' (Just body') (Just (t $> t')) catchall
   where
-    toPat i (Arg (ArgInfo h r) x) = Arg (ArgInfo h r) $ namedDBVarP i x
-    removeHiddenLambdas :: ClauseBody -> ([Arg ArgName], ClauseBody)
-    removeHiddenLambdas = underBinds $ hlamsToBinds
+    toPat i (Arg info x) = Arg info $ namedDBVarP i x
 
-    hlamsToBinds :: Term -> ([Arg ArgName], ClauseBody)
-    hlamsToBinds v =
+    removeHiddenLambdas :: Term -> ([Arg ArgName], Term)
+    removeHiddenLambdas v =
       case ignoreSharing v of
         Lam info b | getHiding info == Hidden ->
-          let (xs, b') = hlamsToBinds $ unAbs b
-          in  (Arg info (absName b) : xs, Bind $ b' <$ b)
-        _ -> ([], Body v)
-    underBinds :: (Term -> ([a], ClauseBody)) -> ClauseBody -> ([a], ClauseBody)
-    underBinds k body = loop body where
-      loop (Bind b) =
-        let (res, b') = loop $ unAbs b
-        in  (res, Bind $ b' <$ b)
-      loop NoBody = ([], NoBody)
-      loop (Body v) = k v
+          let (xs, b') = removeHiddenLambdas $ unAbs b
+          in  (Arg info (absName b) : xs, b')
+        _ -> ([], v)
 
 -- | Extract recursive calls from expressions.
 class ExtractCalls a where
@@ -723,32 +706,8 @@ instance ExtractCalls Sort where
 instance ExtractCalls Type where
   extract (El s t) = extract (s, t)
 
-{-
--- | Auxiliary type to write an instance of 'ExtractCalls'.
-
-data TerConstructor = TerConstructor
-  { terConsName      :: QName
-    -- ^ Constructor name.
-  , terConsInduction :: Induction
-    -- ^ Should the constructor be treated as inductive or coinductive?
-  , terConsArgs      :: [(Arg Term, Bool)]
-    -- ^ All the arguments,
-    --   and for every argument a boolean which is 'True' iff the
-    --   argument should be viewed as preserving guardedness.
-  }
-
 -- | Extract recursive calls from a constructor application.
 
-instance ExtractCalls TerConstructor where
-  extract (TerConstructor c ind args) = mapM' loopArg args where
-    loopArg (arg, preserves) = terModifyGuarded g' $ extract arg where
-      g' = case (preserves, ind) of
-             (True,  Inductive)   -> id
-             (True,  CoInductive) -> (Order.lt .*.)
-             (False, _)           -> const Order.unknown
--}
-
--- | Extract recursive calls from a constructor application.
 constructor
   :: QName
     -- ^ Constructor name.
@@ -771,7 +730,6 @@ constructor c ind args = do
              (False, _)           -> const Order.unknown
 
 
-
 -- | Handle guardedness preserving type constructor.
 
 guardPresTyCon :: QName -> Elims -> (QName -> Elims -> TerM Calls) -> TerM Calls
@@ -823,8 +781,8 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
     -- thus, we need to use traverseTermM.  Sharing is handled by traverseTermM,
     -- so no ignoreSharing needed here.
     let reduceCon = traverseTermM $ \ t -> case t of
-           Con c vs -> (`apply` vs) <$> reduce (Con c [])  -- make sure we don't reduce the arguments
-           _        -> return t
+           Con c ci vs -> (`apply` vs) <$> reduce (Con c ci [])  -- make sure we don't reduce the arguments
+           _ -> return t
 
     -- Reduce constructors only when this call is actually a recursive one.
     -- es <- liftTCM $ billTo [Benchmark.Termination, Benchmark.Reduce] $ forM es $
@@ -872,8 +830,8 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
          let ifDelayed o | Order.decreasing o && delayed == NotDelayed = Order.le
                          | otherwise                                  = o
          liftTCM $ reportSLn "term.guardedness" 20 $
-           "composing with guardedness " ++ show guarded ++
-           " counting as " ++ show (ifDelayed guarded)
+           "composing with guardedness " ++ prettyShow guarded ++
+           " counting as " ++ prettyShow (ifDelayed guarded)
          cutoff <- terGetCutOff
          let ?cutoff = cutoff
          let matrix' = composeGuardedness (ifDelayed guarded) matrix
@@ -922,7 +880,7 @@ instance ExtractCalls Term where
     case ignoreSharing t of
 
       -- Constructed value.
-      Con ConHead{conName = c} args -> do
+      Con ConHead{conName = c} _ args -> do
 
         -- A constructor preserves the guardedness of all its arguments.
         let argsg = zip args $ repeat True
@@ -1029,9 +987,9 @@ compareArgs es = do
   -- Count the number of coinductive projection(pattern)s in caller and callee.
   -- Only recursive coinductive projections are eligible (Issue 1209).
   projsCaller <- genericLength <$> do
-    filterM (isCoinductiveProjection True) $ mapMaybe (isProjP . getMasked) pats
+    filterM (isCoinductiveProjection True) $ mapMaybe (fmap (head . unAmbQ . snd) . isProjP . getMasked) pats
   projsCallee <- genericLength <$> do
-    filterM (isCoinductiveProjection True) $ mapMaybe isProjElim es
+    filterM (isCoinductiveProjection True) $ mapMaybe (fmap snd . isProjElim) es
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
   let guardedness = decr $ projsCaller - projsCallee
@@ -1053,7 +1011,7 @@ compareArgs es = do
 annotatePatsWithUseSizeLt :: [DeBruijnPat] -> TerM [(Bool,DeBruijnPat)]
 annotatePatsWithUseSizeLt = loop where
   loop [] = return []
-  loop (p@(ProjDBP q) : pats) = ((False,p) :) <$> do projUseSizeLt q $ loop pats
+  loop (p@(ProjDBP _ q) : pats) = ((False,p) :) <$> do projUseSizeLt q $ loop pats
   loop (p : pats) = (\ b ps -> (b,p) : ps) <$> terGetUseSizeLt <*> loop pats
 
 
@@ -1072,7 +1030,10 @@ compareElim e p = do
       , nest 2 $ text $ "p = " ++ show p
       ]
   case (e, getMasked p) of
-    (Proj d, ProjDBP d')           -> compareProj d d'
+    (Proj _ d, ProjDBP _ d')       -> do
+      d  <- liftTCM $ getOriginalProjection d
+      d' <- liftTCM $ getOriginalProjection d'
+      compareProj d d'
     (Proj{}, _         )           -> return Order.unknown
     (Apply{}, ProjDBP{})           -> return Order.unknown
     (Apply arg, _)                 -> compareTerm (unArg arg) p
@@ -1147,7 +1108,7 @@ subPatterns p = case p of
   VarDBP _    -> []
   LitDBP _    -> []
   TermDBP _   -> []
-  ProjDBP _   -> []
+  ProjDBP{}   -> []
 
 compareTerm :: Term -> Masked DeBruijnPat -> TerM Order
 compareTerm t p = do
@@ -1191,9 +1152,9 @@ instance StripAllProjections Elims where
       []             -> return []
       (Apply a : es) -> do
         (:) <$> (Apply <$> stripAllProjections a) <*> stripAllProjections es
-      (Proj p  : es) -> do
+      (Proj o p  : es) -> do
         isP <- isProjectionButNotCoinductive p
-        applyUnless isP (Proj p :) <$> stripAllProjections es
+        applyUnless isP (Proj o p :) <$> stripAllProjections es
 
 instance StripAllProjections Args where
   stripAllProjections = mapM stripAllProjections
@@ -1202,7 +1163,7 @@ instance StripAllProjections Term where
   stripAllProjections t = do
     case ignoreSharing t of
       Var i es   -> Var i <$> stripAllProjections es
-      Con c ts   -> Con c <$> stripAllProjections ts
+      Con c ci ts -> Con c ci <$> stripAllProjections ts
       Def d es   -> Def d <$> stripAllProjections es
       DontCare t -> stripAllProjections t
       _ -> return t
@@ -1268,14 +1229,14 @@ compareTerm' v mp@(Masked m p) = do
     (Con{}, ConDBP c ps) | any (isSubTerm v) ps ->
       decrease <$> offsetFromConstructor c <*> return Order.le
 
-    (Con c ts, ConDBP c' ps) | conName c == c'->
+    (Con c ci ts, ConDBP c' ps) | conName c == c'->
       compareConArgs ts ps
 
-    (Con c [], _) -> return Order.le
+    (Con c ci [], _) -> return Order.le
 
     -- new case for counting constructors / projections
     -- register also increase
-    (Con c ts, _) -> do
+    (Con c ci ts, _) -> do
       increase <$> offsetFromConstructor (conName c)
                <*> (infimum <$> mapM (\ t -> compareTerm' (unArg t) mp) ts)
 
@@ -1286,7 +1247,7 @@ subTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPat -> Order
 subTerm t p = if equal t p then Order.le else properSubTerm t p
   where
     equal (Shared p) dbp = equal (derefPtr p) dbp
-    equal (Con c ts) (ConDBP c' ps) =
+    equal (Con c ci ts) (ConDBP c' ps) =
       and $ (conName c == c')
           : (length ts == length ps)
           : zipWith equal (map unArg ts) ps
