@@ -1,10 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
-
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
 
 module Agda.TypeChecking.Conversion where
 
@@ -23,7 +17,7 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Translation.InternalToAbstract (reify)
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin (constructorForm)
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.CompiledClause (CompiledClauses(Fail))
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.MetaVars.Occurs (killArgs,PruneResult(..))
@@ -59,11 +53,6 @@ import Agda.Utils.Lens
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-{- MOVED to TypeChecking.Level
-mlevel :: TCM (Maybe Term)
-mlevel = liftTCM $ (Just <$> primLevel) `catchError` \_ -> return Nothing
--}
-
 -- | Try whether a computation runs without errors or new constraints
 --   (may create new metas, though).
 --   Restores state upon failure.
@@ -75,8 +64,7 @@ tryConversion = isJust <.> tryConversion'
 --   Return 'Just' the result upon success.
 --   Return 'Nothing' and restore state upon failure.
 tryConversion' :: TCM a -> TCM (Maybe a)
-tryConversion' m = (Just <$> do disableDestructiveUpdate $ noConstraints m)
-  `catchError` \ _ -> return Nothing
+tryConversion' m = tryMaybe $ disableDestructiveUpdate $ noConstraints m
 
 -- | Check if to lists of arguments are the same (and all variables).
 --   Precondition: the lists have the same length.
@@ -247,7 +235,7 @@ compareTerm' cmp a m n =
     proofIrr <- proofIrrelevance
     isSize   <- isJust <$> isSizeType a'
     s        <- reduce $ getSort a'
-    mlvl     <- mlevel
+    mlvl     <- tryMaybe primLevel
     reportSDoc "tc.conv.level" 60 $ nest 2 $ sep
       [ text $ "a'   = " ++ show a'
       , text $ "mlvl = " ++ show mlvl
@@ -309,7 +297,7 @@ compareTerm' cmp a m n =
                   -- No subtyping on record terms
                   c <- getRecordConstructor r
                   -- Record constructors are covariant (see test/succeed/CovariantConstructors).
-                  compareArgs (repeat $ polFromCmp cmp) (telePi_ tel $ sort Prop) (Con c []) m' n'
+                  compareArgs (repeat $ polFromCmp cmp) (telePi_ tel $ sort Prop) (Con c ConOSystem []) m' n'
 
             else compareAtom cmp a' m n
         _ -> compareAtom cmp a' m n
@@ -335,35 +323,24 @@ compareTel t1 t2 cmp tel1 tel2 =
     (EmptyTel, EmptyTel) -> return ()
     (EmptyTel, _)        -> bad
     (_, EmptyTel)        -> bad
-    (ExtendTel dom1@(Dom i1 a1) tel1, ExtendTel dom2@(Dom i2 a2) tel2)
-      | getHiding i1 /= getHiding i2 -> bad
-        -- Andreas, 2011-09-11 do not test r1 == r2 because they could differ
-        -- e.g. one could be Forced and the other Relevant (see fail/UncurryMeta)
-      | otherwise -> do
-          name <- freshName_ (suggest (absName tel1) (absName tel2))
-          let r = max (getRelevance i1) (getRelevance i2) -- take "most irrelevant"
-              dependent = (r /= Irrelevant) && isBinderUsed tel2
--- NEW
-          pid <- newProblem_ $ compareType cmp a1 a2
-          dom <- if dependent
-                 then Dom i1 <$> blockTypeOnProblem a1 pid
-                 else return dom1
-          addContext (name, dom) $ compareTel t1 t2 cmp (absBody tel1) (absBody tel2)
-          stealConstraints pid
+    (ExtendTel dom1@(Dom i1 a1) tel1, ExtendTel dom2@(Dom i2 a2) tel2) -> do
+      compareDom cmp dom1 dom2 tel1 tel2 bad bad $
+        compareTel t1 t2 cmp (absBody tel1) (absBody tel2)
 
 {- OLD, before 2013-05-15
           let checkDom = escapeContext 1 $ compareType cmp a1 a2
               c = TelCmp t1 t2 cmp (absBody tel1) (absBody tel2)
 
-          addCtx name dom1 $
+          addContext (name, dom1) $
             if dependent
             then guardConstraint c checkDom
             else checkDom >> solveConstraint_ c
 -}
   where
     -- Andreas, 2011-05-10 better report message about types
-    bad = typeError $ UnequalTypes cmp t2 t1 -- switch t2 and t1 because of contravariance!
---    bad = typeError $ UnequalTelescopes cmp tel1 tel2
+    bad = typeError $ UnequalTypes cmp t2 t1
+      -- switch t2 and t1 because of contravariance!
+
 
 
 -- | Raise 'UnequalTerms' if there is no hope that by
@@ -542,13 +519,13 @@ compareAtom cmp t m n =
             (Def f es, Def f' es') ->
               unlessM (bothAbsurd f f') $ do
                 trySizeUniv cmp t m n f es f' es'
-            (Con x xArgs, Con y yArgs)
+            (Con x ci xArgs, Con y _ yArgs)
                 | x == y -> do
                     -- Get the type of the constructor instantiated to the datatype parameters.
                     a' <- conType x t
                     -- Constructors are covariant in their arguments
                     -- (see test/succeed/CovariantConstructors).
-                    compareArgs (repeat $ polFromCmp cmp) a' (Con x []) xArgs yArgs
+                    compareArgs (repeat $ polFromCmp cmp) a' (Con x ci []) xArgs yArgs
             _ -> etaInequal cmp t m n -- fixes issue 856 (unsound conversion error)
     where
         -- Andreas, 2013-05-15 due to new postponement strategy, type can now be blocked
@@ -566,44 +543,64 @@ compareAtom cmp t m n =
                 patternViolation
           maybe impossible return =<< getConType c t
         equalFun t1 t2 = case (ignoreSharing t1, ignoreSharing t2) of
-          (Pi dom1@(Dom i1 a1@(El a1s a1t)) b1, Pi (Dom i2 a2) b2)
-            | argInfoHiding i1 /= argInfoHiding i2 -> typeError $ UnequalHiding t1 t2
-            -- Andreas 2010-09-21 compare r1 and r2, but ignore forcing annotations!
-            | not (compareRelevance cmp (ignoreForced $ argInfoRelevance i2)
-                                        (ignoreForced $ argInfoRelevance i1))
-                -> typeError $ UnequalRelevance cmp t1 t2
-            | otherwise -> verboseBracket "tc.conv.fun" 15 "compare function types" $ do
-                reportSDoc "tc.conv.fun" 20 $ nest 2 $ vcat
-                  [ text "t1 =" <+> prettyTCM t1
-                  , text "t2 =" <+> prettyTCM t2 ]
+          (Pi dom1 b1, Pi dom2 b2) -> do
+            verboseBracket "tc.conv.fun" 15 "compare function types" $ do
+              reportSDoc "tc.conv.fun" 20 $ nest 2 $ vcat
+                [ text "t1 =" <+> prettyTCM t1
+                , text "t2 =" <+> prettyTCM t2 ]
+              compareDom cmp dom2 dom1 b1 b2 errH errR $
+                compareType cmp (absBody b1) (absBody b2)
+            where
+            errH = typeError $ UnequalHiding t1 t2
+            errR = typeError $ UnequalRelevance cmp t1 t2
 
-                -- We only need to require a1 == a2 if t2 is a dependent function type.
-                -- If it's non-dependent it doesn't matter what we add to the context.
-                pid <- newProblem_ $ compareType cmp a2 a1
-                dom <- if isBinderUsed b2
-                       then Dom i1 <$> blockTypeOnProblem a1 pid
-                       -- then Dom i1 . El a1s <$> blockTermOnProblem (El Inf $ Sort a1s) a1t pid
-                       else return dom1
-                name <- freshName_ (suggest b1 b2)
-                addContext (name, dom) $ compareType cmp (absBody b1) (absBody b2)
-                stealConstraints pid
-                -- Andreas, 2013-05-15 Now, comparison of codomains is not
-                -- blocked any more by getting stuck on domains.
-                -- Only the domain type in context will be blocked.
-
-{- OLD
+{- OLD, before 2013-05-15
                 let checkDom = escapeContext 1 $ compareType cmp a2 a1
                     conCoDom = TypeCmp cmp (absBody b1) (absBody b2)
                 -- We only need to require a1 == a2 if t2 is a dependent function type.
                 -- If it's non-dependent it doesn't matter what we add to the context.
                 name <- freshName_ (suggest b1 b2)
-                addCtx name dom1 $
+                addContext (name, dom1) $
                   if isBinderUsed b2 -- dependent function type?
                   then guardConstraint conCoDom checkDom
                   else checkDom >> solveConstraint_ conCoDom
 -}
           _ -> __IMPOSSIBLE__
 
+-- | Check whether @a1 `cmp` a2@ and continue in context extended by @a1@.
+compareDom :: Free c
+  => Comparison -- ^ @cmp@ The comparison direction
+  -> Dom Type   -- ^ @a1@  The smaller domain.
+  -> Dom Type   -- ^ @a2@  The other domain.
+  -> Abs b      -- ^ @b1@  The smaller codomain.
+  -> Abs c      -- ^ @b2@  The bigger codomain.
+  -> TCM ()     -- ^ Continuation if mismatch in 'Hiding'.
+  -> TCM ()     -- ^ Continuation if mismatch in 'Relevance'.
+  -> TCM ()     -- ^ Continuation if comparison is successful.
+  -> TCM ()
+compareDom cmp dom1@(Dom i1 a1) dom2@(Dom i2 a2) b1 b2 errH errR cont
+  | getHiding dom1 /= getHiding dom2 = errH
+  -- Andreas 2010-09-21 compare r1 and r2, but ignore forcing annotations!
+  | not $ compareRelevance cmp (ignoreForced $ getRelevance dom1)
+                               (ignoreForced $ getRelevance dom2) = errR
+  | otherwise = do
+      let r = max (getRelevance dom1) (getRelevance dom2)
+              -- take "most irrelevant"
+          dependent = (r /= Irrelevant) && isBinderUsed b2
+      pid <- newProblem_ $ compareType cmp a1 a2
+      dom <- if dependent
+             then Dom i1 <$> blockTypeOnProblem a1 pid
+             else return dom1
+        -- We only need to require a1 == a2 if b2 is dependent
+        -- If it's non-dependent it doesn't matter what we add to the context.
+      name <- freshName_ $ suggest b1 b2
+      addContext (name, dom) $ cont
+      stealConstraints pid
+        -- Andreas, 2013-05-15 Now, comparison of codomains is not
+        -- blocked any more by getting stuck on domains.
+        -- Only the domain type in context will be blocked.
+        -- But see issue #1258.
+
 compareRelevance :: Comparison -> Relevance -> Relevance -> Bool
 compareRelevance CmpEq  = (==)
 compareRelevance CmpLeq = (<=)
@@ -652,7 +649,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
       ifBlockedType a (\ m t -> patternViolation) $ \ a -> do
         case ignoreSharing . unEl $ a of
           (Pi (Dom info b) codom) -> do
-            mlvl <- mlevel
+            mlvl <- tryMaybe primLevel
             let freeInCoDom (Abs _ c) = 0 `freeInIgnoringSorts` c
                 freeInCoDom _         = False
                 dependent = (Just (unEl b) /= mlvl) && freeInCoDom codom
@@ -719,12 +716,12 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
             -- __IMPOSSIBLE__
 
     -- case: f == f' are projections
-    (Proj f : els1, Proj f' : els2)
+    (Proj o f : els1, Proj _ f' : els2)
       | f /= f'   -> typeError . GenericError . show =<< prettyTCM f <+> text "/=" <+> prettyTCM f'
       | otherwise -> ifBlockedType a (\ m t -> patternViolation) $ \ a -> do
-        res <- projectTyped v a f -- fails only if f is proj.like but parameters cannot be retrieved
+        res <- projectTyped v a o f -- fails only if f is proj.like but parameters cannot be retrieved
         case res of
-          Just (u, t) -> do
+          Just (_, u, t) -> do
             -- Andreas, 2015-07-01:
             -- The arguments following the principal argument of a projection
             -- are invariant.  (At least as long as we have no explicit polarity
@@ -766,7 +763,6 @@ compareIrrelevant t v w = do
       let rel  = getMetaRelevance mv
           inst = case mvInstantiation mv of
                    InstV{} -> True
-                   InstS{} -> True
                    _       -> False
       reportSDoc "tc.conv.irr" 20 $ vcat
         [ nest 2 $ text $ "rel  = " ++ show rel
@@ -774,7 +770,10 @@ compareIrrelevant t v w = do
         ]
       if not (irrelevantOrUnused rel) || inst
         then fallback
-        else assignE DirEq x es w $ compareIrrelevant t
+        -- Andreas, 2016-08-08, issue #2131:
+        -- Mining for solutions for irrelevant metas is not definite.
+        -- Thus, in case of error, leave meta unsolved.
+        else (assignE DirEq x es w $ compareIrrelevant t) `catchError` \ _ -> fallback
         -- the value of irrelevant or unused meta does not matter
     try v w fallback = fallback
 
@@ -979,9 +978,11 @@ leqLevel a b = liftTCM $ do
   -- See case for `same term` below.
   a <- normalise a
   b <- normalise b
-  catchConstraint (LevelCmp CmpLeq a b) $ leqView a b
+  leqView a b
   where
-    leqView a@(Max as) b@(Max bs) = do
+    -- Andreas, 2016-09-28
+    -- If we have to postpone a constraint, then its simplified form!
+    leqView a@(Max as) b@(Max bs) = catchConstraint (LevelCmp CmpLeq a b) $ do
       reportSDoc "tc.conv.nat" 30 $
         text "compareLevelView" <+>
           sep [ text (show a) <+> text "=<"
@@ -995,7 +996,11 @@ leqLevel a b = liftTCM $ do
         ([], _) -> ok
 
         -- as ≤ 0
-        (as, [])  -> sequence_ [ equalLevel' (Max [a]) (Max []) | a <- as ]
+        (as, [])              -> sequence_ [ equalLevel' (Max [a]) (Max []) | a <- as ]
+        (as, [ClosedLevel 0]) -> sequence_ [ equalLevel' (Max [a]) (Max []) | a <- as ]
+           -- Andreas, 2016-09-28, @[ClosedLevel 0]@ is possible if we come from case
+           -- "reduce constants" where we run @subtr@ on both sides.
+           -- See test/Succeed/LevelMetaLeqZero.agda.
 
         -- as ≤ [b]
         (as@(_:_:_), [b]) -> sequence_ [ leqView (Max [a]) (Max [b]) | a <- as ]
@@ -1041,11 +1046,14 @@ leqLevel a b = liftTCM $ do
                         [n] -> n
                         _   -> __IMPOSSIBLE__
 
-        -- [a] ≤ [neutral]
-        ([a@(Plus n _)], [b@(Plus m NeutralLevel{})])
-          | m == n -> equalLevel' (Max [a]) (Max [b])
-          -- Andreas, 2014-04-07: This call to equalLevel is ok even if we removed
-          -- subsumed terms from the lhs.
+        -- Andreas, 2016-09-28: This simplification loses the solution lzero.
+        -- Thus, it is invalid.
+        -- See test/Succeed/LevelMetaLeqNeutralLevel.agda.
+        -- -- [a] ≤ [neutral]
+        -- ([a@(Plus n _)], [b@(Plus m NeutralLevel{})])
+        --   | m == n -> equalLevel' (Max [a]) (Max [b])
+        --   -- Andreas, 2014-04-07: This call to equalLevel is ok even if we removed
+        --   -- subsumed terms from the lhs.
 
         -- anything else
         _ -> postpone
