@@ -1,9 +1,5 @@
 {-# LANGUAGE CPP              #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE PatternGuards    #-}
-{-# LANGUAGE TupleSections    #-}
-
-{-# OPTIONS_GHC -fwarn-unused-imports #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 {-| Coverage checking, case splitting, and splitting for refine tactics.
 
@@ -13,9 +9,11 @@ module Agda.TypeChecking.Coverage
   ( SplitClause(..), clauseToSplitClause, fixTarget
   , Covering(..), splitClauses
   , coverageCheck
+  , isCovered
   , splitClauseWithAbsurd
   , splitLast
   , splitResult
+  , normaliseProjP
   ) where
 
 import Prelude hiding (null)
@@ -28,18 +26,23 @@ import Control.Applicative hiding (empty)
 #endif
 
 import Data.List hiding (null)
+import Data.Monoid (Any(..))
+import Data.Map (Map)
+import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Common
+import Agda.Syntax.Position
+import Agda.Syntax.Literal
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Exception
 
-import Agda.TypeChecking.Rules.LHS.Problem (flexibleVarFromHiding)
+import Agda.TypeChecking.Rules.LHS.Problem (allFlexVars)
 import Agda.TypeChecking.Rules.LHS.Unify
 
 import Agda.TypeChecking.Coverage.Match
@@ -48,10 +51,10 @@ import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Datatypes (getConForm)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Substitute.Pattern
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Records (isRecordType)
+import Agda.TypeChecking.Records
 import Agda.TypeChecking.Telescope
+import Agda.TypeChecking.MetaVars
 
 import Agda.Interaction.Options
 
@@ -64,6 +67,7 @@ import Agda.Utils.Null
 import Agda.Utils.Permutation
 import Agda.Utils.Size
 import Agda.Utils.Tuple
+import Agda.Utils.Lens
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -87,6 +91,9 @@ data SplitClause = SClause
     --   'splitResult', which does not split on a variable,
     --   should reset it to the identity 'idS', lest it be
     --   applied to 'scTarget' again, leading to Issue 1294.
+  , scModuleParameterSub :: ModuleParamDict
+    -- ^ We need to keep track of the module parameter substitutions for the
+    -- clause for the purpose of inferring missing instance clauses.
   , scTarget :: Maybe (Arg Type)
     -- ^ The type of the rhs, living in context 'scTel'.
     --   This invariant is broken before calls to 'fixTarget';
@@ -107,12 +114,13 @@ data Covering = Covering
 splitClauses :: Covering -> [SplitClause]
 splitClauses (Covering _ qcs) = map snd qcs
 
--- | Create a split clause from a clause in internal syntax.
+-- | Create a split clause from a clause in internal syntax. Used by make-case.
 clauseToSplitClause :: Clause -> SplitClause
 clauseToSplitClause cl = SClause
   { scTel    = clauseTel  cl
   , scPats   = namedClausePats cl
   , scSubst  = idS  -- Andreas, 2014-07-15  TODO: Is this ok?
+  , scModuleParameterSub = Map.empty
   , scTarget = clauseType cl
   }
 
@@ -125,13 +133,22 @@ coverageCheck f t cs = do
   let -- n             = arity
       -- xs            = variable patterns fitting lgamma
       n            = size gamma
-      xs           = teleNamedArgs gamma
+      xs           =  map (setOrigin Inserted) $ teleNamedArgs gamma
+      -- The initial module parameter substitutions need to be weakened by the
+      -- number of arguments that aren't module parameters.
+  fv           <- getDefFreeVars f
+  moduleParams <- raise (n - fv) <$> use stModuleParameters
       -- construct the initial split clause
-      sc           = SClause gamma xs idS $ Just $ defaultArg a
-  reportSDoc "tc.cover.top" 10 $ vcat
-    [ text $ "Coverage checking " ++ show f
-    , nest 2 $ vcat $ map (text . show . clausePats) cs
-    ]
+  let sc = SClause gamma xs idS moduleParams $ Just $ defaultArg a
+
+  reportSDoc "tc.cover.top" 10 $ do
+    let prCl cl = addContext (clauseTel cl) $
+                  prettyTCMPatternList $ namedClausePats cl
+    vcat
+      [ text $ "Coverage checking " ++ show f ++ " with patterns:"
+      , nest 2 $ vcat $ map prCl cs
+      ]
+
   -- used = actually used clauses for cover
   -- pss  = uncovered cases
   (splitTree, used, pss) <- cover f cs sc
@@ -142,39 +159,49 @@ coverageCheck f t cs = do
   -- report an error if there are uncovered cases
   unless (null pss) $
       setCurrentRange cs $
-        typeError $ CoverageFailure f (map (map (fmap namedThing)) pss)
+        typeError $ CoverageFailure f pss
   -- is = indices of unreachable clauses
   let is = Set.toList $ Set.difference (Set.fromList [0..genericLength cs - 1]) used
   -- report an error if there are unreachable clauses
   unless (null is) $ do
       let unreached = map (cs !!) is
       setCurrentRange unreached $
-        typeError $ UnreachableClauses f (map clausePats unreached)
+        typeError $ UnreachableClauses f $ map namedClausePats unreached
   return splitTree
 
+-- | Top-level function for eliminating redundant clauses in the interactive
+--   case splitter
+isCovered :: QName -> [Clause] -> SplitClause -> TCM Bool
+isCovered f cs sc = do
+  (_, _, missing) <- cover f cs sc
+  return $ null missing
 
 -- | @cover f cs (SClause _ _ ps _) = return (splitTree, used, pss)@.
 --   checks that the list of clauses @cs@ covers the given split clause.
 --   Returns the @splitTree@, the @used@ clauses, and missing cases @pss@.
 cover :: QName -> [Clause] -> SplitClause ->
          TCM (SplitTree, Set Nat, [[NamedArg DeBruijnPattern]])
-cover f cs sc@(SClause tel ps _ target) = do
+cover f cs sc@(SClause tel ps _ _ target) = do
   reportSDoc "tc.cover.cover" 10 $ vcat
     [ text "checking coverage of pattern:"
     , nest 2 $ text "tel  =" <+> prettyTCM tel
-    , nest 2 $ text "ps   =" <+> text (show ps)
+    , nest 2 $ text "ps   =" <+> do addContext tel $ prettyTCMPatternList ps
     ]
-  let ups = map (fmap namedThing) ps
   exactSplitEnabled <- optExactSplit <$> pragmaOptions
-  case match cs ups of
-    Yes (i,mps)
-     | not exactSplitEnabled || (clauseCatchall (cs !! i) || all isTrivialMPattern mps)
+  cs' <- normaliseProjP cs
+  case match cs' ps of
+    Yes (i,(mps,ls0))
+     | not exactSplitEnabled || (clauseCatchall (cs !! i) || all isTrivialPattern mps)
      -> do
       reportSLn "tc.cover.cover" 10 $ "pattern covered by clause " ++ show i
       -- Check if any earlier clauses could match with appropriate literals
-      let is = [ j | (j, c) <- zip [0..i-1] cs, matchLits c ups ]
-      reportSLn "tc.cover.cover"  10 $ "literal matches: " ++ show is
-      return (SplittingDone (size tel), Set.fromList (i : is), [])
+      let lsis = mapMaybe (\(j,c) -> (,j) <$> matchLits c ps) $ zip [0..i-1] cs
+      reportSLn "tc.cover.cover"  10 $ "literal matches: " ++ show lsis
+      -- Andreas, 2016-10-08, issue #2243 (#708)
+      -- If we have several literal matches with the same literals
+      -- only take the first matching clause of these.
+      let is = Map.elems $ Map.fromListWith min $ (ls0,i) : lsis
+      return (SplittingDone (size tel), Set.fromList is, [])
 
      | otherwise -> do
          reportSDoc "tc.cover.cover" 10 $ vcat
@@ -186,25 +213,24 @@ cover f cs sc@(SClause tel ps _ target) = do
 
     No        ->  do
       reportSLn "tc.cover" 20 $ "pattern is not covered"
-      return (SplittingDone (size tel), Set.empty, [ps])
+      case fmap getHiding target of
+        Just h | h == Instance -> do
+          -- Ulf, 2016-10-31: For now we only infer instance clauses. It would
+          -- make sense to do it also for hidden, but since the value of a
+          -- hidden clause is expected to be forced by later clauses, it's too
+          -- late to add it now. If it was inferrable we would have gotten a
+          -- type error before getting to this point.
+          inferMissingClause f sc
+          return (SplittingDone (size tel), Set.empty, [])
+        _ -> return (SplittingDone (size tel), Set.empty, [ps])
 
-    -- case: split into projection patterns
-    BlockP   -> do
-      reportSLn "tc.cover" 20 $ "blocked by projection pattern"
-      -- if we want to split projections, but have no target type, we give up
+    -- We need to split!
+    -- If all clauses have an unsplit copattern, we try that first.
+    Block res bs -> tryIf (getAny res) splitRes $ do
       let done = return (SplittingDone (size tel), Set.empty, [ps])
-      caseMaybeM (splitResult f sc) done $ \ (Covering n scs) -> do
-        (projs, (trees, useds, psss)) <- mapSnd unzip3 . unzip <$> do
-          mapM (traverseF $ cover f cs <=< (snd <.> fixTarget)) scs
-          -- OR:
-          -- forM scs $ \ (proj, sc') -> (proj,) <$> do
-          --   cover f cs =<< do
-          --     snd <$> fixTarget sc'
-        let tree = SplitAt n $ zip projs trees
-        return (tree, Set.unions useds, concat psss)
-
-    -- case: split on variable
-    Block bs -> do
+      if null bs then done else do
+      -- Otherwise, if there are variables to split, we try them
+      -- in the order determined by a split strategy.
       reportSLn "tc.cover.strategy" 20 $ "blocking vars = " ++ show bs
       -- xs is a non-empty lists of blocking variables
       -- try splitting on one of them
@@ -234,53 +260,88 @@ cover f cs sc@(SClause tel ps _ target) = do
             , nest 2 $ vcat
               [ text "n   = " <+> text (show n)
               , text "scs = " <+> prettyTCM scs
-              , text "ups = " <+> text (show ups)
+              , text "ps  = " <+> text (show ps)
               ]
             ]
-          let trees' = zipWith (etaRecordSplits (unArg n) ups) scs trees
+          let trees' = zipWith (etaRecordSplits (unArg n) ps) scs trees
               tree   = SplitAt n trees'
           return (tree, Set.unions useds, concat psss)
 
   where
+    tryIf :: Monad m => Bool -> m (Maybe a) -> m a -> m a
+    tryIf True  me m = fromMaybeM m me
+    tryIf False me m = m
+
+    -- Try to split result
+    splitRes :: TCM (Maybe (SplitTree, Set Nat, [[NamedArg DeBruijnPattern]]))
+    splitRes = do
+      reportSLn "tc.cover" 20 $ "blocked by projection pattern"
+      -- forM is a monadic map over a Maybe here
+      mcov <- splitResult f sc
+      Trav.forM mcov $ \ (Covering n scs) -> do
+        -- If result splitting was successful, continue coverage checking.
+        (projs, (trees, useds, psss)) <- mapSnd unzip3 . unzip <$> do
+          mapM (traverseF $ cover f cs <=< (snd <.> fixTarget)) scs
+          -- OR:
+          -- forM scs $ \ (proj, sc') -> (proj,) <$> do
+          --   cover f cs =<< do
+          --     snd <$> fixTarget sc'
+        let tree = SplitAt n $ zip projs trees
+        return (tree, Set.unions useds, concat psss)
+
     gatherEtaSplits :: Int -> SplitClause
-                    -> [Arg DeBruijnPattern] -> [Arg DeBruijnPattern]
+                    -> [NamedArg DeBruijnPattern] -> [NamedArg DeBruijnPattern]
     gatherEtaSplits n sc []
        | n >= 0    = __IMPOSSIBLE__ -- we should have encountered the main
                                     -- split by now already
        | otherwise = []
-    gatherEtaSplits n sc (p:ps) = case unArg p of
-      VarP  (i,_)
-       | n == 0    -> case lookupS (scSubst sc) i of -- this is the main split
+    gatherEtaSplits n sc (p:ps) = case namedArg p of
+      VarP x
+       | n == 0    -> case p' of -- this is the main split
            VarP  _      -> __IMPOSSIBLE__
            DotP  _      -> __IMPOSSIBLE__
-           ConP  _ _ qs ->
-             map (fmap namedThing) qs ++ gatherEtaSplits (-1) sc ps
+           ConP  _ _ qs -> qs ++ gatherEtaSplits (-1) sc ps
            LitP  _      -> __IMPOSSIBLE__
-           ProjP _      -> __IMPOSSIBLE__
+           ProjP{}      -> __IMPOSSIBLE__
        | otherwise ->
-           (p $> lookupS (scSubst sc) i) : gatherEtaSplits (n-1) sc ps
+           updateNamedArg (\ _ -> p') p : gatherEtaSplits (n-1) sc ps
+        where p' = lookupS (scSubst sc) $ dbPatVarIndex x
       DotP  _      -> p : gatherEtaSplits (n-1) sc ps -- count dot patterns
-      ConP  _ _ qs -> gatherEtaSplits n sc (map (fmap namedThing) qs ++ ps)
+      ConP  _ _ qs -> gatherEtaSplits n sc (qs ++ ps)
       LitP  _      -> gatherEtaSplits n sc ps
-      ProjP _      -> gatherEtaSplits n sc ps
+      ProjP{}      -> gatherEtaSplits n sc ps
 
-    addEtaSplits :: Int -> [Arg DeBruijnPattern] -> SplitTree -> SplitTree
+    addEtaSplits :: Int -> [NamedArg DeBruijnPattern] -> SplitTree -> SplitTree
     addEtaSplits k []     t = t
-    addEtaSplits k (p:ps) t = case unArg p of
+    addEtaSplits k (p:ps) t = case namedArg p of
       VarP  _       -> addEtaSplits (k+1) ps t
       DotP  _       -> addEtaSplits (k+1) ps t
-      ConP c cpi nqs ->
-        let qs = map (fmap namedThing) nqs
-            t' = [(conName c , addEtaSplits k (qs ++ ps) t)]
-        in  SplitAt (p $> k) t'
+      ConP c cpi qs -> SplitAt (p $> k) [(conName c , addEtaSplits k (qs ++ ps) t)]
       LitP  _       -> __IMPOSSIBLE__
-      ProjP _       -> __IMPOSSIBLE__
+      ProjP{}       -> __IMPOSSIBLE__
 
-    etaRecordSplits :: Int -> [Arg DeBruijnPattern] -> (QName,SplitClause)
+    etaRecordSplits :: Int -> [NamedArg DeBruijnPattern] -> (QName,SplitClause)
                     -> SplitTree -> (QName,SplitTree)
     etaRecordSplits n ps (q , sc) t =
       (q , addEtaSplits 0 (gatherEtaSplits n sc ps) t)
 
+inferMissingClause :: QName -> SplitClause -> TCM ()
+inferMissingClause f (SClause tel ps _ mpsub (Just t)) = setCurrentRange f $ do
+  reportSDoc "tc.cover.infer" 20 $ addContext tel $ text "Trying to infer right-hand side of type" <+> prettyTCM t
+  cl <- addContext tel $ withModuleParameters mpsub $ do
+    (x, rhs) <- case getHiding t of
+                  Instance  -> newIFSMeta "" (unArg t)
+                  Hidden    -> __IMPOSSIBLE__
+                  NotHidden -> __IMPOSSIBLE__
+    return $ Clause { clauseRange     = noRange
+                    , clauseTel       = tel
+                    , namedClausePats = ps
+                    , clauseBody      = Just rhs
+                    , clauseType      = Just t
+                    , clauseCatchall  = False }
+  addClauses f [cl]
+inferMissingClause _ (SClause _ _ _ _ Nothing) = __IMPOSSIBLE__
+
 splitStrategy :: BlockingVars -> Telescope -> TCM BlockingVars
 splitStrategy bs tel = return $ updateLast clearBlockingVarCons xs
   -- Make sure we do not insists on precomputed coverage when
@@ -335,26 +396,31 @@ isDatatype ind at = do
 --   if target becomes a function type.
 --   Returns the domains of the function type (if any).
 fixTarget :: SplitClause -> TCM (Telescope, SplitClause)
-fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scTarget = target } =
+fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParameterSub = mpsub, scTarget = target } =
   caseMaybe target (return (empty, sc)) $ \ a -> do
     reportSDoc "tc.cover.target" 20 $ sep
       [ text "split clause telescope: " <+> prettyTCM sctel
-      , text "old patterns          : " <+> sep (map (prettyTCM . namedArg) ps)
-      , text "substitution          : " <+> text (show sigma)
+      , text "old patterns          : " <+> do
+          addContext sctel $ prettyTCMPatternList ps
+      ]
+    reportSDoc "tc.cover.target" 60 $ sep
+      [ text "substitution          : " <+> text (show sigma)
       ]
     reportSDoc "tc.cover.target" 30 $ sep
       [ text "target type before substitution (variables may be wrong): " <+> do
           addContext sctel $ prettyTCM a
       ]
     TelV tel b <- telView $ applyPatSubst sigma $ unArg a
-    reportSDoc "tc.cover.target" 10 $ sep
+    reportSDoc "tc.cover.target" 15 $ sep
       [ text "target type telescope (after substitution): " <+> do
           addContext sctel $ prettyTCM tel
       , text "target type core      (after substitution): " <+> do
           addContext sctel $ addContext tel $ prettyTCM b
       ]
     let n         = size tel
-        xs        = teleNamedArgs tel
+        -- Andreas, 2016-10-04 issue #2236
+        -- Need to set origin to "Inserted" to avoid printing of hidden patterns.
+        xs        = map (setOrigin Inserted) $ teleNamedArgs tel
         -- Compute new split clause
         sctel'    = telFromList $ telToList (raise n sctel) ++ telToList tel
         -- Dot patterns in @ps@ need to be raised!  (Issue 1298)
@@ -365,6 +431,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scTarget = ta
           , scPats   = ps'
           , scSubst  = wkS n $ sigma -- Should be wkS instead of liftS since
                                      -- variables are only added to new tel.
+          , scModuleParameterSub = applySubst (raiseS n) mpsub
           , scTarget = newTarget
           }
     -- Separate debug printing to find cause of crash (Issue 1374)
@@ -372,9 +439,10 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scTarget = ta
       [ text "new split clause telescope   : " <+> prettyTCM sctel'
       ]
     reportSDoc "tc.cover.target" 30 $ sep
-      [ text "new split clause patterns    : " <+> sep (map (prettyTCM . namedArg) ps')
+      [ text "new split clause patterns    : " <+> do
+          addContext sctel' $ prettyTCMPatternList ps'
       ]
-    reportSDoc "tc.cover.target" 30 $ sep
+    reportSDoc "tc.cover.target" 60 $ sep
       [ text "new split clause substitution: " <+> text (show $ scSubst sc')
       ]
     reportSDoc "tc.cover.target" 30 $ sep
@@ -387,7 +455,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scTarget = ta
       ]
     return $ if n == 0 then (empty, sc { scTarget = newTarget }) else (tel, sc')
 
--- | @computeNeighbourhood delta1 delta2 d pars ixs hix hps con@
+-- | @computeNeighbourhood delta1 delta2 d pars ixs hix tel ps mpsub con@
 --
 --   @
 --      delta1   Telescope before split point
@@ -397,7 +465,9 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scTarget = ta
 --      pars     Data type parameters
 --      ixs      Data type indices
 --      hix      Index of split variable
+--      tel      Telescope for patterns ps
 --      ps       Patterns before doing the split
+--      mpsub    Current module parameter substitutions
 --      con      Constructor to fit into hole
 --   @
 --   @dtype == d pars ixs@
@@ -409,10 +479,12 @@ computeNeighbourhood
   -> Args                         -- ^ Data type parameters.
   -> Args                         -- ^ Data type indices.
   -> Nat                          -- ^ Index of split variable.
-  -> [NamedArg DeBruijnPattern] -- ^ Patterns before doing the split.
+  -> Telescope                    -- ^ Telescope for the patterns.
+  -> [NamedArg DeBruijnPattern]   -- ^ Patterns before doing the split.
+  -> ModuleParamDict              -- ^ Current module parameter substitution.
   -> QName                        -- ^ Constructor to fit into hole.
   -> CoverM (Maybe SplitClause)   -- ^ New split clause if successful.
-computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
+computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
 
   -- Get the type of the datatype
   dtype <- liftTCM $ (`piApply` pars) . defType <$> getConstInfo d
@@ -423,7 +495,7 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
                                        -- Andreas, 2013-11-29 changes nothing!
 {-
   con <- conSrcCon . theDef <$> getConstInfo con
-  Con con [] <- liftTCM $ ignoreSharing <$> (constructorForm =<< normalise (Con con []))
+  Con con ci [] <- liftTCM $ ignoreSharing <$> (constructorForm =<< normalise (Con con ci []))
 -}
 
   -- Get the type of the constructor
@@ -444,21 +516,19 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
       preserve p = p
       gammal = map (fmap preserve) . telToList $ gamma0
       gamma  = telFromList gammal
+      delta1Gamma = delta1 `abstract` gamma
 
-  debugInit con ctype d pars ixs cixs delta1 delta2 gamma ps hix
+  debugInit con ctype d pars ixs cixs delta1 delta2 gamma tel ps hix
 
   -- All variables are flexible
-  -- let flex = [0..size delta1 + size gamma - 1]
-  let gammaDelta1  = gammal ++ telToList delta1
-      makeFlex i d = flexibleVarFromHiding (getHiding d) i
-      flex = zipWith makeFlex [0..] gammaDelta1
+  let flex = allFlexVars delta1Gamma
 
   -- Unify constructor target and given type (in Δ₁Γ)
   let conIxs   = drop (size pars) cixs
       givenIxs = raise (size gamma) ixs
 
   r <- unifyIndices
-         (delta1 `abstract` gamma)
+         delta1Gamma
          flex
          (raise (size gamma) dtype)
          conIxs
@@ -470,7 +540,7 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
     DontKnow{} -> do
       debugCantSplit
       throwException $ CantSplit (conName con) (delta1 `abstract` gamma) conIxs givenIxs
-    Unifies (delta1',rho0) -> do
+    Unifies (delta1',rho0,_) -> do
       debugSubst "rho0" rho0
 
       -- We have Δ₁' ⊢ ρ₀ : Δ₁Γ, so split it into the part for Δ₁ and the part for Γ
@@ -479,7 +549,9 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
       -- Andreas, 2015-05-01  I guess it is fine to use @noConPatternInfo@
       -- as the result of splitting is never used further down the pipeline.
       -- After splitting, Agda reloads the file.
-      let conp    = ConP con noConPatternInfo $ applySubst rho2 $ teleNamedArgs gamma
+      let conp    = ConP con noConPatternInfo $ applySubst rho2 $
+                      map (setOrigin Inserted) $ tele2NamedArgs gamma0 gamma
+          -- Andreas, 2016-09-08, issue #2166: use gamma0 for correct argument names
 
       -- Compute final context and substitution
       let rho3    = consS conp rho1            -- Δ₁' ⊢ ρ₃ : Δ₁(x:D)
@@ -489,26 +561,29 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
 
       debugTel "delta'" delta'
       debugSubst "rho" rho
+      addContext tel $ debugPs ps
 
       -- Apply the substitution
       let ps' = applySubst rho ps
-      debugPlugged ps ps'
+      addContext delta' $ debugPlugged ps'
 
-      return $ Just $ SClause delta' ps' rho Nothing -- target fixed later
+      let mpsub' = applySubst (fromPatternSubstitution rho) mpsub
+
+      return $ Just $ SClause delta' ps' rho mpsub' Nothing -- target fixed later
 
   where
-    debugInit con ctype d pars ixs cixs delta1 delta2 gamma hps hix =
+    debugInit con ctype d pars ixs cixs delta1 delta2 gamma tel ps hix =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $ vcat
         [ text "computeNeighbourhood"
         , nest 2 $ vcat
           [ text "context=" <+> (inTopContext . prettyTCM =<< getContextTelescope)
           , text "con    =" <+> prettyTCM con
           , text "ctype  =" <+> prettyTCM ctype
-          , text "hps    =" <+> text (show hps)
+          , text "ps     =" <+> do addContext tel $ prettyTCMPatternList ps
           , text "d      =" <+> prettyTCM d
           , text "pars   =" <+> prettyList (map prettyTCM pars)
-          , text "ixs    =" <+> addCtxTel delta1 (prettyList (map prettyTCM ixs))
-          , text "cixs   =" <+> do addCtxTel gamma $ prettyList (map prettyTCM cixs)
+          , text "ixs    =" <+> addContext delta1 (prettyList (map prettyTCM ixs))
+          , text "cixs   =" <+> do addContext gamma $ prettyList (map prettyTCM cixs)
           , text "delta1 =" <+> prettyTCM delta1
           , text "delta2 =" <+> prettyTCM delta2
           , text "gamma  =" <+> prettyTCM gamma
@@ -532,21 +607,14 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix ps c = do
         [ text (s ++ " =") <+> prettyTCM tel
         ]
 
-    debugShow s x =
+    debugPs ps =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $ nest 2 $ vcat
-        [ text (s ++ " =") <+> text (show x)
+        [ text "ps     =" <+> prettyTCMPatternList ps
         ]
 
-    debugPlugged ps ps' =
+    debugPlugged ps' =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $ nest 2 $ vcat
-        [ text "ps     =" <+> text (show ps)
-        , text "ps'    =" <+> text (show ps')
-        ]
-
-    debugFinal tel ps =
-      liftTCM $ reportSDoc "tc.cover.split.con" 20 $ nest 2 $ vcat
-        [ text "rtel   =" <+> prettyTCM tel
-        , text "rps    =" <+> text (show ps)
+        [ text "ps'    =" <+> do prettyTCMPatternList ps'
         ]
 
 -- | Entry point from @Interaction.MakeCase@.
@@ -561,7 +629,7 @@ splitClauseWithAbsurd c x = split' Inductive False c (BlockingVar x Nothing)
 
 splitLast :: Induction -> Telescope -> [NamedArg DeBruijnPattern] -> TCM (Either SplitError Covering)
 splitLast ind tel ps = split ind sc (BlockingVar 0 Nothing)
-  where sc = SClause tel ps __IMPOSSIBLE__ Nothing
+  where sc = SClause tel ps empty empty Nothing
 
 -- | @split ind splitClause x = return res@
 --   splits @splitClause@ at pattern var @x@ (de Bruijn index).
@@ -595,7 +663,8 @@ lookupPatternVar SClause{ scTel = tel, scPats = pats } x = arg $>
     if n < 0 then __IMPOSSIBLE__ else n
   where n = if k < 0
             then __IMPOSSIBLE__
-            else fromMaybe __IMPOSSIBLE__ $ permPicks (dbPatPerm pats) !!! k
+            else fromMaybe __IMPOSSIBLE__ $ permPicks perm !!! k
+        perm = fromMaybe __IMPOSSIBLE__ $ dbPatPerm pats
         k = size tel - x - 1
         arg = telVars (size tel) tel !! k
 
@@ -622,9 +691,9 @@ split' :: Induction
        -> SplitClause
        -> BlockingVar
        -> TCM (Either SplitError (Either SplitClause Covering))
-split' ind fixtarget sc@(SClause tel ps _ target) (BlockingVar x mcons) = liftTCM $ runExceptionT $ do
+split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) = liftTCM $ runExceptionT $ do
 
-  debugInit tel x ps
+  debugInit tel x ps mpsub
 
   -- Split the telescope at the variable
   -- t = type of the variable,  Δ₁ ⊢ t
@@ -641,19 +710,24 @@ split' ind fixtarget sc@(SClause tel ps _ target) (BlockingVar x mcons) = liftTC
   ns <- catMaybes <$> do
     forM cons $ \ con ->
       fmap (con,) <$> do
-        msc <- computeNeighbourhood delta1 n delta2 d pars ixs x ps con
+        msc <- computeNeighbourhood delta1 n delta2 d pars ixs x tel ps mpsub con
         if not fixtarget then return msc else do
         Trav.forM msc $ \ sc -> lift $ snd <$> fixTarget sc{ scTarget = target }
 
   case ns of
     []  -> do
-      let ps' = (fmap . fmap . fmap . fmap) (\(y,name) -> if (x==y) then (y,"()") else (y,name)) ps
+      let ps' = (fmap . fmap . fmap . fmap)
+                  (\(DBPatVar name y) -> if (x==y)
+                                         then DBPatVar absurdPatternName y
+                                         else DBPatVar name y)
+                  ps
       return $ Left $ SClause
                { scTel  = telFromList $ telToList delta1 ++
                                         [fmap ((,) "()") t] ++ -- add name "()"
                                         telToList delta2
                , scPats = ps
                , scSubst = idS -- not used anyway
+               , scModuleParameterSub = __IMPOSSIBLE__ -- not used
                , scTarget = Nothing -- not used
                }
 
@@ -681,19 +755,20 @@ split' ind fixtarget sc@(SClause tel ps _ target) (BlockingVar x mcons) = liftTC
 
   where
     inContextOfT :: MonadTCM tcm => tcm a -> tcm a
-    inContextOfT = addCtxTel tel . escapeContext (x + 1)
+    inContextOfT = addContext tel . escapeContext (x + 1)
 
     inContextOfDelta2 :: MonadTCM tcm => tcm a -> tcm a
-    inContextOfDelta2 = addCtxTel tel . escapeContext x
+    inContextOfDelta2 = addContext tel . escapeContext x
 
     -- Debug printing
-    debugInit tel x ps =
-      liftTCM $ reportSDoc "tc.cover.top" 10 $ vcat
+    debugInit tel x ps mpsub = liftTCM $ do
+      reportSDoc "tc.cover.top" 10 $ vcat
         [ text "TypeChecking.Coverage.split': split"
         , nest 2 $ vcat
           [ text "tel     =" <+> prettyTCM tel
           , text "x       =" <+> text (show x)
-          , text "ps      =" <+> text (show ps)
+          , text "ps      =" <+> do addContext tel $ prettyTCMPatternList ps
+          , text "mpsub   =" <+> prettyTCM mpsub
           ]
         ]
 
@@ -713,7 +788,7 @@ split' ind fixtarget sc@(SClause tel ps _ target) (BlockingVar x mcons) = liftTC
 --   otherwise @res == Nothing@.
 --   Note that the empty set of split clauses is returned if the record has no fields.
 splitResult :: QName -> SplitClause -> TCM (Maybe Covering)
-splitResult f sc@(SClause tel ps _ target) = do
+splitResult f sc@(SClause tel ps _ _ target) = do
   reportSDoc "tc.cover.split" 10 $ vcat
     [ text "splitting result:"
     , nest 2 $ text "f      =" <+> text (show f)
@@ -722,33 +797,36 @@ splitResult f sc@(SClause tel ps _ target) = do
   -- if we want to split projections, but have no target type, we give up
   let done = return Nothing
   caseMaybe target done $ \ t -> do
-    isR <- addCtxTel tel $ isRecordType $ unArg t
+    isR <- addContext tel $ isRecordType $ unArg t
     case isR of
       Just (_r, vs, Record{ recFields = fs }) -> do
         reportSDoc "tc.cover" 20 $ sep
           [ text $ "we are of record type _r = " ++ show _r
-          , text   "applied to parameters vs = " <+> (addCtxTel tel $ prettyTCM vs)
+          , text   "applied to parameters vs = " <+> (addContext tel $ prettyTCM vs)
           , text $ "and have fields       fs = " ++ show fs
           ]
-        fvs <- freeVarsToApply f
         let es = patternsToElims ps
-        let self  = defaultArg $ Def f (map Apply fvs) `applyE` es
+        -- Note: module parameters are part of ps
+        let self  = defaultArg $ Def f [] `applyE` es
             pargs = vs ++ [self]
         reportSDoc "tc.cover" 20 $ sep
-          [ text   "we are              self = " <+> (addCtxTel tel $ prettyTCM $ unArg self)
+          [ text   "we are              self = " <+> (addContext tel $ prettyTCM $ unArg self)
           ]
-        let n = defaultArg $ permRange (dbPatPerm ps)
+        let n = defaultArg $ permRange $ fromMaybe __IMPOSSIBLE__ $ dbPatPerm ps
             -- Andreas & James, 2013-11-19 includes the dot patterns!
             -- See test/succeed/CopatternsAndDotPatterns.agda for a case with dot patterns
             -- and copatterns which fails for @n = size tel@ with a broken case tree.
 
+        -- Andreas, 2016-07-22 read the style of projections from the user's lips
+        projOrigin <- ifM (optPostfixProjections <$> pragmaOptions) (return ProjPostfix) (return ProjPrefix)
         Just . Covering n <$> do
           forM fs $ \ proj -> do
             -- compute the new target
             dType <- defType <$> do getConstInfo $ unArg proj -- WRONG: typeOfConst $ unArg proj
             let -- type of projection instantiated at self
-                target' = Just $ proj $> dType `piApply` pargs
-                sc' = sc { scPats   = scPats sc ++ [fmap (Named Nothing . ProjP) proj]
+                target' = Just $ proj $> dType `piApply` pargs      -- Always visible (#2287)
+                projArg = fmap (Named Nothing . ProjP projOrigin) $ setHiding NotHidden proj
+                sc' = sc { scPats   = scPats sc ++ [projArg]
                          , scSubst  = idS
                          , scTarget = target'
                          }
@@ -759,13 +837,14 @@ splitResult f sc@(SClause tel ps _ target) = do
 
 -- | For debugging only.
 instance PrettyTCM SplitClause where
-  prettyTCM (SClause tel pats sigma target) = sep
+  prettyTCM (SClause tel pats sigma mpsub target) = sep
     [ text "SplitClause"
     , nest 2 $ vcat
-      [ text "tel          = " <+> prettyTCM tel
-      , text "pats         = " <+> sep (map (prettyTCM . namedArg) pats)
-      , text "subst        = " <+> (text . show) sigma
-      , text "target       = " <+> do
+      [ text "tel          =" <+> prettyTCM tel
+      , text "pats         =" <+> sep (map (prettyTCM . namedArg) pats)
+      , text "subst        =" <+> (text . show) sigma
+      , text "mpsub        =" <+> prettyTCM mpsub
+      , text "target       =" <+> do
           caseMaybe target empty $ \ t -> do
             addContext tel $ prettyTCM t
       -- Triggers crash (see Issue 1374).
