@@ -26,15 +26,12 @@ import Agda.Utils.Impossible
 getConHead :: QName -> TCM ConHead
 getConHead c = conSrcCon . theDef <$> getConstInfo c
 
--- | Get true constructor as term.
-getConTerm :: QName -> TCM Term
-getConTerm c = flip Con [] <$> getConHead c
-
 -- | Get true constructor with fields, expanding literals to constructors
 --   if possible.
 getConForm :: QName -> TCM ConHead
 getConForm c = do
-  Con con [] <- ignoreSharing <$> do constructorForm =<< getConTerm c
+  ch <- getConHead c
+  Con con _ [] <- ignoreSharing <$> constructorForm (Con ch ConOCon [])
   return con
 
 -- | Augment constructor with record fields (preserve constructor name).
@@ -42,10 +39,6 @@ getConForm c = do
 getOrigConHead :: QName -> TCM ConHead
 getOrigConHead c = setConName c <$> getConHead c
 
--- | Analogous to 'getConTerm'.
-getOrigConTerm :: QName -> TCM Term
-getOrigConTerm c = flip Con [] <$> getOrigConHead c
-
 -- | Get the name of the datatype constructed by a given constructor.
 --   Precondition: The argument must refer to a constructor
 {-# SPECIALIZE getConstructorData :: QName -> TCM QName #-}
@@ -153,6 +146,15 @@ getNumberOfParameters d = do
     Constructor{ conPars = n } -> return $ Just n
     _                          -> return Nothing
 
+-- | Precondition: Name is a data or record type.
+getConstructors :: QName -> TCM [QName]
+getConstructors d = do
+  def <- theDef <$> getConstInfo d
+  case def of
+    Datatype{dataCons = cs} -> return cs
+    Record{recConHead = h}  -> return [conName h]
+    _                       -> __IMPOSSIBLE__
+
 {- UNUSED
 data DatatypeInfo = DataInfo
   { datatypeName   :: QName
