@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleInstances #-}
 
 -- | Tools for 'DisplayTerm' and 'DisplayForm'.
 
@@ -8,6 +7,7 @@ module Agda.TypeChecking.DisplayForm where
 import Prelude hiding (all)
 import Control.Applicative
 import Control.Monad
+import Control.Monad.Trans (lift)
 import Control.Monad.Trans.Maybe
 import Data.Foldable (all)
 import qualified Data.Set as Set
@@ -20,10 +20,12 @@ import Agda.Syntax.Scope.Base (inverseScopeLookupName)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Level
+import Agda.TypeChecking.Reduce (instantiate)
 
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Except
+import Agda.Utils.Functor
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -31,43 +33,44 @@ import Agda.Utils.Impossible
 -- | Convert a 'DisplayTerm' into a 'Term'.
 dtermToTerm :: DisplayTerm -> Term
 dtermToTerm dt = case dt of
-  DWithApp d ds vs -> dtermToTerm d `apply` (map (defaultArg . dtermToTerm) ds ++ vs)
-  DCon c args      -> Con c $ map (fmap dtermToTerm) args
+  DWithApp d ds es ->
+    dtermToTerm d `apply` map (defaultArg . dtermToTerm) ds `applyE` es
+  DCon c ci args   -> Con c ci $ map (fmap dtermToTerm) args
   DDef f es        -> Def f $ map (fmap dtermToTerm) es
   DDot v           -> v
   DTerm v          -> v
 
 -- | Get the arities of all display forms for a name.
 displayFormArities :: QName -> TCM [Int]
-displayFormArities q = map (length . dfPats . openThing) <$> getDisplayForms q
+displayFormArities q = map (length . dfPats . dget) <$> getDisplayForms q
 
--- | Find a matching display form for @q vs@.
---   In essence this tries to reqwrite @q vs@ with any
+-- | Find a matching display form for @q es@.
+--   In essence this tries to rewrite @q es@ with any
 --   display form @q ps --> dt@ and returns the instantiated
 --   @dt@ if successful.  First match wins.
-displayForm :: QName -> Args -> TCM (Maybe DisplayTerm)
-displayForm q vs = do
+displayForm :: QName -> Elims -> TCM (Maybe DisplayTerm)
+displayForm q es = do
     -- Get display forms for name q.
     odfs  <- getDisplayForms q `catchError` \_ -> return []
     -- Display debug info about the @Open@s.
     unless (null odfs) $ verboseS "tc.display.top" 100 $ do
       n <- getContextId
       reportSLn "tc.display.top" 100 $
-        "displayForm: context = " ++ show n ++
-        ", dfs = " ++ show (map openThingCtxIds odfs)
+        "displayForm for " ++ show q ++ ": context = " ++ show n ++
+        ", dfs = " ++ show odfs
     -- Use only the display forms that can be opened in the current context.
-    dfs   <- catMaybes <$> mapM tryOpen odfs
+    dfs   <- catMaybes <$> mapM getLocal odfs
     scope <- getScope
-    -- Keep the display forms that match the application @c vs@.
+    -- Keep the display forms that match the application @q es@.
     ms <- do
-      ms <- mapM (runMaybeT . (`matchDisplayForm` vs)) dfs
+      ms <- mapM (runMaybeT . (`matchDisplayForm` es)) dfs
       return [ m | Just (d, m) <- ms, wellScoped scope d ]
     -- Not safe when printing non-terminating terms.
-    -- (nfdfs, us) <- normalise (dfs, vs)
+    -- (nfdfs, us) <- normalise (dfs, es)
     unless (null odfs) $ reportSLn "tc.display.top" 100 $ unlines
       [ "name        : " ++ show q
       , "displayForms: " ++ show dfs
-      , "arguments   : " ++ show vs
+      , "arguments   : " ++ show es
       , "matches     : " ++ show ms
       , "result      : " ++ show (headMaybe ms)
       ]
@@ -92,16 +95,16 @@ displayForm q vs = do
     isWithDisplay DWithApp{} = True
     isWithDisplay _          = False
 
--- | Match a 'DisplayForm' @q ps = v@ against @q vs@.
+-- | Match a 'DisplayForm' @q ps = v@ against @q es@.
 --   Return the 'DisplayTerm' @v[us]@ if the match was successful,
---   i.e., @vs / ps = Just us@.
-matchDisplayForm :: DisplayForm -> Args -> MaybeT TCM (DisplayForm, DisplayTerm)
-matchDisplayForm d@(Display _ ps v) vs
-  | length ps > length vs = mzero
+--   i.e., @es / ps = Just us@.
+matchDisplayForm :: DisplayForm -> Elims -> MaybeT TCM (DisplayForm, DisplayTerm)
+matchDisplayForm d@(Display _ ps v) es
+  | length ps > length es = mzero
   | otherwise             = do
-      let (vs0, vs1) = splitAt (length ps) vs
-      us <- match ps $ raise 1 $ map unArg vs0
-      return (d, applySubst (parallelS $ reverse us) v `apply` vs1)
+      let (es0, es1) = splitAt (length ps) es
+      us <- match ps $ raise 1 es0
+      return (d, applySubst (parallelS $ reverse us) v `applyE` es1)
 
 -- | Class @Match@ for matching a term @p@ in the role of a pattern
 --   against a term @v@.
@@ -130,16 +133,16 @@ instance Match a => Match (Arg a) where
 instance Match a => Match (Elim' a) where
   match p v =
     case (p, v) of
-      (Proj f, Proj f') | f == f' -> return []
+      (Proj _ f, Proj _ f') | f == f' -> return []
       (Apply a, Apply a')         -> match a a'
       _                           -> mzero
 
 instance Match Term where
-  match p v = case (ignoreSharing p, ignoreSharing v) of
+  match p v = lift (instantiate v) >>= \ v -> case (ignoreSharing p, ignoreSharing v) of
     (Var 0 [], v)                  -> return [strengthen __IMPOSSIBLE__ v]
     (Var i ps, Var j vs) | i == j  -> match ps vs
     (Def c ps, Def d vs) | c == d  -> match ps vs
-    (Con c ps, Con d vs) | c == d  -> match ps vs
+    (Con c _ ps, Con d _ vs) | c == d -> match ps vs
     (Lit l, Lit l')      | l == l' -> return []
     (p, v)               | p == v  -> return []
     (p, Level l)                   -> match p =<< reallyUnLevelView l
