@@ -1,18 +1,19 @@
-{-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE TupleSections     #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.TypeChecking.Errors
   ( prettyError
   , tcErrString
-  , Warnings(..)
-  , warningsToError
+  , prettyTCWarnings
+  , tcWarningsToError
+  , applyFlagsToTCWarnings
   ) where
 
 import Prelude hiding (null)
 
+import Control.Monad.Reader
 import Control.Monad.State
 
 import Data.Function
@@ -23,6 +24,7 @@ import qualified Data.Set as Set
 import qualified Data.Map as Map
 import qualified Text.PrettyPrint.Boxes as Boxes
 
+import Agda.Interaction.Options
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
@@ -43,13 +45,16 @@ import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Positivity
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce (instantiate)
 
-import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.FileName
 import Agda.Utils.Function
 import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Size
@@ -80,24 +85,96 @@ prettyError err = liftTCM $ show <$> prettyError' err []
 -- * Warnings
 ---------------------------------------------------------------------------
 
--- | Warnings.
---
--- Invariant: The fields are never empty at the same time.
+instance PrettyTCM TCWarning where
+  prettyTCM (TCWarning tcst clw) = localState $ do
+    put tcst
+    sayWhen (envRange  $ clEnv clw)
+            (envCall   $ clEnv clw)
+            (prettyTCM $ clValue clw)
 
-data Warnings = Warnings
-  { unsolvedMetaVariables :: [Range]
-    -- ^ Meta-variable problems are reported as type errors unless
-    -- 'optAllowUnsolved' is 'True'.
-  , unsolvedConstraints   :: Constraints
-    -- ^ Same as 'unsolvedMetaVariables'.
-  }
+instance PrettyTCM Warning where
+  prettyTCM wng = case wng of
 
--- | Turns warnings into an error. Even if several errors are possible
---   only one is raised.
-warningsToError :: Warnings -> TCM a
-warningsToError (Warnings [] [])     = typeError $ SolvedButOpenHoles
-warningsToError (Warnings w@(_:_) _) = typeError $ UnsolvedMetas w
-warningsToError (Warnings _ w@(_:_)) = typeError $ UnsolvedConstraints w
+    UnsolvedMetaVariables ms  ->
+      fsep ( pwords "Unsolved metas at the following locations:" )
+      $$ nest 2 (vcat $ map prettyTCM ms)
+
+    UnsolvedInteractionMetas is ->
+      fsep ( pwords "Unsolved interaction metas at the following locations:" )
+      $$ nest 2 (vcat $ map prettyTCM is)
+
+    UnsolvedConstraints cs ->
+      fsep ( pwords "Failed to solve the following constraints:" )
+      $$ nest 2 (P.vcat . nub <$> mapM prettyConstraint cs)
+
+      where prettyConstraint :: ProblemConstraint -> TCM Doc
+            prettyConstraint c = f (prettyTCM c)
+              where
+              r   = getRange c
+              f d = if null $ P.pretty r
+                    then d
+                    else d $$ nest 4 (text "[ at" <+> prettyTCM r <+> text "]")
+
+    TerminationIssue tes ->
+      fwords "Termination checking failed for the following functions:"
+      $$ (nest 2 $ fsep $ punctuate comma $
+           map (pretty . dropTopLevelModule) $
+             concatMap termErrFunctions tes)
+      $$ fwords "Problematic calls:"
+      $$ (nest 2 $ fmap (P.vcat . nub) $
+            mapM prettyTCM $ sortBy (compare `on` callInfoRange) $
+            concatMap termErrCalls tes)
+
+    NotStrictlyPositive d ocs -> fsep $
+      [prettyTCM (dropTopLevelModule d)] ++
+      pwords "is not strictly positive, because it occurs"
+      ++ [prettyTCM ocs]
+
+    OldBuiltin old new -> fwords $
+      "Builtin " ++ old ++ " does no longer exist. " ++
+      "It is now bound by BUILTIN " ++ new
+
+    EmptyRewritePragma -> fsep . pwords $ "Empty REWRITE pragma"
+
+    ParseWarning pw -> pretty pw
+
+prettyTCWarnings :: [TCWarning] -> TCM String
+prettyTCWarnings = fmap (unlines . intersperse " ") . prettyTCWarnings'
+
+prettyTCWarnings' :: [TCWarning] -> TCM [String]
+prettyTCWarnings' = mapM (fmap show . prettyTCM)
+
+-- | Turns all warnings into errors.
+tcWarningsToError :: [TCWarning] -> TCM a
+tcWarningsToError ws = typeError $ case ws of
+  [] -> SolvedButOpenHoles
+  _  -> NonFatalErrors ws
+
+
+-- | Depending which flags are set, one may happily ignore some
+-- warnings.
+
+applyFlagsToTCWarnings :: IgnoreFlags -> [TCWarning] -> TCM [TCWarning]
+applyFlagsToTCWarnings ifs ws = do
+
+  unsolvedNotOK <- not . optAllowUnsolved <$> pragmaOptions
+  negativeNotOK <- not . optDisablePositivity <$> pragmaOptions
+  loopingNotOK  <- optTerminationCheck <$> pragmaOptions
+
+  let cleanUp w =
+        let ignore = ifs == IgnoreFlags
+            keepUnsolved us = not (null us) && (ignore || unsolvedNotOK)
+        in case w of
+          TerminationIssue{}           -> ignore || loopingNotOK
+          NotStrictlyPositive{}        -> ignore || negativeNotOK
+          UnsolvedMetaVariables ums    -> keepUnsolved ums
+          UnsolvedInteractionMetas uis -> keepUnsolved uis
+          UnsolvedConstraints ucs      -> keepUnsolved ucs
+          OldBuiltin{}                 -> True
+          EmptyRewritePragma           -> True
+          ParseWarning{}               -> True
+
+  return $ filter (cleanUp . tcWarning) ws
 
 ---------------------------------------------------------------------------
 -- * Helpers
@@ -126,7 +203,6 @@ tcErrString err = show (getRange err) ++ " " ++ case err of
   Exception r s   -> show r ++ " " ++ show s
   IOException r e -> show r ++ " " ++ show e
   PatternErr{}    -> "PatternErr"
-  {- AbortAssign _   -> "AbortAssign" -- UNUSED -}
 
 errorString :: TypeError -> String
 errorString err = case err of
@@ -184,6 +260,7 @@ errorString err = case err of
   ModuleArityMismatch{}                    -> "ModuleArityMismatch"
   ModuleDefinedInOtherFile {}              -> "ModuleDefinedInOtherFile"
   ModuleDoesntExport{}                     -> "ModuleDoesntExport"
+  ModuleNameUnexpected{}                   -> "ModuleNameUnexpected"
   ModuleNameDoesntMatchFileName {}         -> "ModuleNameDoesntMatchFileName"
   NeedOptionCopatterns{}                   -> "NeedOptionCopatterns"
   NeedOptionRewriting{}                    -> "NeedOptionRewriting"
@@ -204,12 +281,12 @@ errorString err = case err of
   InvalidTypeSort{}                        -> "InvalidTypeSort"
   FunctionTypeInSizeUniv{}                 -> "FunctionTypeInSizeUniv"
   NotAValidLetBinding{}                    -> "NotAValidLetBinding"
+  NotValidBeforeField{}                    -> "NotValidBeforeField"
   NotAnExpression{}                        -> "NotAnExpression"
   NotImplemented{}                         -> "NotImplemented"
   NotSupported{}                           -> "NotSupported"
   NotInScope{}                             -> "NotInScope"
   NotLeqSort{}                             -> "NotLeqSort"
-  NotStrictlyPositive{}                    -> "NotStrictlyPositive"
   NothingAppliedToHiddenArg{}              -> "NothingAppliedToHiddenArg"
   NothingAppliedToInstanceArg{}            -> "NothingAppliedToInstanceArg"
   OverlappingProjects {}                   -> "OverlappingProjects"
@@ -223,7 +300,8 @@ errorString err = case err of
   SafeFlagNonTerminating{}                 -> "SafeFlagNonTerminating"
   SafeFlagTerminating{}                    -> "SafeFlagTerminating"
   SafeFlagPrimTrustMe{}                    -> "SafeFlagPrimTrustMe"
-  SafeFlagNoPositivityCheck{}              -> "SafeNoPositivityCheck"
+  SafeFlagNoPositivityCheck{}              -> "SafeFlagNoPositivityCheck"
+  SafeFlagPolarity{}                       -> "SafeFlagPolarity"
   ShadowedModule{}                         -> "ShadowedModule"
   ShouldBeASort{}                          -> "ShouldBeASort"
   ShouldBeApplicationOf{}                  -> "ShouldBeApplicationOf"
@@ -239,6 +317,7 @@ errorString err = case err of
   TooFewFields{}                           -> "TooFewFields"
   TooManyArgumentsInLHS{}                  -> "TooManyArgumentsInLHS"
   TooManyFields{}                          -> "TooManyFields"
+  TooManyPolarities{}                      -> "TooManyPolarities"
   SplitOnIrrelevant{}                      -> "SplitOnIrrelevant"
   DefinitionIsIrrelevant{}                 -> "DefinitionIsIrrelevant"
   VariableIsIrrelevant{}                   -> "VariableIsIrrelevant"
@@ -261,11 +340,10 @@ errorString err = case err of
   UninstantiatedDotPattern{}               -> "UninstantiatedDotPattern"
   UninstantiatedModule{}                   -> "UninstantiatedModule"
   UnreachableClauses{}                     -> "UnreachableClauses"
-  UnsolvedConstraints{}                    -> "UnsolvedConstraints"
-  UnsolvedMetas{}                          -> "UnsolvedMetas"
   SolvedButOpenHoles{}                     -> "SolvedButOpenHoles"
   UnusedVariableInPatternSynonym           -> "UnusedVariableInPatternSynonym"
   UnquoteFailed{}                          -> "UnquoteFailed"
+  DeBruijnIndexOutOfScope{}                -> "DeBruijnIndexOutOfScope"
   WithClausePatternMismatch{}              -> "WithClausePatternMismatch"
   WithoutKError{}                          -> "WithoutKError"
   WrongHidingInApplication{}               -> "WrongHidingInApplication"
@@ -277,9 +355,15 @@ errorString err = case err of
   WrongNumberOfConstructorArguments{}      -> "WrongNumberOfConstructorArguments"
   HidingMismatch{}                         -> "HidingMismatch"
   RelevanceMismatch{}                      -> "RelevanceMismatch"
+  NonFatalErrors{}                         -> "NonFatalErrors"
+  InstanceSearchDepthExhausted{}           -> "InstanceSearchDepthExhausted"
 
 instance PrettyTCM TCErr where
   prettyTCM err = case err of
+    -- Gallais, 2016-05-14
+    -- Given where `NonFatalErrors` are created, we know for a
+    -- fact that ̀ws` is non-empty.
+    TypeError _ (Closure _ _ _ _ (NonFatalErrors ws)) -> foldr1 ($$) $ fmap prettyTCM ws
     -- Andreas, 2014-03-23
     -- This use of localState seems ok since we do not collect
     -- Benchmark info during printing errors.
@@ -289,7 +373,6 @@ instance PrettyTCM TCErr where
     Exception r s   -> sayWhere r $ return s
     IOException r e -> sayWhere r $ fwords $ show e
     PatternErr{}    -> sayWhere err $ panic "uncaught pattern violation"
-    {- AbortAssign _   -> sayWhere err $ panic "uncaught aborted assignment" -- UNUSED -}
 
 instance PrettyTCM CallInfo where
   prettyTCM c = do
@@ -300,8 +383,11 @@ instance PrettyTCM CallInfo where
       else call $$ nest 2 (text "(at" <+> prettyTCM r <> text ")")
 
 -- | Drops the filename component of the qualified name.
+dropTopLevelModule' :: Int -> QName -> QName
+dropTopLevelModule' k (QName (MName ns) n) = QName (MName (drop k ns)) n
+
 dropTopLevelModule :: QName -> QName
-dropTopLevelModule (QName (MName ns) n) = QName (MName (drop 1 ns)) n
+dropTopLevelModule = dropTopLevelModule' 1
 
 instance PrettyTCM TypeError where
   prettyTCM err = case err of
@@ -317,15 +403,19 @@ instance PrettyTCM TypeError where
 
     GenericDocError d -> return d
 
-    TerminationCheckFailed because ->
+    TerminationCheckFailed because -> do
+      dropTopLevelModule <- do
+        caseMaybeM (asks envCurrentPath) (return id) $ \ f -> do
+        m <- fromMaybe __IMPOSSIBLE__ <$> lookupModuleFromSource f
+        return $ dropTopLevelModule' $ size m
       fwords "Termination checking failed for the following functions:"
-      $$ (nest 2 $ fsep $ punctuate comma $
-           map (pretty . dropTopLevelModule) $
-             concatMap termErrFunctions because)
-      $$ fwords "Problematic calls:"
-      $$ (nest 2 $ fmap (P.vcat . nub) $
-            mapM prettyTCM $ sortBy (compare `on` callInfoRange) $
-            concatMap termErrCalls because)
+        $$ (nest 2 $ fsep $ punctuate comma $
+             map (pretty . dropTopLevelModule) $
+               concatMap termErrFunctions because)
+        $$ fwords "Problematic calls:"
+        $$ (nest 2 $ fmap (P.vcat . nub) $
+              mapM prettyTCM $ sortBy (compare `on` callInfoRange) $
+              concatMap termErrCalls because)
 
     PropMustBeSingleton -> fwords
       "Datatypes in Prop must have at most one constructor when proof irrelevance is enabled"
@@ -575,9 +665,16 @@ instance PrettyTCM TypeError where
       pwords "Duplicate fields" ++ punctuate comma (map pretty xs) ++
       pwords "in record"
 
-    WithOnFreeVariable e -> fsep $
-      pwords "Cannot `with` on variable " ++ [prettyA e] ++
-      pwords " bound in a module telescope (or patterns of a parent clause)"
+    WithOnFreeVariable e v -> do
+      de <- prettyA e
+      dv <- prettyTCM v
+      if show de == show dv
+        then fsep $
+          pwords "Cannot `with` on variable" ++ [return dv] ++
+          pwords " bound in a module telescope (or patterns of a parent clause)"
+        else fsep $
+          pwords "Cannot `with` on expression" ++ [return de] ++ pwords "which reduces to variable" ++ [return dv] ++
+          pwords " bound in a module telescope (or patterns of a parent clause)"
 
     UnexpectedWithPatterns ps -> fsep $
       pwords "Unexpected with patterns" ++ (punctuate (text " |") $ map prettyA ps)
@@ -647,22 +744,6 @@ instance PrettyTCM TypeError where
     SolvedButOpenHoles ->
       text "Module cannot be imported since it has open interaction points"
 
-    UnsolvedMetas rs ->
-      fsep ( pwords "Unsolved metas at the following locations:" )
-      $$ nest 2 (vcat $ map prettyTCM rs)
-
-    UnsolvedConstraints cs ->
-      fsep ( pwords "Failed to solve the following constraints:" )
-      $$ nest 2 (vcat $ map prettyConstraint cs)
-
-      where prettyConstraint :: ProblemConstraint -> TCM Doc
-            prettyConstraint c = f (prettyTCM c)
-              where
-              r   = getRange c
-              f d = if null $ P.pretty r
-                    then d
-                    else d $$ nest 4 (text "[ at" <+> prettyTCM r <+> text "]")
-
     CyclicModuleDependency ms ->
       fsep (pwords "cyclic module dependency:")
       $$ nest 2 (vcat $ map pretty ms)
@@ -696,6 +777,12 @@ instance PrettyTCM TypeError where
       pwords "However, according to the include path this module should" ++
       pwords "be defined in" ++ [text (filePath file') <> text "."]
 
+    ModuleNameUnexpected given expected -> fsep $
+      pwords "The name of the top level module does not match the file name. The module" ++
+      [ pretty given ] ++
+      pwords "should probably be named" ++
+      [ pretty expected ]
+
     ModuleNameDoesntMatchFileName given files ->
       fsep (pwords "The name of the top level module does not match the file name. The module" ++
            [ pretty given ] ++ pwords "should be defined in one of the following files:")
@@ -803,6 +890,9 @@ instance PrettyTCM TypeError where
     NotAValidLetBinding nd -> fwords $
       "Not a valid let-declaration"
 
+    NotValidBeforeField nd -> fwords $
+      "This declaration is illegal in a record before the last field"
+
     NothingAppliedToHiddenArg e -> fsep $
       [pretty e] ++ pwords "cannot appear by itself. It needs to be the argument to" ++
       pwords "a function expecting an implicit argument."
@@ -954,7 +1044,15 @@ instance PrettyTCM TypeError where
         )
         where
         nota    = sectNotation sect
-        section = trim (notation nota)
+        section = qualifyFirstIdPart
+                    (foldr (\x s -> C.nameToRawName x ++ "." ++ s)
+                           ""
+                           (init (C.qnameParts (notaName nota))))
+                    (trim (notation nota))
+
+        qualifyFirstIdPart _ []              = []
+        qualifyFirstIdPart q (IdPart x : ps) = IdPart (q ++ x) : ps
+        qualifyFirstIdPart q (p : ps)        = p : qualifyFirstIdPart q ps
 
         trim = case sectKind sect of
           InfixNotation   -> trimLeft . trimRight
@@ -1012,7 +1110,7 @@ instance PrettyTCM TypeError where
         where
           display ps = do
             ps <- nicify f ps
-            prettyTCM f <+> fsep (map prettyArg ps)
+            prettyTCM f <+> fsep (map (prettyArg . fmap namedThing) ps)
 
           nicify f ps = do
             showImp <- showImplicitArguments
@@ -1022,7 +1120,7 @@ instance PrettyTCM TypeError where
 
     CoverageCantSplitOn c tel cIxs gIxs
       | length cIxs /= length gIxs -> __IMPOSSIBLE__
-      | otherwise                  -> addCtxTel tel $ vcat (
+      | otherwise                  -> addContext tel $ vcat (
           [ fsep $ pwords "I'm not sure if there should be a case for the constructor" ++
                    [prettyTCM c <> text ","] ++
                    pwords "because I get stuck when trying to solve the following" ++
@@ -1048,12 +1146,14 @@ instance PrettyTCM TypeError where
       pwords "because K has been disabled."
 
     UnifyConflict c c' -> fsep $
-      pwords "There was a conflict between the constructors " ++
-      [prettyTCM c] ++ pwords " and " ++ [prettyTCM c']
+      pwords "This case is impossible because of a conflict between the constructors " ++
+      [prettyTCM c] ++ pwords " and " ++ [prettyTCM c' <> text "."] ++
+      pwords "Possible solution: remove the clause, or use an absurd pattern ()."
 
     UnifyCycle i u -> fsep $
-      pwords "The variable " ++ [prettyTCM (var i)] ++
-      pwords "occurs strongly rigid in" ++ [prettyTCM u]
+      pwords "This case is impossible because the variable " ++ [prettyTCM (var i)] ++
+      pwords "occurs strongly rigid in" ++ [prettyTCM u <> text "."] ++
+      pwords "Possible solution: remove the clause, or use an absurd pattern ()."
 
     UnifyIndicesNotVars a u v ixs -> fsep $
       pwords "Cannot apply injectivity to the equation" ++ [prettyTCM u] ++
@@ -1073,35 +1173,10 @@ instance PrettyTCM TypeError where
       pwords "with" ++ [prettyList (map prettyTCM vs)] ++
       pwords "in the telescope" ++ [prettyTCM tel]
 
-    NotStrictlyPositive d ocs -> fsep $
-      pwords "The datatype" ++ [prettyTCM d] ++
-      pwords "is not strictly positive, because"
-      ++ prettyOcc "it" ocs
-      where
-        prettyOcc _ [] = []
-        prettyOcc it (OccCon d c r : ocs) = concat
-          [ pwords it, pwords "occurs", prettyR r
-          , pwords "in the constructor", [prettyTCM c], pwords "of"
-          , [prettyTCM d <> com ocs], prettyOcc "which" ocs
-          ]
-        prettyOcc it (OccClause f n r : ocs) = concat
-          [ pwords it, pwords "occurs", prettyR r
-          , pwords "in the", [th n], pwords "clause of"
-          , [prettyTCM f <> com ocs], prettyOcc "which" ocs
-          ]
-
-        prettyR NonPositively = pwords "negatively"
-        prettyR (ArgumentTo i q) =
-          pwords "as the" ++ [th i] ++
-          pwords "argument to" ++ [prettyTCM q]
-
-        th 0 = text "first"
-        th 1 = text "second"
-        th 2 = text "third"
-        th n = prettyTCM (n - 1) <> text "th"
-
-        com []    = empty
-        com (_:_) = comma
+    TooManyPolarities x n -> fsep $
+      pwords "Too many polarities given in the POLARITY pragma for" ++
+      [prettyTCM x] ++
+      pwords "(at most" ++ [text (show n)] ++ pwords "allowed)."
 
     IFSNoCandidateInScope t -> fsep $
       pwords "No instance of type" ++ [prettyTCM t] ++ pwords "was found in scope."
@@ -1127,6 +1202,16 @@ instance PrettyTCM TypeError where
 
       UnquotePanic err -> __IMPOSSIBLE__
 
+    DeBruijnIndexOutOfScope i EmptyTel [] -> fsep $
+        pwords $ "de Bruijn index " ++ show i ++ " is not in scope in the empty context"
+    DeBruijnIndexOutOfScope i cxt names ->
+        sep [ text ("de Bruijn index " ++ show i ++ " is not in scope in the context")
+            , inTopContext $ addContext "_" $ prettyTCM cxt' ]
+      where
+        cxt' = cxt `abstract` raise (size cxt) (nameCxt names)
+        nameCxt [] = EmptyTel
+        nameCxt (x : xs) = ExtendTel (defaultDom (El I.Prop $ I.Var 0 [])) $ NoAbs (show x) $ nameCxt xs
+
     SafeFlagPostulate e -> fsep $
       pwords "Cannot postulate" ++ [pretty e] ++ pwords "with safe flag"
 
@@ -1147,11 +1232,21 @@ instance PrettyTCM TypeError where
     SafeFlagNoPositivityCheck -> fsep $
       pwords "Cannot use NO_POSITIVITY_CHECK pragma with safe flag."
 
+    SafeFlagPolarity -> fsep $
+      pwords "The POLARITY pragma must not be used in safe mode."
+
     NeedOptionCopatterns -> fsep $
       pwords "Option --copatterns needed to enable destructor patterns"
+
     NeedOptionRewriting  -> fsep $
       pwords "Option --rewriting needed to add and use rewrite rules"
 
+    NonFatalErrors ws -> foldr1 ($$) $ fmap prettyTCM ws
+
+    InstanceSearchDepthExhausted c a d -> fsep $
+      pwords ("Instance search depth exhaused (max depth: " ++ show d ++ ") for candidate") ++
+      [hang (prettyTCM c <+> text ":") 2 (prettyTCM a)]
+
     where
     mpar n args
       | n > 0 && not (null args) = parens
@@ -1170,7 +1265,7 @@ instance PrettyTCM TypeError where
       mpar n args $
         prettyTCM c <+> fsep (map (prettyArg . fmap namedThing) args)
     prettyPat _ (I.LitP l) = prettyTCM l
-    prettyPat _ (I.ProjP p) = prettyTCM p
+    prettyPat _ (I.ProjP _ p) = text "." <> prettyTCM p
 
 notCmp :: Comparison -> TCM Doc
 notCmp cmp = text "!" <> prettyTCM cmp
@@ -1205,7 +1300,7 @@ prettyInEqual t1 t2 = do
       "they contain different but identically rendered identifiers somewhere"
     varVar :: Int -> Int -> TCM Doc
     varVar i j = parens $ fwords $
-                   "because one has deBruijn index " ++ show i
+                   "because one has de Bruijn index " ++ show i
                    ++ " and the other " ++ show j
 
 class PrettyUnequal a where
@@ -1252,7 +1347,7 @@ instance PrettyTCM Call where
         pwords "when checking that the clause"
         ++ [prettyA cl] ++ pwords "has type" ++ [prettyTCM t]
 
-    CheckPattern p tel t -> addCtxTel tel $ fsep $
+    CheckPattern p tel t -> addContext tel $ fsep $
       pwords "when checking that the pattern"
       ++ [prettyA p] ++ pwords "has type" ++ [prettyTCM t]
 
@@ -1291,7 +1386,7 @@ instance PrettyTCM Call where
     CheckDataDef _ x ps cs ->
       fsep $ pwords "when checking the definition of" ++ [prettyTCM x]
 
-    CheckConstructor d _ _ (A.Axiom _ _ _ c _) -> fsep $
+    CheckConstructor d _ _ (A.Axiom _ _ _ _ c _) -> fsep $
       pwords "when checking the constructor" ++ [prettyTCM c] ++
       pwords "in the declaration of" ++ [prettyTCM d]
 
@@ -1322,7 +1417,7 @@ instance PrettyTCM Call where
     InferVar x ->
       fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
 
-    InferDef _ x ->
+    InferDef x ->
       fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
 
     CheckIsEmpty r t ->
@@ -1332,8 +1427,13 @@ instance PrettyTCM Call where
     ScopeCheckExpr e -> fsep $ pwords "when scope checking" ++ [pretty e]
 
     ScopeCheckDeclaration d ->
-      fwords "when scope checking the declaration" $$
-      nest 2 (pretty $ simpleDecl d)
+      fwords ("when scope checking the declaration" ++ suffix) $$
+      nest 2 (vcat $ map pretty ds)
+      where
+      ds     = D.notSoNiceDeclarations d
+      suffix = case ds of
+        [_] -> ""
+        _   -> "s"
 
     ScopeCheckLHS x p ->
       fsep $ pwords "when scope checking the left-hand side" ++ [pretty p] ++
@@ -1345,7 +1445,7 @@ instance PrettyTCM Call where
 
     CheckSectionApplication _ m1 modapp -> fsep $
       pwords "when checking the module application" ++
-      [prettyA $ A.Apply info m1 modapp empty empty defaultImportDir]
+      [prettyA $ A.Apply info m1 modapp initCopyInfo defaultImportDir]
       where
       info = A.ModuleInfo noRange noRange Nothing Nothing Nothing
 
@@ -1355,8 +1455,6 @@ instance PrettyTCM Call where
     hPretty :: Arg (Named_ Expr) -> TCM Doc
     hPretty a = pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (getHiding a)) a
 
-    simpleDecl = D.notSoNiceDeclaration
-
 ---------------------------------------------------------------------------
 -- * Natural language
 ---------------------------------------------------------------------------
