@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE PatternGuards #-}
 
 -- | Compute eta short normal forms.
 module Agda.TypeChecking.EtaContract where
@@ -34,8 +32,8 @@ binAppView t = case t of
   -- (Cf. also issue 889 (fixed differently).)
   -- At least record constructors should be fully applied where possible!
   -- TODO: also for ordinary constructors (\ x -> suc x  vs.  suc)?
-  Con c xs
-    | null (conFields c) -> app (Con c) xs
+  Con c ci xs
+    | null (conFields c) -> app (Con c ci) xs
     | otherwise          -> noApp
   Lit _      -> noApp
   Level _    -> noApp   -- could be an application, but let's not eta contract levels
@@ -69,12 +67,11 @@ etaOnce v = case v of
   -- reportSDoc "tc.eta" 70 $ text "eta-contracting" <+> prettyTCM v
   Shared{} -> updateSharedTerm etaOnce v
   Lam i (Abs _ b) -> do  -- NoAbs can't be eta'd
-      imp <- shouldEtaContractImplicit
       tyty <- typeInType
       case binAppView b of
         App u (Arg info v)
           | (isIrrelevant info || isVar0 tyty v)
-                    && allowed imp info
+                    && getHiding i == getHiding info
                     && not (freeIn 0 u) ->
             return $ strengthen __IMPOSSIBLE__ u
         _ -> return v
@@ -89,16 +86,15 @@ etaOnce v = case v of
         BlockedLevel{}   -> False
         MetaLevel{}      -> False
       isVar0 _ _ = False
-      allowed imp i' = getHiding i == getHiding i' && (imp || notHidden i)
 
   -- Andreas, 2012-12-18:  Abstract definitions could contain
   -- abstract records whose constructors are not in scope.
   -- To be able to eta-contract them, we ignore abstract.
-  Con c args -> ignoreAbstractMode $ do
+  Con c ci args -> ignoreAbstractMode $ do
       -- reportSDoc "tc.eta" 20 $ text "eta-contracting record" <+> prettyTCM t
       r <- getConstructorData $ conName c -- fails in ConcreteMode if c is abstract
       ifM (isEtaRecord r)
           (do -- reportSDoc "tc.eta" 20 $ text "eta-contracting record" <+> prettyTCM t
-              etaContractRecord r c args)
+              etaContractRecord r c ci args)
           (return v)
   v -> return v
