@@ -1,6 +1,3 @@
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 -- | Computing the free variables of a term lazily.
@@ -36,9 +33,8 @@ import Control.Monad.Reader
 
 import Data.Foldable (foldMap)
 import Data.IntMap (IntMap)
-import Data.Monoid
-
-import Test.QuickCheck
+import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, mconcat)
+import Data.Set (Set)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -48,6 +44,9 @@ import Agda.Syntax.Internal
 import Agda.Utils.Functor
 import Agda.Utils.Monad
 import Agda.Utils.Singleton
+import Agda.Utils.Size
+
+type MetaSet = Set MetaId
 
 -- | Depending on the surrounding context of a variable,
 --   it's occurrence can be classified as flexible or rigid,
@@ -55,7 +54,7 @@ import Agda.Utils.Singleton
 --
 --   The constructors are listed in increasing order (wrt. information content).
 data FlexRig
-  = Flexible [MetaId] -- ^ In arguments of metas.
+  = Flexible MetaSet  -- ^ In arguments of metas.
   | WeaklyRigid       -- ^ In arguments to variables and definitions.
   | Unguarded         -- ^ In top position, or only under inductive record constructors.
   | StronglyRigid     -- ^ Under at least one and only inductive constructors.
@@ -74,7 +73,7 @@ data FlexRig
 composeFlexRig :: FlexRig -> FlexRig -> FlexRig
 composeFlexRig o o' =
   case (o, o') of
-    (Flexible ms1, Flexible ms2) -> Flexible $ ms1 ++ ms2
+    (Flexible ms1, Flexible ms2) -> Flexible $ ms1 `mappend` ms2
     (Flexible ms1, _) -> Flexible ms1
     (_, Flexible ms2) -> Flexible ms2
     (WeaklyRigid, _) -> WeaklyRigid
@@ -107,7 +106,7 @@ topVarOcc :: VarOcc
 topVarOcc = VarOcc StronglyRigid Relevant
 
 botVarOcc :: VarOcc
-botVarOcc = VarOcc (Flexible []) Irrelevant
+botVarOcc = VarOcc (Flexible mempty) Irrelevant
 
 type VarMap = IntMap VarOcc
 
@@ -152,9 +151,11 @@ initFreeEnv sing = FreeEnv
 
 type FreeM c = Reader (FreeEnv c) c
 
+instance Monoid c => Semigroup (FreeM c) where
+  (<>) = liftA2 mappend
 instance Monoid c => Monoid (FreeM c) where
   mempty  = pure mempty
-  mappend = liftA2 mappend
+  mappend = (<>)
   mconcat = mconcat <.> sequence
 
 -- instance Singleton a c => Singleton a (FreeM c) where
@@ -172,7 +173,10 @@ variable n = do
 
 -- | Going under a binder.
 bind :: FreeM a -> FreeM a
-bind = local $ \ e -> e { feBinders = 1 + feBinders e }
+bind = bind' 1
+
+bind' :: Nat -> FreeM a -> FreeM a
+bind' n = local $ \ e -> e { feBinders = n + feBinders e }
 
 -- | Changing the 'FlexRig' context.
 go :: FlexRig -> FreeM a -> FreeM a
@@ -219,7 +223,7 @@ instance Free' Term c where
       -- we cannot query whether we are dealing with a data/record (strongly r.)
       -- or a definition by pattern matching (weakly rigid)
       -- thus, we approximate, losing that x = List x is unsolvable
-    Con c ts     -> underConstructor c $ freeVars' ts
+    Con c _ ts   -> underConstructor c $ freeVars' ts
     -- Pi is not guarding, since we cannot prove that A ≡ B → A is impossible.
     -- Even as we do not permit infinite type expressions,
     -- we cannot prove their absence (as Set is not inductive).
@@ -231,7 +235,7 @@ instance Free' Term c where
     DontCare mt  -> goRel Irrelevant $ freeVars' mt
     Shared p     -> freeVars' (derefPtr p)
 
-instance Free' Type c where
+instance Free' a c => Free' (Type' a) c where
   -- {-# SPECIALIZE instance Free' Type All #-}
   -- {-# SPECIALIZE freeVars' :: Type -> FreeM Any #-}
   -- {-# SPECIALIZE freeVars' :: Type -> FreeM All #-}
@@ -353,39 +357,14 @@ instance Free' a c => Free' (Tele a) c where
   freeVars' EmptyTel          = mempty
   freeVars' (ExtendTel a tel) = freeVars' (a, tel)
 
-instance Free' ClauseBody c where
-  -- {-# SPECIALIZE instance Free' ClauseBody All #-}
-  -- {-# SPECIALIZE freeVars' :: ClauseBody -> FreeM Any #-}
-  -- {-# SPECIALIZE freeVars' :: ClauseBody -> FreeM All #-}
-  -- {-# SPECIALIZE freeVars' :: ClauseBody -> FreeM VarSet #-}
-  -- {-# SPECIALIZE freeVars' :: ClauseBody -> FreeM VarMap #-}
-  freeVars' (Body t)   = freeVars' t
-  freeVars' (Bind b)   = freeVars' b
-  freeVars'  NoBody    = mempty
-
 instance Free' Clause c where
   -- {-# SPECIALIZE instance Free' Clause All #-}
   -- {-# SPECIALIZE freeVars' :: Clause -> FreeM Any #-}
   -- {-# SPECIALIZE freeVars' :: Clause -> FreeM All #-}
   -- {-# SPECIALIZE freeVars' :: Clause -> FreeM VarSet #-}
   -- {-# SPECIALIZE freeVars' :: Clause -> FreeM VarMap #-}
-  freeVars' = freeVars' . clauseBody
+  freeVars' cl = bind' (size $ clauseTel cl) $ freeVars' $ clauseBody cl
 
 instance Free' EqualityView c where
   freeVars' (OtherType t) = freeVars' t
   freeVars' (EqualityType s _eq l t a b) = freeVars' s `mappend` freeVars' [l, t, a, b]
-
--- Generators
-
-instance Arbitrary FlexRig where
-  arbitrary = oneof
-    [ pure $ Flexible [] -- TODO
-    , pure WeaklyRigid
-    , pure Unguarded
-    , pure StronglyRigid
-    ]
-
-instance Arbitrary VarOcc where
-  arbitrary = VarOcc <$> arbitrary <*> arbitrary
-
--- -}
