@@ -1,4 +1,3 @@
-{-# LANGUAGE FlexibleInstances #-}
 
 -- | Computing the free variables of a term.
 --
@@ -41,13 +40,14 @@ import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 
 import Data.Foldable (foldMap)
-import Data.Monoid
+import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, mconcat)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
 import Agda.Utils.Functor
 import Agda.Utils.Monad
+import Agda.Utils.Size
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as Set
 
@@ -173,9 +173,12 @@ empty :: FreeVars
 empty = FV Set.empty Set.empty Set.empty Set.empty Set.empty Set.empty
 
 -- | Free variable sets form a monoid under 'union'.
+instance Semigroup FreeVars where
+  (<>) = union
+
 instance Monoid FreeVars where
   mempty  = empty
-  mappend = union
+  mappend = (<>)
   mconcat = unions
 
 -- | @delete x fv@ deletes variable @x@ from variable set @fv@.
@@ -223,9 +226,12 @@ freeVarsIgnore i t = freeVars' t `runReader` initFreeConf{ fcIgnoreSorts = i }
 -- | Return type of fold over syntax.
 type FreeT = Reader FreeConf FreeVars
 
+instance Semigroup FreeT where
+  (<>) = liftA2 mappend
+
 instance Monoid FreeT where
   mempty  = pure mempty
-  mappend = liftA2 mappend
+  mappend = (<>)
   mconcat = mconcat <.> sequence
 
 -- | Base case: a variable.
@@ -236,7 +242,11 @@ variable n = do
 
 -- | Going under a binder.
 bind :: FreeT -> FreeT
-bind = local $ \ e -> e { fcContext = 1 + fcContext e }
+bind = bind' 1
+
+-- | Going under n binders.
+bind' :: Nat -> FreeT -> FreeT
+bind' n = local $ \ e -> e { fcContext = n + fcContext e }
 
 class Free a where
   freeVars'   :: a -> FreeT
@@ -252,7 +262,7 @@ instance Free Term where
       -- we cannot query whether we are dealing with a data/record (strongly r.)
       -- or a definition by pattern matching (weakly rigid)
       -- thus, we approximate, losing that x = List x is unsolvable
-    Con c ts   -> underConstructor c <$> freeVars' ts
+    Con c _ ts -> underConstructor c <$> freeVars' ts
     -- Pi is not guarding, since we cannot prove that A ≡ B → A is impossible.
     -- Even as we do not permit infinite type expressions,
     -- we cannot prove their absence (as Set is not inductive).
@@ -326,13 +336,8 @@ instance Free a => Free (Tele a) where
   freeVars' EmptyTel          = mempty
   freeVars' (ExtendTel a tel) = freeVars' (a, tel)
 
-instance Free ClauseBody where
-  freeVars' (Body t)   = freeVars' t
-  freeVars' (Bind b)   = freeVars' b
-  freeVars'  NoBody    = mempty
-
 instance Free Clause where
-  freeVars' = freeVars' . clauseBody
+  freeVars' cl = bind' (size $ clauseTel cl) $ freeVars' $ clauseBody cl
 
 freeIn :: Free a => Nat -> a -> Bool
 freeIn v t = v `Set.member` allVars (freeVars t)
