@@ -9,16 +9,13 @@ import Control.Monad.Reader
 
 import qualified Data.Map as Map
 
-import Agda.Interaction.Options hiding (tests)
+import Agda.Interaction.Options
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad
 
-import Agda.Utils.QuickCheck
-import Agda.Utils.TestHelpers
-
 -- | data 'Relevance'
 --   see "Agda.Syntax.Common".
 
@@ -48,17 +45,13 @@ workOnTypes cont = do
   allowed <- optExperimentalIrrelevance <$> pragmaOptions
   verboseBracket "tc.irr" 20 "workOnTypes" $ workOnTypes' allowed cont
 
--- | Call me if --experimental-irrelevance is set.
-doWorkOnTypes :: TCM a -> TCM a
-doWorkOnTypes = verboseBracket "tc.irr" 20 "workOnTypes" . workOnTypes' True
-
 -- | Internal workhorse, expects value of --experimental-irrelevance flag
 --   as argument.
 workOnTypes' :: Bool -> TCM a -> TCM a
-workOnTypes' allowed cont =
-  if allowed then
-    liftTCM $ modifyContext (modifyContextEntries $ mapRelevance $ irrToNonStrict) cont
-   else cont
+workOnTypes' experimental cont = modifyContext (modifyContextEntries $ mapRelevance f) cont
+  where
+    f | experimental = irrToNonStrict
+      | otherwise    = nonStrictToRel
 
 -- | (Conditionally) wake up irrelevant variables and make them relevant.
 --   For instance,
@@ -84,18 +77,3 @@ applyRelevanceToContext rel =
 --   may be used, so they are awoken before type checking the argument.
 wakeIrrelevantVars :: TCM a -> TCM a
 wakeIrrelevantVars = applyRelevanceToContext Irrelevant
-
-
-------------------------------------------------------------------------
--- * Tests
-------------------------------------------------------------------------
-
-prop_galois :: Relevance -> Relevance -> Relevance -> Bool
-prop_galois r x y =
-  x `moreRelevant` (r `composeRelevance` y) ==
-  (r `inverseComposeRelevance` x) `moreRelevant` y
-
-tests :: IO Bool
-tests = runTests "Agda.TypeChecking.Irrelevance"
-  [ quickCheck' prop_galois
-  ]
