@@ -1,23 +1,23 @@
-{-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE FlexibleContexts  #-}
-{-# LANGUAGE PatternGuards     #-}
-{-# LANGUAGE RelaxedPolyRec    #-}
-{-# LANGUAGE TupleSections     #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.MetaVars where
 
+import Prelude hiding (null)
+
 import Control.Monad.Reader
 
 import Data.Function
-import Data.List hiding (sort)
+import Data.List hiding (sort, null)
 import qualified Data.List as List
 import qualified Data.Map as Map
 import qualified Data.Foldable as Fold
 
+import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Generic
+import Agda.Syntax.Position (killRange)
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
@@ -47,13 +47,15 @@ import Agda.Utils.Except
   , runExceptT
   )
 
+import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import Agda.Utils.Permutation
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Utils.Pretty ( prettyShow, render )
 import qualified Agda.Utils.VarSet as Set
 
 #include "undefined.h"
@@ -76,7 +78,6 @@ isBlockedTerm x = do
             BlockedConst{}                 -> True
             PostponedTypeCheckingProblem{} -> True
             InstV{}                        -> False
-            InstS{}                        -> False
             Open{}                         -> False
             OpenIFS{}                      -> False
     reportSLn "tc.meta.blocked" 12 $
@@ -90,7 +91,6 @@ isEtaExpandable kinds x = do
       Open{}                         -> True
       OpenIFS{}                      -> notElem Records kinds
       InstV{}                        -> False
-      InstS{}                        -> False
       BlockedConst{}                 -> False
       PostponedTypeCheckingProblem{} -> False
 
@@ -98,7 +98,7 @@ isEtaExpandable kinds x = do
 
 -- | Performing the meta variable assignment.
 --
---   The instantiation should not be an 'InstV' or 'InstS' and the 'MetaId'
+--   The instantiation should not be an 'InstV' and the 'MetaId'
 --   should point to something 'Open' or a 'BlockedConst'.
 --   Further, the meta variable may not be 'Frozen'.
 assignTerm :: MetaId -> [Arg ArgName] -> Term -> TCM ()
@@ -128,7 +128,7 @@ assignTerm' x tel v = do
     -- dontAssignMetas $ do
     --   checkInternal t . jMetaType . mvJudgement =<< lookupMeta x
 
-    let i = metaInstance tel v
+    let i = metaInstance tel $ killRange v
     verboseS "profile.metas" 10 $ liftTCM $ tickMax "max-open-metas" . size =<< getOpenMetas
     modifyMetaStore $ ins x i
     etaExpandListeners x
@@ -144,30 +144,42 @@ assignTerm' x tel v = do
 
 -- * Creating meta variables.
 
+-- | Create a sort meta that cannot be instantiated with 'Inf' (Setω).
+newSortMetaBelowInf :: TCM Sort
+newSortMetaBelowInf = newSortMeta' $ HasType ()
+
+-- | Create a sort meta that may be instantiated with 'Inf' (Setω).
 newSortMeta :: TCM Sort
-newSortMeta =
+newSortMeta = newSortMeta' $ IsSort ()
+
+newSortMeta' :: (Type -> Judgement ()) -> TCM Sort
+newSortMeta' judge =
   ifM typeInType (return $ mkType 0) $ {- else -}
-  ifM hasUniversePolymorphism (newSortMetaCtx =<< getContextArgs)
+  ifM hasUniversePolymorphism (newSortMetaCtx' judge =<< getContextArgs)
   -- else (no universe polymorphism)
   $ do i   <- createMetaInfo
        lvl <- levelType
-       x   <- newMeta i normalMetaPriority (idP 0) $ IsSort () lvl -- WAS: topSort
+       x   <- newMeta i normalMetaPriority (idP 0) $ judge lvl
        return $ Type $ Max [Plus 0 $ MetaLevel x []]
 
+-- | Create a sort meta that may be instantiated with 'Inf' (Setω).
 newSortMetaCtx :: Args -> TCM Sort
-newSortMetaCtx vs =
+newSortMetaCtx = newSortMetaCtx' $ IsSort ()
+
+newSortMetaCtx' :: (Type -> Judgement ()) -> Args -> TCM Sort
+newSortMetaCtx' judge vs = do
   ifM typeInType (return $ mkType 0) $ {- else -} do
     i   <- createMetaInfo
     tel <- getContextTelescope
     lvl <- levelType
-    let t = telePi_ tel lvl -- WAS: topSort
-    x   <- newMeta i normalMetaPriority (idP 0) (IsSort () t)
+    let t = telePi_ tel lvl
+    x   <- newMeta i normalMetaPriority (idP 0) $ judge t
     reportSDoc "tc.meta.new" 50 $
       text "new sort meta" <+> prettyTCM x <+> text ":" <+> prettyTCM t
     return $ Type $ Max [Plus 0 $ MetaLevel x $ map Apply vs]
 
 newTypeMeta :: Sort -> TCM Type
-newTypeMeta s = El s <$> newValueMeta RunMetaOccursCheck (sort s)
+newTypeMeta s = El s . snd <$> newValueMeta RunMetaOccursCheck (sort s)
 
 newTypeMeta_ ::  TCM Type
 newTypeMeta_  = newTypeMeta =<< (workOnTypes $ newSortMeta)
@@ -180,15 +192,15 @@ newTypeMeta_  = newTypeMeta =<< (workOnTypes $ newSortMeta)
 --   of type the output type of @t@ with name suggestion @s@.
 --   If @t@ is a function type, then insert enough
 --   lambdas in front of it.
-newIFSMeta :: MetaNameSuggestion -> Type -> TCM Term
+newIFSMeta :: MetaNameSuggestion -> Type -> TCM (MetaId, Term)
 newIFSMeta s t = do
   TelV tel t' <- telView t
-  addCtxTel tel $ do
+  addContext tel $ do
     vs  <- getContextArgs
     ctx <- getContextTelescope
-    teleLam tel <$> newIFSMetaCtx s (telePi_ ctx t') vs
+    mapSnd (teleLam tel) <$> newIFSMetaCtx s (telePi_ ctx t') vs
 
-newIFSMetaCtx :: MetaNameSuggestion -> Type -> Args -> TCM Term
+newIFSMetaCtx :: MetaNameSuggestion -> Type -> Args -> TCM (MetaId, Term)
 newIFSMetaCtx s t vs = do
   reportSDoc "tc.meta.new" 50 $ fsep
     [ text "new ifs meta:"
@@ -204,35 +216,42 @@ newIFSMetaCtx s t vs = do
     ]
   addConstraint $ FindInScope x Nothing Nothing
   etaExpandMetaSafe x
-  return $ MetaV x $ map Apply vs
+  return (x, MetaV x $ map Apply vs)
 
--- | Create a new value meta with specific dependencies.
-newNamedValueMeta :: RunMetaOccursCheck -> MetaNameSuggestion -> Type -> TCM Term
+-- | Create a new value meta with specific dependencies, possibly η-expanding in the process.
+newNamedValueMeta :: RunMetaOccursCheck -> MetaNameSuggestion -> Type -> TCM (MetaId, Term)
 newNamedValueMeta b s t = do
-  v <- newValueMeta b t
-  setValueMetaName v s
-  return v
+  (x, v) <- newValueMeta b t
+  setMetaNameSuggestion x s
+  return (x, v)
+
+-- | Create a new value meta with specific dependencies without η-expanding.
+newNamedValueMeta' :: RunMetaOccursCheck -> MetaNameSuggestion -> Type -> TCM (MetaId, Term)
+newNamedValueMeta' b s t = do
+  (x, v) <- newValueMeta' b t
+  setMetaNameSuggestion x s
+  return (x, v)
 
 -- | Create a new metavariable, possibly η-expanding in the process.
-newValueMeta :: RunMetaOccursCheck -> Type -> TCM Term
+newValueMeta :: RunMetaOccursCheck -> Type -> TCM (MetaId, Term)
 newValueMeta b t = do
   vs  <- getContextArgs
   tel <- getContextTelescope
   newValueMetaCtx b t tel (idP $ size tel) vs
 
-newValueMetaCtx :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM Term
+newValueMetaCtx :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM (MetaId, Term)
 newValueMetaCtx b t tel perm ctx =
-  instantiateFull =<< newValueMetaCtx' b t tel perm ctx
+  mapSndM instantiateFull =<< newValueMetaCtx' b t tel perm ctx
 
 -- | Create a new value meta without η-expanding.
-newValueMeta' :: RunMetaOccursCheck -> Type -> TCM Term
+newValueMeta' :: RunMetaOccursCheck -> Type -> TCM (MetaId, Term)
 newValueMeta' b t = do
   vs  <- getContextArgs
   tel <- getContextTelescope
   newValueMetaCtx' b t tel (idP $ size tel) vs
 
 -- | Create a new value meta with specific dependencies.
-newValueMetaCtx' :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM Term
+newValueMetaCtx' :: RunMetaOccursCheck -> Type -> Telescope -> Permutation -> Args -> TCM (MetaId, Term)
 newValueMetaCtx' b a tel perm vs = do
   i <- createMetaInfo' b
   let t     = telePi_ tel a
@@ -246,7 +265,7 @@ newValueMetaCtx' b a tel perm vs = do
   -- Andreas, 2012-09-24: for Metas X : Size< u add constraint X+1 <= u
   u <- shared $ MetaV x $ map Apply vs
   boundedSizeMetaHook u tel a
-  return u
+  return (x, u)
 
 newTelMeta :: Telescope -> TCM Args
 newTelMeta tel = newArgsMeta (abstract tel $ typeDontCare)
@@ -273,7 +292,7 @@ newArgsMetaCtx' condition (El s tm) tel perm ctx = do
   tm <- reduce tm
   case ignoreSharing tm of
     Pi dom@(Dom info a) codom | condition dom codom -> do
-      u <- applyRelevanceToContext (getRelevance info) $
+      (_, u) <- applyRelevanceToContext (getRelevance info) $
                {-
                  -- Andreas, 2010-09-24 skip irrelevant record fields when eta-expanding a meta var
                  -- Andreas, 2010-10-11 this is WRONG, see Issue 347
@@ -297,16 +316,26 @@ newRecordMetaCtx r pars tel perm ctx = do
   ftel   <- flip apply pars <$> getRecordFieldTypes r
   fields <- newArgsMetaCtx (telePi_ ftel $ sort Prop) tel perm ctx
   con    <- getRecordConstructor r
-  return $ Con con fields
+  return $ Con con ConOSystem fields
+
+newQuestionMark :: InteractionId -> Type -> TCM (MetaId, Term)
+newQuestionMark = newQuestionMark' $ newValueMeta' DontRunMetaOccursCheck
+
+newQuestionMark' :: (Type -> TCM (MetaId, Term)) -> InteractionId -> Type -> TCM (MetaId, Term)
+newQuestionMark' new ii t = do
+  -- Andreas, 2016-07-29, issue 1720-2
+  -- This is slightly risky, as the same interaction id
+  -- maybe be shared between different contexts.
+  -- Blame goes to the record processing hack, see issue #424
+  -- and @ConcreteToAbstract.recordConstructorType@.
+  let existing x = (x,) . MetaV x . map Apply <$> getContextArgs
+  flip (caseMaybeM $ lookupInteractionMeta ii) existing $ {-else-} do
 
-newQuestionMark :: InteractionId -> Type -> TCM Term
-newQuestionMark ii t = do
   -- Do not run check for recursive occurrence of meta in definitions,
   -- because we want to give the recursive solution interactively (Issue 589)
-  m  <- newValueMeta' DontRunMetaOccursCheck t
-  MetaV x _ <- return $ ignoreSharing m   -- needs to be strict!
+  (x, m) <- new t
   connectInteractionPoint ii x
-  return m
+  return (x, m)
 
 -- | Construct a blocked constant if there are constraints.
 blockTerm :: Type -> TCM Term -> TCM Term
@@ -338,10 +367,10 @@ blockTermOnProblem t v pid =
         -- blocked terms can be instantiated before they are unblocked, thus making
         -- constraint solving a bit more robust against instantiation order.
         -- Andreas, 2015-05-22: DontRunMetaOccursCheck to avoid Issue585-17.
-        v   <- newValueMeta DontRunMetaOccursCheck t
+        (_, v) <- newValueMeta DontRunMetaOccursCheck t
         i   <- liftTCM fresh
         -- This constraint is woken up when unblocking, so it doesn't need a problem id.
-        cmp <- buildProblemConstraint 0 (ValueCmp CmpEq t v (MetaV x es))
+        cmp <- buildProblemConstraint_ (ValueCmp CmpEq t v (MetaV x es))
         listenToMeta (CheckConstraint i cmp) x
         return v
 
@@ -387,8 +416,8 @@ postponeTypeCheckingProblem p unblock = do
   -- to run the extended occurs check (metaOccurs) to exclude
   -- non-terminating solutions.
   es  <- map Apply <$> getContextArgs
-  v   <- newValueMeta DontRunMetaOccursCheck t
-  cmp <- buildProblemConstraint 0 (ValueCmp CmpEq t v (MetaV m es))
+  (_, v) <- newValueMeta DontRunMetaOccursCheck t
+  cmp <- buildProblemConstraint_ (ValueCmp CmpEq t v (MetaV m es))
   i   <- liftTCM fresh
   listenToMeta (CheckConstraint i cmp) m
   addConstraint (UnBlock m)
@@ -609,23 +638,33 @@ assign dir x args v = do
       -- Andreas, 2011-04-21 do the occurs check first
       -- e.g. _1 x (suc x) = suc (_2 x y)
       -- even though the lhs is not a pattern, we can prune the y from _2
-      let relVL = Set.toList $ allRelevantVars args
-{- Andreas, 2012-04-02: DontCare no longer present
-      -- take away top-level DontCare constructors
-      args <- return $ map (fmap stripDontCare) args
--}
-      -- Andreas, 2011-10-06 only irrelevant vars that are direct
-      -- arguments to the meta, hence, can be abstracted over, may
-      -- appear on the rhs.  (test/fail/Issue483b)
-      -- Update 2011-03-27: Also irr. vars under record constructors.
-      let fromIrrVar (Var i [])   = return [i]
-          fromIrrVar (Con c vs)   =
-            ifM (isNothing <$> isRecordConstructor (conName c)) (return []) $
-              concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) vs
-          fromIrrVar (Shared p)   = fromIrrVar (derefPtr p)
-          fromIrrVar _ = return []
-      irrVL <- concat <$> mapM fromIrrVar
-                 [ v | Arg info v <- args, irrelevantOrUnused (getRelevance info) ]
+
+      (relVL, irrVL) <- do
+        -- Andreas, 2016-11-03 #2211 attempt to do s.th. for unused
+        if False -- irrelevantOrUnused $ getMetaRelevance mvar
+          then do
+            reportSDoc "tc.meta.assign" 25 $ text "meta is irrelevant or unused"
+            return (Set.toList $ allFreeVars args, empty)
+          else do
+            -- Andreas, 2016-11-03, issue #2211
+            -- treating UnusedArg as Irrelevant bears trouble
+            -- since the UnusedArg info is not consistently present
+            -- Thus, make sure we include the "unused" variables.
+            let relVL = Set.toList $ allRelevantOrUnusedVars args
+            -- Andreas, 2011-10-06 only irrelevant vars that are direct
+            -- arguments to the meta, hence, can be abstracted over, may
+            -- appear on the rhs.  (test/fail/Issue483b)
+            -- Update 2011-03-27: Also irr. vars under record constructors.
+            let fromIrrVar (Var i [])   = return [i]
+                fromIrrVar (Con c _ vs)   =
+                  ifM (isNothing <$> isRecordConstructor (conName c)) (return []) $
+                    concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) vs
+                fromIrrVar (Shared p)   = fromIrrVar (derefPtr p)
+                fromIrrVar _ = return []
+            irrVL <- concat <$> mapM fromIrrVar
+                       [ v | Arg info v <- args, isIrrelevant info ]
+                          -- irrelevantOrUnused (getRelevance info) ]
+            return (relVL, irrVL)
       reportSDoc "tc.meta.assign" 20 $
           let pr (Var n []) = text (show n)
               pr (Def c []) = prettyTCM c
@@ -677,7 +716,7 @@ assign dir x args v = do
           Left NeutralArg  -> Just <$> attemptPruning x args fvs
           -- we have a projected variable which could not be eta-expanded away:
           -- same as neutral
-          Left (ProjectedVar i qs) -> Just <$> attemptPruning x args fvs
+          Left ProjectedVar{} -> Just <$> attemptPruning x args fvs
 
       case mids of
         Nothing  -> patternViolation -- Ulf 2014-07-13: actually not needed after all: attemptInertRHSImprovement x args v
@@ -732,7 +771,7 @@ attemptInertRHSImprovement m args v = do
     patternViolation
   -- Solve the meta with _M := \ xs -> D (_Y1 xs) .. (_Yn xs), for fresh metas
   -- _Yi.
-  metaArgs <- inTopContext $ addCtxTel tel $ newArgsMeta a
+  metaArgs <- inTopContext $ addContext tel $ newArgsMeta a
   let varArgs = map Apply $ reverse $ zipWith (\i a -> var i <$ a) [0..] (reverse args)
       sol     = mkRHS metaArgs
       argTel  = map ("x" <$) args
@@ -759,7 +798,7 @@ attemptInertRHSImprovement m args v = do
               Just args -> return args
       case ignoreSharing v of
         Var x elims -> (, Var x . map Apply) <$> typeOfBV x
-        Con c args  -> notInert -- (, Con c) <$> defType <$> getConstInfo (conName c)
+        Con c ci args  -> notInert -- (, Con c ci) <$> defType <$> getConstInfo (conName c)
         Def f elims -> do
           def <- getConstInfo f
           let good = return (defType def, Def f . map Apply)
@@ -950,7 +989,7 @@ etaExpandProjectedVar i v fail succeed = do
 class NoProjectedVar a where
   noProjectedVar :: a -> Either ProjVarExc ()
 
-data ProjVarExc = ProjVarExc Int [QName]
+data ProjVarExc = ProjVarExc Int [(ProjOrigin, QName)]
 
 -- ASR (17 June 2015). Unused Error instance.
 -- instance Error ProjVarExc where
@@ -963,7 +1002,7 @@ instance NoProjectedVar Term where
         | qs@(_:_) <- takeWhileJust id $ map isProjElim es -> Left $ ProjVarExc i qs
       -- Andreas, 2015-09-12 Issue 1316:
       -- Also look in inductive record constructors
-      Con (ConHead _ Inductive (_:_)) vs -> noProjectedVar vs
+      Con (ConHead _ Inductive (_:_)) _ vs -> noProjectedVar vs
       _ -> return ()
 
 instance NoProjectedVar a => NoProjectedVar (Arg a) where
@@ -1026,7 +1065,7 @@ etaExpandProjectedVar mvar x t n qs = inTopContext $ do
 -}
 
 type FVs = Set.VarSet
-type SubstCand = [(Nat,Term)] -- ^ a possibly non-deterministic substitution
+type SubstCand = [(Int,Term)] -- ^ a possibly non-deterministic substitution
 
 -- | Turn non-det substitution into proper substitution, if possible.
 --   Otherwise, raise the error.
@@ -1053,7 +1092,7 @@ type Res = [(Arg Nat, Term)]
 data InvertExcept
   = CantInvert                -- ^ Cannot recover.
   | NeutralArg                -- ^ A potentially neutral arg: can't invert, but can try pruning.
-  | ProjectedVar Int [QName]  -- ^ Try to eta-expand var to remove projs.
+  | ProjectedVar Int [(ProjOrigin, QName)]  -- ^ Try to eta-expand var to remove projs.
 
 #if !MIN_VERSION_transformers(0,4,1)
 instance Error InvertExcept where
@@ -1095,18 +1134,24 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
         -- (i, j) := x  becomes  [i := fst x, j := snd x]
         -- Andreas, 2013-09-17 but only if constructor is fully applied
-        Arg info (Con c vs) -> do
+        Arg info (Con c ci vs) -> do
           let fallback
-               | irrelevantOrUnused (getRelevance info) = return vars
+               | isIrrelevant info = return vars
+                 -- Andreas, 2016-11-03, issue #2211
+                 -- treating UnusedArg as Irrelevant bears trouble
+                 -- since the UnusedArg info is not consistently present
+                 -- irrelevantOrUnused (getRelevance info) = return vars
                | otherwise                              = failure
           isRC <- lift $ isRecordConstructor $ conName c
           case isRC of
             Just (_, Record{ recFields = fs })
               | length fs == length vs -> do
-                let aux (Arg _ v) (Arg info' f) = (Arg ai v,) $ t `applyE` [Proj f] where
+                let aux (Arg _ v) (Arg info' f) = (Arg ai v,) $ t `applyE` [Proj ProjSystem f] where
                      ai = ArgInfo
-                       { argInfoHiding    = min (getHiding info) (getHiding info')
-                       , argInfoRelevance = max (getRelevance info) (getRelevance info')
+                       { argInfoHiding       = min (getHiding info) (getHiding info')
+                       , argInfoRelevance    = max (getRelevance info) (getRelevance info')
+                       , argInfoOrigin       = min (getOrigin info) (getOrigin info')
+                       , argInfoOverlappable = False
                        }
                 res <- loop $ zipWith aux vs fs
                 return $ res `append` vars
@@ -1115,7 +1160,11 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
             Nothing -> fallback
 
         -- An irrelevant argument which is not an irrefutable pattern is dropped
-        Arg info _ | irrelevantOrUnused (getRelevance info) -> return vars
+        Arg info _ | isIrrelevant info -> return vars
+          -- Andreas, 2016-11-03, issue #2211
+          -- treating UnusedArg as Irrelevant bears trouble
+          -- since the UnusedArg info is not consistently present
+          -- irrelevantOrUnused (getRelevance info) -> return vars
         -- Andreas, 2013-10-29
         -- An irrelevant part can also be marked by a DontCare
         -- (coming from an irrelevant projection), see Issue 927:
@@ -1140,7 +1189,7 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
     -- adding an irrelevant entry only if not present
     cons :: (Arg Nat, Term) -> Res -> Res
-    cons a@(Arg (ArgInfo _ Irrelevant) i, t) vars    -- TODO? UnusedArg?!
+    cons a@(Arg (ArgInfo _ Irrelevant _ _) i, t) vars    -- TODO? UnusedArg?!
       | any ((i==) . unArg . fst) vars  = vars
       | otherwise                       = a : vars
     -- adding a relevant entry:
@@ -1148,19 +1197,54 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
       -- filter out duplicate irrelevants
       filter (not . (\ a@(Arg info j, t) -> isIrrelevant info && i == j)) vars
 
--- | Used in 'Agda.Interaction.BasicOps.giveExpr'.
-updateMeta :: MetaId -> Term -> TCM ()
-updateMeta mI v = do
-    mv <- lookupMeta mI
-    withMetaInfo' mv $ do
-      args <- getContextArgs
-      noConstraints $ assignV DirEq mI args v
+-- UNUSED
+-- -- | Used in 'Agda.Interaction.BasicOps.giveExpr'.
+-- updateMeta :: MetaId -> Term -> TCM ()
+-- updateMeta mI v = do
+--     mv <- lookupMeta mI
+--     withMetaInfo' mv $ do
+--       args <- getContextArgs
+--       noConstraints $ assignV DirEq mI args v
 
--- | Returns every meta-variable occurrence in the given type, except
--- for those in 'Sort's.
+-- | Turn open metas into postulates.
+--
+--   Preconditions:
+--
+--   1. We are 'inTopContext'.
+--
+--   2. 'envCurrentModule' is set to the top-level module.
+--
+openMetasToPostulates :: TCM ()
+openMetasToPostulates = do
+  m <- asks envCurrentModule
 
-allMetas :: TermLike a => a -> [MetaId]
-allMetas = foldTerm metas
-  where
-  metas (MetaV m _) = [m]
-  metas _           = []
+  -- Go through all open metas.
+  ms <- Map.assocs <$> use stMetaStore
+  forM_ ms $ \ (x, mv) -> do
+    when (isOpenMeta $ mvInstantiation mv) $ do
+      let t = jMetaType $ mvJudgement mv
+
+      -- Create a name for the new postulate.
+      let r = clValue $ miClosRange $ mvInfo mv
+      -- s <- render <$> prettyTCM x -- Using _ is a bad idea, as it prints as prefix op
+      let s = "unsolved#meta." ++ show (metaId x)
+      n <- freshName r s
+      let q = A.QName m n
+
+      -- Debug.
+      reportSDoc "meta.postulate" 20 $ vcat
+        [ text ("Turning " ++ if isSortMeta_ mv then "sort" else "value" ++ " meta ")
+            <+> prettyTCM x <+> text " into postulate."
+        , nest 2 $ vcat
+          [ text "Name: " <+> prettyTCM q
+          , text "Type: " <+> prettyTCM t
+          ]
+        ]
+
+      -- Add the new postulate to the signature.
+      addConstant q $ defaultDefn defaultArgInfo q t Axiom
+
+      -- Solve the meta.
+      let inst = InstV [] $ Def q []
+      stMetaStore %= Map.adjust (\ mv0 -> mv0 { mvInstantiation = inst }) x
+      return ()
