@@ -1,7 +1,6 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 {- | The occurs check for unification.  Does pruning on the fly.
@@ -25,6 +24,7 @@ import Data.Foldable (foldMap)
 import Data.Monoid
 import Data.Set (Set)
 import qualified Data.Set as Set
+import Data.Traversable (traverse)
 
 import qualified Agda.Benchmarking as Bench
 
@@ -37,6 +37,7 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Free hiding (Occurrence(..))
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Records
 import {-# SOURCE #-} Agda.TypeChecking.MetaVars
 -- import Agda.TypeChecking.MetaVars
@@ -116,7 +117,7 @@ initOccursCheck mv = modifyOccursCheckDefs . const =<<
          reportSLn "tc.meta.occurs" 20 $ "(none)"
          return Set.empty
        Just b  -> do
-         ds <- lookupMutualBlock b
+         ds <- mutualNames <$> lookupMutualBlock b
          reportSDoc "tc.meta.occurs" 20 $ sep $ map prettyTCM $ Set.toList ds
          return ds
 
@@ -130,11 +131,11 @@ tallyDef :: QName -> TCM ()
 tallyDef d = modifyOccursCheckDefs $ \ s -> Set.delete d s
 
 data OccursCtx
-  = Flex          -- ^ we are in arguments of a meta
-  | Rigid         -- ^ we are not in arguments of a meta but a bound var
-  | StronglyRigid -- ^ we are at the start or in the arguments of a constructor
-  | Top           -- ^ we are at the term root (this turns into @StronglyRigid@)
-  | Irrel         -- ^ we are in an irrelevant argument
+  = Flex          -- ^ We are in arguments of a meta.
+  | Rigid         -- ^ We are not in arguments of a meta but a bound var.
+  | StronglyRigid -- ^ We are at the start or in the arguments of a constructor.
+  | Top           -- ^ We are at the term root (this turns into @StronglyRigid@).
+  | Irrel         -- ^ We are in an irrelevant argument.
   deriving (Eq, Show)
 
 data UnfoldStrategy = YesUnfold | NoUnfold
@@ -269,7 +270,7 @@ instance Occurs Term where
           Lit l       -> return v
           DontCare v  -> dontCare <$> occurs red Irrel m (goIrrelevant xs) v
           Def d es    -> Def d <$> occDef d (leaveTop ctx) es
-          Con c vs    -> Con c <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
+          Con c ci vs -> Con c ci <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
           Pi a b      -> uncurry Pi <$> occ (leaveTop ctx) (a,b)
           Sort s      -> Sort <$> occ (leaveTop ctx) s
           v@Shared{}  -> updateSharedTerm (occ ctx) v
@@ -316,12 +317,10 @@ instance Occurs Term where
             -- a data or record type constructor propagates strong occurrences
             -- since e.g. x = List x is unsolvable
             occDef d ctx vs = do
-              def <- theDef <$> getConstInfo d
-              whenM (defNeedsChecking d) $ do
-                tallyDef d
-                reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
-                metaOccurs m def
-              if (defIsDataOrRecord def) then (occ ctx vs) else (occ (defArgs red ctx) vs)
+              metaOccurs m d
+              ifM (isJust <$> isDataOrRecordType d)
+                {-then-} (occ ctx vs)
+                {-else-} (occ (defArgs red ctx) vs)
 
   metaOccurs m v = do
     v <- instantiate v
@@ -332,7 +331,7 @@ instance Occurs Term where
       Lit l      -> return ()
       DontCare v -> metaOccurs m v
       Def d vs   -> metaOccurs m d >> metaOccurs m vs
-      Con c vs   -> metaOccurs m vs
+      Con c _ vs -> metaOccurs m vs
       Pi a b     -> metaOccurs m (a,b)
       Sort s     -> metaOccurs m s
       Shared p   -> metaOccurs m $ derefPtr p
@@ -345,7 +344,7 @@ instance Occurs QName where
   metaOccurs m d = whenM (defNeedsChecking d) $ do
     tallyDef d
     reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
-    metaOccurs m . theDef =<< getConstInfo d
+    metaOccurs m . theDef =<< ignoreAbstractMode (getConstInfo d)
 
 instance Occurs Defn where
   occurs red ctx m xs def = __IMPOSSIBLE__
@@ -356,17 +355,15 @@ instance Occurs Defn where
   -- we check the constructor types
   metaOccurs m Datatype{ dataCons = cs }    = mapM_ mocc cs
     where mocc c = metaOccurs m . defType =<< getConstInfo c
-  metaOccurs m Record{ recConType = v }     = metaOccurs m v
+  metaOccurs m Record{ recConHead = c }     = metaOccurs m . defType =<< getConstInfo (conName c)
   metaOccurs m Constructor{}                = return ()
   metaOccurs m Primitive{}                  = return ()
+  metaOccurs m AbstractDefn{}               = __IMPOSSIBLE__
 
 instance Occurs Clause where
   occurs red ctx m xs cl = __IMPOSSIBLE__
 
-  metaOccurs m (Clause { clauseBody = body }) = walk body
-    where walk NoBody   = return ()
-          walk (Body v) = metaOccurs m v
-          walk (Bind b) = underAbstraction_ b walk
+  metaOccurs m = metaOccurs m . clauseBody
 
 instance Occurs Level where
   occurs red ctx m xs (Max as) = Max <$> occurs red ctx m xs as
@@ -465,6 +462,11 @@ instance Occurs a => Occurs [a] where
 
   metaOccurs m ys = mapM_ (metaOccurs m) ys
 
+instance Occurs a => Occurs (Maybe a) where
+  occurs red ctx m mx my = traverse (occurs red ctx m mx) my
+
+  metaOccurs m = maybe (return ()) (metaOccurs m)
+
 -- * Getting rid of flexible occurrences
 
 -- | @prune m' vs xs@ attempts to remove all arguments from @vs@ whose
@@ -525,7 +527,7 @@ hasBadRigid xs t = do
     -- offending variables under a constructor could be removed by
     -- the right instantiation of the meta variable.
     -- Thus, they are not rigid.
-    Con c args   -> do
+    Con c _ args -> do
       ifM (liftTCM $ isEtaCon (conName c))
         -- in case of a record con, we can in principle prune
         -- (but not this argument; the meta could become a projection!)
@@ -542,7 +544,9 @@ isNeutral b f es = liftTCM $ do
   let yes = return True
       no  = return False
   def <- getConstInfo f
+  if defMatchable def then no else do
   case theDef def of
+    AbstractDefn -> yes
     Axiom{}    -> yes
     Datatype{} -> yes
     Record{}   -> yes
@@ -593,7 +597,7 @@ instance FoldRigid Term where
         Blocked{}                   -> mempty
         NotBlocked MissingClauses _ -> mempty
         _        -> fold es
-      Con _ ts   -> fold ts
+      Con _ _ ts -> fold ts
       Pi a b     -> fold (a,b)
       Sort s     -> fold s
       Level l    -> fold l
