@@ -5,6 +5,7 @@ module Agda.TypeChecking.Monad
     , module Agda.TypeChecking.Monad.Context
     , module Agda.TypeChecking.Monad.Env
     , module Agda.TypeChecking.Monad.Imports
+    , module Agda.TypeChecking.Monad.Local
     , module Agda.TypeChecking.Monad.MetaVars
     , module Agda.TypeChecking.Monad.Mutual
     , module Agda.TypeChecking.Monad.Open
@@ -24,6 +25,7 @@ import Agda.TypeChecking.Monad.Constraints
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Imports
+import Agda.TypeChecking.Monad.Local
 import Agda.TypeChecking.Monad.MetaVars
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Options
