@@ -4,6 +4,7 @@ module Agda.TypeChecking.Monad.Builtin where
 
 import Control.Applicative
 import Control.Monad.State
+import Control.Monad.Trans.Maybe
 
 import qualified Data.Map as Map
 
@@ -16,7 +17,7 @@ import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Lens
-import Agda.Utils.Monad (when_)
+import Agda.Utils.Monad
 import Agda.Utils.Maybe
 import Agda.Utils.Tuple
 
@@ -26,6 +27,9 @@ import Agda.Utils.Impossible
 class (Functor m, Applicative m, Monad m) => HasBuiltins m where
   getBuiltinThing :: String -> m (Maybe (Builtin PrimFun))
 
+instance HasBuiltins m => HasBuiltins (MaybeT m) where
+  getBuiltinThing b = lift $ getBuiltinThing b
+
 litType :: Literal -> TCM Type
 litType l = case l of
   LitNat _ n    -> do
@@ -116,6 +120,9 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaErrorPart, primAgdaErrorPartString, primAgdaErrorPartTerm, primAgdaErrorPartName,
     primHiding, primHidden, primInstance, primVisible,
     primRelevance, primRelevant, primIrrelevant,
+    primAssoc, primAssocLeft, primAssocRight, primAssocNon,
+    primPrecedence, primPrecRelated, primPrecUnrelated,
+    primFixity, primFixityFixity,
     primAgdaLiteral, primAgdaLitNat, primAgdaLitFloat, primAgdaLitString, primAgdaLitChar, primAgdaLitQName, primAgdaLitMeta,
     primAgdaSort, primAgdaSortSet, primAgdaSortLit, primAgdaSortUnsupported,
     primAgdaDefinition, primAgdaDefinitionFunDef, primAgdaDefinitionDataDef, primAgdaDefinitionRecordDef,
@@ -127,11 +134,13 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaMeta,
     primAgdaTCM, primAgdaTCMReturn, primAgdaTCMBind, primAgdaTCMUnify,
     primAgdaTCMTypeError, primAgdaTCMInferType, primAgdaTCMCheckType,
-    primAgdaTCMNormalise, primAgdaTCMCatchError, primAgdaTCMGetContext, primAgdaTCMExtendContext, primAgdaTCMInContext,
+    primAgdaTCMNormalise, primAgdaTCMReduce,
+    primAgdaTCMCatchError, primAgdaTCMGetContext, primAgdaTCMExtendContext, primAgdaTCMInContext,
     primAgdaTCMFreshName, primAgdaTCMDeclareDef, primAgdaTCMDefineFun,
     primAgdaTCMGetType, primAgdaTCMGetDefinition,
     primAgdaTCMQuoteTerm, primAgdaTCMUnquoteTerm,
-    primAgdaTCMBlockOnMeta, primAgdaTCMCommit
+    primAgdaTCMBlockOnMeta, primAgdaTCMCommit, primAgdaTCMIsMacro,
+    primAgdaTCMWithNormalisation
     :: TCM Term
 
 primInteger      = getBuiltin builtinInteger
@@ -191,6 +200,15 @@ primVisible      = getBuiltin builtinVisible
 primRelevance    = getBuiltin builtinRelevance
 primRelevant     = getBuiltin builtinRelevant
 primIrrelevant   = getBuiltin builtinIrrelevant
+primAssoc        = getBuiltin builtinAssoc
+primAssocLeft    = getBuiltin builtinAssocLeft
+primAssocRight   = getBuiltin builtinAssocRight
+primAssocNon     = getBuiltin builtinAssocNon
+primPrecedence    = getBuiltin builtinPrecedence
+primPrecRelated   = getBuiltin builtinPrecRelated
+primPrecUnrelated = getBuiltin builtinPrecUnrelated
+primFixity        = getBuiltin builtinFixity
+primFixityFixity  = getBuiltin builtinFixityFixity
 primArgInfo      = getBuiltin builtinArgInfo
 primArgArgInfo   = getBuiltin builtinArgArgInfo
 primAgdaSortSet  = getBuiltin builtinAgdaSortSet
@@ -244,6 +262,7 @@ primAgdaTCMTypeError  = getBuiltin builtinAgdaTCMTypeError
 primAgdaTCMInferType  = getBuiltin builtinAgdaTCMInferType
 primAgdaTCMCheckType  = getBuiltin builtinAgdaTCMCheckType
 primAgdaTCMNormalise  = getBuiltin builtinAgdaTCMNormalise
+primAgdaTCMReduce     = getBuiltin builtinAgdaTCMReduce
 primAgdaTCMCatchError = getBuiltin builtinAgdaTCMCatchError
 primAgdaTCMGetContext = getBuiltin builtinAgdaTCMGetContext
 primAgdaTCMExtendContext = getBuiltin builtinAgdaTCMExtendContext
@@ -257,6 +276,8 @@ primAgdaTCMQuoteTerm          = getBuiltin builtinAgdaTCMQuoteTerm
 primAgdaTCMUnquoteTerm        = getBuiltin builtinAgdaTCMUnquoteTerm
 primAgdaTCMBlockOnMeta        = getBuiltin builtinAgdaTCMBlockOnMeta
 primAgdaTCMCommit             = getBuiltin builtinAgdaTCMCommit
+primAgdaTCMIsMacro            = getBuiltin builtinAgdaTCMIsMacro
+primAgdaTCMWithNormalisation  = getBuiltin builtinAgdaTCMWithNormalisation
 
 builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinNatTimes, builtinNatDivSucAux, builtinNatModSucAux, builtinNatEquals,
@@ -274,6 +295,9 @@ builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinAgdaSortUnsupported,
   builtinHiding, builtinHidden, builtinInstance, builtinVisible,
   builtinRelevance, builtinRelevant, builtinIrrelevant, builtinArg,
+  builtinAssoc, builtinAssocLeft, builtinAssocRight, builtinAssocNon,
+  builtinPrecedence, builtinPrecRelated, builtinPrecUnrelated,
+  builtinFixity, builtinFixityFixity,
   builtinArgInfo, builtinArgArgInfo, builtinArgArg,
   builtinAbs, builtinAbsAbs, builtinAgdaTerm,
   builtinAgdaTermVar, builtinAgdaTermLam, builtinAgdaTermExtLam,
@@ -292,12 +316,14 @@ builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinAgdaMeta,
   builtinAgdaTCM, builtinAgdaTCMReturn, builtinAgdaTCMBind, builtinAgdaTCMUnify,
   builtinAgdaTCMTypeError, builtinAgdaTCMInferType,
-  builtinAgdaTCMCheckType, builtinAgdaTCMNormalise, builtinAgdaTCMCatchError,
+  builtinAgdaTCMCheckType, builtinAgdaTCMNormalise, builtinAgdaTCMReduce,
+  builtinAgdaTCMCatchError,
   builtinAgdaTCMGetContext, builtinAgdaTCMExtendContext, builtinAgdaTCMInContext,
   builtinAgdaTCMFreshName, builtinAgdaTCMDeclareDef, builtinAgdaTCMDefineFun,
   builtinAgdaTCMGetType, builtinAgdaTCMGetDefinition,
   builtinAgdaTCMQuoteTerm, builtinAgdaTCMUnquoteTerm,
-  builtinAgdaTCMBlockOnMeta, builtinAgdaTCMCommit
+  builtinAgdaTCMBlockOnMeta, builtinAgdaTCMCommit, builtinAgdaTCMIsMacro,
+  builtinAgdaTCMWithNormalisation
   :: String
 
 builtinNat                           = "NATURAL"
@@ -356,6 +382,15 @@ builtinVisible                       = "VISIBLE"
 builtinRelevance                     = "RELEVANCE"
 builtinRelevant                      = "RELEVANT"
 builtinIrrelevant                    = "IRRELEVANT"
+builtinAssoc                         = "ASSOC"
+builtinAssocLeft                     = "ASSOCLEFT"
+builtinAssocRight                    = "ASSOCRIGHT"
+builtinAssocNon                      = "ASSOCNON"
+builtinPrecedence                    = "PRECEDENCE"
+builtinPrecRelated                   = "PRECRELATED"
+builtinPrecUnrelated                 = "PRECUNRELATED"
+builtinFixity                        = "FIXITY"
+builtinFixityFixity                  = "FIXITYFIXITY"
 builtinArg                           = "ARG"
 builtinArgInfo                       = "ARGINFO"
 builtinArgArgInfo                    = "ARGARGINFO"
@@ -410,6 +445,7 @@ builtinAgdaTCMTypeError  = "AGDATCMTYPEERROR"
 builtinAgdaTCMInferType  = "AGDATCMINFERTYPE"
 builtinAgdaTCMCheckType  = "AGDATCMCHECKTYPE"
 builtinAgdaTCMNormalise  = "AGDATCMNORMALISE"
+builtinAgdaTCMReduce     = "AGDATCMREDUCE"
 builtinAgdaTCMCatchError = "AGDATCMCATCHERROR"
 builtinAgdaTCMGetContext = "AGDATCMGETCONTEXT"
 builtinAgdaTCMExtendContext = "AGDATCMEXTENDCONTEXT"
@@ -423,6 +459,8 @@ builtinAgdaTCMBlockOnMeta   = "AGDATCMBLOCKONMETA"
 builtinAgdaTCMCommit        = "AGDATCMCOMMIT"
 builtinAgdaTCMQuoteTerm     = "AGDATCMQUOTETERM"
 builtinAgdaTCMUnquoteTerm   = "AGDATCMUNQUOTETERM"
+builtinAgdaTCMIsMacro       = "AGDATCMISMACRO"
+builtinAgdaTCMWithNormalisation = "AGDATCMWITHNORMALISATION"
 
 -- | Builtins that come without a definition in Agda syntax.
 --   These are giving names to Agda internal concepts which
@@ -467,9 +505,7 @@ coinductionKit' = do
     }
 
 coinductionKit :: TCM (Maybe CoinductionKit)
-coinductionKit =
-  (Just <$> coinductionKit')
-  `catchError` \_ -> return Nothing
+coinductionKit = tryMaybe coinductionKit'
 
 ------------------------------------------------------------------------
 -- * Builtin equality
