@@ -23,7 +23,7 @@ import Agda.Utils.Impossible
 
 -- | Get the current problem
 currentProblem :: TCM ProblemId
-currentProblem = headWithDefault __IMPOSSIBLE__ <$> asks envActiveProblems
+currentProblem = headWithDefault 0 <$> asks envActiveProblems
 
 -- | Steal all constraints belonging to the given problem and add them to the current problem.
 stealConstraints :: ProblemId -> TCM ()
@@ -78,13 +78,6 @@ wakeConstraints wake = do
   modifySleepingConstraints $ const sleepin
   modifyAwakeConstraints (++ wakeup)
 
-partitionM :: (a -> TCM Bool) -> [a] -> TCM ([a], [a])
-partitionM f [] = return ([], [])
-partitionM f (x:xs) = do
-  b <- f x
-  (l, r) <- partitionM f xs
-  if b then return (x:l, r) else return (l, x:r)
-
 -- danger...
 dropConstraints :: (ProblemConstraint -> Bool) -> TCM ()
 dropConstraints crit = do
@@ -92,11 +85,15 @@ dropConstraints crit = do
   modifySleepingConstraints filt
   modifyAwakeConstraints    filt
 
-putAllConstraintsToSleep :: TCM ()
-putAllConstraintsToSleep = do
+putConstraintsToSleep :: (ProblemConstraint -> Bool) -> TCM ()
+putConstraintsToSleep sleepy = do
   awakeOnes <- use stAwakeConstraints
-  modifySleepingConstraints $ (++ awakeOnes)
-  modifyAwakeConstraints    $ const []
+  let (gotoSleep, stayAwake) = partition sleepy awakeOnes
+  modifySleepingConstraints $ (++ gotoSleep)
+  modifyAwakeConstraints    $ const stayAwake
+
+putAllConstraintsToSleep :: TCM ()
+putAllConstraintsToSleep = putConstraintsToSleep (const True)
 
 data ConstraintStatus = AwakeConstraint | SleepingConstraint
   deriving (Eq, Show)
@@ -135,11 +132,14 @@ withConstraint f (PConstr pids c) = do
     local (\e -> e { envActiveProblems = pids', envSolvingConstraints = isSolving }) $
     solvingProblems pids (f c)
 
-buildProblemConstraint :: ProblemId -> Constraint -> TCM ProblemConstraint
-buildProblemConstraint pid c = PConstr [pid] <$> buildClosure c
+buildProblemConstraint :: [ProblemId] -> Constraint -> TCM ProblemConstraint
+buildProblemConstraint pids c = PConstr pids <$> buildClosure c
+
+buildProblemConstraint_ :: Constraint -> TCM ProblemConstraint
+buildProblemConstraint_ = buildProblemConstraint []
 
 buildConstraint :: Constraint -> TCM ProblemConstraint
-buildConstraint c = flip buildProblemConstraint c =<< currentProblem
+buildConstraint c = flip buildProblemConstraint c . nub . filter (> 0) =<< asks envActiveProblems
 
 -- | Add new a constraint
 addConstraint' :: Constraint -> TCM ()
@@ -149,7 +149,7 @@ addConstraint' c = do
     stSleepingConstraints %= (pc :)
   where
     build | isBlocking c = buildConstraint c
-          | otherwise    = buildProblemConstraint 0 c
+          | otherwise    = buildProblemConstraint_ c
     isBlocking SortCmp{}     = False
     isBlocking LevelCmp{}    = False
     isBlocking ValueCmp{}    = True
