@@ -1,7 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleContexts  #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE TupleSections     #-}
 
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
@@ -11,8 +8,10 @@ module Agda.TypeChecking.Monad.Context where
 
 import Control.Applicative
 import Control.Monad.Reader
+import Control.Monad.State
 
 import Data.List hiding (sort)
+import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Monoid
 
@@ -24,10 +23,13 @@ import Agda.Syntax.Scope.Monad (getLocalVars, setLocalVars)
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad.Open
+import Agda.TypeChecking.Monad.Options
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Functor
 import Agda.Utils.List ((!!!), downFrom)
+import Agda.Utils.Size
+import Agda.Utils.Lens
 
 -- * Modifying the context
 
@@ -50,14 +52,9 @@ mkContextEntry x = do
   i <- fresh
   return $ Ctx i x
 
--- | Change the context.
-{-# SPECIALIZE inContext :: [Dom (Name, Type)] -> TCM a -> TCM a #-}
-inContext :: MonadTCM tcm => [Dom (Name, Type)] -> tcm a -> tcm a
-inContext xs ret = do
-  ctx <- mapM mkContextEntry xs
-  modifyContext (const ctx) ret
-
 -- | Change to top (=empty) context.
+--
+--   TODO: currently, this makes the @ModuleParamDict@ ill-formed!
 {-# SPECIALIZE inTopContext :: TCM a -> TCM a #-}
 inTopContext :: MonadTCM tcm => tcm a -> tcm a
 inTopContext cont = do
@@ -68,30 +65,107 @@ inTopContext cont = do
   return a
 
 -- | Delete the last @n@ bindings from the context.
+--
+--   TODO: currently, this makes the @ModuleParamDict@ ill-formed!
 {-# SPECIALIZE escapeContext :: Int -> TCM a -> TCM a #-}
 escapeContext :: MonadTCM tcm => Int -> tcm a -> tcm a
 escapeContext n = modifyContext $ drop n
 
+-- * Manipulating module parameters --
+
+-- | Locally set module parameters for a computation.
+
+withModuleParameters :: ModuleParamDict -> TCM a -> TCM a
+withModuleParameters mp ret = do
+  old <- use stModuleParameters
+  stModuleParameters .= mp
+  x <- ret
+  stModuleParameters .= old
+  return x
+
+-- | Apply a substitution to all module parameters.
+
+updateModuleParameters :: MonadTCM tcm => Substitution -> tcm a -> tcm a
+updateModuleParameters sub ret = do
+  pm <- use stModuleParameters
+  let showMP pref mps = intercalate "\n" $
+        [ p ++ show m ++ " : " ++ show (mpSubstitution mp)
+        | (p, (m, mp)) <- zip (pref : repeat (map (const ' ') pref))
+                              (Map.toList mps)
+        ]
+  cxt <- reverse <$> getContext
+  reportSLn "tc.cxt.param" 90 $ unlines $
+    [ "updatingModuleParameters"
+    , "  sub = " ++ show sub
+    , "  cxt = " ++ unwords (map (show . fst . unDom) cxt)
+    , showMP "  old = " pm
+    ]
+  let pm' = applySubst sub pm
+  reportSLn "tc.cxt.param" 90 $ showMP "  new = " pm'
+  stModuleParameters .= pm'
+  x <- ret              -- We need to keep introduced modules around
+  pm1 <- use stModuleParameters
+  let pm'' = Map.union pm (defaultModuleParameters <$ Map.difference pm1 pm)
+  stModuleParameters .= pm''
+  reportSLn "tc.cxt.param" 90 $ showMP "  restored = " pm''
+  return x
+
+-- | Since the @ModuleParamDict@ is relative to the current context,
+--   this function should be called everytime the context is extended.
+--
+weakenModuleParameters :: MonadTCM tcm => Nat -> tcm a -> tcm a
+weakenModuleParameters n = updateModuleParameters (raiseS n)
+
+-- | Get substitution @Γ ⊢ ρ : Γm@ where @Γ@ is the current context
+--   and @Γm@ is the module parameter telescope of module @m@.
+--
+--   In case the current 'ModuleParamDict' does not know @m@,
+--   we return the identity substitution.
+--   This is ok for instance if we are outside module @m@
+--   (in which case we have to supply all module parameters to any
+--   symbol defined within @m@ we want to refer).
+getModuleParameterSub :: MonadTCM tcm => ModuleName -> tcm Substitution
+getModuleParameterSub m = do
+  r <- use stModuleParameters
+  case Map.lookup m r of
+    Nothing -> return IdS
+    Just mp -> return $ mpSubstitution mp
+
+
 -- * Adding to the context
 
 -- | @addCtx x arg cont@ add a variable to the context.
 --
 --   Chooses an unused 'Name'.
+--
+--   Warning: Does not update module parameter substitution!
 {-# SPECIALIZE addCtx :: Name -> Dom Type -> TCM a -> TCM a #-}
 addCtx :: MonadTCM tcm => Name -> Dom Type -> tcm a -> tcm a
 addCtx x a ret = do
-  ctx <- map (nameConcrete . fst . unDom) <$> getContext
-  let x' = head $ filter (notTaken ctx) $ iterate nextName x
-  ce <- mkContextEntry $ (x',) <$> a
+  ce <- mkContextEntry $ (x,) <$> a
   modifyContext (ce :) ret
       -- let-bindings keep track of own their context
+
+-- | Pick a concrete name that doesn't shadow anything in the context.
+unshadowName :: MonadTCM tcm => Name -> tcm Name
+unshadowName x = do
+  ctx <- map (nameConcrete . fst . unDom) <$> getContext
+  return $ head $ filter (notTaken ctx) $ iterate nextName x
   where
     notTaken xs x = isNoName x || nameConcrete x `notElem` xs
 
 -- | Various specializations of @addCtx@.
 {-# SPECIALIZE addContext :: b -> TCM a -> TCM a #-}
 class AddContext b where
-  addContext :: MonadTCM tcm => b -> tcm a -> tcm a
+  addContext  :: MonadTCM tcm => b -> tcm a -> tcm a
+  contextSize :: b -> Nat
+
+-- | Since the module parameter substitution is relative to
+--   the current context, we need to weaken it when we
+--   extend the context.  This function takes care of that.
+--
+addContext' :: (MonadTCM tcm, AddContext b) => b -> tcm a -> tcm a
+addContext' cxt = addContext cxt . weakenModuleParameters (contextSize cxt)
 
 #if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} AddContext a => AddContext [a] where
@@ -99,37 +173,46 @@ instance {-# OVERLAPPABLE #-} AddContext a => AddContext [a] where
 instance AddContext a => AddContext [a] where
 #endif
   addContext = flip (foldr addContext)
+  contextSize = sum . map contextSize
 
 instance AddContext (Name, Dom Type) where
   addContext = uncurry addCtx
+  contextSize _ = 1
 
 instance AddContext (Dom (Name, Type)) where
   addContext = addContext . distributeF
   -- addContext dom = addCtx (fst $ unDom dom) (snd <$> dom)
+  contextSize _ = 1
 
 instance AddContext ([Name], Dom Type) where
   addContext (xs, dom) = addContext (bindsToTel' id xs dom)
+  contextSize (xs, _) = length xs
 
 instance AddContext ([WithHiding Name], Dom Type) where
   addContext ([]                 , dom) = id
   addContext (WithHiding h x : xs, dom) =
     addContext (x , mapHiding (mappend h) dom) .
     addContext (xs, raise 1 dom)
+  contextSize (xs, _) = length xs
 
 instance AddContext (String, Dom Type) where
   addContext (s, dom) ret = do
-    x <- freshName_ s
+    x <- unshadowName =<< freshName_ s
     addCtx x dom ret
+  contextSize _ = 1
 
 instance AddContext (Dom (String, Type)) where
   addContext = addContext . distributeF
   -- addContext dom = addContext (fst $ unDom dom, snd <$> dom)
+  contextSize _ = 1
 
 instance AddContext (Dom Type) where
   addContext dom = addContext ("_", dom)
+  contextSize _ = 1
 
 instance AddContext Name where
   addContext x = addContext (x, dummyDom)
+  contextSize _ = 1
 
 #if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} AddContext String where
@@ -137,41 +220,13 @@ instance {-# OVERLAPPING #-} AddContext String where
 instance AddContext String where
 #endif
   addContext s = addContext (s, dummyDom)
+  contextSize _ = 1
 
 instance AddContext Telescope where
   addContext tel ret = loop tel where
     loop EmptyTel          = ret
     loop (ExtendTel t tel) = underAbstraction t tel loop
-
-{-
--- | N-ary variant of @addCtx@.
-{-# SPECIALIZE addContext :: [Dom (Name, Type)] -> TCM a -> TCM a #-}
-addContext :: MonadTCM tcm => [Dom (Name, Type)] -> tcm a -> tcm a
-addContext ctx m =
-  foldr (\arg -> addCtx (fst $ unDom arg) (snd <$> arg)) m ctx
--}
-
--- | add a bunch of variables with the same type to the context
-{-# SPECIALIZE addCtxs :: [Name] -> Dom Type -> TCM a -> TCM a #-}
-addCtxs :: MonadTCM tcm => [Name] -> Dom Type -> tcm a -> tcm a
-addCtxs []     _ k = k
-addCtxs (x:xs) t k = addCtx x t $ addCtxs xs (raise 1 t) k
-
--- | Turns the string into a name and adds it to the context.
-{-# SPECIALIZE addCtxString :: String -> Dom Type -> TCM a -> TCM a #-}
-addCtxString :: MonadTCM tcm => String -> Dom Type -> tcm a -> tcm a
-addCtxString s a m = do
-  x <- freshName_ s
-  addCtx x a m
-
--- | Turns the string into a name and adds it to the context, with dummy type.
-{-# SPECIALIZE addCtxString_ :: String -> TCM a -> TCM a #-}
-addCtxString_ :: MonadTCM tcm => String -> tcm a -> tcm a
-addCtxString_ s = addCtxString s dummyDom
-
-{-# SPECIALIZE addCtxStrings_ :: [String] -> TCM a -> TCM a #-}
-addCtxStrings_ :: MonadTCM tcm => [String] -> tcm a -> tcm a
-addCtxStrings_ = flip (foldr addCtxString_)
+  contextSize = size
 
 -- | Context entries without a type have this dummy type.
 dummyDom :: Dom Type
@@ -182,8 +237,8 @@ dummyDom = defaultDom typeDontCare
 underAbstraction :: (Subst t a, MonadTCM tcm) => Dom Type -> Abs a -> (a -> tcm b) -> tcm b
 underAbstraction _ (NoAbs _ v) k = k v
 underAbstraction t a           k = do
-    x <- freshName_ $ realName $ absName a
-    addCtx x t $ k $ absBody a
+    x <- unshadowName =<< freshName_ (realName $ absName a)
+    addContext (x, t) $ k $ absBody a
   where
     realName s = if isNoName s then "x" else argNameToString s
 
@@ -192,14 +247,7 @@ underAbstraction t a           k = do
 underAbstraction_ :: (Subst t a, MonadTCM tcm) => Abs a -> (a -> tcm b) -> tcm b
 underAbstraction_ = underAbstraction dummyDom
 
--- | Add a telescope to the context.
-{-# SPECIALIZE addCtxTel :: Telescope -> TCM a -> TCM a #-}
-addCtxTel :: MonadTCM tcm => Telescope -> tcm a -> tcm a
-addCtxTel tel ret = loop tel where
-  loop EmptyTel          = ret
-  loop (ExtendTel t tel) = underAbstraction t tel loop
-
--- | Add a let bound variable
+-- | Add a let bound variable.
 {-# SPECIALIZE addLetBinding :: ArgInfo -> Name -> Term -> Type -> TCM a -> TCM a #-}
 addLetBinding :: MonadTCM tcm => ArgInfo -> Name -> Term -> Type -> tcm a -> tcm a
 addLetBinding info x v t0 ret = do
@@ -252,7 +300,7 @@ getContextNames = map (fst . unDom) <$> getContext
 lookupBV :: MonadReader TCEnv m => Nat -> m (Dom (Name, Type))
 lookupBV n = do
   ctx <- getContext
-  let failure = fail $ "deBruijn index out of scope: " ++ show n ++
+  let failure = fail $ "de Bruijn index out of scope: " ++ show n ++
                        " in context " ++ show (map (fst . unDom) ctx)
   maybe failure (return . fmap (raise $ n + 1)) $ ctx !!! n
 
