@@ -1,9 +1,5 @@
-{-# LANGUAGE CPP               #-}
-{-# LANGUAGE DoAndIfThenElse   #-}
-{-# LANGUAGE FlexibleContexts  #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE LambdaCase        #-}
-{-# LANGUAGE PatternGuards     #-}
+{-# LANGUAGE CPP                      #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Monad.Signature where
 
@@ -13,6 +9,7 @@ import Control.Arrow (first, second, (***))
 import Control.Applicative hiding (empty)
 import Control.Monad.State
 import Control.Monad.Reader
+import Control.Monad.Trans.Maybe
 
 import Data.List hiding (null)
 import Data.Set (Set)
@@ -23,14 +20,13 @@ import Data.Maybe
 import Data.Monoid
 
 import Agda.Syntax.Abstract.Name
-import Agda.Syntax.Abstract (Ren)
+import Agda.Syntax.Abstract (Ren, ScopeCopyInfo(..))
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Names
 import Agda.Syntax.Position
 import Agda.Syntax.Treeless (Compiled(..), TTerm)
 
-import qualified Agda.Compiler.JS.Parser as JS
 import qualified Agda.Compiler.UHC.Pragmas.Base as CR
 
 import Agda.TypeChecking.Monad.Base
@@ -40,6 +36,7 @@ import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Exception ( ExceptionT )
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
+import Agda.TypeChecking.Monad.Local
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
@@ -70,8 +67,15 @@ addConstant q d = do
   reportSLn "tc.signature" 20 $ "adding constant " ++ show q ++ " to signature"
   tel <- getContextTelescope
   let tel' = replaceEmptyName "r" $ killRange $ case theDef d of
-              Constructor{} -> fmap (setHiding Hidden) tel
-              _             -> tel
+              Constructor{} -> fmap hideOrKeepInstance tel
+              Function{ funProjection = Just Projection{ projProper = True, projIndex = n } } ->
+                let fallback = fmap hideOrKeepInstance tel in
+                if n > 0 then fallback else
+                -- if the record value is part of the telescope, its hiding should left unchanged
+                  case initLast $ telToList tel of
+                    Nothing -> fallback
+                    Just (doms, dom) -> telFromList $ fmap hideOrKeepInstance doms ++ [dom]
+              _ -> tel
   let d' = abstract tel' $ d { defName = q }
   reportSLn "tc.signature" 30 $ "lambda-lifted definition = " ++ show d'
   modifySignature $ updateDefinitions $ HMap.insertWith (+++) q d'
@@ -134,14 +138,9 @@ addEpicCode q epDef = modifySignature $ updateDefinition q $ updateDefCompiledRe
     addEp crep = crep { compiledEpic = Just epDef }
 
 addJSCode :: QName -> String -> TCM ()
-addJSCode q jsDef =
-  case JS.parse jsDef of
-    Left e ->
-      modifySignature $ updateDefinition q $ updateDefCompiledRep $ addJS (Just e)
-    Right s ->
-      typeError (CompilationError ("Failed to parse ECMAScript (..." ++ s ++ ") for " ++ show q))
+addJSCode q jsDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addJS
   where
-    addJS e crep = crep { compiledJS = e }
+    addJS crep = crep { compiledJS = Just jsDef }
 
 addCoreCode :: QName -> CR.CoreExpr -> TCM ()
 addCoreCode q crDef =  modifySignature $ updateDefinition q $ updateDefCompiledRep $ addCore crDef
@@ -159,26 +158,14 @@ addCoreType q crTy = modifySignature $ updateDefinition q $ updateDefCompiledRep
   where
     addCr crep = crep { compiledCore = Just $ CrType crTy }
 
-markNoSmashing :: QName -> TCM ()
-markNoSmashing q = modifySignature $ updateDefinition q $ mark
-  where
-    mark def@Defn{theDef = fun@Function{}} =
-      def{theDef = fun{funSmashable = False}}
-    mark def = def
+setFunctionFlag :: FunctionFlag -> Bool -> QName -> TCM ()
+setFunctionFlag flag val q = modifyGlobalDefinition q $ set (theDefLens . funFlag flag) val
 
 markStatic :: QName -> TCM ()
-markStatic q = modifySignature $ updateDefinition q $ mark
-  where
-    mark def@Defn{theDef = fun@Function{}} =
-      def{theDef = fun{funStatic = True}}
-    mark def = def
+markStatic = setFunctionFlag FunStatic True
 
 markInline :: QName -> TCM ()
-markInline q = modifySignature $ updateDefinition q $ mark
-  where
-    mark def@Defn{theDef = fun@Function{}} =
-      def{theDef = fun{funInline = True}}
-    mark def = def
+markInline = setFunctionFlag FunInline True
 
 unionSignatures :: [Signature] -> Signature
 unionSignatures ss = foldr unionSignature emptySignature ss
@@ -210,67 +197,84 @@ addSection m = do
       reportSLn "impossible" 60 $ "with content " ++ show sec
       __IMPOSSIBLE__
   -- Add the new section.
+  setDefaultModuleParameters m
   modifySignature $ over sigSections $ Map.insert m sec
 
--- | Lookup a section. If it doesn't exist that just means that the module
---   wasn't parameterised.
+setDefaultModuleParameters :: ModuleName -> TCM ()
+setDefaultModuleParameters m =
+  stModuleParameters %= Map.insert m defaultModuleParameters
+
+-- | Get a section.
+--
+--   Why Maybe? The reason is that we look up all prefixes of a module to
+--   compute number of parameters, and for hierarchical top-level modules,
+--   A.B.C say, A and A.B do not exist.
+{-# SPECIALIZE getSection :: ModuleName -> TCM (Maybe Section) #-}
+{-# SPECIALIZE getSection :: ModuleName -> ReduceM (Maybe Section) #-}
+getSection :: (Functor m, ReadTCState m) => ModuleName -> m (Maybe Section)
+getSection m = do
+  sig  <- (^. stSignature . sigSections) <$> getTCState
+  isig <- (^. stImports   . sigSections) <$> getTCState
+  return $ Map.lookup m sig `mplus` Map.lookup m isig
+
+-- | Lookup a section telescope.
+--
+--   If it doesn't exist, like in hierarchical top-level modules,
+--   the section telescope is empty.
 {-# SPECIALIZE lookupSection :: ModuleName -> TCM Telescope #-}
 {-# SPECIALIZE lookupSection :: ModuleName -> ReduceM Telescope #-}
 lookupSection :: (Functor m, ReadTCState m) => ModuleName -> m Telescope
-lookupSection m = do
-  sig  <- (^. stSignature . sigSections) <$> getTCState
-  isig <- (^. stImports   . sigSections) <$> getTCState
-  return $ maybe EmptyTel (^. secTelescope) $ Map.lookup m sig `mplus` Map.lookup m isig
+lookupSection m = maybe EmptyTel (^. secTelescope) <$> getSection m
 
 -- Add display forms to all names @xn@ such that @x = x1 es1@, ... @xn-1 = xn esn@.
 addDisplayForms :: QName -> TCM ()
 addDisplayForms x = do
   def  <- getConstInfo x
-  args <- getContextArgs
-  add (drop (projectionArgs $ theDef def) args) x x []
+  args <- drop (projectionArgs $ theDef def) <$> getContextArgs
+  add args x x $ map Apply $ raise 1 args -- make room for the single match variable of the display form
   where
-    add args top x vs0 = do
+    add args top x es0 = do
       def <- getConstInfo x
       let cs = defClauses def
           isCopy = defCopy def
       case cs of
-        [ Clause{ namedClausePats = pats, clauseBody = b } ]
-          | isCopy
-          , all (isVar . namedArg) pats
-          , Just (m, Def y es) <- strip (b `apply` vs0)
-          , Just vs <- mapM isApplyElim es -> do
-              let ps = raise 1 $ map unArg vs
-                  df = Display 0 ps $ DTerm $ Def top $ map Apply args
+        [ cl ] -> do
+          if not isCopy
+            then noDispForm x "not a copy" else do
+          if not $ all (isVar . namedArg) $ namedClausePats cl
+            then noDispForm x "properly matching patterns" else do
+          -- We have
+          --    x ps = e
+          -- and we're trying to generate a display form
+          --    x es0 <-- e[es0/ps]
+          -- Of course x es0 might be an over- or underapplication, hence the
+          -- n/m arithmetic.
+          let n          = size $ namedClausePats cl
+              (es1, es2) = splitAt n es0
+              m          = n - size es1
+              vs1 = map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es1
+              sub = parallelS $ reverse $ vs1 ++ replicate m (var 0)
+              body = applySubst sub (compiledClauseBody cl) `applyE` es2
+          case unSpine <$> body of
+            Just (Def y es) -> do
+              let df = Display m es $ DTerm $ Def top $ map Apply args
               reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
                                                 ++ "\n  " ++ show df
               addDisplayForm y df
-              add args top y vs
+              add args top y es
+            Just v          -> noDispForm x $ "not a def body, but " ++ show v
+            Nothing         -> noDispForm x $ "bad body"
         [] | Constructor{ conSrcCon = h } <- theDef def -> do
               let y  = conName h
-                  df = Display 0 [] $ DTerm $ Con (h {conName = top }) []
+                  df = Display 0 [] $ DTerm $ Con (h {conName = top }) ConOSystem []
               reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
                                                 ++ "\n  " ++ show df
               addDisplayForm y df
-        _ -> do
-          let reason = if not isCopy then "not a copy" else
-                  case cs of
-                    []    -> "no clauses"
-                    _:_:_ -> "many clauses"
-                    [ Clause{ clauseBody = b } ] -> case strip b of
-                      Nothing -> "bad body"
-                      Just (m, Def y es)
-                        | m < length args -> "too few args"
-                        | m > length args -> "too many args"
-                        | otherwise       -> "args=" ++ show args ++ " es=" ++ show es
-                      Just (m, v) -> "not a def body, but " ++ show v
-          reportSLn "tc.display.section" 30 $
-            "no display form from " ++ show x ++ " because " ++ reason
+        [] -> noDispForm x "no clauses"
+        (_:_:_) -> noDispForm x "many clauses"
 
-    strip (Body v)   = return (0, unSpine v)
-    strip  NoBody    = Nothing
-    strip (Bind b)   = do
-      (n, v) <- strip $ absBody b
-      return (n + 1, ignoreSharing v)
+    noDispForm x reason = reportSLn "tc.display.section" 30 $
+      "no display form from " ++ show x ++ " because " ++ reason
 
     isVar VarP{} = True
     isVar _      = False
@@ -281,12 +285,11 @@ applySection
   -> Telescope      -- ^ Parameters of new module.
   -> ModuleName     -- ^ Name of old module applied to arguments.
   -> Args           -- ^ Arguments of module application.
-  -> Ren QName      -- ^ Imported names (given as renaming).
-  -> Ren ModuleName -- ^ Imported modules (given as renaming).
+  -> ScopeCopyInfo  -- ^ Imported names and modules
   -> TCM ()
-applySection new ptel old ts rd rm = do
+applySection new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd } = do
   rd <- closeConstructors rd
-  applySection' new ptel old ts rd rm
+  applySection' new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd }
   where
     -- If a datatype is being copied, all its constructors need to be copied,
     -- and if a constructor is copied its datatype needs to be.
@@ -316,8 +319,8 @@ applySection new ptel old ts rd rm = do
             Record{ recConHead = h }      -> [conName h]
             _                         -> []
 
-applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> Ren QName -> Ren ModuleName -> TCM ()
-applySection' new ptel old ts rd rm = do
+applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> ScopeCopyInfo -> TCM ()
+applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } = do
   reportSLn "tc.mod.apply" 10 $ render $ vcat
     [ text "applySection"
     , text "new  =" <+> text (show new)
@@ -354,10 +357,11 @@ applySection' new ptel old ts rd rm = do
       copyDef' np def
       where
         copyDef' np d = do
-          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ show y ++ " from " ++ show x ++ " with " ++ show np ++ " args " ++ show abstr
+          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ show y ++ " from " ++ show x ++ " with " ++ show np ++ " args " ++ show (defAbstract d)
           reportSLn "tc.mod.apply" 80 $
             "args = " ++ show ts' ++ "\n" ++
-            "old type = " ++ prettyShow (defType d) ++ "\n" ++
+            "old type = " ++ prettyShow (defType d)
+          reportSLn "tc.mod.apply" 80 $
             "new type = " ++ prettyShow t
           addConstant y =<< nd y
           makeProjection y
@@ -382,7 +386,6 @@ applySection' new ptel old ts rd rm = do
             pol = defPolarity d `apply` ts'
             occ = defArgOccurrences d `apply` ts'
             inst = defInstance d
-            abstr = defAbstract d
             -- the name is set by the addConstant function
             nd :: QName -> TCM Definition
             nd y = for def $ \ df -> Defn
@@ -396,6 +399,7 @@ applySection' new ptel old ts rd rm = do
                     , defCompiledRep    = noCompiledRep
                     , defInstance       = inst
                     , defCopy           = True
+                    , defMatchable      = False
                     , theDef            = df }
             oldDef = theDef d
             isCon  = case oldDef of { Constructor{} -> True ; _ -> False }
@@ -409,9 +413,9 @@ applySection' new ptel old ts rd rm = do
             isVar0 t = case ignoreSharing $ unArg t of Var 0 [] -> True; _ -> False
             proj   = case oldDef of
               Function{funProjection = Just p@Projection{projIndex = n}}
-                | size ts < n || (size ts == n && isVar0 (last ts))
-                -> Just $ p { projIndex    = n - size ts
-                            , projDropPars = projDropPars p `apply` ts
+                | size ts' < n || (size ts' == n && maybe True isVar0 (lastMaybe ts'))
+                -> Just $ p { projIndex = n - size ts'
+                            , projLams  = projLams p `apply` ts'
                             }
               _ -> Nothing
             def =
@@ -425,26 +429,22 @@ applySection' new ptel old ts rd rm = do
                          , dataClause = Just cl
                          , dataCons   = map copyName cs
                          }
-                Record{ recPars = np, recConType = t, recTel = tel } -> return $
+                Record{ recPars = np, recTel = tel } -> return $
                   oldDef { recPars    = np - size ts'
                          , recClause  = Just cl
-                         , recConType = piApply t ts'
                          , recTel     = apply tel ts'
                          }
                 _ -> do
                   cc <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
-                  let newDef = Function
+                  let newDef =
+                        set funMacro  (oldDef ^. funMacro) $
+                        set funStatic (oldDef ^. funStatic) $
+                        set funInline True $
+                        emptyFunction
                         { funClauses        = [cl]
-                        , funCompiled       = Just $ cc
-                        , funTreeless       = Nothing
-                        , funDelayed        = NotDelayed
-                        , funInv            = NotInjective
+                        , funCompiled       = Just cc
                         , funMutual         = mutual
-                        , funAbstr          = ConcreteDef -- OR: abstr -- ?!
                         , funProjection     = proj
-                        , funStatic         = False
-                        , funInline         = False
-                        , funSmashable      = True
                         , funTerminates     = Just True
                         , funExtLam         = extlam
                         , funWith           = with
@@ -453,14 +453,12 @@ applySection' new ptel old ts rd rm = do
                   reportSLn "tc.mod.apply" 80 $ "new def for " ++ show x ++ "\n  " ++ show newDef
                   return newDef
 
-            head = case oldDef of
-                     Function{funProjection = Just Projection{ projDropPars = f}}
-                       -> f
-                     _ -> Def x []
             cl = Clause { clauseRange     = getRange $ defClauses d
                         , clauseTel       = EmptyTel
                         , namedClausePats = []
-                        , clauseBody      = Body $ head `apply` ts'
+                        , clauseBody      = Just $ case oldDef of
+                            Function{funProjection = Just p} -> projDropParsApply p ProjSystem ts'
+                            _ -> Def x $ map Apply ts'
                         , clauseType      = Just $ defaultArg t
                         , clauseCatchall  = False
                         }
@@ -505,26 +503,29 @@ applySection' new ptel old ts rd rm = do
       reportSLn "tc.mod.apply" 80 $ "  totalArgs    = " ++ show totalArgs
       reportSLn "tc.mod.apply" 80 $ "  tel          = " ++ intercalate " " (map (fst . unDom) $ telToList tel)  -- only names
       reportSLn "tc.mod.apply" 80 $ "  sectionTel   = " ++ intercalate " " (map (fst . unDom) $ telToList ptel) -- only names
-      addCtxTel sectionTel $ addSection y
+      addContext sectionTel $ addSection y
 
 -- | Add a display form to a definition (could be in this or imported signature).
 addDisplayForm :: QName -> DisplayForm -> TCM ()
 addDisplayForm x df = do
-  d <- makeOpen df
+  d <- makeLocal df
   let add = updateDefinition x $ \ def -> def{ defDisplay = d : defDisplay def }
-  inCurrentSig <- isJust . HMap.lookup x <$> use (stSignature . sigDefinitions)
-  if inCurrentSig
-     then modifySignature add
-     else stImportsDisplayForms %= HMap.insertWith (++) x [d]
+  ifM (isLocal x)
+    {-then-} (modifySignature add)
+    {-else-} (stImportsDisplayForms %= HMap.insertWith (++) x [d])
   whenM (hasLoopingDisplayForm x) $
     typeError . GenericDocError $ text "Cannot add recursive display form for" <+> pretty x
 
-getDisplayForms :: QName -> TCM [Open DisplayForm]
+isLocal :: QName -> TCM Bool
+isLocal x = isJust . HMap.lookup x <$> use (stSignature . sigDefinitions)
+
+getDisplayForms :: QName -> TCM [LocalDisplayForm]
 getDisplayForms q = do
   ds  <- defDisplay <$> getConstInfo q
   ds1 <- maybe [] id . HMap.lookup q <$> use stImportsDisplayForms
   ds2 <- maybe [] id . HMap.lookup q <$> use stImportedDisplayForms
-  return $ ds ++ ds1 ++ ds2
+  ifM (isLocal q) (return $ ds ++ ds1 ++ ds2)
+                  (return $ ds1 ++ ds ++ ds2)
 
 -- | Find all names used (recursively) by display forms of a given name.
 chaseDisplayForms :: QName -> TCM (Set QName)
@@ -533,7 +534,7 @@ chaseDisplayForms q = go Set.empty [q]
     go used []       = pure used
     go used (q : qs) = do
       let rhs (Display _ _ e) = e   -- Only look at names in the right-hand side (#1870)
-      ds <- (`Set.difference` used) . Set.unions . map (namesIn . rhs . openThing)
+      ds <- (`Set.difference` used) . Set.unions . map (namesIn . rhs . dget)
             <$> (getDisplayForms q `catchError_` \ _ -> pure [])  -- might be a pattern synonym
       go (Set.union ds used) (Set.toList ds ++ qs)
 
@@ -546,15 +547,14 @@ canonicalName x = do
   def <- theDef <$> getConstInfo x
   case def of
     Constructor{conSrcCon = c}                                -> return $ conName c
-    Record{recClause = Just (Clause{ clauseBody = body })}    -> canonicalName $ extract body
-    Datatype{dataClause = Just (Clause{ clauseBody = body })} -> canonicalName $ extract body
+    Record{recClause = Just (Clause{ clauseBody = body })}    -> can body
+    Datatype{dataClause = Just (Clause{ clauseBody = body })} -> can body
     _                                                         -> return x
   where
-    extract NoBody           = __IMPOSSIBLE__
-    extract (Body (Def x _)) = x
-    extract (Body (Shared p)) = extract (Body $ derefPtr p)
-    extract (Body _)         = __IMPOSSIBLE__
-    extract (Bind b)         = extract (unAbs b)
+    can body = canonicalName $ extract $ fromMaybe __IMPOSSIBLE__ body
+    extract (Def x _)  = x
+    extract (Shared p) = extract $ derefPtr p
+    extract _          = __IMPOSSIBLE__
 
 sameDef :: QName -> QName -> TCM (Maybe QName)
 sameDef d1 d2 = do
@@ -607,6 +607,11 @@ defaultGetRewriteRulesFor getTCState q = do
       look s = HMap.lookup q $ s ^. sigRewriteRules
   return $ mconcat $ catMaybes [look sig, look imp]
 
+-- | Get the original name of the projection
+--   (the current one could be from a module application).
+getOriginalProjection :: HasConstInfo m => QName -> m QName
+getOriginalProjection q = projOrig . fromMaybe __IMPOSSIBLE__ <$> isProjection q
+
 instance HasConstInfo (TCMT IO) where
   getRewriteRulesFor = defaultGetRewriteRulesFor get
   getConstInfo q = join $ pureTCM $ \st env ->
@@ -635,6 +640,10 @@ instance HasConstInfo (TCMT IO) where
           dropLastModule q@QName{ qnameModule = m } =
             q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
 
+instance (HasConstInfo m) => HasConstInfo (MaybeT m) where
+  getConstInfo = lift . getConstInfo
+  getRewriteRulesFor = lift . getRewriteRulesFor
+
 instance (HasConstInfo m, Error err) => HasConstInfo (ExceptionT err m) where
   getConstInfo = lift . getConstInfo
   getRewriteRulesFor = lift . getRewriteRulesFor
@@ -655,7 +664,10 @@ getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 
 -- | Set the polarity of a definition.
 setPolarity :: QName -> [Polarity] -> TCM ()
-setPolarity q pol = modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
+setPolarity q pol = do
+  reportSLn "tc.polarity.set" 20 $
+    "Setting polarity of " ++ show q ++ " to " ++ show pol ++ "."
+  modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
 -- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
 getArgOccurrence :: QName -> Nat -> TCM Occurrence
@@ -665,6 +677,8 @@ getArgOccurrence d i = do
     Constructor{} -> StrictPos
     _             -> fromMaybe Mixed $ defArgOccurrences def !!! i
 
+-- | Sets the 'defArgOccurrences' for the given identifier (which
+-- should already exist in the signature).
 setArgOccurrences :: QName -> [Occurrence] -> TCM ()
 setArgOccurrences d os = modifyArgOccurrences d $ const os
 
@@ -693,6 +707,27 @@ getCompiled q = do
     Function{ funTreeless = t } -> t
     _                           -> Nothing
 
+getErasedConArgs :: QName -> TCM [Bool]
+getErasedConArgs q = do
+  def <- getConstInfo q
+  case theDef def of
+    Constructor{ conData = d, conPars = np, conErased = es } -> do
+      ddef <- getConstInfo d
+      case compiledHaskell $ defCompiledRep ddef of
+        Nothing -> return es
+        Just _  -> do
+          -- Can't erase arguments of COMPILED_DATA constructors yet
+          TelV tel _ <- telView $ defType def
+          return $ replicate (size tel - np) False
+    _ -> __IMPOSSIBLE__
+
+setErasedConArgs :: QName -> [Bool] -> TCM ()
+setErasedConArgs q args = modifyGlobalDefinition q setArgs
+  where
+    setArgs def@Defn{theDef = con@Constructor{}} =
+      def{ theDef = con{ conErased = args } }
+    setArgs def = def   -- no-op for non-constructors
+
 getTreeless :: QName -> TCM (Maybe TTerm)
 getTreeless q = fmap cTreeless <$> getCompiled q
 
@@ -722,15 +757,6 @@ setMutual d m = modifySignature $ updateDefinition d $ updateTheDef $ \ def ->
 mutuallyRecursive :: QName -> QName -> TCM Bool
 mutuallyRecursive d d' = (d `elem`) <$> getMutual d'
 
--- | Why Maybe? The reason is that we look up all prefixes of a module to
---   compute number of parameters, and for hierarchical top-level modules,
---   A.B.C say, A and A.B do not exist.
-getSection :: ModuleName -> TCM (Maybe Section)
-getSection m = do
-  sig  <- use $ stSignature . sigSections
-  isig <- use $ stImports   . sigSections
-  return $ Map.lookup m sig <|> Map.lookup m isig
-
 -- | Get the number of parameters to the current module.
 getCurrentModuleFreeVars :: TCM Nat
 getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
@@ -738,11 +764,16 @@ getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
 -- | Compute the number of free variables of a defined name. This is the sum of
 --   number of parameters shared with the current module and the number of
 --   anonymous variables (if the name comes from a let-bound module).
-{-# SPECIALIZE getDefFreeVars :: QName -> TCM Nat #-}
-{-# SPECIALIZE getDefFreeVars :: QName -> ReduceM Nat #-}
 getDefFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadReader TCEnv m) => QName -> m Nat
-getDefFreeVars q = do
-  let m = qnameModule q
+getDefFreeVars = getModuleFreeVars . qnameModule
+
+freeVarsToApply :: QName -> TCM Args
+freeVarsToApply = moduleParamsToApply . qnameModule
+
+{-# SPECIALIZE getModuleFreeVars :: ModuleName -> TCM Nat #-}
+{-# SPECIALIZE getModuleFreeVars :: ModuleName -> ReduceM Nat #-}
+getModuleFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadReader TCEnv m) => ModuleName -> m Nat
+getModuleFreeVars m = do
   m0   <- commonParentModule m <$> currentModule
   (+) <$> getAnonymousVariables m <*> (size <$> lookupSection m0)
 
@@ -760,40 +791,51 @@ getDefFreeVars q = do
 --        module M₃ Θ where
 --          ... M₁.M₂.f [insert Γ raised by Θ]
 --   @
-freeVarsToApply :: QName -> TCM Args
-freeVarsToApply x = do
-  -- Get the correct number of free variables (correctly raised) of @x@.
+moduleParamsToApply :: ModuleName -> TCM Args
+moduleParamsToApply m = do
+  -- Get the correct number of free variables (correctly raised) of @m@.
 
-  args <- take <$> getDefFreeVars x <*> getContextArgs
+  reportSLn "tc.sig.param" 90 $ "computing module parameters of " ++ show m
+  cxt <- getContext
+  n   <- getModuleFreeVars m
+  tel <- take n . telToList <$> lookupSection m
+  sub <- getModuleParameterSub m
+  reportSLn "tc.sig.param" 60 $ unlines $
+    [ "  n    = " ++ show n
+    , "  cxt  = " ++ show (map (fmap fst) cxt)
+    , "  sub  = " ++ show sub
+    ]
+  unless (size tel == n) __IMPOSSIBLE__
+  let args = applySubst sub $ zipWith (\ i a -> var i <$ argFromDom a) (downFrom n) tel
+  reportSLn "tc.sig.param" 60 $ "  args = " ++ show args
 
   -- Apply the original ArgInfo, as the hiding information in the current
-  -- context might be different from the hiding information expected by @x@.
+  -- context might be different from the hiding information expected by @m@.
 
-  getSection (qnameModule x) >>= \case
+  getSection m >>= \case
     Nothing -> do
-      -- We have no section for @x@.
+      -- We have no section for @m@.
       -- This should only happen for toplevel definitions, and then there
       -- are no free vars to apply, or?
       -- unless (null args) __IMPOSSIBLE__
       -- No, this invariant is violated by private modules, see Issue1701a.
       return args
-    Just (Section tel) -> do
-      -- The section telescope of the home of @x@ should be as least
-      -- as long as the number of free vars @x@ is applied to.
+    Just (Section stel) -> do
+      -- The section telescope of @m@ should be as least
+      -- as long as the number of free vars @m@ is applied to.
       -- We still check here as in no case, we want @zipWith@ to silently
       -- drop some @args@.
       -- And there are also anonymous modules, thus, the invariant is not trivial.
-      when (size tel < size args) __IMPOSSIBLE__
-      return $ zipWith (\ (Dom ai _) (Arg _ v) -> Arg ai v) (telToList tel) args
+      when (size stel < size args) __IMPOSSIBLE__
+      return $ zipWith (\ (Dom ai _) (Arg _ v) -> Arg ai v) (telToList stel) args
 
 -- | Unless all variables in the context are module parameters, create a fresh
 --   module to capture the non-module parameters. Used when unquoting to make
 --   sure generated definitions work properly.
 inFreshModuleIfFreeParams :: TCM a -> TCM a
 inFreshModuleIfFreeParams k = do
-  a <- getCurrentModuleFreeVars
-  b <- size <$> getContext
-  if a == b then k else do
+  sub <- getModuleParameterSub =<< currentModule
+  if sub == IdS then k else do
     m  <- currentModule
     m' <- qualifyM m . mnameFromList . (:[]) <$> freshName_ "_"
     addSection m'
@@ -824,23 +866,26 @@ makeAbstract d =
                , theDef = def
                }
   where
-    makeAbs Datatype   {} = Just Axiom
-    makeAbs Function   {} = Just Axiom
+    makeAbs Axiom         = Just Axiom
+    makeAbs Datatype   {} = Just AbstractDefn
+    makeAbs Function   {} = Just AbstractDefn
     makeAbs Constructor{} = Nothing
     -- Andreas, 2012-11-18:  Make record constructor and projections abstract.
-    makeAbs d@Record{}    = Just Axiom
-    -- Q: what about primitive?
-    makeAbs d             = Just d
+    makeAbs d@Record{}    = Just AbstractDefn
+    makeAbs Primitive{}   = __IMPOSSIBLE__
+    makeAbs AbstractDefn  = __IMPOSSIBLE__
 
 -- | Enter abstract mode. Abstract definition in the current module are transparent.
-inAbstractMode :: TCM a -> TCM a
+{-# SPECIALIZE inAbstractMode :: TCM a -> TCM a #-}
+inAbstractMode :: MonadReader TCEnv m => m a -> m a
 inAbstractMode = local $ \e -> e { envAbstractMode = AbstractMode,
                                    envAllowDestructiveUpdate = False }
                                     -- Allowing destructive updates when seeing through
                                     -- abstract may break the abstraction.
 
 -- | Not in abstract mode. All abstract definitions are opaque.
-inConcreteMode :: TCM a -> TCM a
+{-# SPECIALIZE inConcreteMode :: TCM a -> TCM a #-}
+inConcreteMode :: MonadReader TCEnv m => m a -> m a
 inConcreteMode = local $ \e -> e { envAbstractMode = ConcreteMode }
 
 -- | Ignore abstract mode. All abstract definitions are transparent.
@@ -852,14 +897,15 @@ ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
 
 -- | Enter concrete or abstract mode depending on whether the given identifier
 --   is concrete or abstract.
-inConcreteOrAbstractMode :: QName -> TCM a -> TCM a
+{-# SPECIALIZE inConcreteOrAbstractMode :: QName -> (Definition -> TCM a) -> TCM a #-}
+inConcreteOrAbstractMode :: (MonadReader TCEnv m, HasConstInfo m) => QName -> (Definition -> m a) -> m a
 inConcreteOrAbstractMode q cont = do
   -- Andreas, 2015-07-01: If we do not ignoreAbstractMode here,
   -- we will get ConcreteDef for abstract things, as they are turned into axioms.
-  a <- ignoreAbstractMode $ defAbstract <$> getConstInfo q
-  case a of
-    AbstractDef -> inAbstractMode cont
-    ConcreteDef -> inConcreteMode cont
+  def <- ignoreAbstractMode $ getConstInfo q
+  case defAbstract def of
+    AbstractDef -> inAbstractMode $ cont def
+    ConcreteDef -> inConcreteMode $ cont def
 
 -- | Check whether a name might have to be treated abstractly (either if we're
 --   'inAbstractMode' or it's not a local name). Returns true for things not
@@ -904,16 +950,6 @@ sortOfConst q =
             Datatype{dataSort = s} -> return s
             _                      -> fail $ "Expected " ++ show q ++ " to be a datatype."
 
--- | The number of parameters of a definition.
-defPars :: Definition -> Int
-defPars d = case theDef d of
-    Axiom{}                  -> 0
-    def@Function{}           -> projectionArgs def
-    Datatype  {dataPars = n} -> n
-    Record     {recPars = n} -> n
-    Constructor{conPars = n} -> n
-    Primitive{}              -> 0
-
 -- | The number of dropped parameters for a definition.
 --   0 except for projection(-like) functions and constructors.
 droppedPars :: Definition -> Int
@@ -924,6 +960,7 @@ droppedPars d = case theDef d of
     Record     {recPars = _} -> 0  -- not dropped
     Constructor{conPars = n} -> n
     Primitive{}              -> 0
+    AbstractDefn             -> __IMPOSSIBLE__
 
 -- | Is it the name of a record projection?
 {-# SPECIALIZE isProjection :: QName -> TCM (Maybe Projection) #-}
@@ -938,20 +975,18 @@ isProjection_ def =
 
 -- | Is it a function marked STATIC?
 isStaticFun :: Defn -> Bool
-isStaticFun Function{ funStatic = b } = b
-isStaticFun _ = False
+isStaticFun = (^. funStatic)
 
 -- | Is it a function marked INLINE?
 isInlineFun :: Defn -> Bool
-isInlineFun Function{ funInline = b } = b
-isInlineFun _ = False
+isInlineFun = (^. funInline)
 
 -- | Returns @True@ if we are dealing with a proper projection,
 --   i.e., not a projection-like function nor a record field value
 --   (projection applied to argument).
 isProperProjection :: Defn -> Bool
 isProperProjection d = caseMaybe (isProjection_ d) False $ \ isP ->
-  if projIndex isP <= 0 then False else isJust $ projProper isP
+  if projIndex isP <= 0 then False else projProper isP
 
 -- | Number of dropped initial arguments of a projection(-like) function.
 projectionArgs :: Defn -> Int
@@ -967,11 +1002,11 @@ usesCopatterns q = do
 
 -- | Apply a function @f@ to its first argument, producing the proper
 --   postfix projection if @f@ is a projection.
-applyDef :: QName -> Arg Term -> TCM Term
-applyDef f a = do
+applyDef :: ProjOrigin -> QName -> Arg Term -> TCM Term
+applyDef o f a = do
   let fallback = return $ Def f [Apply a]
   caseMaybeM (isProjection f) fallback $ \ isP -> do
     if projIndex isP <= 0 then fallback else do
       -- Get the original projection, if existing.
-      caseMaybe (projProper isP) fallback $ \ f' -> do
-        return $ unArg a `applyE` [Proj f']
+      if not (projProper isP) then fallback else do
+        return $ unArg a `applyE` [Proj o $ projOrig isP]
