@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE TupleSections #-}
 
 -- | Stuff for sized types that does not require modules
 --   "Agda.TypeChecking.Reduce" or "Agda.TypeChecking.Constraints"
@@ -131,11 +129,10 @@ sizeType = El sizeSort <$> primSize
 
 -- | The name of @SIZESUC@.
 sizeSucName :: TCM (Maybe QName)
-sizeSucName = liftTCM $
-  ifM (not . optSizedTypes <$> pragmaOptions) (return Nothing) $ do
+sizeSucName = do
+  ifM (not . optSizedTypes <$> pragmaOptions) (return Nothing) $ tryMaybe $ do
     Def x [] <- ignoreSharing <$> primSizeSuc
-    return $ Just x
-  `catchError` \_ -> return Nothing
+    return x
 
 sizeSuc :: Nat -> Term -> TCM Term
 sizeSuc n v | n < 0     = __IMPOSSIBLE__
