@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP                 #-}
-{-# LANGUAGE FlexibleInstances   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 -- | Tools to manipulate patterns in abstract syntax
@@ -38,12 +37,12 @@ expandLitPattern p = traverse (traverse expand) p
         | n < 0     -> __IMPOSSIBLE__
         | n > 20    -> tooBig
         | otherwise -> do
-          Con z _ <- ignoreSharing <$> primZero
-          Con s _ <- ignoreSharing <$> primSuc
+          Con z _ _ <- ignoreSharing <$> primZero
+          Con s _ _ <- ignoreSharing <$> primSuc
           let zero  = A.ConP cinfo (A.AmbQ [setRange r $ conName z]) []
               suc p = A.ConP cinfo (A.AmbQ [setRange r $ conName s]) [defaultNamedArg p]
               info  = A.PatRange r
-              cinfo = A.ConPatInfo ConPCon info
+              cinfo = A.ConPatInfo ConOCon info
               p'    = foldr ($) zero $ genericReplicate n suc
           return $ foldr (A.AsP info) p' xs
       _ -> return p
@@ -80,6 +79,7 @@ instance ExpandPatternSynonyms A.Pattern where
     A.DotP{}             -> return p
     A.LitP{}             -> return p
     A.AbsurdP{}          -> return p
+    A.ProjP{}            -> return p
     A.ConP i ds as       -> A.ConP i ds <$> expandPatternSynonyms as
     A.DefP i q as        -> A.DefP i q <$> expandPatternSynonyms as
     A.AsP i x p          -> A.AsP i x <$> expandPatternSynonyms p
