@@ -1,6 +1,6 @@
 {-# LANGUAGE CPP                 #-}
-{-# LANGUAGE DeriveFunctor       #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 -- | Pattern matcher used in the reducer for clauses that
 --   have not been compiled to case trees yet.
@@ -9,6 +9,8 @@ module Agda.TypeChecking.Patterns.Match where
 
 import Prelude hiding (null)
 
+import Data.IntMap (IntMap)
+import qualified Data.IntMap as IntMap
 import Data.Monoid
 import Data.Traversable (traverse)
 
@@ -20,10 +22,16 @@ import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad hiding (reportSDoc)
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Records
+import Agda.TypeChecking.Datatypes
 
+import Agda.Utils.Empty
 import Agda.Utils.Functor (for, ($>))
+import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
+import Agda.Utils.Singleton
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 
@@ -32,7 +40,7 @@ import Agda.Utils.Impossible
 
 -- | If matching is inconclusive (@DontKnow@) we want to know whether
 --   it is due to a particular meta variable.
-data Match a = Yes Simplification [a]
+data Match a = Yes Simplification (IntMap (Arg a))
              | No
              | DontKnow (Blocked ())
   deriving Functor
@@ -42,6 +50,17 @@ instance Null (Match a) where
   null (Yes simpl as) = null simpl && null as
   null _              = False
 
+matchedArgs :: Empty -> Int -> IntMap (Arg a) -> [Arg a]
+matchedArgs err n vs = map get [0..n-1]
+  where
+    get k = fromMaybe (absurd err) $ IntMap.lookup k vs
+
+-- | Builds a proper substitution from an IntMap produced by match(Co)patterns
+buildSubstitution :: (DeBruijn a)
+                  => Empty -> Int -> IntMap (Arg a) -> Substitution' a
+buildSubstitution err n vs = parallelS $ map unArg $ matchedArgs err n vs
+
+
 -- 'mappend' is UNUSED.
 --
 -- instance Monoid (Match a) where
@@ -96,15 +115,15 @@ foldMatch match = loop where
             (r', vs') <- loop ps vs
             let vs1 = v' : vs'
             case r' of
-              Yes s' us' -> return (Yes (s `mappend` s') (us ++ us'), vs1)
-              No         -> return (No                              , vs1)
-              DontKnow m -> return (DontKnow m                      , vs1)
+              Yes s' us' -> return (Yes (s `mappend` s') (us `mappend` us'), vs1)
+              No         -> return (No                                     , vs1)
+              DontKnow m -> return (DontKnow m                             , vs1)
       _ -> __IMPOSSIBLE__
 
 -- | @matchCopatterns ps es@ matches spine @es@ against copattern spine @ps@.
 --
 --   Returns 'Yes' and a substitution for the pattern variables
---   (in form of [Term]) if matching was successful.
+--   (in form of IntMap Term) if matching was successful.
 --
 --   Returns 'No' if there was a constructor or projection mismatch.
 --
@@ -114,7 +133,9 @@ foldMatch match = loop where
 --   In any case, also returns spine @es@ in reduced form
 --   (with all the weak head reductions performed that were necessary
 --   to come to a decision).
-matchCopatterns :: [NamedArg Pattern] -> [Elim] -> ReduceM (Match Term, [Elim])
+matchCopatterns :: [NamedArg DeBruijnPattern]
+                -> [Elim]
+                -> ReduceM (Match Term, [Elim])
 matchCopatterns ps vs = do
   traceSDoc "tc.match" 50
     (vcat [ text "matchCopatterns"
@@ -126,15 +147,21 @@ matchCopatterns ps vs = do
      foldMatch (matchCopattern . namedArg) ps vs
 
 -- | Match a single copattern.
-matchCopattern :: Pattern -> Elim -> ReduceM (Match Term, Elim)
-matchCopattern (ProjP p) elim@(Proj q)
-  | p == q    = return (Yes YesSimplification [], elim)
-  | otherwise = return (No                      , elim)
+matchCopattern :: DeBruijnPattern
+               -> Elim
+               -> ReduceM (Match Term, Elim)
+matchCopattern pat@ProjP{} elim@(Proj _ q) = do
+  ProjP _ p <- normaliseProjP pat
+  q         <- getOriginalProjection q
+  return $ if p == q then (Yes YesSimplification empty, elim)
+                     else (No,                          elim)
 matchCopattern ProjP{} Apply{}   = __IMPOSSIBLE__
 matchCopattern _       Proj{}    = __IMPOSSIBLE__
 matchCopattern p       (Apply v) = mapSnd Apply <$> matchPattern p v
 
-matchPatterns :: [NamedArg Pattern] -> [Arg Term] -> ReduceM (Match Term, [Arg Term])
+matchPatterns :: [NamedArg DeBruijnPattern]
+              -> [Arg Term]
+              -> ReduceM (Match Term, [Arg Term])
 matchPatterns ps vs = do
   traceSDoc "tc.match" 50
     (vcat [ text "matchPatterns"
@@ -147,36 +174,51 @@ matchPatterns ps vs = do
      foldMatch (matchPattern . namedArg) ps vs
 
 -- | Match a single pattern.
-matchPattern :: Pattern -> Arg Term -> ReduceM (Match Term, Arg Term)
+matchPattern :: DeBruijnPattern
+             -> Arg Term
+             -> ReduceM (Match Term, Arg Term)
 matchPattern p u = case (p, u) of
   (ProjP{}, _            ) -> __IMPOSSIBLE__
-  (VarP _ , arg@(Arg _ v)) -> return (Yes NoSimplification [v], arg)
-  (DotP _ , arg@(Arg _ v)) -> return (Yes NoSimplification [v], arg)
+  (VarP x , arg          ) -> return (Yes NoSimplification entry, arg)
+    where entry = singleton (dbPatVarIndex x, arg)
+  (DotP _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
   (LitP l , arg@(Arg _ v)) -> do
     w <- reduceB' v
     let arg' = arg $> ignoreBlocking w
     case ignoreSharing <$> w of
       NotBlocked _ (Lit l')
-          | l == l'            -> return (Yes YesSimplification []    , arg')
+          | l == l'            -> return (Yes YesSimplification empty , arg')
           | otherwise          -> return (No                          , arg')
       NotBlocked _ (MetaV x _) -> return (DontKnow $ Blocked x ()     , arg')
       Blocked x _              -> return (DontKnow $ Blocked x ()     , arg')
       NotBlocked r t           -> return (DontKnow $ NotBlocked r' () , arg')
         where r' = stuckOn (Apply arg') r
 
-  -- Case record pattern: always succeed!
-  -- This case is necessary if we want to use the clauses before
-  -- record pattern translation (e.g., in type-checking definitions by copatterns).
-  (ConP con@(ConHead c _ ds) ConPatternInfo{conPRecord = Just{}} ps, arg@(Arg info v))
-     -- precondition: con actually comes with the record fields
-     | size ds == size ps -> mapSnd (Arg info . Con con) <$> do
-         matchPatterns ps $ for ds $ \ d -> Arg info $ v `applyE` [Proj d]
-           -- TODO: correct info for projected terms
-     | otherwise -> __IMPOSSIBLE__
+  -- Case constructor pattern.
+  (ConP c cpi ps, Arg info v) -> do
+    if isNothing $ conPRecord cpi then fallback else do
+    isEtaRecordCon (conName c) >>= \case
+      Nothing -> fallback
+      Just fs -> do
+        -- Case: Eta record constructor.
+        -- This case is necessary if we want to use the clauses before
+        -- record pattern translation (e.g., in type-checking definitions by copatterns).
+        unless (size fs == size ps) __IMPOSSIBLE__
+        mapSnd (Arg info . Con c (fromConPatternInfo cpi)) <$> do
+          matchPatterns ps $ for fs $ \ (Arg ai f) -> Arg ai $ v `applyE` [Proj ProjSystem f]
+    where
+    isEtaRecordCon :: QName -> ReduceM (Maybe [Arg QName])
+    isEtaRecordCon c = do
+      (theDef <$> getConstInfo c) >>= \case
+        Constructor{ conData = d } -> do
+          (theDef <$> getConstInfo d) >>= \case
+            r@Record{ recFields = fs } | recEtaEquality r -> return $ Just fs
+            _ -> return Nothing
+        _ -> __IMPOSSIBLE__
 
-  -- Case data constructor pattern.
-  (ConP c _ ps, Arg info v) ->
-    do  w <- traverse constructorForm =<< reduceB' v
+    -- Default: not an eta record constructor.
+    fallback = do
+        w <- reduceB' v
         -- Unfold delayed (corecursive) definitions one step. This is
         -- only necessary if c is a coinductive constructor, but
         -- 1) it does not hurt to do it all the time, and
@@ -192,21 +234,24 @@ matchPattern p u = case (p, u) of
                    -- unfolded (due to open public).
                _ -> return w
 -}
-        w <- case w of
+        -- Jesper, 23-06-2016: Note that unfoldCorecursion may destroy
+        -- constructor forms, so we only call constructorForm after.
+        w <- traverse constructorForm =<< case w of
                NotBlocked r u -> unfoldCorecursion u  -- Andreas, 2014-06-12 TODO: r == ReallyNotBlocked sufficient?
                _ -> return w
         let v = ignoreBlocking w
             arg = Arg info v  -- the reduced argument
         case ignoreSharing <$> w of
-          NotBlocked _ (Con c' vs)
+          NotBlocked _ (Con c' ci vs)
             | c == c'               -> do
                 (m, vs) <- yesSimplification <$> matchPatterns ps vs
-                return (m, Arg info $ Con c' vs)
+                return (m, Arg info $ Con c' ci vs)
             | otherwise             -> return (No                          , arg)
           NotBlocked _ (MetaV x vs) -> return (DontKnow $ Blocked x ()     , arg)
           Blocked x _               -> return (DontKnow $ Blocked x ()     , arg)
           NotBlocked r _            -> return (DontKnow $ NotBlocked r' () , arg)
             where r' = stuckOn (Apply arg) r
+
 -- ASR (08 November 2014). The type of the function could be
 --
 -- @(Match Term, [Arg Term]) -> (Match Term, [Arg Term])@.
