@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
 
 module Agda.TypeChecking.Polarity where
 
@@ -92,11 +90,10 @@ purgeNonvariant = map (\ p -> if p == Nonvariant then Covariant else p)
 
 -- | Main function of this module.
 computePolarity :: QName -> TCM ()
-computePolarity x = inConcreteOrAbstractMode x $ do
+computePolarity x = inConcreteOrAbstractMode x $ \ def -> do
   reportSLn "tc.polarity.set" 25 $ "Computing polarity of " ++ show x
 
   -- get basic polarity from positivity analysis
-  def      <- getConstInfo x
   let npars = droppedPars def
   let pol0 = replicate npars Nonvariant ++ map polFromOcc (defArgOccurrences def)
   reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ show x ++ " from positivity: " ++ show pol0
@@ -165,13 +162,13 @@ usagePolarity :: Defn -> [Polarity]
 usagePolarity def = case def of
     Axiom{}                                 -> []
     Function{ funClauses = [] }             -> []
-    Function{ funClauses = cs }             -> usage $ map clausePats cs
+    Function{ funClauses = cs }             -> usage $ map namedClausePats cs
     Datatype{ dataPars = np, dataIxs = ni } -> genericReplicate np Nonvariant
     Record{ recPars = n }                   -> genericReplicate n Nonvariant
     Constructor{}                           -> []
     Primitive{}                             -> []
   where
-    usage = foldr1 (zipWith (/\)) . map (map (usagePat . unArg))
+    usage = foldr1 (zipWith (/\)) . map (map (usagePat . namedArg))
     usagePat VarP{} = Nonvariant
     usagePat DotP{} = Nonvariant
     usagePat ConP{} = Invariant
@@ -258,7 +255,7 @@ nonvariantToUnusedArgInClause pol cl@Clause{clauseTel = tel, namedClausePats = p
       -- get a list of 'Relevance's for the variables bound in the pattern
       rels0 = getRelevance <$> (concatMap (patternVars . fmap namedThing) ps')
       -- this is the order the variables appear in the telescope
-      perm  = clausePerm cl
+      perm  = fromMaybe __IMPOSSIBLE__ $ clausePerm cl
       rels  = permute perm rels0
       -- now improve 'Relevance' in 'Telescope' by pattern relevance
       updateDom UnusedArg = mapRelevance mkUnused
@@ -293,7 +290,7 @@ sizePolarity d pol0 = do
             -- and seek confirm it by looking at the constructor types
             let check c = do
                   t <- defType <$> getConstInfo c
-                  addCtxTel (telFromList parTel) $ do
+                  addContext (telFromList parTel) $ do
                     let pars = map (defaultArg . var) $ downFrom np
                     TelV conTel target <- telView =<< (t `piApplyM` pars)
                     case conTel of
@@ -387,7 +384,7 @@ instance HasPolarity Type where
   polarities i (El _ v) = polarities i v
 
 instance HasPolarity a => HasPolarity (Elim' a) where
-  polarities i (Proj p)  = return []
+  polarities i Proj{}    = return []
   polarities i (Apply a) = polarities i a
 
 instance HasPolarity Term where
@@ -405,7 +402,7 @@ instance HasPolarity Term where
       pols <- getPolarity x
       let compose p ps = map (composePol p) ps
       concat . zipWith compose (pols ++ repeat Invariant) <$> mapM (polarities i) ts
-    Con _ ts   -> polarities i ts -- constructors can be seen as monotone in all args.
+    Con _ _ ts -> polarities i ts -- constructors can be seen as monotone in all args.
     Pi a b     -> (++) <$> (map neg <$> polarities i a) <*> polarities i b
     Sort s     -> return [] -- polarities i s -- return []
     MetaV _ ts -> map (const Invariant) <$> polarities i ts
