@@ -1,14 +1,15 @@
 {-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE TemplateHaskell    #-}
+
 
 -- | Occurrences.
 
 module Agda.TypeChecking.Positivity.Occurrence
   ( Occurrence(..)
+  , OccursWhere(..)
+  , Where(..)
   , boundToEverySome
   , productOfEdgesInBoundedWalk
-  , Agda.TypeChecking.Positivity.Occurrence.tests
   ) where
 
 import Control.Applicative
@@ -19,8 +20,10 @@ import Data.Maybe
 import Data.Typeable (Typeable)
 import Data.Map.Strict (Map)
 import qualified Data.Map.Strict as Map
-import Test.QuickCheck
+import qualified Data.Sequence as DS
 
+import Agda.Syntax.Common
+import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Position
 import Agda.Utils.Graph.AdjacencyMap.Unidirectional (Graph)
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as Graph
@@ -31,6 +34,33 @@ import Agda.Utils.SemiRing
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+-- Specification of occurrences -------------------------------------------
+
+-- Operations and instances in Agda.TypeChecking.Positivity.
+
+-- | Description of an occurrence.
+data OccursWhere
+  = Unknown
+    -- ^ an unknown position (treated as negative)
+  | Known Range (DS.Seq Where)
+    -- ^ The elements of the sequence, from left to right, explain how
+    -- to get to the occurrence.
+  deriving (Show, Eq, Ord)
+
+-- | One part of the description of an occurrence.
+data Where
+  = LeftOfArrow
+  | DefArg QName Nat -- ^ in the nth argument of a define constant
+  | UnderInf         -- ^ in the principal argument of built-in ∞
+  | VarArg           -- ^ as an argument to a bound variable
+  | MetaArg          -- ^ as an argument of a metavariable
+  | ConArgType QName -- ^ in the type of a constructor
+  | IndArgType QName -- ^ in a datatype index of a constructor
+  | InClause Nat     -- ^ in the nth clause of a defined function
+  | Matched          -- ^ matched against in a clause of a defined function
+  | InDefOf QName    -- ^ in the definition of a constant
+  deriving (Show, Eq, Ord)
+
 -- | Subterm occurrences for positivity checking.
 --   The constructors are listed in increasing information they provide:
 --   @Mixed <= JustPos <= StrictPos <= GuardPos <= Unused@
@@ -49,15 +79,6 @@ instance NFData Occurrence where rnf x = seq x ()
 instance KillRange Occurrence where
   killRange = id
 
-instance Arbitrary Occurrence where
-  arbitrary = elements [minBound .. maxBound]
-
-  shrink Unused = []
-  shrink _      = [Unused]
-
-instance CoArbitrary Occurrence where
-  coarbitrary = coarbitrary . fromEnum
-
 -- | 'Occurrence' is a complete lattice with least element 'Mixed'
 --   and greatest element 'Unused'.
 --
@@ -154,128 +175,3 @@ productOfEdgesInBoundedWalk occ g u v bound =
         Just es@(_ : _) -> Just (foldr1 otimes (map Graph.label es))
         Just []         -> __IMPOSSIBLE__
         Nothing         -> Nothing
-
-------------------------------------------------------------------------
--- Tests
-
-prop_Occurrence_oplus_associative ::
-  Occurrence -> Occurrence -> Occurrence -> Bool
-prop_Occurrence_oplus_associative x y z =
-  oplus x (oplus y z) == oplus (oplus x y) z
-
-prop_Occurrence_oplus_ozero :: Occurrence -> Bool
-prop_Occurrence_oplus_ozero x =
-  oplus ozero x == x
-
-prop_Occurrence_oplus_commutative :: Occurrence -> Occurrence -> Bool
-prop_Occurrence_oplus_commutative x y =
-  oplus x y == oplus y x
-
-prop_Occurrence_otimes_associative ::
-  Occurrence -> Occurrence -> Occurrence -> Bool
-prop_Occurrence_otimes_associative x y z =
-  otimes x (otimes y z) == otimes (otimes x y) z
-
-prop_Occurrence_otimes_oone :: Occurrence -> Bool
-prop_Occurrence_otimes_oone x =
-  otimes oone x == x
-    &&
-  otimes x oone == x
-
-prop_Occurrence_distributive ::
-  Occurrence -> Occurrence -> Occurrence -> Bool
-prop_Occurrence_distributive x y z =
-  otimes x (oplus y z) == oplus (otimes x y) (otimes x z)
-    &&
-  otimes (oplus x y) z == oplus (otimes x z) (otimes y z)
-
-prop_Occurrence_otimes_ozero :: Occurrence -> Bool
-prop_Occurrence_otimes_ozero x =
-  otimes ozero x == ozero
-    &&
-  otimes x ozero == ozero
-
-prop_Occurrence_ostar :: Occurrence -> Bool
-prop_Occurrence_ostar x =
-  ostar x == oplus oone (otimes x (ostar x))
-    &&
-  ostar x == oplus oone (otimes (ostar x) x)
-
--- | Is the given predicate satisfiable?
-
-satisfiable :: (Occurrence -> Bool) -> Bool
-satisfiable p = or [ p o | o <- [minBound .. maxBound] ]
-
--- Some properties that are used in the implementation of
--- prop_boundToEverySome2.
-
-prop_boundToEverySome0 :: Bool
-prop_boundToEverySome0 = and
-  [ length ess >= 1
-      &&
-    all satisfiable ps
-      &&
-    and [ p minBound | p <- ps ]
-      &&
-    all (\p -> satisfiable (not . p)) [ e | (e, _) <- ess ]
-      &&
-    and [ not (p maxBound) | p <- ps, satisfiable (not . p) ]
-  | (_, ess) <- Map.toList boundToEverySome
-  , let ps = concat [ [e, s] | (e, s) <- ess ]
-  ]
-
--- A simple property that does not always generate enough interesting
--- test cases.
-
-prop_boundToEverySome1 :: NonEmptyList Occurrence -> Property
-prop_boundToEverySome1 (NonEmpty w) =
-  forAll (elements $ Map.toList boundToEverySome) $ \(bound, ess) ->
-    (foldr1 otimes w <= bound)
-      ==
-    or [ all every w && any some w | (every, some) <- ess ]
-
--- A more complicated property that does not always generate enough
--- interesting test cases.
-
-prop_boundToEverySome2 :: Property
-prop_boundToEverySome2 =
-  forAll (elements $ Map.toList boundToEverySome) $ \(bound, ess) ->
-    (forAll (oneof [ do os1 <- listOf (arbitrary `suchThat` every)
-                        o   <- arbitrary
-                                 `suchThat` (\o -> every o && some o)
-                        os2 <- listOf (arbitrary `suchThat` every)
-                        return (os1 ++ [o] ++ os2)
-                   | (every, some) <- ess
-                   ]) $ \w ->
-       foldr1 otimes w <= bound)
-      .&&.
-    (forAll (do
-         ess <- mapM (\(e, s) ->
-                         elements
-                           (Left e :
-                            [ Right s | satisfiable (not . s) ])) ess
-         let (es, ss) = partitionEithers ess
-             every    = \o -> and [ not (s o) | s <- ss ]
-             some e   = \o -> every o && not (e o)
-             everyG   = arbitrary `suchThat` every
-             segment  = listOf everyG
-         os <- uniqOn id <$> mapM (\e -> arbitrary `suchThat` some e) es
-         if Prelude.null os
-           then listOf1 everyG
-           else (++) <$> listOf everyG
-                     <*> (concat <$>
-                            mapM (\o -> (o :) <$> listOf everyG) os))
-       (\w -> not (foldr1 otimes w <= bound)))
-
-------------------------------------------------------------------------
-
--- Template Haskell hack to make the following $quickCheckAll work
--- under GHC 7.8.
-return []
-
--- | Tests.
-
-tests :: IO Bool
-tests = do
-  putStrLn "Agda.TypeChecking.Positivity.Occurrence"
-  $quickCheckAll
