@@ -1,9 +1,17 @@
 {-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleContexts     #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TupleSections        #-}
 {-# LANGUAGE UndecidableInstances #-}
 
+-- To define <>, we will probably need to add:
+--import Prelude hiding ((<>))
+-- but using that now gives warnings and doesn't silence -Wsemigroup
+#if __GLASGOW_HASKELL__ >= 800
+{-# OPTIONS_GHC -Wno-semigroup    #-}
+#endif
+#if __GLASGOW_HASKELL__ <= 708
+{-# LANGUAGE OverlappingInstances #-}
+#endif
+
+
 module Agda.TypeChecking.Pretty where
 
 import Prelude hiding (null)
@@ -11,6 +19,7 @@ import Prelude hiding (null)
 import Control.Applicative hiding (empty)
 import Control.Monad
 
+import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe
 
@@ -69,6 +78,9 @@ prettyAs x = AP.prettyAs x
 text :: String -> TCM Doc
 text s = return $ P.text s
 
+multiLineText :: String -> TCM Doc
+multiLineText s = return $ P.multiLineText s
+
 pwords :: String -> [TCM Doc]
 pwords s = map return $ P.pwords s
 
@@ -82,6 +94,9 @@ hsep ds = P.hsep <$> sequence ds
 hcat ds = P.hcat <$> sequence ds
 vcat ds = P.vcat <$> sequence ds
 
+hang :: TCM Doc -> Int -> TCM Doc -> TCM Doc
+hang p n q = P.hang <$> p <*> pure n <*> q
+
 ($$), ($+$), (<>), (<+>) :: TCM Doc -> TCM Doc -> TCM Doc
 d1 $$ d2  = (P.$$) <$> d1 <*> d2
 d1 $+$ d2 = (P.$+$) <$> d1 <*> d2
@@ -132,7 +147,11 @@ instance PrettyTCM Range       where prettyTCM = pretty
 instance PrettyTCM a => PrettyTCM (Closure a) where
   prettyTCM cl = enterClosure cl prettyTCM
 
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPABLE #-} PrettyTCM a => PrettyTCM [a] where
+#else
 instance PrettyTCM a => PrettyTCM [a] where
+#endif
   prettyTCM = prettyList . map prettyTCM
 
 instance (PrettyTCM a, PrettyTCM b) => PrettyTCM (a,b) where
@@ -147,6 +166,7 @@ instance PrettyTCM Type         where prettyTCM = prettyA <=< reify
 instance PrettyTCM Sort         where prettyTCM = prettyA <=< reify
 instance PrettyTCM DisplayTerm  where prettyTCM = prettyA <=< reify
 instance PrettyTCM NamedClause  where prettyTCM = prettyA <=< reify
+instance PrettyTCM (QNamed Clause) where prettyTCM = prettyA <=< reify
 instance PrettyTCM Level        where prettyTCM = prettyA <=< reify . Level
 instance PrettyTCM Permutation  where prettyTCM = text . show
 instance PrettyTCM Polarity     where prettyTCM = text . show
@@ -161,22 +181,14 @@ instance (Show a, PrettyTCM a, Subst a a) => PrettyTCM (Substitution' a) where
       (rho1, rho2) = splitS 1 rho
       u            = lookupS rho2 0
 
+instance PrettyTCM ModuleParameters where
+  prettyTCM = prettyTCM . mpSubstitution
+
 instance PrettyTCM Clause where
   prettyTCM cl = do
     x <- qualify_ <$> freshName_ "<unnamedclause>"
     prettyTCM (QNamed x cl)
 
-instance PrettyTCM ClauseBody where
-  prettyTCM b = do
-    (binds, body) <- walk b
-    sep [ brackets (fsep binds), return body ]
-    where
-      walk NoBody = return ([], P.text "()")
-      walk (Body v) = ([],) <$> prettyTCM v
-      walk (Bind b) = do
-        (bs, v) <- underAbstraction_ b walk
-        return (text (argNameToString $ absName b) : bs, v)
-
 instance PrettyTCM a => PrettyTCM (Judgement a) where
   prettyTCM (HasType a t) = prettyTCM a <+> text ":" <+> prettyTCM t
   prettyTCM (IsSort  a t) = text "Sort" <+> prettyTCM a <+> text ":" <+> prettyTCM t
@@ -199,10 +211,21 @@ instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Arg a) where
 instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Dom a) where
   prettyTCM x = prettyA =<< reify x
 
+instance (PrettyTCM k, PrettyTCM v) => PrettyTCM (Map k v) where
+  prettyTCM m = text "Map" <> braces (sep $ punctuate comma
+    [ hang (prettyTCM k <+> text "=") 2 (prettyTCM v) | (k, v) <- Map.toList m ])
+
+#if __GLASGOW_HASKELL__ >= 710
+instance {-# OVERLAPPING #-} PrettyTCM ArgName where
+#else
+instance PrettyTCM ArgName where
+#endif
+  prettyTCM = text . show
+
 -- instance (Reify a e, ToConcrete e c, P.Pretty c, PrettyTCM a) => PrettyTCM (Elim' a) where
 instance PrettyTCM Elim where
   prettyTCM (Apply v) = text "$" <+> prettyTCM v
-  prettyTCM (Proj f)  = text "." <> prettyTCM f
+  prettyTCM (Proj _ f)= text "." <> prettyTCM f
 
 instance PrettyTCM a => PrettyTCM (MaybeReduced a) where
   prettyTCM = prettyTCM . ignoreReduced
@@ -213,6 +236,9 @@ instance PrettyTCM EqualityView where
 instance PrettyTCM A.Expr where
   prettyTCM = prettyA
 
+instance PrettyTCM A.TypedBinding where
+  prettyTCM = prettyA
+
 instance PrettyTCM Relevance where
   prettyTCM Irrelevant = text "."
   prettyTCM NonStrict  = text ".."
@@ -221,7 +247,8 @@ instance PrettyTCM Relevance where
   prettyTCM UnusedArg  = empty
 
 instance PrettyTCM ProblemConstraint where
-  prettyTCM (PConstr pid c) = prettyList (map prettyTCM pid) <+> prettyTCM c
+  prettyTCM (PConstr []   c) = prettyTCM c
+  prettyTCM (PConstr pids c) = prettyList (map prettyTCM pids) <+> prettyTCM c
 
 instance PrettyTCM Constraint where
     prettyTCM c = case c of
@@ -269,23 +296,26 @@ instance PrettyTCM Constraint where
                     , nest 2 $ prettyTCM p ]
               Open{}  -> __IMPOSSIBLE__
               OpenIFS{}  -> __IMPOSSIBLE__
-              InstS{} -> __IMPOSSIBLE__
               InstV{} -> __IMPOSSIBLE__
-        FindInScope m mb Nothing -> do
+        FindInScope m mb mcands -> do
             t <- getMetaType m
-            sep [ text "Find in scope" <+> pretty m
-                  <+> maybe (text ":") (\ b -> text "blocked on" <+> pretty b <+> text ":") mb
-                , prettyTCM t
-                , text " (no candidate for now)"
-                ]
-        FindInScope m mb (Just cands) -> do
-            t <- getMetaType m
-            sep [ text "Find in scope" <+> pretty m
-                  <+> maybe (text ":") (\ b -> text "blocked on" <+> pretty b <+> text ":") mb
-                , nest 2 $ prettyTCM t
-                , sep $ flip map cands $ \cand ->
-                           prettyTCM (candidateTerm cand) <+> text ": " <+> prettyTCM (candidateType cand)
+            sep [ hang (text "Resolve instance argument" <+> blk) 2 $
+                  hang (pretty m <+> text ":") 2 $ prettyTCM t
+                , cands
                 ]
+          where
+            blk = case mb of
+                    Nothing -> empty
+                    Just b  -> parens $ text "blocked on" <+> pretty b
+            cands =
+              case mcands of
+                Nothing -> text "No candidates yet"
+                Just cnds ->
+                  hang (text "Candidates") 2 $
+                    vcat [ hang (overlap c <+> prettyTCM (candidateTerm c) <+> text ":") 2 $
+                            prettyTCM (candidateType c) | c <- cnds ]
+              where overlap c | candidateOverlappable c = text "overlap"
+                              | otherwise               = empty
         IsEmpty r t ->
             sep [ text "Is empty:", nest 2 $ prettyTCM t ]
         CheckSizeLtSat t ->
@@ -340,15 +370,7 @@ instance PrettyTCM Telescope where
 newtype PrettyContext = PrettyContext Context
 
 instance PrettyTCM PrettyContext where
-  prettyTCM (PrettyContext ctx) = P.fsep . reverse <$> pr (map ctxEntry ctx)
-    where
-    pr :: [Dom (Name, Type)] -> TCM [P.Doc]
-    pr []                            = return []
-    pr (Dom info (x,t) : ctx) = escapeContext 1 $ do
-      -- TODO guilhem: show colors
-      d <- CP.prettyRelevance info . CP.prettyHiding info P.parens <$> do
-             prettyTCM x <+> text ":" <+> prettyTCM t
-      (d :) <$> pr ctx
+  prettyTCM (PrettyContext ctx) = prettyTCM $ telFromList' nameToArgName $ map ctxEntry $ reverse ctx
 
 instance PrettyTCM Context where
   prettyTCM = prettyTCM . PrettyContext
@@ -356,72 +378,82 @@ instance PrettyTCM Context where
 instance PrettyTCM CtxId where
   prettyTCM (CtxId x) = prettyTCM x
 
-instance PrettyTCM Pattern where
-  prettyTCM = showPat' (text . patVarNameToString)
+instance PrettyTCM DBPatVar where
+  prettyTCM = prettyTCM . var . dbPatVarIndex
 
-instance PrettyTCM DeBruijnPattern where
-  prettyTCM = showPat' $ \ (i, x) -> prettyTCM $ var i
-
--- | Show a pattern, given a method how to show pattern variables.
-showPat' :: (a -> TCM Doc) -> Pattern' a -> TCM Doc
-showPat' showVar = showPat
-  where
-      showPat (VarP x)      = showVar x
-      showPat (DotP t)      = text ".(" <> prettyTCM t <> text ")"
-      showPat (ConP c i ps) = (if b then braces else parens) $ prTy $
-        prettyTCM c <+> fsep (map (showPat . namedArg) ps)
+instance PrettyTCM a => PrettyTCM (Pattern' a) where
+  prettyTCM (VarP x)      = prettyTCM x
+  prettyTCM (DotP t)      = text ".(" <> prettyTCM t <> text ")"
+  prettyTCM (ConP c i ps) = (if b then braces else parens) $ prTy $
+        prettyTCM c <+> fsep (map (prettyTCM . namedArg) ps)
         where
-        b = maybe False (/= ConPCon) $ conPRecord i
+        b = maybe False (/= ConOCon) $ conPRecord i
         showRec :: TCM Doc
         showRec = sep
           [ text "record"
           , bracesAndSemicolons <$> zipWithM showField (conFields c) ps
           ]
         showField x p =
-          sep [ prettyTCM (A.qnameName x) <+> text "=" , nest 2 $ showPat $ namedArg p ]
-        showCon = parens $ prTy $ prettyTCM c <+> fsep (map (showPat . namedArg) ps)
+          sep [ prettyTCM (A.qnameName x) <+> text "=" , nest 2 $ prettyTCM $ namedArg p ]
+        showCon = parens $ prTy $ prettyTCM c <+> fsep (map (prettyTCM . namedArg) ps)
         prTy d = d -- caseMaybe (conPType i) d $ \ t -> d  <+> text ":" <+> prettyTCM t
-      showPat (LitP l)      = text (show l)
-      showPat (ProjP q)     = text (show q)
+  prettyTCM (LitP l)      = text (show l)
+  prettyTCM (ProjP _ q)   = text ("." ++ show q)
+
+-- | Proper pretty printing of patterns:
+prettyTCMPatternList :: [NamedArg DeBruijnPattern] -> TCM Doc
+prettyTCMPatternList = prettyList . map prettyA <=< reifyPatterns
 
 instance PrettyTCM (Elim' DisplayTerm) where
   prettyTCM (Apply v) = text "$" <+> prettyTCM (unArg v)
-  prettyTCM (Proj f)  = text "." <> prettyTCM f
+  prettyTCM (Proj _ f)= text "." <> prettyTCM f
 
 raisePatVars :: Int -> NLPat -> NLPat
-raisePatVars k (PVar id x) = PVar id (k+x)
+raisePatVars k (PVar id x bvs) = PVar id (k+x) bvs
 raisePatVars k (PWild)     = PWild
 raisePatVars k (PDef f es) = PDef f $ (fmap . fmap) (raisePatVars k) es
 raisePatVars k (PLam i u)  = PLam i $ fmap (raisePatVars k) u
-raisePatVars k (PPi a b)   = PPi ((fmap . fmap) (raisePatVars k) a) ((fmap . fmap) (raisePatVars k) b)
+raisePatVars k (PPi a b)   =
+  PPi (fmap (raisePatVarsInType k) a) (fmap (raisePatVarsInType k) b)
 raisePatVars k (PBoundVar i es) = PBoundVar i $ (fmap . fmap) (raisePatVars k) es
 raisePatVars k (PTerm t)   = PTerm t
 
+raisePatVarsInType :: Int -> NLPType -> NLPType
+raisePatVarsInType k (NLPType l a) =
+  NLPType (raisePatVars k l) (raisePatVars k a)
+
 instance PrettyTCM NLPat where
-  prettyTCM (PVar id x) = prettyTCM (var x)
+  prettyTCM (PVar id x bvs) = prettyTCM (Var x (map (Apply . fmap var) bvs))
   prettyTCM (PWild)     = text $ "_"
   prettyTCM (PDef f es) = parens $
     prettyTCM f <+> fsep (map prettyTCM es)
-  prettyTCM (PLam i u)  = text ("λ " ++ absName u ++ " →") <+>
-                          (addContext (absName u) $ prettyTCM (raisePatVars 1 $ absBody u))
-  prettyTCM (PPi a b)   = text "Π" <+> prettyTCM (unDom a) <+>
-                          (addContext (absName b) $ prettyTCM (fmap (raisePatVars 1) $ unAbs b))
+  prettyTCM (PLam i u)  = parens $
+    text ("λ " ++ absName u ++ " →") <+>
+    (addContext (absName u) $ prettyTCM (raisePatVars 1 $ absBody u))
+  prettyTCM (PPi a b)   = parens $
+    text ("(" ++ absName b ++ " :") <+> prettyTCM (unDom a) <> text ") →" <+>
+    (addContext (absName b) $ prettyTCM (raisePatVarsInType 1 $ unAbs b))
+  prettyTCM (PBoundVar i []) = prettyTCM (var i)
   prettyTCM (PBoundVar i es) = parens $ prettyTCM (var i) <+> fsep (map prettyTCM es)
   prettyTCM (PTerm t)   = text "." <> parens (prettyTCM t)
 
+instance PrettyTCM NLPType where
+  prettyTCM (NLPType PWild a) = prettyTCM a
+  prettyTCM (NLPType l     a) = text "{" <> prettyTCM l <> text "}" <> prettyTCM a
+
 instance PrettyTCM (Elim' NLPat) where
-  prettyTCM (Apply v) = text "$" <+> prettyTCM (unArg v)
-  prettyTCM (Proj f)  = text "." <> prettyTCM f
+  prettyTCM (Apply v) = prettyTCM (unArg v)
+  prettyTCM (Proj _ f)= text "." <> prettyTCM f
 
 instance PrettyTCM (Type' NLPat) where
   prettyTCM = prettyTCM . unEl
 
 instance PrettyTCM RewriteRule where
-  prettyTCM (RewriteRule q gamma lhs rhs b) = sep
-    [ prettyTCM q <+> text " rule "
+  prettyTCM (RewriteRule q gamma f ps rhs b) = fsep
+    [ prettyTCM q
     , prettyTCM gamma <+> text " |- "
-    , addContext gamma $ hsep
-      [ prettyTCM lhs
+    , addContext gamma $ sep
+      [ prettyTCM (PDef f ps)
       , text " --> "
       , prettyTCM rhs
       , text " : "
