@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP                        #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ScopedTypeVariables        #-}
 {-# LANGUAGE UndecidableInstances       #-}
@@ -19,6 +18,8 @@ import Data.Maybe
 import Data.Traversable (traverse)
 import Data.Monoid (mempty)
 
+import Numeric.IEEE ( IEEE(identicalIEEE) )
+
 import Agda.Interaction.Options
 
 import Agda.Syntax.Position
@@ -27,6 +28,7 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Generic (TermLike(..))
 import Agda.Syntax.Literal
 import Agda.Syntax.Concrete.Pretty ()
+import Agda.Syntax.Fixity
 
 import Agda.TypeChecking.Monad hiding (getConstInfo, typeOfConst)
 import qualified Agda.TypeChecking.Monad as TCM
@@ -38,7 +40,6 @@ import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Quote (QuotingKit, quoteTermWithKit, quoteTypeWithKit, quoteClauseWithKit, quotingKit)
 import Agda.TypeChecking.Pretty ()  -- instances only
-import Agda.TypeChecking.MetaVars (allMetas)
 
 import Agda.Utils.Monad
 import Agda.Utils.Pretty (pretty)
@@ -102,6 +103,8 @@ instance PrimTerm QName   where primTerm _ = primQName
 instance PrimTerm MetaId  where primTerm _ = primAgdaMeta
 instance PrimTerm Type    where primTerm _ = primAgdaTerm
 
+instance PrimTerm Fixity' where primTerm _ = primFixity
+
 instance PrimTerm a => PrimTerm [a] where
   primTerm _ = list (primTerm (undefined :: a))
 
@@ -159,12 +162,12 @@ instance ToTerm ArgInfo where
     ins  <- primInstance
     rel  <- primRelevant
     irr  <- primIrrelevant
-    return $ \ (ArgInfo h r) -> info `applys`
-      [ case h of
+    return $ \ i -> info `applys`
+      [ case getHiding i of
           NotHidden -> vis
           Hidden    -> hid
           Instance  -> ins
-      , case r of
+      , case getRelevance i of
           Relevant   -> rel
           Irrelevant -> irr
           NonStrict  -> rel
@@ -172,6 +175,38 @@ instance ToTerm ArgInfo where
           UnusedArg  -> irr
       ]
 
+instance ToTerm Fixity' where
+  toTerm = (. theFixity) <$> toTerm
+
+instance ToTerm Fixity where
+  toTerm = do
+    lToTm  <- toTerm
+    aToTm  <- toTerm
+    fixity <- primFixityFixity
+    return $ \ Fixity{fixityAssoc = a, fixityLevel = l} ->
+      fixity `apply` [defaultArg (aToTm a), defaultArg (lToTm l)]
+
+instance ToTerm Associativity where
+  toTerm = do
+    lassoc <- primAssocLeft
+    rassoc <- primAssocRight
+    nassoc <- primAssocNon
+    return $ \ a ->
+      case a of
+        NonAssoc   -> nassoc
+        LeftAssoc  -> lassoc
+        RightAssoc -> rassoc
+
+instance ToTerm PrecedenceLevel where
+  toTerm = do
+    (iToTm :: Integer -> Term) <- toTerm
+    related   <- primPrecRelated
+    unrelated <- primPrecUnrelated
+    return $ \ p ->
+      case p of
+        Unrelated -> unrelated
+        Related n -> related `apply` [defaultArg $ iToTm n]
+
 -- | @buildList A ts@ builds a list of type @List A@. Assumes that the terms
 --   @ts@ all have type @A@.
 buildList :: TCM ([Term] -> Term)
@@ -198,22 +233,22 @@ class FromTerm a where
 
 instance FromTerm Integer where
   fromTerm = do
-    Con pos    [] <- ignoreSharing <$> primIntegerPos
-    Con negsuc [] <- ignoreSharing <$> primIntegerNegSuc
+    Con pos _    [] <- ignoreSharing <$> primIntegerPos
+    Con negsuc _ [] <- ignoreSharing <$> primIntegerNegSuc
     toNat         <- fromTerm :: TCM (FromTermFunction Nat)
     return $ \ v -> do
       b <- reduceB' v
       let v'  = ignoreBlocking b
           arg = (<$ v')
       case ignoreSharing $ unArg (ignoreBlocking b) of
-        Con c [u]
+        Con c ci [u]
           | c == pos    ->
             redBind (toNat u)
-              (\ u' -> notReduced $ arg $ Con c [ignoreReduced u']) $ \ n ->
+              (\ u' -> notReduced $ arg $ Con c ci [ignoreReduced u']) $ \ n ->
             redReturn $ fromIntegral n
           | c == negsuc ->
             redBind (toNat u)
-              (\ u' -> notReduced $ arg $ Con c [ignoreReduced u']) $ \ n ->
+              (\ u' -> notReduced $ arg $ Con c ci [ignoreReduced u']) $ \ n ->
             redReturn $ fromIntegral $ -n - 1
         _ -> return $ NoReduction (reduced b)
 
@@ -263,7 +298,7 @@ instance FromTerm Bool where
         where
             a =?= b = ignoreSharing a === ignoreSharing b
             Def x [] === Def y []   = x == y
-            Con x [] === Con y []   = x == y
+            Con x _ [] === Con y _ [] = x == y
             Var n [] === Var m []   = n == m
             _        === _          = False
 
@@ -278,7 +313,7 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
     return $ mkList nil cons toA fromA
     where
       isCon (Lam _ b)  = isCon $ absBody b
-      isCon (Con c _)  = return c
+      isCon (Con c _ _)= return c
       isCon (Shared p) = isCon (derefPtr p)
       isCon v          = __IMPOSSIBLE__
 
@@ -287,15 +322,15 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
         let t = ignoreBlocking b
         let arg = (<$ t)
         case ignoreSharing $ unArg t of
-          Con c []
+          Con c ci []
             | c == nil  -> return $ YesReduction NoSimplification []
-          Con c [x,xs]
+          Con c ci [x,xs]
             | c == cons ->
               redBind (toA x)
-                  (\x' -> notReduced $ arg $ Con c [ignoreReduced x',xs]) $ \y ->
+                  (\x' -> notReduced $ arg $ Con c ci [ignoreReduced x',xs]) $ \y ->
               redBind
                   (mkList nil cons toA fromA xs)
-                  (fmap $ \xs' -> arg $ Con c [defaultArg $ fromA y, xs']) $ \ys ->
+                  (fmap $ \xs' -> arg $ Con c ci [defaultArg $ fromA y, xs']) $ \ys ->
               redReturn (y : ys)
           _ -> return $ NoReduction (reduced b)
 
@@ -334,11 +369,11 @@ primTrustMe = do
           hPi "y" (El (varSort 2) <$> varM 1) $
           El (varSort 3) <$>
             primEquality <#> varM 3 <#> varM 2 <@> varM 1 <@> varM 0
-  Con rf [] <- ignoreSharing <$> primRefl
+  Con rf ci [] <- ignoreSharing <$> primRefl
   n         <- conPars . theDef <$> getConInfo rf
   -- Andreas, 2015-02-27  Forced Big vs. Forced Small should not matter here
-  let refl x | n == 2    = Con rf [setRelevance (Forced Small) $ hide $ defaultArg x]
-             | n == 3    = Con rf []
+  let refl x | n == 2    = Con rf ci [setRelevance (Forced Small) $ hide $ defaultArg x]
+             | n == 3    = Con rf ci []
              | otherwise = __IMPOSSIBLE__
   return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 4 $ \ts ->
       case ts of
@@ -670,21 +705,32 @@ primitiveFunctions = Map.fromList
   , "primLevelMax"        |-> mkPrimLevelMax
 
   -- Floating point functions
-  , "primNatToFloat"      |-> mkPrimFun1 (fromIntegral :: Nat -> Double)
-  , "primFloatPlus"       |-> mkPrimFun2 ((+)          :: Op Double)
-  , "primFloatMinus"      |-> mkPrimFun2 ((-)          :: Op Double)
-  , "primFloatTimes"      |-> mkPrimFun2 ((*)          :: Op Double)
-  , "primFloatDiv"        |-> mkPrimFun2 ((/)          :: Op Double)
-  , "primFloatEquality"   |-> mkPrimFun2 (floatEq      :: Rel Double)
-  , "primFloatLess"       |-> mkPrimFun2 (floatLt      :: Rel Double)
-  , "primFloatSqrt"       |-> mkPrimFun1 (sqrt         :: Double -> Double)
-  , "primRound"           |-> mkPrimFun1 (round        :: Double -> Integer)
-  , "primFloor"           |-> mkPrimFun1 (floor        :: Double -> Integer)
-  , "primCeiling"         |-> mkPrimFun1 (ceiling      :: Double -> Integer)
-  , "primExp"             |-> mkPrimFun1 (exp          :: Fun Double)
-  , "primLog"             |-> mkPrimFun1 (log          :: Fun Double)
-  , "primSin"             |-> mkPrimFun1 (sin          :: Fun Double)
-  , "primShowFloat"       |-> mkPrimFun1 (Str . floatShow :: Double -> Str)
+  , "primNatToFloat"      |-> mkPrimFun1 (fromIntegral    :: Nat -> Double)
+  , "primFloatPlus"       |-> mkPrimFun2 ((+)             :: Op Double)
+  , "primFloatMinus"      |-> mkPrimFun2 ((-)             :: Op Double)
+  , "primFloatTimes"      |-> mkPrimFun2 ((*)             :: Op Double)
+  , "primFloatNegate"     |-> mkPrimFun1 (negate          :: Fun Double)
+  , "primFloatDiv"        |-> mkPrimFun2 ((/)             :: Op Double)
+  -- ASR (2016-09-29). We use bitwise equality for comparing Double
+  -- because Haskell's Eq, which equates 0.0 and -0.0, allows to prove
+  -- a contradiction (see Issue #2169).
+  , "primFloatEquality"   |-> mkPrimFun2 (floatEq         :: Rel Double)
+  , "primFloatNumericalEquality" |-> mkPrimFun2 ((==)     :: Rel Double)
+  , "primFloatNumericalLess"     |-> mkPrimFun2 (floatLt  :: Rel Double)
+  , "primFloatSqrt"       |-> mkPrimFun1 (sqrt            :: Double -> Double)
+  , "primRound"           |-> mkPrimFun1 (round           :: Double -> Integer)
+  , "primFloor"           |-> mkPrimFun1 (floor           :: Double -> Integer)
+  , "primCeiling"         |-> mkPrimFun1 (ceiling         :: Double -> Integer)
+  , "primExp"             |-> mkPrimFun1 (exp             :: Fun Double)
+  , "primLog"             |-> mkPrimFun1 (log             :: Fun Double)
+  , "primSin"             |-> mkPrimFun1 (sin             :: Fun Double)
+  , "primCos"             |-> mkPrimFun1 (cos             :: Fun Double)
+  , "primTan"             |-> mkPrimFun1 (tan             :: Fun Double)
+  , "primASin"            |-> mkPrimFun1 (asin            :: Fun Double)
+  , "primACos"            |-> mkPrimFun1 (acos            :: Fun Double)
+  , "primATan"            |-> mkPrimFun1 (atan            :: Fun Double)
+  , "primATan2"           |-> mkPrimFun2 (atan2           :: Double -> Double -> Double)
+  , "primShowFloat"       |-> mkPrimFun1 (Str . show      :: Double -> Str)
 
   -- Character functions
   , "primCharEquality"    |-> mkPrimFun2 ((==) :: Rel Char)
@@ -717,6 +763,7 @@ primitiveFunctions = Map.fromList
   , "primQNameEquality"   |-> mkPrimFun2 ((==) :: Rel QName)
   , "primQNameLess"       |-> mkPrimFun2 ((<) :: Rel QName)
   , "primShowQName"       |-> mkPrimFun1 (Str . show :: QName -> Str)
+  , "primQNameFixity"     |-> mkPrimFun1 (nameFixity . qnameName)
   , "primMetaEquality"    |-> mkPrimFun2 ((==) :: Rel MetaId)
   , "primMetaLess"        |-> mkPrimFun2 ((<) :: Rel MetaId)
   , "primShowMeta"        |-> mkPrimFun1 (Str . show . pretty :: MetaId -> Str)
@@ -725,22 +772,26 @@ primitiveFunctions = Map.fromList
     (|->) = (,)
 
 floatEq :: Double -> Double -> Bool
-floatEq x y | isNaN x && isNaN y = True
-            | otherwise          = x == y
+floatEq x y = identicalIEEE x y || (isNaN x && isNaN y)
 
 floatLt :: Double -> Double -> Bool
-floatLt x y
-  | isNegInf y = False
-  | isNegInf x = True
-  | isNaN x    = True
-  | otherwise  = x < y
+floatLt x y =
+  case compareFloat x y of
+    LT -> True
+    _  -> False
   where
+    -- Also implemented in the GHC/UHC backends
+    compareFloat :: Double -> Double -> Ordering
+    compareFloat x y
+      | identicalIEEE x y          = EQ
+      | isNegInf x                 = LT
+      | isNegInf y                 = GT
+      | isNaN x && isNaN y         = EQ
+      | isNaN x                    = LT
+      | isNaN y                    = GT
+      | otherwise                  = compare x y
     isNegInf z = z < 0 && isInfinite z
 
-floatShow :: Double -> String
-floatShow x | isNegativeZero x = "0.0"
-            | otherwise        = show x
-
 lookupPrimitiveFunction :: String -> TCM PrimitiveImpl
 lookupPrimitiveFunction x =
   fromMaybe (typeError $ NoSuchPrimitiveFunction x)
@@ -758,7 +809,7 @@ getBuiltinName b = do
   caseMaybeM (getBuiltin' b) (return Nothing) $ \v -> do
     v <- normalise v
     let getName (Def x _) = x
-        getName (Con x _) = conName x
+        getName (Con x _ _) = conName x
         getName (Lam _ b) = getName $ ignoreSharing $ unAbs b
         getName _         = __IMPOSSIBLE__
     return $ Just $ getName (ignoreSharing v)
