@@ -1,16 +1,11 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
-
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
 
 module Agda.TypeChecking.ProjectionLike where
 
 import Control.Monad
 
 import qualified Data.Map as Map
+import Data.Monoid (Any(..), getAny)
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
@@ -18,7 +13,7 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Free (isBinderUsed)
+import Agda.TypeChecking.Free (runFree, IgnoreSorts(..))
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Positivity
 import Agda.TypeChecking.Pretty
@@ -111,9 +106,9 @@ elimView loneProjToLambda v = do
   case pv of
     NoProjection{}        -> return v
     LoneProjectionLike f ai
-      | loneProjToLambda  -> return $ Lam ai $ Abs "r" $ Var 0 [Proj f]
+      | loneProjToLambda  -> return $ Lam ai $ Abs "r" $ Var 0 [Proj ProjPrefix f]
       | otherwise         -> return v
-    ProjectionView f a es -> (`applyE` (Proj f : es)) <$> elimView loneProjToLambda (unArg a)
+    ProjectionView f a es -> (`applyE` (Proj ProjPrefix f : es)) <$> elimView loneProjToLambda (unArg a)
 
 -- | Which @Def@types are eligible for the principle argument
 --   of a projection-like function?
@@ -124,11 +119,18 @@ eligibleForProjectionLike d = do
     Datatype{} -> True
     Record{}   -> True
     Axiom{}    -> True
-    _          -> False
+    Function{}    -> False
+    Primitive{}   -> False
+    Constructor{} -> __IMPOSSIBLE__
+    AbstractDefn  -> False
+      -- Andreas, 2016-10-11 AIM XXIV
+      -- Projection-like at abstract types violates the parameter reconstructibility property.
+      -- See test/Fail/AbstractTypeProjectionLike.
 
 -- | Turn a definition into a projection if it looks like a projection.
 makeProjection :: QName -> TCM ()
-makeProjection x = inTopContext $ do
+makeProjection x = -- if True then return () else do
+ inTopContext $ do
   -- reportSLn "tc.proj.like" 30 $ "Considering " ++ show x ++ " for projection likeness"
   defn <- getConstInfo x
   let t = defType defn
@@ -138,7 +140,7 @@ makeProjection x = inTopContext $ do
     ]
   case theDef defn of
     Function{funClauses = cls}
-      | any (isNothing . getBodyUnraised) cls ->
+      | any (isNothing . clauseBody) cls ->
         reportSLn "tc.proj.like" 30 $ "  projection-like functions cannot have absurd clauses"
     -- Constructor-headed functions can't be projection-like (at the moment). The reason
     -- for this is that invoking constructor-headedness will circumvent the inference of
@@ -177,17 +179,15 @@ makeProjection x = inTopContext $ do
               reportSLn "tc.proj.like" 60 $ "  rewrote clauses to\n    " ++ show cc
 
               -- Andreas, 2013-10-20 build parameter dropping function
-              let (ptel, Dom ai _ : _) = splitAt n $ telToList $ theTel $ telView' t
-                  -- leading lambdas are to ignore parameter applications
-                  proj = teleNoAbs ptel $ Def x []
-                  -- proj = foldr (\ (Dom ai (y, _)) -> Lam ai . NoAbs y) (Def x []) ptel
-
+              let pIndex = n + 1
+                  tel = take pIndex $ telToList $ theTel $ telView' t
+              unless (length tel == pIndex) __IMPOSSIBLE__
               let projection = Projection
-                    { projProper   = Nothing
+                    { projProper   = False
+                    , projOrig     = x
                     , projFromType = d
-                    , projIndex    = n + 1
-                    , projDropPars = proj
-                    , projArgInfo  = ai
+                    , projIndex    = pIndex
+                    , projLams     = ProjLams $ map (\ (Dom ai (y, _)) -> Arg ai y) tel
                     }
               let newDef = def
                            { funProjection     = Just projection
@@ -210,9 +210,9 @@ makeProjection x = inTopContext $ do
   where
     -- @validProj (d,n)@ checks whether the head @d@ of the type of the
     -- @n@th argument is injective in all args (i.d. being name of data/record/axiom).
-    validProj :: (QName, Int) -> TCM Bool
+    validProj :: (Arg QName, Int) -> TCM Bool
     validProj (_, 0) = return False
-    validProj (d, _) = eligibleForProjectionLike d
+    validProj (d, _) = eligibleForProjectionLike (unArg d)
 
     -- NOTE: If the following definition turns out to be slow, then
     -- one could perhaps reuse information computed by the termination
@@ -231,11 +231,12 @@ makeProjection x = inTopContext $ do
           , onlyMatch n ps  -- projection-like functions are only allowed to match on the eliminatee
                             -- otherwise we may end up projecting from constructor applications, in
                             -- which case we can't reconstruct the dropped parameters
-          , checkBody n b ]
+          , checkBody m n b ]
       where
-        Perm _ p = clausePerm cl
+        Perm _ p = fromMaybe __IMPOSSIBLE__ $ clausePerm cl
         ps       = namedClausePats cl
-        b        = clauseBody cl
+        b        = compiledClauseBody cl
+        m        = size $ concatMap patternVars $ clausePats cl
 
 
     onlyMatch n ps = all (shallowMatch . namedArg) (take 1 ps1) &&
@@ -251,10 +252,8 @@ makeProjection x = inTopContext $ do
         noMatch VarP{} = True
         noMatch DotP{} = True
 
-    checkBody 0 _          = True
-    checkBody _ NoBody     = __IMPOSSIBLE__ -- we check this earlier
-    checkBody n (Bind b)   = not (isBinderUsed b) && checkBody (n - 1) (unAbs b)
-    checkBody _ Body{}     = __IMPOSSIBLE__
+    checkBody m n b = not . getAny $ runFree badVar IgnoreNot b
+      where badVar (x,_) = Any $ m-1-n < x && x < m
 
     -- @candidateArgs [var 0,...,var(n-1)] t@ adds @(n,d)@ to the output,
     -- if @t@ is a function-type with domain @t 0 .. (n-1)@
@@ -267,13 +266,13 @@ makeProjection x = inTopContext $ do
     -- E.g. f : {x : _}(y : _){z : _} -> D x y z -> ...
     -- will return (D,3) as a candidate (amongst maybe others).
     --
-    candidateArgs :: [Term] -> Type -> [(QName,Int)]
+    candidateArgs :: [Term] -> Type -> [(Arg QName, Int)]
     candidateArgs vs t =
       case ignoreSharing $ unEl t of
         Pi a b
           | Def d es <- ignoreSharing $ unEl $ unDom a,
             Just us  <- allApplyElims es,
-            vs == map unArg us -> (d, length vs) : candidateRec b
+            vs == map unArg us -> (d <$ argFromDom a, length vs) : candidateRec b
           | otherwise          -> candidateRec b
         _                      -> []
       where
