@@ -1,8 +1,5 @@
 {-# LANGUAGE CPP                        #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE PatternGuards              #-}
-{-# LANGUAGE TupleSections              #-}
 
 -- | Code which replaces pattern matching on record constructors with
 -- uses of projection functions.
@@ -61,7 +58,7 @@ import Agda.Utils.Impossible
 --   E.g. for @(x , (y , z))@ we return @[ fst, fst . snd, snd . snd ]@.
 --
 --   If it is not a record pattern, error 'ShouldBeRecordPattern' is raised.
-recordPatternToProjections :: Pattern -> TCM [Term -> Term]
+recordPatternToProjections :: DeBruijnPattern -> TCM [Term -> Term]
 recordPatternToProjections p =
   case p of
     VarP{}       -> return [ \ x -> x ]
@@ -75,8 +72,8 @@ recordPatternToProjections p =
       concat <$> zipWithM comb (map proj fields) (map namedArg ps)
     ProjP{}      -> __IMPOSSIBLE__ -- copattern cannot appear here
   where
-    proj p = (`applyE` [Proj $ unArg p])
-    comb :: (Term -> Term) -> Pattern -> TCM [Term -> Term]
+    proj p = (`applyE` [Proj ProjSystem $ unArg p])
+    comb :: (Term -> Term) -> DeBruijnPattern -> TCM [Term -> Term]
     comb prj p = map (\ f -> f . prj) <$> recordPatternToProjections p
 
 
@@ -249,7 +246,7 @@ replaceByProjections (Arg ai i) projs cc =
                     | otherwise = map unArg xs1
               x                 = Arg ai $ foldr1 appendArgNames names
               xs'               = xs0 ++ x : xs2
-              us                = map (\ p -> Var 0 [Proj p]) (reverse projs)
+              us                = map (\ p -> Var 0 [Proj ProjSystem p]) (reverse projs)
               -- go from level (i + n - 1) to index (subtract from |xs|-1)
               index             = length xs - (i + n)
           in  Done xs' $ applySubst (liftS (length xs2) $ us ++# raiseS 1) v
@@ -455,7 +452,8 @@ translateRecordPatterns clause = do
 
       -- Substitution used to convert terms in the old telescope's
       -- context to terms in the new RHS's context.
-      rhsSubst' = mkSub $ permute (reverseP $ clausePerm clause) s'
+      perm = fromMaybe __IMPOSSIBLE__ $ clausePerm clause
+      rhsSubst' = mkSub $ permute (reverseP perm) s'
       -- TODO: Is it OK to replace the definition above with the
       -- following one?
       --
@@ -465,7 +463,7 @@ translateRecordPatterns clause = do
       -- order (i.e. the type signature for the variable which occurs
       -- first in the list of patterns comes first).
       flattenedOldTel =
-        permute (invertP __IMPOSSIBLE__ $ compactP $ clausePerm clause) $
+        permute (invertP __IMPOSSIBLE__ $ compactP perm) $
         zip (teleNames $ clauseTel clause) $
         flattenTel $
         clauseTel clause
@@ -495,7 +493,7 @@ translateRecordPatterns clause = do
 
       -- Substitution used to convert terms in the new RHS's context
       -- to terms in the new telescope's context.
-      lhsSubst' = renaming (reverseP newPerm)
+      lhsSubst' = {-'-} renaming __IMPOSSIBLE__ (reverseP newPerm)
 
       -- Substitution used to convert terms in the old telescope's
       -- context to terms in the new telescope's context.
@@ -512,8 +510,8 @@ translateRecordPatterns clause = do
       -- New clause.
       c = clause
             { clauseTel       = newTel
-            , namedClausePats = numberPatVars newPerm $ applySubst lhsSubst ps
-            , clauseBody      = translateBody cs rhsSubst $ clauseBody clause
+            , namedClausePats = numberPatVars __IMPOSSIBLE__ newPerm $ applySubst lhsSubst ps
+            , clauseBody      = applySubst lhsSubst $ clauseBody clause
             }
 
   reportSDoc "tc.lhs.recpat" 20 $ vcat
@@ -546,7 +544,7 @@ translateRecordPatterns clause = do
         [ text "delta =" <+> prettyTCM (clauseTel c)
         , text "ps    =" <+> text (show $ clausePats c)
         , text "body  =" <+> text (show $ clauseBody c)
-        , text "body  =" <+> prettyTCM (clauseBody c)
+        , text "body  =" <+> addContext (clauseTel c) (maybe (text "_|_") prettyTCM (clauseBody c))
         ]
       ]
 
@@ -584,7 +582,7 @@ nextVar = RecPatM $ do
   n <- lift get
   lift $ put $ succ n
   noVars <- lift ask
-  return (VarP "r", var $ noVars - n - 1)
+  return (varP "r", var $ noVars - n - 1)
 
 ------------------------------------------------------------------------
 -- Types used to record changes to a clause
@@ -689,7 +687,7 @@ removeTree tree = do
 translatePattern :: Pattern -> RecPatM (Pattern, [Term], Changes)
 translatePattern p@(ConP c ci ps)
   -- Andreas, 2015-05-28 only translate implicit record patterns
-  | Just ConPImplicit <- conPRecord ci = do
+  | Just ConOSystem <- conPRecord ci = do
       r <- recordTree p
       case r of
         Left  r -> r
@@ -725,7 +723,7 @@ recordTree ::
   Pattern ->
   RecPatM (Either (RecPatM (Pattern, [Term], Changes)) RecordTree)
 -- Andreas, 2015-05-28 only translate implicit record patterns
-recordTree (ConP c ci ps) | Just ConPImplicit <- conPRecord ci = do
+recordTree (ConP c ci ps) | Just ConOSystem <- conPRecord ci = do
   let t = fromMaybe __IMPOSSIBLE__ $ conPType ci
   rs <- mapM (recordTree . namedArg) ps
   case allRight rs of
@@ -738,7 +736,7 @@ recordTree (ConP c ci ps) | Just ConPImplicit <- conPRecord ci = do
       t <- reduce t
       fields <- getRecordTypeFields (unArg t)
 --      let proj p = \x -> Def (unArg p) [defaultArg x]
-      let proj p = (`applyE` [Proj $ unArg p])
+      let proj p = (`applyE` [Proj ProjSystem $ unArg p])
       return $ Right $ RecCon t $ zip (map proj fields) ts
 recordTree p@(ConP _ ci _) = return $ Left $ translatePattern p
 recordTree p@VarP{} = return (Right (Leaf p))
@@ -769,27 +767,3 @@ translateTel (Left _ : rest) (t : tel)    = Just t : translateTel rest tel
 translateTel []              []           = []
 translateTel (Left _ : _)    []           = __IMPOSSIBLE__
 translateTel []              (_ : _)      = __IMPOSSIBLE__
-
--- | Translates the clause body. The substitution should take things
--- in the context of the old RHS to the new RHS's context.
-
-translateBody :: Changes -> Substitution -> ClauseBody -> ClauseBody
-translateBody _                        s NoBody = NoBody
-translateBody (Right (n, x, _) : rest) s b      =
-  Bind $ Abs x $ translateBody rest s $ dropBinds n' b
-  where n' = sum $ map n [VarPat, DotPat]
-translateBody (Left _ : rest) s (Bind b)   = Bind $ fmap (translateBody rest s) b
-translateBody []              s (Body t)   = Body $ applySubst s t
-translateBody _               _ _          = __IMPOSSIBLE__
-
-------------------------------------------------------------------------
--- Helper functions
-
--- | @dropBinds n b@ drops the initial @n@ occurrences of 'Bind' from @b@.
---
--- Precondition: @b@ has to start with @n@ occurrences of 'Bind'.
-
-dropBinds :: Nat -> ClauseBody -> ClauseBody
-dropBinds n b          | n == 0 = b
-dropBinds n (Bind b)   | n > 0  = dropBinds (pred n) (absBody b)
-dropBinds _ _                   = __IMPOSSIBLE__
