@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP           #-}
-{-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE CPP                      #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Records where
 
@@ -11,6 +10,7 @@ import Data.Function
 import Data.List
 import Data.Maybe
 import qualified Data.Set as Set
+import Data.Traversable (traverse)
 
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete.Name as C
@@ -83,7 +83,7 @@ insertMissingFields r placeholder fs axs = do
       givenFields = [ (x, Just $ arg x e) | FieldAssignment x e <- fs ]
   -- Compute a list of p[aceholders for the missing visible fields.
   let missingExplicits =
-       [ (x, Just $ unnamed . placeholder <$> a)
+       [ (x, Just $ setOrigin Inserted $ unnamed . placeholder <$> a)
        | a <- filter notHidden axs
        , let x = unArg a
        , x `notElem` map (view nameFieldA) fs
@@ -109,8 +109,9 @@ getRecordDef r = maybe err return =<< isRecord r
 
 -- | Get the record name belonging to a field name.
 getRecordOfField :: QName -> TCM (Maybe QName)
-getRecordOfField d = maybe Nothing fromP <$> isProjection d
-  where fromP Projection{ projProper = mp, projFromType = r} = mp $> r
+getRecordOfField d = caseMaybeM (isProjection d) (return Nothing) $
+  \ Projection{ projProper = proper, projFromType = r} ->
+    return $ if proper then Just (unArg r) else Nothing
 
 -- | Get the field names of a record.
 getRecordFieldNames :: QName -> TCM [Arg C.Name]
@@ -122,13 +123,18 @@ recordFieldNames = map (fmap (nameConcrete . qnameName)) . recFields
 -- | Find all records with at least the given fields.
 findPossibleRecords :: [C.Name] -> TCM [QName]
 findPossibleRecords fields = do
-  defs <- (HMap.union `on` (^. sigDefinitions)) <$> getSignature <*> getImportedSignature
-  let possible def = case theDef def of
-        Record{ recFields = fs } -> Set.isSubsetOf given inrecord
-          where inrecord = Set.fromList $ map (nameConcrete . qnameName . unArg) fs
-        _ -> False
-  return [ defName d | d <- HMap.elems defs, possible d ]
+  defs  <- HMap.elems <$> use (stSignature . sigDefinitions)
+  idefs <- HMap.elems <$> use (stImports   . sigDefinitions)
+  return $ cands defs ++ cands idefs
   where
+    cands defs = [ defName d | d <- defs, possible d ]
+    possible def =
+      -- Check whether the given fields are contained
+      -- in the fields of record @def@ (if it is a record).
+      case theDef def of
+        Record{ recFields = fs } -> Set.isSubsetOf given $
+          Set.fromList $ map (nameConcrete . qnameName . unArg) fs
+        _ -> False
     given = Set.fromList fields
 
 -- | Get the field types of a record.
@@ -146,17 +152,6 @@ getRecordTypeFields t =
         _ -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
--- | Get the original name of the projection
---   (the current one could be from a module application).
-getOriginalProjection :: QName -> TCM QName
-getOriginalProjection q = do
-  proj <- fromMaybe __IMPOSSIBLE__ <$> isProjection q
-  return $ fromMaybe __IMPOSSIBLE__ $ projProper proj
-
--- | Get the type of the record constructor.
-getRecordConstructorType :: QName -> TCM Type
-getRecordConstructorType r = recConType <$> getRecordDef r
-
 -- | Returns the given record type's constructor name (with an empty
 -- range).
 getRecordConstructor :: QName -> TCM ConHead
@@ -192,6 +187,19 @@ tryRecordType t = ifBlockedType t (\ _ _ -> return $ Left Nothing) $ \ t -> do
       caseMaybeM (isRecord r) no $ \ def -> return $ Right (r,vs,def)
     _ -> no
 
+-- | Get the original projection info for name.
+{-# SPECIALIZE origProjection :: QName -> TCM (QName, Definition, Maybe Projection) #-}
+origProjection ::  HasConstInfo m => QName -> m (QName, Definition, Maybe Projection)
+origProjection f = do
+  def <- getConstInfo f
+  let proj     = isProjection_ $ theDef def
+      fallback = return (f, def, proj)
+  caseMaybe proj fallback $
+    \ p@Projection{ projProper = proper, projOrig = f' } ->
+      if not proper || f == f' then fallback else do
+        def <- getConstInfo f'
+        return (f', def, isProjection_ $ theDef def)
+
 -- | @getDefType f t@ computes the type of (possibly projection-(like))
 --   function @f@ whose first argument has type @t@.
 --   The `parameters' for @f@ are extracted from @t@.
@@ -203,15 +211,27 @@ tryRecordType t = ifBlockedType t (\ _ _ -> return $ Left Nothing) $ \ t -> do
 --   See also: 'Agda.TypeChecking.Datatypes.getConType'
 getDefType :: QName -> Type -> TCM (Maybe Type)
 getDefType f t = do
-  def <- getConstInfo f
+  -- Andreas, Issue #1973: we need to take the original projection
+  -- since the parameters from the reduced type t are correct for
+  -- the original projection only.
+  -- Due to module application, the given (non-original) projection f
+  -- may expect less parameters, those corresponding to a unreduced
+  -- version of t (which we cannot obtain here).
+  (f, def, mp) <- origProjection f
   let a = defType def
   -- if @f@ is not a projection (like) function, @a@ is the correct type
       fallback = return $ Just a
-  caseMaybe (isProjection_ $ theDef def) fallback $
+  reportSDoc "tc.deftype" 20 $ vcat
+    [ text "definition f = " <> prettyTCM f <+> text ("raw: " ++ show f)
+    , text "has type   a = " <> prettyTCM a
+    , text "principal  t = " <> prettyTCM t
+    ]
+  caseMaybe mp fallback $
     \ (Projection{ projIndex = n }) -> if n <= 0 then fallback else do
       -- otherwise, we have to instantiate @a@ to the "parameters" of @f@
       let npars | n == 0    = __IMPOSSIBLE__
                 | otherwise = n - 1
+      reportSLn "tc.deftype" 20 $ "projIndex    = " ++ show n
       -- we get the parameters from type @t@
       case ignoreSharing $ unEl t of
         Def d es -> do
@@ -223,7 +243,13 @@ getDefType f t = do
           ifNotM (eligibleForProjectionLike d) failNotElig $ {- else -} do
             -- now we know it is reduced, we can safely take the parameters
             let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take npars es
-            Just <$> a `piApplyM` pars
+            reportSDoc "tc.deftype" 20 $ vcat
+              [ text $ "head d     = " ++ show d
+              , text "parameters =" <+> sep (map prettyTCM pars)
+              ]
+            reportSLn "tc.deftype" 60 $ "parameters = " ++ show pars
+            if length pars < npars then failure "does not supply enough parameters"
+            else Just <$> a `piApplyM` pars
         _ -> failNotDef
   where
     failNotElig = failure "is not eligible for projection-likeness"
@@ -239,15 +265,17 @@ getDefType f t = do
 
 -- | The analogue of 'piApply'.  If @v@ is a value of record type @t@
 --   with field @f@, then @projectTyped v t f@ returns the type of @f v@.
+--   And also the record type (as first result).
 --
 --   Works also for projection-like definitions @f@.
+--   In this case, the first result is not a record type.
 --
 --   Precondition: @t@ is reduced.
-projectTyped :: Term -> Type -> QName -> TCM (Maybe (Term, Type))
-projectTyped v t f = caseMaybeM (getDefType f t) (return Nothing) $ \ tf -> do
-  (dom, b) <- mustBePi tf
-  u <- f `applyDef` (argFromDom dom $> v)
-  return $ Just (u, b `absApp` v)
+projectTyped :: Term -> Type -> ProjOrigin -> QName -> TCM (Maybe (Dom Type, Term, Type))
+projectTyped v t o f = caseMaybeM (getDefType f t) (return Nothing) $ \ tf -> do
+  ifNotPiType tf (const $ return Nothing) {- else -} $ \ dom b -> do
+  u <- applyDef o f (argFromDom dom $> v)
+  return $ Just (dom, u, b `absApp` v)
 
 -- | Check if a name refers to an eta expandable record.
 {-# SPECIALIZE isEtaRecord :: QName -> TCM Bool #-}
@@ -276,8 +304,8 @@ isEtaRecordType a = case ignoreSharing $ unEl a of
 
 -- | Check if a name refers to a record constructor.
 --   If yes, return record definition.
-isRecordConstructor :: MonadTCM tcm => QName -> tcm (Maybe (QName, Defn))
-isRecordConstructor c = liftTCM $ do
+isRecordConstructor :: HasConstInfo m => QName -> m (Maybe (QName, Defn))
+isRecordConstructor c = do
   def <- theDef <$> getConstInfo c
   case def of
     Constructor{ conData = r } -> fmap (r,) <$> isRecord r
@@ -298,16 +326,34 @@ isGeneratedRecordConstructor c = do
 -- | Mark record type as unguarded.
 --   No eta-expansion.  Projections do not preserve guardedness.
 unguardedRecord :: QName -> TCM ()
-unguardedRecord q = modifySignature $ updateDefinition q $ updateTheDef $ updateRecord
-  where updateRecord r@Record{} = r { recEtaEquality' = setEtaEquality (recEtaEquality' r) False, recRecursive = True }
-        updateRecord _          = __IMPOSSIBLE__
+unguardedRecord q = modifySignature $ updateDefinition q $ updateTheDef $ \case
+  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) False, recRecursive = True }
+  _ -> __IMPOSSIBLE__
 
 -- | Mark record type as recursive.
 --   Projections do not preserve guardedness.
 recursiveRecord :: QName -> TCM ()
-recursiveRecord q = modifySignature $ updateDefinition q $ updateTheDef $ updateRecord
-  where updateRecord r@Record{} = r { recRecursive = True }
-        updateRecord _          = __IMPOSSIBLE__
+recursiveRecord q = do
+  ok <- etaEnabled
+  modifySignature $ updateDefinition q $ updateTheDef $ \case
+    r@Record{ recInduction = ind, recEtaEquality' = eta } ->
+      r { recRecursive = True, recEtaEquality' = eta' }
+      where
+      eta' | ok, eta == Inferred False, ind /= Just CoInductive = Inferred True
+           | otherwise = eta
+    _ -> __IMPOSSIBLE__
+
+-- | Turn on eta for non-recursive record, unless user declared otherwise.
+nonRecursiveRecord :: QName -> TCM ()
+nonRecursiveRecord q = whenM etaEnabled $ do
+  -- Do nothing if eta is disabled by option.
+  modifySignature $ updateDefinition q $ updateTheDef $ \case
+    r@Record{ recInduction = ind, recEtaEquality' = Inferred False }
+      | ind /= Just CoInductive ->
+      r { recEtaEquality' = Inferred True }
+    r@Record{} -> r
+    _          -> __IMPOSSIBLE__
+
 
 -- | Check whether record type is marked as recursive.
 --
@@ -365,14 +411,14 @@ expandRecordVar i gamma0 = do
           fs  = recFields def
       -- Construct the record pattern @Γ₁, Γ' ⊢ u := c ys@.
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = Con (recConHead def) ys
+          u   = Con (recConHead def) ConOSystem ys
       -- @Γ₁, Γ' ⊢ τ₀ : Γ₁, x:_@
           tau0 = consS u $ raiseS m
       -- @Γ₁, Γ', Γ₂ ⊢ τ₀ : Γ₁, x:_, Γ₂@
           tau  = liftS (size gamma2) tau0
 
       --  Fields are in order first-first.
-          zs  = for fs $ fmap $ \ f -> Var 0 [Proj f]
+          zs  = for fs $ fmap $ \ f -> Var 0 [Proj ProjSystem f]
       --  We need to reverse the field sequence to build the substitution.
       -- @Γ₁, x:_ ⊢ σ₀ : Γ₁, Γ'@
           sigma0 = reverse (map unArg zs) ++# raiseS 1
@@ -427,14 +473,14 @@ curryAt t n = do
           m   = size tel
           fs  = recFields def
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = Con (recConHead def) ys
+          u   = Con (recConHead def) ConOSystem ys
           b'  = raise m b `absApp` u
           t'  = gamma `telePi` (tel `telePi` b')
           gammai = map domInfo $ telToList gamma
           xs  = reverse $ zipWith (\ ai i -> Arg ai $ var i) gammai [m..]
           curry v = teleLam gamma $ teleLam tel $
                       raise (n+m) v `apply` (xs ++ [Arg ai u])
-          zs  = for fs $ fmap $ \ f -> Var 0 [Proj f]
+          zs  = for fs $ fmap $ \ f -> Var 0 [Proj ProjSystem f]
           atel = sgTel $ Dom ai (absName b, a)
           uncurry v = teleLam gamma $ teleLam atel $
                         raise (n + 1) v `apply` (xs ++ zs)
@@ -444,7 +490,8 @@ curryAt t n = do
 {-| @etaExpand r pars u@ computes the eta expansion of record value @u@
     at record type @r pars@.
 
-    The first argument @r@ should be the name of a record type. Given
+    The first argument @r@ should be the name of an eta-expandable record type.
+    Given
 
       @record R : Set where field x : A; y : B; .z : C@
 
@@ -455,30 +502,48 @@ curryAt t n = do
     where @tel@ is the record telescope instantiated at the parameters @pars@.
 -}
 etaExpandRecord :: QName -> Args -> Term -> TCM (Telescope, Args)
-etaExpandRecord r pars u = do
+etaExpandRecord = etaExpandRecord' False
+
+-- | Eta expand a record regardless of whether it's an eta-record or not.
+forceEtaExpandRecord :: QName -> Args -> Term -> TCM (Telescope, Args)
+forceEtaExpandRecord = etaExpandRecord' True
+
+etaExpandRecord' :: Bool -> QName -> Args -> Term -> TCM (Telescope, Args)
+etaExpandRecord' forceEta r pars u = do
   def <- getRecordDef r
-  (tel, _, args) <- etaExpandRecord_ r pars def u
+  (tel, _, _, args) <- etaExpandRecord'_ forceEta r pars def u
   return (tel, args)
 
-etaExpandRecord_ :: QName -> Args -> Defn -> Term -> TCM (Telescope, ConHead, Args)
-etaExpandRecord_ r pars def u = do
+etaExpandRecord_ :: QName -> Args -> Defn -> Term -> TCM (Telescope, ConHead, ConInfo, Args)
+etaExpandRecord_ = etaExpandRecord'_ False
+
+etaExpandRecord'_ :: Bool -> QName -> Args -> Defn -> Term -> TCM (Telescope, ConHead, ConInfo, Args)
+etaExpandRecord'_ forceEta r pars def u = do
   let Record{ recConHead     = con
             , recFields      = xs
             , recTel         = tel
             } = def
       eta = recEtaEquality def
       tel' = apply tel pars
-  unless eta __IMPOSSIBLE__ -- make sure we do not expand non-eta records
+  unless (eta || forceEta) __IMPOSSIBLE__ -- make sure we do not expand non-eta records (unless forced to)
   case ignoreSharing u of
+
     -- Already expanded.
-    Con con_ args -> do
-      when (con /= con_) __IMPOSSIBLE__
-      return (tel', con, args)
+    Con con_ ci args -> do
+      when (con /= con_) $ do
+        reportSDoc "impossible" 10 $ vcat
+          [ text "etaExpandRecord_: the following two constructors should be identical"
+          , nest 2 $ text $ "con  = " ++ show con
+          , nest 2 $ text $ "con_ = " ++ show con_
+          ]
+        __IMPOSSIBLE__
+      return (tel', con, ci, args)
+
     -- Not yet expanded.
-    _             -> do
+    _ -> do
       -- Andreas, < 2016-01-18: Note: recFields are always the original projections,
       -- thus, we can use them in Proj directly.
-      let xs' = for xs $ fmap $ \ x -> u `applyE` [Proj x]
+      let xs' = for xs $ fmap $ \ x -> u `applyE` [Proj ProjSystem x]
       reportSDoc "tc.record.eta" 20 $ vcat
         [ text "eta expanding" <+> prettyTCM u <+> text ":" <+> prettyTCM r
         , nest 2 $ vcat
@@ -486,13 +551,13 @@ etaExpandRecord_ r pars def u = do
           , text "args =" <+> prettyTCM xs'
           ]
         ]
-      return (tel', con, xs')
+      return (tel', con, ConOSystem, xs')
 
 etaExpandAtRecordType :: Type -> Term -> TCM (Telescope, Term)
 etaExpandAtRecordType t u = do
   (r, pars, def) <- fromMaybe __IMPOSSIBLE__ <$> isRecordType t
-  (tel, con, args) <- etaExpandRecord_ r pars def u
-  return (tel, Con con args)
+  (tel, con, ci, args) <- etaExpandRecord_ r pars def u
+  return (tel, Con con ci args)
 
 -- | The fields should be eta contracted already.
 --
@@ -502,10 +567,10 @@ etaExpandAtRecordType t u = do
 --
 --   TODO: this can be moved out of TCM (but only if ConHead
 --   stores also the Arg-decoration of the record fields.
-{-# SPECIALIZE etaContractRecord :: QName -> ConHead -> Args -> TCM Term #-}
-{-# SPECIALIZE etaContractRecord :: QName -> ConHead -> Args -> ReduceM Term #-}
-etaContractRecord :: HasConstInfo m => QName -> ConHead -> Args -> m Term
-etaContractRecord r c args = do
+{-# SPECIALIZE etaContractRecord :: QName -> ConHead -> ConInfo -> Args -> TCM Term #-}
+{-# SPECIALIZE etaContractRecord :: QName -> ConHead -> ConInfo -> Args -> ReduceM Term #-}
+etaContractRecord :: HasConstInfo m => QName -> ConHead -> ConInfo -> Args -> m Term
+etaContractRecord r c ci args = do
   Just Record{ recFields = xs } <- isRecord r
   let check :: Arg Term -> Arg QName -> Maybe (Maybe Term)
       check a ax = do
@@ -516,10 +581,10 @@ etaContractRecord r c args = do
           -- if @a@ is the record field name applied to a single argument
           -- then it passes the check
           (_, Just (_, [])) -> Nothing  -- not a projection
-          (_, Just (h, es)) | Proj f <- last es, unArg ax == f
+          (_, Just (h, es)) | Proj _o f <- last es, unArg ax == f
                             -> Just $ Just $ h $ init es
           _                 -> Nothing
-      fallBack = return (Con c args)
+      fallBack = return (Con c ci args)
   case compare (length args) (length xs) of
     LT -> fallBack       -- Not fully applied
     GT -> __IMPOSSIBLE__ -- Too many arguments. Impossible.
@@ -551,7 +616,7 @@ isSingletonRecord' :: Bool -> QName -> Args -> TCM (Either MetaId (Maybe Term))
 isSingletonRecord' regardIrrelevance r ps = do
   reportSLn "tc.meta.eta" 30 $ "Is " ++ show r ++ " a singleton record type?"
   def <- getRecordDef r
-  emap (Con $ recConHead def) <$> check (recTel def `apply` ps)
+  emap (Con (recConHead def) ConOSystem) <$> check (recTel def `apply` ps)
   where
   check :: Telescope -> TCM (Either MetaId (Maybe [Arg Term]))
   check tel = do
@@ -597,3 +662,27 @@ isSingletonType' regardIrrelevance t = do
 -- | Auxiliary function.
 emap :: (a -> b) -> Either c (Maybe a) -> Either c (Maybe b)
 emap = mapRight . fmap
+
+class NormaliseProjP a where
+  normaliseProjP :: HasConstInfo m => a -> m a
+
+instance NormaliseProjP Clause where
+  normaliseProjP cl = do
+    ps <- normaliseProjP $ namedClausePats cl
+    return $ cl { namedClausePats = ps }
+
+instance NormaliseProjP a => NormaliseProjP [a] where
+  normaliseProjP = traverse normaliseProjP
+
+instance NormaliseProjP a => NormaliseProjP (Arg a) where
+  normaliseProjP = traverse normaliseProjP
+
+instance NormaliseProjP a => NormaliseProjP (Named_ a) where
+  normaliseProjP = traverse normaliseProjP
+
+instance NormaliseProjP (Pattern' x) where
+  normaliseProjP p@VarP{}        = return p
+  normaliseProjP p@DotP{}        = return p
+  normaliseProjP (ConP c cpi ps) = ConP c cpi <$> normaliseProjP ps
+  normaliseProjP p@LitP{}        = return p
+  normaliseProjP (ProjP o d0)    = ProjP o <$> getOriginalProjection d0
