@@ -1,12 +1,6 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE UndecidableInstances #-}
-
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
+{-# LANGUAGE CPP                      #-}
+{-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE UndecidableInstances     #-}
 
 module Agda.TypeChecking.Reduce where
 
@@ -20,6 +14,8 @@ import Data.Map (Map)
 import Data.Traversable
 import Data.Hashable
 
+import Agda.Interaction.Options
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -42,11 +38,13 @@ import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Match
 import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
 import {-# SOURCE #-} Agda.TypeChecking.Pretty
 import {-# SOURCE #-} Agda.TypeChecking.Rewriting
+import {-# SOURCE #-} Agda.TypeChecking.Reduce.Fast
 
 import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Monad
 import Agda.Utils.HashMap (HashMap)
+import Agda.Utils.Size
 import Agda.Utils.Tuple
 
 #include "undefined.h"
@@ -70,6 +68,14 @@ normalise = runReduceM . normalise'
 simplify :: Simplify a => a -> TCM a
 simplify = runReduceM . simplify'
 
+-- | Meaning no metas left in the instantiation.
+isFullyInstantiatedMeta :: MetaId -> TCM Bool
+isFullyInstantiatedMeta m = do
+  mv <- TCM.lookupMeta m
+  case mvInstantiation mv of
+    InstV sub v -> null . allMetas <$> instantiateFull (sub, v)
+    _ -> return False
+
 -- | Instantiate something.
 --   Results in an open meta variable or a non meta.
 --   Doesn't do any reduction, and preserves blocking tags (when blocking meta
@@ -98,7 +104,6 @@ instance Instantiate Term where
       OpenIFS                          -> return t
       BlockedConst _                   -> return t
       PostponedTypeCheckingProblem _ _ -> return t
-      InstS _                          -> __IMPOSSIBLE__
   instantiate' (Level l) = levelTm <$> instantiate' l
   instantiate' (Sort s) = sortTm <$> instantiate' s
   instantiate' v@Shared{} =
@@ -132,7 +137,6 @@ instance Instantiate a => Instantiate (Blocked a) where
       OpenIFS                        -> return v
       BlockedConst{}                 -> return v
       PostponedTypeCheckingProblem{} -> return v
-      InstS{}                        -> __IMPOSSIBLE__
 
 instance Instantiate Type where
     instantiate' (El s t) = El <$> instantiate' s <*> instantiate' t
@@ -144,7 +148,7 @@ instance Instantiate Sort where
 
 instance Instantiate Elim where
   instantiate' (Apply v) = Apply <$> instantiate' v
-  instantiate' (Proj f)  = pure $ Proj f
+  instantiate' (Proj o f)= pure $ Proj o f
 
 instance Instantiate t => Instantiate (Abs t) where
   instantiate' = traverse instantiate'
@@ -195,7 +199,7 @@ instance Instantiate e => Instantiate (Map k e) where
     instantiate' = traverse instantiate'
 
 instance Instantiate Candidate where
-  instantiate' (Candidate u t eti) = Candidate <$> instantiate' u <*> instantiate' t <*> return eti
+  instantiate' (Candidate u t eti ov) = Candidate <$> instantiate' u <*> instantiate' t <*> pure eti <*> pure ov
 
 instance Instantiate EqualityView where
   instantiate' (OtherType t)            = OtherType
@@ -257,7 +261,7 @@ instance Reduce Sort where
 
 instance Reduce Elim where
   reduce' (Apply v) = Apply <$> reduce' v
-  reduce' (Proj f)  = pure $ Proj f
+  reduce' (Proj o f)= pure $ Proj o f
 
 instance Reduce Level where
   reduce'  (Max as) = levelMax <$> mapM reduce' as
@@ -311,7 +315,23 @@ instance (Reduce a, Reduce b,Reduce c) => Reduce (a,b,c) where
     reduce' (x,y,z) = (,,) <$> reduce' x <*> reduce' y <*> reduce' z
 
 instance Reduce Term where
-  reduceB' = {-# SCC "reduce'<Term>" #-} rewriteAfter $ \ v -> do
+  reduceB' = {-# SCC "reduce'<Term>" #-} maybeFastReduceTerm
+
+maybeFastReduceTerm :: Term -> ReduceM (Blocked Term)
+maybeFastReduceTerm v = do
+  let tryFast = case v of
+                  Def{} -> True
+                  Con{} -> True
+                  _     -> False
+  if not tryFast then slowReduceTerm v
+                 else do
+    s <- optSharing   <$> commandLineOptions
+    allowed <- asks envAllowedReductions
+    let notAll = delete NonTerminatingReductions allowed /= allReductions
+    if s || notAll then slowReduceTerm v else fastReduce (elem NonTerminatingReductions allowed) v
+
+slowReduceTerm :: Term -> ReduceM (Blocked Term)
+slowReduceTerm v = do
     v <- instantiate' v
     let done = return $ notBlocked v
     case v of
@@ -320,10 +340,10 @@ instance Reduce Term where
 --      MetaV x args -> notBlocked . MetaV x <$> reduce' args
       MetaV x es -> done
       Def f es   -> unfoldDefinitionE False reduceB' (Def f []) f es
-      Con c args -> do
+      Con c ci args -> do
           -- Constructors can reduce' when they come from an
           -- instantiated module.
-          v <- unfoldDefinition False reduceB' (Con c []) (conName c) args
+          v <- unfoldDefinition False reduceB' (Con c ci []) (conName c) args
           traverse reduceNat v
       Sort s   -> fmap sortTm <$> reduceB' s
       Level l  -> ifM (elem LevelReductions <$> asks envAllowedReductions)
@@ -338,64 +358,71 @@ instance Reduce Term where
     where
       -- NOTE: reduceNat can traverse the entire term.
       reduceNat v@Shared{} = updateSharedTerm reduceNat v
-      reduceNat v@(Con c []) = do
+      reduceNat v@(Con c ci []) = do
         mz  <- getBuiltin' builtinZero
         case v of
           _ | Just v == mz  -> return $ Lit $ LitNat (getRange c) 0
           _                 -> return v
-      reduceNat v@(Con c [a]) | notHidden a && isRelevant a = do
+      reduceNat v@(Con c ci [a]) | notHidden a && isRelevant a = do
         ms  <- fmap ignoreSharing <$> getBuiltin' builtinSuc
         case v of
-          _ | Just (Con c []) == ms -> inc <$> reduce' (unArg a)
+          _ | Just (Con c ci []) == ms -> inc <$> reduce' (unArg a)
           _                         -> return v
           where
             inc w = case ignoreSharing w of
               Lit (LitNat r n) -> Lit (LitNat (fuseRange c r) $ n + 1)
-              _                -> Con c [defaultArg w]
+              _                -> Con c ci [defaultArg w]
       reduceNat v = return v
 
-rewriteAfter :: (Term -> ReduceM (Blocked Term)) -> Term -> ReduceM (Blocked Term)
-rewriteAfter f = trampolineM $ rewrite <=< f
-
 -- Andreas, 2013-03-20 recursive invokations of unfoldCorecursion
 -- need also to instantiate metas, see Issue 826.
 unfoldCorecursionE :: Elim -> ReduceM (Blocked Elim)
-unfoldCorecursionE e@(Proj f)           = return $ notBlocked e
+unfoldCorecursionE (Proj o p)           = notBlocked . Proj o <$> getOriginalProjection p
 unfoldCorecursionE (Apply (Arg info v)) = fmap (Apply . Arg info) <$>
   unfoldCorecursion v
 
 unfoldCorecursion :: Term -> ReduceM (Blocked Term)
-unfoldCorecursion = rewriteAfter $ \ v -> do
+unfoldCorecursion v = do
   v <- instantiate' v
   case compressPointerChain v of
     Def f es -> unfoldDefinitionE True unfoldCorecursion (Def f []) f es
     v@(Shared p) ->
       case derefPtr p of
         Def{} -> updateSharedFM unfoldCorecursion v
-        _     -> reduceB' v
-    _ -> reduceB' v
+        _     -> slowReduceTerm v
+    _ -> slowReduceTerm v
 
 -- | If the first argument is 'True', then a single delayed clause may
 -- be unfolded.
 unfoldDefinition ::
   Bool -> (Term -> ReduceM (Blocked Term)) ->
   Term -> QName -> Args -> ReduceM (Blocked Term)
-unfoldDefinition b keepGoing v f args = snd <$> do
-  unfoldDefinition' b (\ t -> (NoSimplification,) <$> keepGoing t) v f $
-    map Apply args
+unfoldDefinition unfoldDelayed keepGoing v f args =
+  unfoldDefinitionE unfoldDelayed keepGoing v f (map Apply args)
 
 unfoldDefinitionE ::
   Bool -> (Term -> ReduceM (Blocked Term)) ->
   Term -> QName -> Elims -> ReduceM (Blocked Term)
-unfoldDefinitionE b keepGoing v f es = snd <$>
-  unfoldDefinition' b (\ t -> (NoSimplification,) <$> keepGoing t) v f es
+unfoldDefinitionE unfoldDelayed keepGoing v f es = do
+  r <- unfoldDefinitionStep unfoldDelayed v f es
+  case r of
+    NoReduction v    -> return v
+    YesReduction _ v -> keepGoing v
 
 unfoldDefinition' ::
-  Bool -> (Term -> ReduceM (Simplification, Blocked Term)) ->
+  Bool -> (Simplification -> Term -> ReduceM (Simplification, Blocked Term)) ->
   Term -> QName -> Elims -> ReduceM (Simplification, Blocked Term)
-unfoldDefinition' unfoldDelayed keepGoing v0 f es =
+unfoldDefinition' unfoldDelayed keepGoing v0 f es = do
+  r <- unfoldDefinitionStep unfoldDelayed v0 f es
+  case r of
+    NoReduction v       -> return (NoSimplification, v)
+    YesReduction simp v -> keepGoing simp v
+
+unfoldDefinitionStep :: Bool -> Term -> QName -> Elims -> ReduceM (Reduced (Blocked Term) Term)
+unfoldDefinitionStep unfoldDelayed v0 f es =
   {-# SCC "reduceDef" #-} do
   info <- getConstInfo f
+  rewr <- getRewriteRulesFor f
   allowed <- asks envAllowedReductions
   let def = theDef info
       v   = v0 `applyE` es
@@ -412,30 +439,29 @@ unfoldDefinition' unfoldDelayed keepGoing v0 f es =
           _                             -> False
   case def of
     Constructor{conSrcCon = c} ->
-      retSimpl $ notBlocked $ Con (c `withRangeOf` f) [] `applyE` es
+      noReduction $ notBlocked $ Con (c `withRangeOf` f) ConOSystem [] `applyE` es
     Primitive{primAbstr = ConcreteDef, primName = x, primClauses = cls} -> do
       pf <- fromMaybe __IMPOSSIBLE__ <$> getPrimitive' x
       if FunctionReductions `elem` allowed
         then reducePrimitive x v0 f es pf dontUnfold
-                             cls (defCompiled info)
-        else retSimpl $ notBlocked v
+                             cls (defCompiled info) rewr
+        else noReduction $ notBlocked v
     _  -> do
       if FunctionReductions `elem` allowed ||
          (isJust (isProjection_ def) && ProjectionReductions `elem` allowed) || -- includes projection-like
          (isInlineFun def && InlineReductions `elem` allowed) ||
          (copatterns && CopatternReductions `elem` allowed)
         then
-          reduceNormalE keepGoing v0 f (map notReduced es)
-                       dontUnfold
-                       (defClauses info) (defCompiled info)
-        else retSimpl $ notBlocked v  -- Andrea(s), 2014-12-05 OK?
+          reduceNormalE v0 f (map notReduced es) dontUnfold
+                       (defClauses info) (defCompiled info) rewr
+        else noReduction $ notBlocked v  -- Andrea(s), 2014-12-05 OK?
 
   where
-    retSimpl v = (,v) <$> getSimplification
-
-    reducePrimitive x v0 f es pf dontUnfold cls mcc
+    noReduction    = return . NoReduction
+    yesReduction s = return . YesReduction s
+    reducePrimitive x v0 f es pf dontUnfold cls mcc rewr
       | genericLength es < ar
-                  = retSimpl $ NotBlocked Underapplied $ v0 `applyE` es -- not fully applied
+                  = noReduction $ NotBlocked Underapplied $ v0 `applyE` es -- not fully applied
       | otherwise = {-# SCC "reducePrimitive" #-} do
           let (es1,es2) = genericSplitAt ar es
               args1     = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es1
@@ -444,63 +470,46 @@ unfoldDefinition' unfoldDelayed keepGoing v0 f es =
             NoReduction args1' -> do
               let es1' = map (fmap Apply) args1'
               if null cls then do
-                retSimpl $ applyE (Def f []) <$> do
+                noReduction $ applyE (Def f []) <$> do
                   traverse id $
                     map mredToBlocked es1' ++ map notBlocked es2
                else
-                reduceNormalE keepGoing v0 f
-                             (es1' ++ map notReduced es2)
-                             dontUnfold cls mcc
-            YesReduction simpl v -> performedSimplification' simpl $
-              keepGoing $ v `applyE` es2
+                reduceNormalE v0 f (es1' ++ map notReduced es2) dontUnfold cls mcc rewr
+            YesReduction simpl v -> yesReduction simpl $ v `applyE` es2
       where
           ar  = primFunArity pf
           mredToBlocked :: MaybeReduced a -> Blocked a
           mredToBlocked (MaybeRed NotReduced  x) = notBlocked x
           mredToBlocked (MaybeRed (Reduced b) x) = x <$ b
 
-    reduceNormalE :: (Term -> ReduceM (Simplification, Blocked Term)) -> Term -> QName -> [MaybeReduced Elim] -> Bool -> [Clause] -> Maybe CompiledClauses -> ReduceM (Simplification, Blocked Term)
-    reduceNormalE keepGoing v0 f es dontUnfold def mcc = {-# SCC "reduceNormal" #-} do
-      case def of
+    reduceNormalE :: Term -> QName -> [MaybeReduced Elim] -> Bool -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> ReduceM (Reduced (Blocked Term) Term)
+    reduceNormalE v0 f es dontUnfold def mcc rewr = {-# SCC "reduceNormal" #-} do
+      case (def,rewr) of
         _ | dontUnfold -> defaultResult -- non-terminating or delayed
-        [] -> defaultResult -- no definition for head
-        cls -> do
-            ev <- appDefE_ f v0 cls mcc es
-            case ev of
-              NoReduction v -> do
-                traceSDoc "tc.reduce'" 90 (vcat
-                  [ text "*** tried to reduce' " <+> prettyTCM f
-                  , text "    es =  " <+> sep (map (prettyTCM . ignoreReduced) es)
---                  [ text "*** tried to reduce' " <+> prettyTCM vfull
-                  , text "    stuck on" <+> prettyTCM (ignoreBlocking v) ]) $ do
-                  retSimpl v
-              YesReduction simpl v -> performedSimplification' simpl $ do
-                traceSDoc "tc.reduce'"  90 (text "*** reduced definition: " <+> prettyTCM f) $ do
-                  traceSDoc "tc.reduce'"  95 (text "    result" <+> prettyTCM v) $ do
-                    traceSDoc "tc.reduce'" 100 (text "    raw   " <+> text (show v)) $ do
-                      keepGoing v
-      where defaultResult = retSimpl $ NotBlocked AbsurdMatch vfull
+        ([],[])        -> defaultResult -- no definition for head
+        (cls,rewr)     -> appDefE_ f v0 cls mcc rewr es
+      where defaultResult = noReduction $ NotBlocked AbsurdMatch vfull
             vfull         = v0 `applyE` map ignoreReduced es
 
 -- | Reduce a non-primitive definition if it is a copy linking to another def.
-reduceDefCopy :: QName -> Args -> TCM (Reduced () Term)
-reduceDefCopy f vs = do
+reduceDefCopy :: QName -> Elims -> TCM (Reduced () Term)
+reduceDefCopy f es = do
   info <- TCM.getConstInfo f
-  if (defCopy info) then reduceDef_ info f vs else return $ NoReduction ()
+  rewr <- TCM.getRewriteRulesFor f
+  if (defCopy info) then reduceDef_ info rewr f es else return $ NoReduction ()
   where
-    reduceDef_ :: Definition -> QName -> Args -> TCM (Reduced () Term)
-    reduceDef_ info f vs = do
+    reduceDef_ :: Definition -> RewriteRules -> QName -> Elims -> TCM (Reduced () Term)
+    reduceDef_ info rewr f es = do
       let v0   = Def f []
-          args = map notReduced vs
           cls  = (defClauses info)
           mcc  = (defCompiled info)
       if (defDelayed info == Delayed) || (defNonterminating info)
        then return $ NoReduction ()
        else do
-          ev <- runReduceM $ appDef_ f v0 cls mcc args
+          ev <- runReduceM $ appDefE_ f v0 cls mcc rewr $ map notReduced es
           case ev of
             YesReduction simpl t -> return $ YesReduction simpl t
-            NoReduction args'    -> return $ NoReduction ()
+            NoReduction{}        -> return $ NoReduction ()
 
 -- | Reduce simple (single clause) definitions.
 reduceHead :: Term -> TCM (Blocked Term)
@@ -542,38 +551,37 @@ reduceHead' v = do -- ignoreAbstractMode $ do
 
 -- | Apply a definition using the compiled clauses, or fall back to
 --   ordinary clauses if no compiled clauses exist.
-appDef_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
-appDef_ f v0 cls mcc args = appDefE_ f v0 cls mcc $ map (fmap Apply) args
+appDef_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
+appDef_ f v0 cls mcc rewr args = appDefE_ f v0 cls mcc rewr $ map (fmap Apply) args
 
-appDefE_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
-appDefE_ f v0 cls mcc args =
+appDefE_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
+appDefE_ f v0 cls mcc rewr args =
   local (\ e -> e { envAppDef = Just f }) $
-  maybe (appDefE' v0 cls args)
-        (\cc -> appDefE v0 cc args) mcc
+  maybe (appDefE' v0 cls rewr args)
+        (\cc -> appDefE v0 cc rewr args) mcc
 
 
 -- | Apply a defined function to it's arguments, using the compiled clauses.
 --   The original term is the first argument applied to the third.
-appDef :: Term -> CompiledClauses -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
-appDef v cc args = appDefE v cc $ map (fmap Apply) args
+appDef :: Term -> CompiledClauses -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
+appDef v cc rewr args = appDefE v cc rewr $ map (fmap Apply) args
 
-appDefE :: Term -> CompiledClauses -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
-appDefE v cc es = do
+appDefE :: Term -> CompiledClauses -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
+appDefE v cc rewr es = do
   r <- matchCompiledE cc es
   case r of
     YesReduction simpl t -> return $ YesReduction simpl t
-    NoReduction es'      -> return $ NoReduction $ applyE v <$> es'
+    NoReduction es'      -> rewrite (void es') v rewr (ignoreBlocking es')
 
 -- | Apply a defined function to it's arguments, using the original clauses.
-appDef' :: Term -> [Clause] -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
-appDef' v cls args = appDefE' v cls $ map (fmap Apply) args
+appDef' :: Term -> [Clause] -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
+appDef' v cls rewr args = appDefE' v cls rewr $ map (fmap Apply) args
 
-appDefE' :: Term -> [Clause] -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
-appDefE' v cls es = goCls cls $ map ignoreReduced es
+appDefE' :: Term -> [Clause] -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
+appDefE' v cls rewr es = goCls cls $ map ignoreReduced es
   where
     goCls :: [Clause] -> [Elim] -> ReduceM (Reduced (Blocked Term) Term)
     goCls cl es = do
-      traceSLn "tc.reduce'" 95 ("Reduce.goCls tries reduction, #clauses = " ++ show (length cl)) $ do
       case cl of
         -- Andreas, 2013-10-26  In case of an incomplete match,
         -- we just do not reduce.  This allows adding single function
@@ -581,43 +589,28 @@ appDefE' v cls es = goCls cls $ map ignoreReduced es
         -- the remaining clauses (see Issue 907).
         -- Andrea(s), 2014-12-05:  We return 'MissingClauses' here, since this
         -- is the most conservative reason.
-        [] -> return $ NoReduction $ NotBlocked MissingClauses $ v `applyE` es
-        Clause{ namedClausePats = pats, clauseBody = body } : cls -> do
-          let n = length pats
+        [] -> rewrite (NotBlocked MissingClauses ()) v rewr es
+        cl : cls -> do
+          let pats = namedClausePats cl
+              body = clauseBody cl
+              npats = length pats
+              nvars = size $ clauseTel cl
           -- if clause is underapplied, skip to next clause
-          if length es < n then goCls cls es else do
-            let (es0, es1) = splitAt n es
-            (m, es0) <- matchCopatterns (unnumberPatVars pats) es0
+          if length es < npats then goCls cls es else do
+            let (es0, es1) = splitAt npats es
+            (m, es0) <- matchCopatterns pats es0
             es <- return $ es0 ++ es1
             case m of
               No         -> goCls cls es
-              DontKnow b -> return $ NoReduction $ b $> v `applyE` es
+              DontKnow b -> rewrite b v rewr es
               Yes simpl vs -- vs is the subst. for the variables bound in body
-                | isJust (getBodyUnraised body)  -- clause has body?
-                                -> return $ YesReduction simpl $
+                | Just w <- body -> do -- clause has body?
                     -- TODO: let matchPatterns also return the reduced forms
                     -- of the original arguments!
                     -- Andreas, 2013-05-19 isn't this done now?
-                    app vs body EmptyS `applyE` es1
-                | otherwise     -> return $ NoReduction $ NotBlocked AbsurdMatch $ v `applyE` es
-
-    -- Build an explicit substitution from arguments
-    -- and execute it using parallel substitution.
-    -- Calculating the de Bruijn indices: ;-) for the Bind case
-    --   Simply-typed version
-    --   (we are not interested in types, only in de Bruijn indices here)
-    --   Γ ⊢ σ : Δ
-    --   Γ ⊢ v : A
-    --   Γ ⊢ (σ,v) : Δ.A
-    --   Δ ⊢ λ b : A → B
-    --   Δ.A ⊢ b : B
-    app :: [Term] -> ClauseBody -> Substitution -> Term
-    app []       (Body v)           sigma = applySubst sigma v
-    app (v : vs) (Bind (Abs   _ b)) sigma = app vs b $ consS v sigma -- CBN
-    app (v : vs) (Bind (NoAbs _ b)) sigma = app vs b sigma
-    app  _        NoBody            sigma = __IMPOSSIBLE__
-    app (_ : _)  (Body _)           sigma = __IMPOSSIBLE__
-    app []       (Bind _)           sigma = __IMPOSSIBLE__
+                    let sigma = buildSubstitution __IMPOSSIBLE__ nvars vs
+                    return $ YesReduction simpl $ applySubst sigma w `applyE` es1
+                | otherwise     -> rewrite (NotBlocked AbsurdMatch ()) v rewr es
 
 instance Reduce a => Reduce (Closure a) where
     reduce' cl = do
@@ -648,7 +641,7 @@ instance Reduce e => Reduce (Map k e) where
     reduce' = traverse reduce'
 
 instance Reduce Candidate where
-  reduce' (Candidate u t eti) = Candidate <$> reduce' u <*> reduce' t <*> return eti
+  reduce' (Candidate u t eti ov) = Candidate <$> reduce' u <*> reduce' t <*> pure eti <*> pure ov
 
 instance Reduce EqualityView where
   reduce' (OtherType t)            = OtherType
@@ -675,7 +668,7 @@ instance Simplify Term where
     v <- instantiate' v
     case v of
       Def f vs   -> do
-        let keepGoing v = (,notBlocked v) <$> getSimplification  -- Andrea(s), 2014-12-05 OK?
+        let keepGoing simp v = return (simp, notBlocked v)
         (simpl, v) <- unfoldDefinition' False keepGoing (Def f []) f vs
         traceSDoc "tc.simplify'" 20 (
           text ("simplify': unfolding definition returns " ++ show simpl)
@@ -684,7 +677,7 @@ instance Simplify Term where
           YesSimplification -> simplifyBlocked' v -- Dangerous, but if @simpl@ then @v /= Def f vs@
           NoSimplification  -> Def f <$> simplify' vs
       MetaV x vs -> MetaV x  <$> simplify' vs
-      Con c vs   -> Con c    <$> simplify' vs
+      Con c ci vs-> Con c ci <$> simplify' vs
       Sort s     -> sortTm   <$> simplify' s
       Level l    -> levelTm  <$> simplify' l
       Pi a b     -> Pi       <$> simplify' a <*> simplify' b
@@ -703,7 +696,7 @@ instance Simplify Type where
 
 instance Simplify Elim where
   simplify' (Apply v) = Apply <$> simplify' v
-  simplify' (Proj f)  = pure $ Proj f
+  simplify' (Proj o f)= pure $ Proj o f
 
 instance Simplify Sort where
     simplify' s = do
@@ -804,16 +797,11 @@ instance Simplify Bool where
 --     DotP v       -> DotP <$> simplify' v
 --     ProjP _      -> return p
 
-instance Simplify ClauseBody where
-    simplify' (Body   t) = Body   <$> simplify' t
-    simplify' (Bind   b) = Bind   <$> simplify' b
-    simplify'  NoBody   = return NoBody
-
 instance Simplify DisplayForm where
   simplify' (Display n ps v) = Display n <$> simplify' ps <*> return v
 
 instance Simplify Candidate where
-  simplify' (Candidate u t eti) = Candidate <$> simplify' u <*> simplify' t <*> return eti
+  simplify' (Candidate u t eti ov) = Candidate <$> simplify' u <*> simplify' t <*> pure eti <*> pure ov
 
 instance Simplify EqualityView where
   simplify' (OtherType t)            = OtherType
@@ -847,12 +835,12 @@ instance Normalise Type where
     normalise' (El s t) = El <$> normalise' s <*> normalise' t
 
 instance Normalise Term where
-    normalise' = ignoreBlocking <.> rewriteAfter (reduceB' >=> traverse normaliseArgs)
+    normalise' = ignoreBlocking <.> (reduceB' >=> traverse normaliseArgs)
       where
         normaliseArgs :: Term -> ReduceM Term
         normaliseArgs v = case v of
                 Var n vs    -> Var n <$> normalise' vs
-                Con c vs    -> Con c <$> normalise' vs
+                Con c ci vs -> Con c ci <$> normalise' vs
                 Def f vs    -> Def f <$> normalise' vs
                 MetaV x vs  -> MetaV x <$> normalise' vs
                 Lit _       -> return v
@@ -865,7 +853,7 @@ instance Normalise Term where
 
 instance Normalise Elim where
   normalise' (Apply v) = Apply <$> normalise' v
-  normalise' (Proj f)  = pure $ Proj f
+  normalise' (Proj o f)= pure $ Proj o f
 
 instance Normalise Level where
   normalise' (Max as) = levelMax <$> normalise' as
@@ -883,11 +871,6 @@ instance Normalise LevelAtom where
       NeutralLevel r v -> NeutralLevel r <$> normalise' v
       UnreducedLevel{} -> __IMPOSSIBLE__    -- I hope
 
-instance Normalise ClauseBody where
-    normalise' (Body   t) = Body   <$> normalise' t
-    normalise' (Bind   b) = Bind   <$> normalise' b
-    normalise'  NoBody   = return NoBody
-
 instance (Subst t a, Normalise a) => Normalise (Abs a) where
     normalise' a@(Abs x _) = Abs x <$> underAbstraction_ a normalise'
     normalise' (NoAbs x v) = NoAbs x <$> normalise' v
@@ -953,13 +936,16 @@ instance Normalise Char where
 instance Normalise ConPatternInfo where
   normalise' (ConPatternInfo mr mt) = ConPatternInfo mr <$> normalise' mt
 
+instance Normalise DBPatVar where
+  normalise' = return
+
 instance Normalise a => Normalise (Pattern' a) where
   normalise' p = case p of
     VarP x       -> VarP <$> normalise' x
     LitP _       -> return p
     ConP c mt ps -> ConP c <$> normalise' mt <*> normalise' ps
     DotP v       -> DotP <$> normalise' v
-    ProjP _      -> return p
+    ProjP{}      -> return p
 
 instance Normalise DisplayForm where
   normalise' (Display n ps v) = Display n <$> normalise' ps <*> return v
@@ -971,7 +957,7 @@ instance Normalise a => Normalise (Maybe a) where
     normalise' = traverse normalise'
 
 instance Normalise Candidate where
-  normalise' (Candidate u t eti) = Candidate <$> normalise' u <*> normalise' t <*> return eti
+  normalise' (Candidate u t eti ov) = Candidate <$> normalise' u <*> normalise' t <*> pure eti <*> pure ov
 
 instance Normalise EqualityView where
   normalise' (OtherType t)            = OtherType
@@ -1019,7 +1005,7 @@ instance InstantiateFull Term where
       v <- instantiate' v
       case v of
           Var n vs    -> Var n <$> instantiateFull' vs
-          Con c vs    -> Con c <$> instantiateFull' vs
+          Con c ci vs -> Con c ci <$> instantiateFull' vs
           Def f vs    -> Def f <$> instantiateFull' vs
           MetaV x vs  -> MetaV x <$> instantiateFull' vs
           Lit _       -> return v
@@ -1071,6 +1057,9 @@ instance InstantiateFull Int where
 instance InstantiateFull ConPatternInfo where
   instantiateFull' (ConPatternInfo mr mt) = ConPatternInfo mr <$> instantiateFull' mt
 
+instance InstantiateFull DBPatVar where
+    instantiateFull' = return
+
 instance InstantiateFull a => InstantiateFull (Pattern' a) where
     instantiateFull' (VarP x)       = VarP <$> instantiateFull' x
     instantiateFull' (DotP t)       = DotP <$> instantiateFull' t
@@ -1078,11 +1067,6 @@ instance InstantiateFull a => InstantiateFull (Pattern' a) where
     instantiateFull' l@LitP{}       = return l
     instantiateFull' p@ProjP{}      = return p
 
-instance InstantiateFull ClauseBody where
-    instantiateFull' (Body   t) = Body   <$> instantiateFull' t
-    instantiateFull' (Bind   b) = Bind   <$> instantiateFull' b
-    instantiateFull'  NoBody    = return NoBody
-
 instance (Subst t a, InstantiateFull a) => InstantiateFull (Abs a) where
     instantiateFull' a@(Abs x _) = Abs x <$> underAbstraction_ a instantiateFull'
     instantiateFull' (NoAbs x a) = NoAbs x <$> instantiateFull' a
@@ -1134,7 +1118,7 @@ instance InstantiateFull Constraint where
 
 instance (InstantiateFull a) => InstantiateFull (Elim' a) where
   instantiateFull' (Apply v) = Apply <$> instantiateFull' v
-  instantiateFull' (Proj f)  = pure $ Proj f
+  instantiateFull' (Proj o f)= pure $ Proj o f
 
 instance InstantiateFull e => InstantiateFull (Map k e) where
     instantiateFull' = traverse instantiateFull'
@@ -1162,12 +1146,12 @@ instance InstantiateFull Char where
     instantiateFull' = return
 
 instance InstantiateFull Definition where
-    instantiateFull' (Defn rel x t pol occ df i c inst copy d) = do
+    instantiateFull' (Defn rel x t pol occ df i c inst copy ma d) = do
       (t, df, d) <- instantiateFull' (t, df, d)
-      return $ Defn rel x t pol occ df i c inst copy d
+      return $ Defn rel x t pol occ df i c inst copy ma d
 
 instance InstantiateFull NLPat where
-  instantiateFull' (PVar x y) = return $ PVar x y
+  instantiateFull' (PVar x y z) = return $ PVar x y z
   instantiateFull' (PWild)    = return PWild
   instantiateFull' (PDef x y) = PDef <$> instantiateFull' x <*> instantiateFull' y
   instantiateFull' (PLam x y) = PLam x <$> instantiateFull' y
@@ -1175,30 +1159,40 @@ instance InstantiateFull NLPat where
   instantiateFull' (PBoundVar x y) = PBoundVar x <$> instantiateFull' y
   instantiateFull' (PTerm x)  = PTerm <$> instantiateFull' x
 
+instance InstantiateFull NLPType where
+  instantiateFull' (NLPType l a) = NLPType
+    <$> instantiateFull' l
+    <*> instantiateFull' a
+
 instance InstantiateFull RewriteRule where
-  instantiateFull' (RewriteRule q gamma lhs rhs t) =
+  instantiateFull' (RewriteRule q gamma f ps rhs t) =
     RewriteRule q
       <$> instantiateFull' gamma
-      <*> instantiateFull' lhs
+      <*> pure f
+      <*> instantiateFull' ps
       <*> instantiateFull' rhs
       <*> instantiateFull' t
 
 instance InstantiateFull a => InstantiateFull (Open a) where
   instantiateFull' (OpenThing n a) = OpenThing n <$> instantiateFull' a
 
+instance InstantiateFull a => InstantiateFull (Local a) where
+  instantiateFull' = traverseF instantiateFull'
+
 instance InstantiateFull DisplayForm where
   instantiateFull' (Display n ps v) = uncurry (Display n) <$> instantiateFull' (ps, v)
 
 instance InstantiateFull DisplayTerm where
   instantiateFull' (DTerm v)       = DTerm <$> instantiateFull' v
   instantiateFull' (DDot  v)       = DDot  <$> instantiateFull' v
-  instantiateFull' (DCon c vs)     = DCon c <$> instantiateFull' vs
+  instantiateFull' (DCon c ci vs)  = DCon c ci <$> instantiateFull' vs
   instantiateFull' (DDef c es)     = DDef c <$> instantiateFull' es
   instantiateFull' (DWithApp v vs ws) = uncurry3 DWithApp <$> instantiateFull' (v, vs, ws)
 
 instance InstantiateFull Defn where
     instantiateFull' d = case d of
       Axiom{} -> return d
+      AbstractDefn -> return d
       Function{ funClauses = cs, funCompiled = cc, funInv = inv } -> do
         (cs, cc, inv) <- instantiateFull' (cs, cc, inv)
         return $ d { funClauses = cs, funCompiled = cc, funInv = inv }
@@ -1206,11 +1200,10 @@ instance InstantiateFull Defn where
         s  <- instantiateFull' s
         cl <- instantiateFull' cl
         return $ d { dataSort = s, dataClause = cl }
-      Record{ recConType = t, recClause = cl, recTel = tel } -> do
-        t   <- instantiateFull' t
+      Record{ recClause = cl, recTel = tel } -> do
         cl  <- instantiateFull' cl
         tel <- instantiateFull' tel
-        return $ d { recConType = t, recClause = cl, recTel = tel }
+        return $ d { recClause = cl, recTel = tel }
       Constructor{} -> return d
       Primitive{ primClauses = cs } -> do
         cs <- instantiateFull' cs
@@ -1268,7 +1261,8 @@ instance InstantiateFull a => InstantiateFull (Maybe a) where
   instantiateFull' = mapM instantiateFull'
 
 instance InstantiateFull Candidate where
-  instantiateFull' (Candidate u t eti) = Candidate <$> instantiateFull' u <*> instantiateFull' t <*> return eti
+  instantiateFull' (Candidate u t eti ov) =
+    Candidate <$> instantiateFull' u <*> instantiateFull' t <*> pure eti <*> pure ov
 
 instance InstantiateFull EqualityView where
   instantiateFull' (OtherType t)            = OtherType
