@@ -1,7 +1,4 @@
 {-# LANGUAGE CPP              #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE RankNTypes       #-}
-{-# LANGUAGE TupleSections    #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -12,7 +9,6 @@ module Agda.TypeChecking.Reduce.Monad
   , getConstInfo
   , isInstantiatedMeta
   , lookupMeta
-  , reportSDoc, reportSLn
   , traceSLn, traceSDoc
   , askR, applyWhenVerboseS
   ) where
@@ -79,10 +75,12 @@ constructorForm v = do
   return $ fromMaybe v $ constructorForm' mz ms v
 
 enterClosure :: Closure a -> (a -> ReduceM b) -> ReduceM b
-enterClosure (Closure sig env scope x) f = localR (mapRedEnvSt inEnv inState) (f x)
+enterClosure (Closure sig env scope pars x) f = localR (mapRedEnvSt inEnv inState) (f x)
   where
     inEnv   e = env { envAllowDestructiveUpdate = envAllowDestructiveUpdate e }
-    inState s = set stScope scope s   -- TODO: use the signature here? would that fix parts of issue 118?
+    inState s =
+      -- TODO: use the signature here? would that fix parts of issue 118?
+      set stScope scope $ set stModuleParameters pars s
 
 withFreshR :: HasFresh i => (i -> ReduceM a) -> ReduceM a
 withFreshR f = do
@@ -127,7 +125,6 @@ isInstantiatedMeta i = do
   mv <- lookupMeta i
   return $ case mvInstantiation mv of
     InstV{} -> True
-    InstS{} -> True
     _       -> False
 
 -- | Run a computation if a certain verbosity level is activated.
@@ -136,12 +133,6 @@ isInstantiatedMeta i = do
 verboseS :: VerboseKey -> Int -> ReduceM () -> ReduceM ()
 verboseS k n action = whenM (hasVerbosity k n) action
 
-reportSDoc :: VerboseKey -> Int -> TCM Doc -> ReduceM ()
-reportSDoc k n doc = return () -- Cannot implement this!
-
-reportSLn :: VerboseKey -> Int -> String -> ReduceM ()
-reportSLn k n s = return () -- Cannot implement this!
-
 -- | Apply a function if a certain verbosity level is activated.
 --
 --   Precondition: The level must be non-negative.
