@@ -1,8 +1,4 @@
 {-# LANGUAGE CPP                  #-}
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE PatternGuards        #-}
-{-# LANGUAGE TupleSections        #-}
-{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.TypeChecking.Rules.LHS where
 
@@ -14,6 +10,7 @@ import Control.Applicative
 import Control.Arrow (first, second, (***))
 import Control.Monad hiding (mapM, forM, sequence)
 import Control.Monad.State hiding (mapM, forM, sequence)
+import Control.Monad.Reader hiding (mapM, forM, sequence)
 import Control.Monad.Trans.Maybe
 
 import Data.Function (on)
@@ -22,6 +19,8 @@ import qualified Data.IntMap as IntMap
 import Data.List (delete, sortBy, stripPrefix)
 import Data.Monoid
 import Data.Traversable
+import Data.Map (Map)
+import qualified Data.Map as Map
 
 import Agda.Interaction.Options
 import Agda.Interaction.Options.Lenses
@@ -49,12 +48,13 @@ import Agda.TypeChecking.Patterns.Abstract
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Rewriting
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Substitute.Pattern
 import Agda.TypeChecking.Telescope
 
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Term (checkExpr)
-import Agda.TypeChecking.Rules.LHS.Problem
+import Agda.TypeChecking.Rules.LHS.AsPatterns
+import Agda.TypeChecking.Rules.LHS.Problem hiding (Substitution)
 import Agda.TypeChecking.Rules.LHS.ProblemRest
 import Agda.TypeChecking.Rules.LHS.Unify
 import Agda.TypeChecking.Rules.LHS.Split
@@ -79,7 +79,7 @@ import Agda.Utils.Impossible
 flexiblePatterns :: [NamedArg A.Pattern] -> TCM FlexibleVars
 flexiblePatterns nps = do
   forMaybeM (zip (downFrom $ length nps) nps) $ \ (i, Arg ai p) -> do
-    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) f i) <$> maybeFlexiblePattern p
+    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) f (Just i) i) <$> maybeFlexiblePattern p
 
 -- | A pattern is flexible if it is dotted or implicit, or a record pattern
 --   with only flexible subpatterns.
@@ -92,10 +92,10 @@ class IsFlexiblePattern a where
 instance IsFlexiblePattern A.Pattern where
   maybeFlexiblePattern p =
     case p of
-      A.DotP{}
-        -> return DotFlex
-      A.WildP{}
-        -> return ImplicitFlex
+      A.DotP{}  -> return DotFlex
+      A.VarP{}  -> return ImplicitFlex
+      A.WildP{} -> return ImplicitFlex
+      A.AsP _ _ p -> maybeFlexiblePattern p
       A.ConP _ (A.AmbQ [c]) qs
         -> ifM (isNothing <$> isRecordConstructor c) mzero {-else-}
              (maybeFlexiblePattern qs)
@@ -106,7 +106,7 @@ instance IsFlexiblePattern (I.Pattern' a) where
     case p of
       I.DotP{}  -> return DotFlex
       I.ConP _ i ps
-        | Just ConPImplicit <- conPRecord i -> return ImplicitFlex  -- expanded from ImplicitP
+        | Just ConOSystem <- conPRecord i -> return ImplicitFlex  -- expanded from ImplicitP
         | Just _            <- conPRecord i -> maybeFlexiblePattern ps
         | otherwise -> mzero
       I.VarP{}  -> mzero
@@ -149,11 +149,12 @@ updateInPatterns as ps qs = do
            -> TCM (IntMap (NamedArg A.Pattern), [DotPatternInst])
     update a p q = case unArg q of
       -- Case: the unifier did not instantiate the variable
-      VarP (i,_) -> return (IntMap.singleton i p, [])
+      VarP x     -> return (IntMap.singleton (dbPatVarIndex x) p, [])
       -- Case: the unifier did instantiate the variable
-      DotP u     -> case namedThing (unArg p) of
-        A.DotP _ e -> return (IntMap.empty,[DPI (Just e) u a])
-        A.WildP _  -> return (IntMap.empty,[DPI Nothing  u a])
+      DotP u     -> case snd $ asView $ namedThing (unArg p) of
+        A.DotP _ _ e -> return (IntMap.empty, [DPI Nothing  (Just e) u a])
+        A.WildP _  -> return (IntMap.empty, [DPI Nothing  Nothing  u a])
+        A.VarP x   -> return (IntMap.empty, [DPI (Just x) Nothing  u a])
         A.ConP _ (A.AmbQ [c]) qs -> do
           Def r es  <- ignoreSharing <$> reduce (unEl $ unDom a)
           let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -166,11 +167,11 @@ updateInPatterns as ps qs = do
               -- Andreas, 2012-09-19 propagate relevance info to dot patterns
               bs  = map (mapRelevance (composeRelevance (getRelevance a))) bs0
           updates bs qs (map (DotP . unArg) us `withArgsFrom` teleArgNames ftel)
-        A.VarP        _     -> __IMPOSSIBLE__
+        A.AsP         _ _ _ -> __IMPOSSIBLE__
         A.ConP        _ _ _ -> __IMPOSSIBLE__
         A.RecP        _ _   -> __IMPOSSIBLE__
+        A.ProjP       _ _ _ -> __IMPOSSIBLE__
         A.DefP        _ _ _ -> __IMPOSSIBLE__
-        A.AsP         _ _ _ -> __IMPOSSIBLE__
         A.AbsurdP     _     -> __IMPOSSIBLE__
         A.LitP        _     -> __IMPOSSIBLE__
         A.PatternSynP _ _ _ -> __IMPOSSIBLE__
@@ -182,21 +183,27 @@ updateInPatterns as ps qs = do
             fs  = killRange $ recFields def
             tel = recTel def `apply` pars
             as  = applyPatSubst (parallelS $ map (namedThing . unArg) qs) $ flattenTel tel
-            -- If the user wrote a dot pattern but the unifier eta-expanded the
-            -- corresponding variable, add the corresponding instantiation
+            -- If the user wrote a dot pattern or variable but the unifier
+            -- eta-expanded it, add the corresponding instantiation.
             dpi :: [DotPatternInst]
-            dpi = maybe [] (\e -> [DPI (Just e) (patternToTerm $ unArg q) a]) (isDotP (namedThing $ unArg p))
+            dpi = mkDPI $ patternToTerm $ unArg q
+              where
+                mkDPI v = case namedThing $ unArg p of
+                  A.DotP _ _ e -> [DPI Nothing (Just e) v a]
+                  A.VarP x     -> [DPI (Just x) Nothing v a]
+                  _            -> []
         second (dpi++) <$>
           updates as (projectInPat p fs) (map (fmap namedThing) qs)
       LitP _     -> __IMPOSSIBLE__
-      ProjP f    -> __IMPOSSIBLE__
+      ProjP{}    -> __IMPOSSIBLE__
 
     projectInPat :: NamedArg A.Pattern -> [Arg QName] -> [NamedArg A.Pattern]
     projectInPat p fs = case namedThing (unArg p) of
-      A.VarP x            -> map makeVarField fs
+      A.VarP x            -> map (makeDotField (PatRange $ getRange x)) fs
       A.ConP cpi _ nps    -> nps
       A.WildP pi          -> map (makeWildField pi) fs
-      A.DotP pi e         -> map (makeDotField pi) fs
+      A.DotP pi _ e       -> map (makeDotField pi) fs
+      A.ProjP _ _ _       -> __IMPOSSIBLE__
       A.DefP _ _ _        -> __IMPOSSIBLE__
       A.AsP _ _ _         -> __IMPOSSIBLE__
       A.AbsurdP _         -> __IMPOSSIBLE__
@@ -204,10 +211,9 @@ updateInPatterns as ps qs = do
       A.PatternSynP _ _ _ -> __IMPOSSIBLE__
       A.RecP _ _          -> __IMPOSSIBLE__
       where
-        makeVarField (Arg fi f) = Arg fi $ unnamed $ A.VarP $ qnameName f
         makeWildField pi (Arg fi f) = Arg fi $ unnamed $ A.WildP pi
         makeDotField pi (Arg fi f) = Arg fi $ unnamed $
-          A.DotP pi $ A.Underscore underscoreInfo
+          A.DotP pi Inserted $ A.Underscore underscoreInfo
           where
             underscoreInfo = A.MetaInfo
               { A.metaRange          = getRange pi
@@ -217,10 +223,6 @@ updateInPatterns as ps qs = do
               }
 
 
-    isDotP :: A.Pattern -> Maybe A.Expr
-    isDotP (A.DotP _ e) = Just e
-    isDotP _            = Nothing
-
 -- | Check if a problem is solved. That is, if the patterns are all variables.
 isSolvedProblem :: Problem -> Bool
 isSolvedProblem problem = null (restPats $ problemRest problem) &&
@@ -229,7 +231,7 @@ isSolvedProblem problem = null (restPats $ problemRest problem) &&
     -- need further splitting:
     isSolved A.ConP{}        = False
     isSolved A.LitP{}        = False
-    isSolved A.DefP{}        = False  -- projection pattern
+    isSolved A.ProjP{}       = False
     isSolved A.RecP{}        = False  -- record pattern
     -- solved:
     isSolved A.VarP{}        = True
@@ -237,6 +239,7 @@ isSolvedProblem problem = null (restPats $ problemRest problem) &&
     isSolved A.DotP{}        = True
     isSolved A.AbsurdP{}     = True
     -- impossible:
+    isSolved A.DefP{}        = __IMPOSSIBLE__
     isSolved A.AsP{}         = __IMPOSSIBLE__  -- removed by asView
     isSolved A.PatternSynP{} = __IMPOSSIBLE__  -- expanded before
 
@@ -262,7 +265,8 @@ noShadowingOfConstructors mkCall problem =
   noShadowing (A.AbsurdP   {}) t = return ()
   noShadowing (A.ConP      {}) t = return ()  -- only happens for eta expanded record patterns
   noShadowing (A.RecP      {}) t = return ()  -- record pattern
-  noShadowing (A.DefP      {}) t = return ()  -- projection pattern
+  noShadowing (A.ProjP     {}) t = return ()  -- projection pattern
+  noShadowing (A.DefP      {}) t = __IMPOSSIBLE__
   noShadowing (A.DotP      {}) t = return ()
   noShadowing (A.AsP       {}) t = __IMPOSSIBLE__
   noShadowing (A.LitP      {}) t = __IMPOSSIBLE__
@@ -283,11 +287,18 @@ noShadowingOfConstructors mkCall problem =
               []      -> return ()
               (c : _) -> setCurrentRange x $
                 typeError $ PatternShadowsConstructor x c
+          AbstractDefn{} -> return ()
+            -- Abstract constructors cannot be brought into scope,
+            -- even by a bigger import list.
+            -- Thus, they cannot be confused with variables.
+            -- Alternatively, we could do getConstInfo in ignoreAbstractMode,
+            -- then Agda would complain if a variable shadowed an abstract constructor.
           Axiom       {} -> return ()
           Function    {} -> return ()
           Record      {} -> return ()
           Constructor {} -> __IMPOSSIBLE__
-          Primitive   {} -> __IMPOSSIBLE__
+          -- TODO: in the future some stuck primitives might allow constructors
+          Primitive   {} -> return ()
       Var   {} -> return ()
       Pi    {} -> return ()
       Sort  {} -> return ()
@@ -306,7 +317,7 @@ noShadowingOfConstructors mkCall problem =
 
 -- | Check that a dot pattern matches it's instantiation.
 checkDotPattern :: DotPatternInst -> TCM ()
-checkDotPattern (DPI (Just e) v (Dom info a)) =
+checkDotPattern (DPI _ (Just e) v (Dom info a)) =
   traceCall (CheckDotPattern e v) $ do
   reportSDoc "tc.lhs.dot" 15 $
     sep [ text "checking dot pattern"
@@ -324,7 +335,11 @@ checkDotPattern (DPI (Just e) v (Dom info a)) =
           ]
     -- Should be ok to do noConstraints here
     noConstraints $ equalTerm a u v
-checkDotPattern (DPI Nothing _ _) = return ()
+checkDotPattern (DPI _ Nothing _ _) = return ()
+
+-- | Temporary data structure for 'checkLeftoverPatterns'
+type Projectn  = (ProjOrigin, QName)
+type Projectns = [Projectn]
 
 -- | Checks whether the dot patterns left over after splitting can be covered
 --   by shuffling around the dots from implicit positions. Returns the updated
@@ -342,13 +357,13 @@ checkLeftoverDotPatterns ps vs as dpi = do
            traverse gatherImplicitDotVars dpi
   reportSDoc "tc.lhs.dot" 30 $ nest 2 $
     text "implicit dotted variables:" <+>
-    prettyList (map (\(i,fs) -> prettyTCM $ Var i (map Proj fs)) idv)
+    prettyList (map (\(i,fs) -> prettyTCM $ Var i (map (uncurry Proj) fs)) idv)
   checkUserDots ps vs as idv
   reportSDoc "tc.lhs.dot" 15 $ text "all leftover dot patterns ok!"
 
   where
     checkUserDots :: [NamedArg A.Pattern] -> [Int] -> [Dom Type]
-                  -> [(Int,[QName])]
+                  -> [(Int,Projectns)]
                   -> TCM ()
     checkUserDots []     []     []     idv = return ()
     checkUserDots []     (_:_)  _      idv = __IMPOSSIBLE__
@@ -360,10 +375,14 @@ checkLeftoverDotPatterns ps vs as dpi = do
       checkUserDots ps vs as idv'
 
     checkUserDot :: NamedArg A.Pattern -> Int -> Dom Type
-                 -> [(Int,[QName])]
-                 -> TCM [(Int,[QName])]
+                 -> [(Int,Projectns)]
+                 -> TCM [(Int,Projectns)]
     checkUserDot p v a idv = case namedArg p of
-      A.DotP i e   -> do
+      A.DotP i o e | o == Inserted -> return idv
+      -- Jesper, 2016-12-08 (Issue 1605): if the origin is Inserted, this
+      -- means the dot pattern was created by expanding '...', so we don't
+      -- have to complain here.
+      A.DotP i o e -> do
         reportSDoc "tc.lhs.dot" 30 $ nest 2 $
           text "checking user dot pattern: " <+> prettyA e
         caseMaybeM (undotImplicitVar (v,[],unDom a) idv)
@@ -380,44 +399,58 @@ checkLeftoverDotPatterns ps vs as dpi = do
       A.AbsurdP _  -> return idv
       A.ConP _ _ _ -> __IMPOSSIBLE__
       A.LitP _     -> __IMPOSSIBLE__
+      A.ProjP _ _ _-> __IMPOSSIBLE__
       A.DefP _ _ _ -> __IMPOSSIBLE__
       A.RecP _ _   -> __IMPOSSIBLE__
       A.AsP  _ _ _ -> __IMPOSSIBLE__
       A.PatternSynP _ _ _ -> __IMPOSSIBLE__
 
-    gatherImplicitDotVars :: DotPatternInst -> TCM [(Int,[QName])]
-    gatherImplicitDotVars (DPI (Just _) _ _) = return [] -- Not implicit
-    gatherImplicitDotVars (DPI Nothing u _)  = gatherVars u
+    gatherImplicitDotVars :: DotPatternInst -> TCM [(Int,Projectns)]
+    gatherImplicitDotVars (DPI _ (Just _) _ _) = return [] -- Not implicit
+    gatherImplicitDotVars (DPI _ Nothing u _)  = gatherVars u
       where
-        gatherVars :: Term -> TCM [(Int,[QName])]
+        gatherVars :: Term -> TCM [(Int,Projectns)]
         gatherVars u = case ignoreSharing u of
           Var i es -> return $ (i,) <$> maybeToList (allProjElims es)
-          Con c us -> ifM (isEtaCon $ conName c)
+          Con c _ us -> ifM (isEtaCon $ conName c)
                       {-then-} (concat <$> traverse (gatherVars . unArg) us)
                       {-else-} (return [])
           _        -> return []
 
-    lookupImplicitDotVar :: (Int,[QName]) -> [(Int,[QName])] -> Maybe [QName]
+    lookupImplicitDotVar :: (Int,Projectns) -> [(Int,Projectns)] -> Maybe Projectns
     lookupImplicitDotVar (i,fs) [] = Nothing
     lookupImplicitDotVar (i,fs) ((j,gs):js)
-     | i == j , Just hs <- stripPrefix fs gs = Just hs
+     -- Andreas, 2016-09-20, issue #2196
+     -- We need to ignore the ProjOrigin!
+     | i == j , Just hs <- stripPrefixBy ((==) `on` snd) fs gs = Just hs
      | otherwise = lookupImplicitDotVar (i,fs) js
 
-    undotImplicitVar :: (Int,[QName],Type) -> [(Int,[QName])]
-                     -> TCM (Maybe [(Int,[QName])])
-    undotImplicitVar (i,fs,a) idv = case lookupImplicitDotVar (i,fs) idv of
+    undotImplicitVar :: (Int,Projectns,Type) -> [(Int,Projectns)]
+                     -> TCM (Maybe [(Int,Projectns)])
+    undotImplicitVar (i,fs,a) idv = do
+     reportSDoc "tc.lhs.dot" 40 $ vcat
+       [ text "undotImplicitVar"
+       , nest 2 $ vcat
+         [ text $ "i  =  " ++ show i
+         , text   "fs = " <+> sep (map (prettyTCM . snd) fs)
+         , text   "a  = " <+> prettyTCM a
+         , text $ "raw=  "  ++ show a
+         , text $ "idv=  "  ++ show idv
+         ]
+       ]
+     case lookupImplicitDotVar (i,fs) idv of
       Nothing -> return Nothing
       Just [] -> return $ Just $ delete (i,fs) idv
       Just rs -> caseMaybeM (isEtaRecordType a) (return Nothing) $ \(d,pars) -> do
         gs <- recFields . theDef <$> getConstInfo d
-        let u = Var i (map Proj fs)
+        let u = Var i (map (uncurry Proj) fs)
         is <- forM gs $ \(Arg _ g) -> do
-                (_,b) <- fromMaybe __IMPOSSIBLE__ <$> projectTyped u a g
-                return (i,fs++[g],b)
+                (_,_,b) <- fromMaybe __IMPOSSIBLE__ <$> projectTyped u a ProjSystem g
+                return (i,fs++[(ProjSystem,g)],b)
         undotImplicitVars is idv
 
-    undotImplicitVars :: [(Int,[QName],Type)] -> [(Int,[QName])]
-                      -> TCM (Maybe [(Int,[QName])])
+    undotImplicitVars :: [(Int,Projectns,Type)] -> [(Int,Projectns)]
+                      -> TCM (Maybe [(Int,Projectns)])
     undotImplicitVars []     idv = return $ Just idv
     undotImplicitVars (i:is) idv =
       caseMaybeM (undotImplicitVar i idv)
@@ -441,14 +474,16 @@ bindLHSVars []        tel@ExtendTel{}  _   = do
 bindLHSVars (_ : _)   EmptyTel         _   = __IMPOSSIBLE__
 bindLHSVars []        EmptyTel         ret = ret
 bindLHSVars (p : ps) (ExtendTel a tel) ret = do
+  -- see test/Fail/WronHidingInLHS:
   unless (getHiding p == getHiding a) $ typeError WrongHidingInLHS
+
   case namedArg p of
     A.VarP x      -> addContext (x, a) $ bindLHSVars ps (absBody tel) ret
     A.WildP _     -> bindDummy (absName tel)
                  -- @bindDummy underscore@ does not fix issue 819, but
                  -- introduces unwanted underscores in error messages
                  -- (Andreas, 2015-05-28)
-    A.DotP _ _    -> bindDummy (absName tel)
+    A.DotP _ _ _  -> bindDummy (absName tel)
     A.AbsurdP pi  -> do
       -- Andreas, 2012-03-15: allow postponement of emptyness check
       isEmptyType (getRange pi) $ unDom a
@@ -456,13 +491,14 @@ bindLHSVars (p : ps) (ExtendTel a tel) ret = do
       bindDummy (absName tel)
     A.ConP{}        -> __IMPOSSIBLE__
     A.RecP{}        -> __IMPOSSIBLE__
+    A.ProjP{}       -> __IMPOSSIBLE__
     A.DefP{}        -> __IMPOSSIBLE__
     A.AsP{}         -> __IMPOSSIBLE__
     A.LitP{}        -> __IMPOSSIBLE__
     A.PatternSynP{} -> __IMPOSSIBLE__
     where
       bindDummy s = do
-        x <- if isUnderscore s then freshNoName_ else freshName_ ("." ++ argNameToString s)
+        x <- if isUnderscore s then freshNoName_ else unshadowName =<< freshName_ ("." ++ argNameToString s)
         addContext (x, a) $ bindLHSVars ps (absBody tel) ret
 
 -- | Bind as patterns
@@ -477,25 +513,35 @@ bindAsPatterns (AsB x v a : asb) ret = do
 
 -- | Result of checking the LHS of a clause.
 data LHSResult = LHSResult
-  { lhsVarTele      :: Telescope
+  { lhsParameters   :: Nat
+    -- ^ The number of original module parameters. These are present in the
+    -- the patterns.
+  , lhsVarTele      :: Telescope
     -- ^ Δ : The types of the pattern variables, in internal dependency order.
     -- Corresponds to 'clauseTel'.
-  , lhsPatterns     :: [NamedArg Pattern]
+  , lhsPatterns     :: [NamedArg DeBruijnPattern]
     -- ^ The patterns in internal syntax.
   , lhsBodyType     :: Arg Type
     -- ^ The type of the body. Is @bσ@ if @Γ@ is defined.
     -- 'Irrelevant' to indicate the rhs must be checked in irrelevant mode.
-  , lhsPermutation  :: Permutation
-    -- ^ The permutation from pattern vars to @Δ@.
-    -- Corresponds to 'clausePerm'.
+  , lhsPatSubst     :: Substitution
+    -- ^ Substitution version of @lhsPatterns@, only up to the first projection
+    -- pattern. @Δ |- lhsPatSubst : Γ@. Where @Γ@ is the argument telescope of
+    -- the function. This is used to update inherited dot patterns in
+    -- with-function clauses.
+  , lhsAsBindings   :: [AsBinding]
+    -- ^ As-bindings from the left-hand side. Return instead of bound since we
+    -- want them in where's and right-hand sides, but not in with-clauses
+    -- (Issue 2303).
   }
 
 instance InstantiateFull LHSResult where
-  instantiateFull' (LHSResult tel ps t perm) = LHSResult
+  instantiateFull' (LHSResult n tel ps t sub as) = LHSResult n
     <$> instantiateFull' tel
     <*> instantiateFull' ps
     <*> instantiateFull' t
-    <*> return perm
+    <*> instantiateFull' sub
+    <*> instantiateFull' as
 
 -- | Check a LHS. Main function.
 --
@@ -512,59 +558,121 @@ checkLeftHandSide
      -- ^ The patterns.
   -> Type
      -- ^ The expected type @a = Γ → b@.
+  -> Maybe Substitution
+     -- ^ Module parameter substitution from with-abstraction.
   -> (LHSResult -> TCM a)
      -- ^ Continuation.
   -> TCM a
-checkLeftHandSide c f ps a ret = Bench.billTo [Bench.Typing, Bench.CheckLHS] $ do
-  problem0 <- problemFromPats ps a
+checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.CheckLHS] $ \ ret -> do
+
+  -- To allow module parameters to be refined by matching, we're adding the
+  -- context arguments as wildcard patterns and extending the type with the
+  -- context telescope.
+  cxt <- reverse <$> getContext
+  let tel = telFromList' show cxt
+      cps = [ unnamed . A.VarP . fst <$> setOrigin Inserted (argFromDom d)
+            | d <- cxt ]
+  problem0 <- problemFromPats (cps ++ ps) (telePi tel a)
   -- Andreas, 2013-03-15 deactivating the following test allows
   -- flexible arity
   -- unless (noProblemRest problem) $ typeError $ TooManyArgumentsInLHS a
 
+  -- We need to grab all let-bindings here (while we still have the old
+  -- context). They will be rebound below once we have the new context set up.
+  -- Subtle: if we're checking a with the context will be empty so we can't use
+  -- 'getOpen'. On the other hand, if we're checking a with the let bindings
+  -- lives in the right context already so we can use 'openThing'.
+  let openLet | isNothing withSub' = getOpen
+              | otherwise          = return . openThing
+  oldLets <- asks $ Map.toList . envLetBindings
+  reportSDoc "tc.lhs.top" 70 $ vcat
+    [ text "context =" <+> inTopContext (prettyTCM tel)
+    , text "cIds    =" <+> (text . show =<< getContextId)
+    , text "oldLets =" <+> text (show oldLets) ]
+  oldLets <- sequence [ (x,) <$> openLet b | (x, b) <- oldLets ]
+
   -- doing the splits:
-  LHSState problem@(Problem ps qs delta rest) sigma dpi asb
-    <- checkLHS f $ LHSState problem0 idS [] []
+  inTopContext $ do
+    LHSState problem@(Problem pxs qs delta rest) dpi
+      <- checkLHS f $ LHSState problem0 []
 
-  unless (null $ restPats rest) $ typeError $ TooManyArgumentsInLHS a
+    unless (null $ restPats rest) $ typeError $ TooManyArgumentsInLHS a
 
-  addCtxTel delta $ do
-    noShadowingOfConstructors c problem
-    noPatternMatchingOnCodata qs
+    addContext delta $ do
+      noShadowingOfConstructors c problem
+      noPatternMatchingOnCodata qs
 
-  reportSDoc "tc.lhs.top" 10 $
-    vcat [ text "checked lhs:"
-         , nest 2 $ vcat
-           [ text "ps    = " <+> fsep (map prettyA ps)
-           , text "delta = " <+> prettyTCM delta
-           , text "dpi   = " <+> addCtxTel delta (brackets $ fsep $ punctuate comma $ map prettyTCM dpi)
-           , text "asb   = " <+> addCtxTel delta (brackets $ fsep $ punctuate comma $ map prettyTCM asb)
-           , text "qs    = " <+> text (show qs)
+    -- f is Nothing when checking let pattern-bindings. In that case there can
+    -- be no copatterns, so we don't need to worry about self.
+    let self = Def (fromMaybe __IMPOSSIBLE__ f) []
+    asb <- addContext delta $ recoverAsPatterns delta (telePi tel a) self (cps ++ ps) qs
+
+    reportSDoc "tc.lhs.top" 10 $
+      vcat [ text "checked lhs:"
+           , nest 2 $ vcat
+             [ text "pxs   = " <+> fsep (map prettyA pxs)
+             , text "delta = " <+> prettyTCM delta
+             , text "dpi   = " <+> addContext delta (brackets $ fsep $ punctuate comma $ map prettyTCM dpi)
+             , text "asb   = " <+> addContext delta (brackets $ fsep $ punctuate comma $ map prettyTCM asb)
+             , text "qs    = " <+> prettyList (map pretty qs)
+             ]
            ]
-         ]
 
-  let b' = restType rest
-  bindLHSVars (filter (isNothing . isProjP) ps) delta $ bindAsPatterns asb $ do
-    reportSDoc "tc.lhs.top" 10 $ text "bound pattern variables"
-    reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "context = " <+> ((text . show) =<< getContext)
-    reportSDoc "tc.lhs.top" 10 $ nest 2 $ text "type  = " <+> prettyTCM b'
-    reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "type  = " <+> text (show b')
+    let b' = restType rest
+    bindLHSVars (filter (isNothing . isProjP) pxs) delta $ do
+      let -- Find the variable patterns that have been refined
+          refinedParams = [ AsB x v (unDom a) | DPI (Just x) _ v a <- dpi ]
+          asb'          = refinedParams ++ asb
 
-    -- Check dot patterns
-    mapM_ checkDotPattern dpi
-    checkLeftoverDotPatterns ps (downFrom $ size delta) (flattenTel delta) dpi
+      reportSDoc "tc.lhs.top" 10 $ text "asb' = " <+> (brackets $ fsep $ punctuate comma $ map prettyTCM asb')
 
-    let qs'  = unnumberPatVars qs
-        perm = dbPatPerm qs
-        lhsResult = LHSResult delta qs' b' perm
-    reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "perm  = " <+> text (show perm)
-    applyRelevanceToContext (getRelevance b') $ ret lhsResult
+      reportSDoc "tc.lhs.top" 10 $ text "bound pattern variables"
+      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "context = " <+> ((text . show) =<< getContext)
+      reportSDoc "tc.lhs.top" 10 $ nest 2 $ text "type  = " <+> prettyTCM b'
+      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "type  = " <+> text (show b')
+
+      let notProj ProjP{} = False
+          notProj _       = True
+                      -- Note: This works because we can't change the number of
+                      --       arguments in the lhs of a with-function relative to
+                      --       the parent function.
+          numPats   = length $ takeWhile (notProj . namedArg) qs
+          -- In the case of a non-with function the pattern substitution
+          -- should be weakened by the number of non-parameter patterns to
+          -- get the paramSub.
+          withSub = fromMaybe (wkS (numPats - length cxt) idS) withSub'
+          -- At this point we need to update the module parameters for all
+          -- parent modules.
+          patSub   = (map (patternToTerm . namedArg) $ reverse $ take numPats qs) ++# EmptyS
+          paramSub = composeS patSub withSub
+          lhsResult = LHSResult (length cxt) delta qs b' patSub asb'
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "patSub   = " <+> text (show patSub)
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "withSub  = " <+> text (show withSub)
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "paramSub = " <+> text (show paramSub)
+
+      let newLets = [ AsB x (applySubst paramSub v) (applySubst paramSub $ unDom a) | (x, (v, a)) <- oldLets ]
+      reportSDoc "tc.lhs.top" 50 $ text "old let-bindings:" <+> text (show oldLets)
+      reportSDoc "tc.lhs.top" 50 $ text "new let-bindings:" <+> (brackets $ fsep $ punctuate comma $ map prettyTCM newLets)
+
+      bindAsPatterns newLets $
+        applyRelevanceToContext (getRelevance b') $ updateModuleParameters paramSub $ do
+        bindAsPatterns asb' $ do
+
+          rebindLocalRewriteRules
+
+          -- Check dot patterns
+          mapM_ checkDotPattern dpi
+          checkLeftoverDotPatterns pxs (downFrom $ size delta) (flattenTel delta) dpi
+
+        -- Issue2303: don't bind asb' for the continuation (return in lhsResult instead)
+        ret lhsResult
 
 -- | The loop (tail-recursive): split at a variable in the problem until problem is solved
 checkLHS
   :: Maybe QName       -- ^ The name of the definition we are checking.
   -> LHSState          -- ^ The current state.
   -> TCM LHSState      -- ^ The final state after all splitting is completed
-checkLHS f st@(LHSState problem sigma dpi asb) = do
+checkLHS f st@(LHSState problem dpi) = do
 
   problem <- insertImplicitProblem problem
   -- Note: inserting implicits no longer preserve solvedness,
@@ -583,17 +691,17 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
 
     -- Split problem rest (projection pattern, does not fail as there is no call to unifier)
 
-    trySplit (SplitRest projPat projType) _ = do
+    trySplit (SplitRest projPat o projType) _ = do
 
       -- Compute the new problem
       let Problem ps1 ip delta (ProblemRest (p:ps2) b) = problem
           -- ps'      = ps1 ++ [p]
           ps'      = ps1 -- drop the projection pattern (already splitted)
           rest     = ProblemRest ps2 (projPat $> projType)
-          ip'      = ip ++ [fmap (Named Nothing . ProjP) projPat]
+          ip'      = ip ++ [fmap (Named Nothing . ProjP o) projPat]
           problem' = Problem ps' ip' delta rest
       -- Jump the trampolin
-      st' <- updateProblemRest (LHSState problem' sigma dpi asb)
+      st' <- updateProblemRest (LHSState problem' dpi)
       -- If the field is irrelevant, we need to continue in irr. cxt.
       -- (see Issue 939).
       applyRelevanceToContext (getRelevance projPat) $ do
@@ -601,9 +709,9 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
 
     -- Split on literal pattern (does not fail as there is no call to unifier)
 
-    trySplit (Split p0 xs (Arg _ (LitFocus lit ip a)) p1) _ = do
+    trySplit (Split p0 (Arg _ (LitFocus lit ip a)) p1) _ = do
 
-      -- substitute the literal in p1 and sigma and dpi and asb
+      -- substitute the literal in p1 and dpi
       let delta1 = problemTel p0
           delta2 = absApp (fmap problemTel p1) (Lit lit)
           rho    = singletonS (size delta2) (LitP lit)
@@ -612,9 +720,7 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
           -- rho    = [ var i | i <- [0..size delta2 - 1] ]
           --       ++ [ raise (size delta2) $ Lit lit ]
           --       ++ [ var i | i <- [size delta2 ..] ]
-          sigma'   = applySubst rho sigma
           dpi'     = applyPatSubst rho dpi
-          asb0     = applyPatSubst rho asb
           ip'      = applySubst rho ip
           rest'    = applyPatSubst rho (problemRest problem)
 
@@ -622,14 +728,13 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
       let ps'      = problemInPat p0 ++ problemInPat (absBody p1)
           delta'   = abstract delta1 delta2
           problem' = Problem ps' ip' delta' rest'
-          asb'     = raise (size delta2) (map (\x -> AsB x (Lit lit) a) xs) ++ asb0
-      st' <- updateProblemRest (LHSState problem' sigma' dpi' asb')
+      st' <- updateProblemRest (LHSState problem' dpi')
       checkLHS f st'
 
     -- Split on constructor pattern (unifier might fail)
 
-    trySplit (Split p0 xs focus@(Arg info Focus{}) p1) tryNextSplit = do
-      res <- trySplitConstructor p0 xs focus p1
+    trySplit (Split p0 focus@(Arg info Focus{}) p1) tryNextSplit = do
+      res <- trySplitConstructor p0 focus p1
       case res of
         -- Success.  Continue checking LHS.
         Unifies st'    -> checkLHS f st'
@@ -651,8 +756,8 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
         NoUnify  tcerr -> return $ NoUnify  tcerr
         DontKnow tcerr -> return $ DontKnow tcerr
 
-    trySplitConstructor p0 xs (Arg info LitFocus{}) p1 = __IMPOSSIBLE__
-    trySplitConstructor p0 xs (Arg info
+    trySplitConstructor p0 (Arg info LitFocus{}) p1 = __IMPOSSIBLE__
+    trySplitConstructor p0 (Arg info
              (Focus { focusCon      = c
                     , focusPatOrigin= porigin
                     , focusConArgs  = qs
@@ -682,9 +787,9 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
           [ text "split problem"
           , nest 2 $ vcat
             [ text "delta1 = " <+> prettyTCM delta1
-            , text "typeOfSplitVar =" <+> addCtxTel delta1 (prettyTCM typeOfSplitVar)
+            , text "typeOfSplitVar =" <+> addContext delta1 (prettyTCM typeOfSplitVar)
             , text "focusOutPat =" <+> (text . show) ip
-            , text "delta2 = " <+> addCtxTel delta1 (addContext ("x",domFromArg typeOfSplitVar) (prettyTCM delta2))
+            , text "delta2 = " <+> addContext delta1 (addContext ("x",domFromArg typeOfSplitVar) (prettyTCM delta2))
             ]
           ]
 
@@ -709,9 +814,9 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
         -- This should be the same datatype as we split on
         unless (d == d') $ typeError $ ShouldBeApplicationOf ca d'
 
-        -- reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
-        --   [ text "gamma' =" <+> text (show gamma')
-        --   ]
+        reportSDoc "tc.lhs.top" 20 $ addContext delta1 $ nest 2 $ vcat
+          [ text "gamma' =" <+> prettyTCM gamma'
+          ]
 
         -- Andreas 2010-09-07  propagate relevance info to new vars
         let updRel = composeRelevance (getRelevance info)
@@ -734,7 +839,7 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
         -- Compute the constructor indices by dropping the parameters
         let us' = drop (size vs) us
 
-        reportSDoc "tc.lhs.top" 15 $ addCtxTel delta1 $
+        reportSDoc "tc.lhs.top" 15 $ addContext delta1 $
           sep [ text "preparing to unify"
               , nest 2 $ vcat
                 [ text "c      =" <+> prettyTCM c <+> text ":" <+> prettyTCM a
@@ -742,7 +847,7 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
                 , text "gamma  =" <+> prettyTCM gamma
                 , text "gamma' =" <+> prettyTCM gamma'
                 , text "vs     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM vs)
-                , text "us'    =" <+> addCtxTel gamma (brackets (fsep $ punctuate comma $ map prettyTCM us'))
+                , text "us'    =" <+> addContext gamma (brackets (fsep $ punctuate comma $ map prettyTCM us'))
                 , text "ws     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM ws)
                 ]
               ]
@@ -780,12 +885,13 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
                  (raise (size gamma) da)
                  us'
                  (raise (size gamma) ws)
-        whenUnifies res $ \ (delta1',rho0) -> do
+        whenUnifies res $ \ (delta1',rho0,es) -> do
 
           reportSDoc "tc.lhs.top" 15 $ text "unification successful"
           reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
             [ text "delta1' =" <+> prettyTCM delta1'
-            , text "rho0    =" <+> addCtxTel delta1' (prettyTCM rho0)
+            , text "rho0    =" <+> addContext delta1' (prettyTCM rho0)
+            , text "es      =" <+> addContext delta1' (prettyTCM $ (fmap . fmap . fmap) patternToTerm es)
             ]
 
           -- Andreas 2014-11-25  clear 'Forced' and 'Unused'
@@ -798,12 +904,12 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
               -- to the *new* telescope delta1'. These are needed to compute the
               -- correct types of new dot pattern instantiations.
               oldTypes = applyPatSubst rho0 $ flattenTel $ delta1 `abstract` gamma
-          (p0',newDpi) <- addCtxTel delta1' $ updateInPatterns
+          (p0',newDpi) <- addContext delta1' $ updateInPatterns
                             oldTypes
                             (problemInPat p0 ++ qs')
                             newPats
 
-          reportSDoc "tc.lhs.top" 20 $ addCtxTel delta1' $ nest 2 $ vcat
+          reportSDoc "tc.lhs.top" 20 $ addContext delta1' $ nest 2 $ vcat
             [ text "p0'     =" <+> text (show p0')
             , text "newDpi  =" <+> brackets (fsep $ punctuate comma $ map prettyTCM newDpi)
             ]
@@ -811,7 +917,7 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
           -- split substitution into part for Δ₁ and part for Γ
           let (rho1,rho2) = splitS (size gamma) rho0
 
-          reportSDoc "tc.lhs.top" 20 $ addCtxTel delta1' $ nest 2 $ vcat
+          reportSDoc "tc.lhs.top" 20 $ addContext delta1' $ nest 2 $ vcat
             [ text "rho1    =" <+> prettyTCM rho1
             , text "rho2    =" <+> prettyTCM rho2
             ]
@@ -824,19 +930,20 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
           let cpi = ConPatternInfo (isRec $> porigin) (Just storedPatternType)
 
           -- compute final context and permutation
-          let crho2   = ConP c cpi $ applySubst rho2 $ teleNamedArgs gamma
+          let crho2   = ConP c cpi $ applySubst rho2 $
+                          teleNamedArgs gamma `useOriginFrom` qs'
               rho3    = consS crho2 rho1
               delta2' = applyPatSubst rho3 delta2
               delta'  = delta1' `abstract` delta2'
               rho     = liftS (size delta2) rho3
 
-          reportSDoc "tc.lhs.top" 20 $ addCtxTel delta1' $ nest 2 $ vcat
+          reportSDoc "tc.lhs.top" 20 $ addContext delta1' $ nest 2 $ vcat
             [ text "crho2   =" <+> prettyTCM crho2
             , text "rho3    =" <+> prettyTCM rho3
             , text "delta2' =" <+> prettyTCM delta2'
             ]
 
-          reportSDoc "tc.lhs.top" 70 $ addCtxTel delta1' $ nest 2 $ vcat
+          reportSDoc "tc.lhs.top" 70 $ addContext delta1' $ nest 2 $ vcat
             [ text "crho2   =" <+> text (show crho2)
             , text "rho3    =" <+> text (show rho3)
             , text "delta2' =" <+> text (show delta2')
@@ -844,46 +951,40 @@ checkLHS f st@(LHSState problem sigma dpi asb) = do
 
           reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
             [ text "delta'  =" <+> prettyTCM delta'
-            , text "rho     =" <+> addCtxTel delta' (prettyTCM rho)
+            , text "rho     =" <+> addContext delta' (prettyTCM rho)
             ]
 
           -- compute new in patterns
           let ps'  = p0' ++ problemInPat (absBody p1)
 
-          reportSDoc "tc.lhs.top" 15 $ addCtxTel delta' $
+          reportSDoc "tc.lhs.top" 15 $ addContext delta' $
             nest 2 $ vcat
               [ text "ps'    =" <+> brackets (fsep $ punctuate comma $ map prettyA ps')
               ]
 
-          -- The final dpis and asbs are the new ones plus the old ones substituted by ρ
+          -- The final dpis are the new ones plus the old ones substituted by ρ
           let dpi' = applyPatSubst rho dpi ++ raise (size delta2') newDpi
-              asb' = applyPatSubst rho asb ++ raise (size delta2') (map (\x -> AsB x (patternToTerm crho2) ca) xs)
 
-          reportSDoc "tc.lhs.top" 15 $ addCtxTel delta' $
+          reportSDoc "tc.lhs.top" 15 $ addContext delta' $
             nest 2 $ vcat
               [ text "dpi'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi')
-              , text "asb'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM asb')
               ]
 
           -- Apply the substitution
-          let sigma'   = applySubst rho sigma
-              ip'      = applySubst rho ip
+          let ip'      = applySubst rho ip
               rest'    = applyPatSubst rho (problemRest problem)
 
-          reportSDoc "tc.lhs.top" 15 $ addCtxTel delta' $
+          reportSDoc "tc.lhs.top" 15 $ addContext delta' $
             nest 2 $ vcat
-              [ text "sigma'  =" <+> prettyTCM sigma'
-              , text "ip'     =" <+> text (show ip)
-              ]
+              [ text "ip' =" <+> text (show ip) ]
 
           -- Construct the new problem
           let problem' = Problem ps' ip' delta' rest'
 
           -- if rest type reduces,
           -- extend the split problem by previously not considered patterns
-          st'@(LHSState problem'@(Problem ps' ip' delta' rest')
-                        sigma' dpi' asb')
-            <- updateProblemRest $ LHSState problem' sigma' dpi' asb'
+          st'@(LHSState problem'@(Problem ps' ip' delta' rest') dpi')
+            <- updateProblemRest $ LHSState problem' dpi'
 
           reportSDoc "tc.lhs.top" 12 $ sep
             [ text "new problem from rest"
