@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE PatternGuards #-}
 
 module Agda.TypeChecking.Rules.LHS.Implicit where
 
@@ -73,16 +72,16 @@ expandImplicitPattern' a p
        -- if not (recNamedCon def) then return Nothing else do
        do
          -- generate one implicit pattern for each field
-         let qs = for (recFields def) $ \ f -> Arg (argInfo f) implicitP
+         let qs = for (recFields def) $ \ f -> implicitP $ argInfo f
          -- generate the pattern (c _ _ ... _)
-         let q  = A.ConP (ConPatInfo ConPImplicit patNoRange) (A.AmbQ [recCon def]) qs
+         let q  = A.ConP (ConPatInfo ConOSystem patNoRange) (A.AmbQ [recCon def]) qs
          -- equip it with the name/arginfo of the original implicit pattern
              p' = updateNamedArg (const q) p   -- WAS: ((q <$) <$> p)  -- Andreas, 2013-03-21 forbiddingly cryptic
          return $ Just p'
   | otherwise = return Nothing
 
-implicitP :: Named_ A.Pattern
-implicitP = unnamed $ A.WildP $ PatRange $ noRange
+implicitP :: ArgInfo -> NamedArg A.Pattern
+implicitP info = Arg (setOrigin Inserted info) $ unnamed $ A.WildP $ PatRange $ noRange
 
 -- | Insert implicit patterns in a list of patterns.
 --   Even if 'DontExpandLast', trailing SIZELT patterns are inserted.
@@ -106,7 +105,7 @@ insertImplicitSizeLtPatterns t = do
   let ts = reverse $ takeWhile (not . visible) $ telToList tel
   keep <- reverse <$> dropWhileM (not <.> isSizeLt . snd . unDom) ts
   -- Insert implicit patterns upto (including) the last SizeLt type.
-  return [ Arg ai implicitP | Dom ai _ <- keep ]
+  return [ implicitP ai | Dom ai _ <- keep ]
 
 -- | Insert implicit patterns in a list of patterns.
 --   Even if 'DontExpandLast', trailing SIZELT patterns are inserted.
@@ -120,7 +119,7 @@ insertImplicitPatternsT exh            ps a = do
         , nest 2 $ text "ps  = " <+> do
             brackets $ fsep $ punctuate comma $ map prettyA ps
         , nest 2 $ text "tel = " <+> prettyTCM tel
-        , nest 2 $ text "b   = " <+> addCtxTel tel (prettyTCM b)
+        , nest 2 $ text "b   = " <+> addContext tel (prettyTCM b)
         ]
   case ps of
     [] -> insImp dummy tel
@@ -148,4 +147,4 @@ insertImplicitPatternsT exh            ps a = do
       ImpInsert n    -> return $ map implicitArg n
       NoInsertNeeded -> return []
 
-    implicitArg h = setHiding h $ defaultArg implicitP
+    implicitArg h = implicitP $ setHiding h $ defaultArgInfo
