@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP                 #-}
-{-# LANGUAGE PatternGuards       #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.TypeChecking.Rules.LHS.Split
@@ -15,6 +14,7 @@ import Control.Monad.Trans.Maybe
 import Data.Maybe (fromMaybe)
 import Data.List hiding (null)
 import Data.Traversable hiding (mapM, sequence)
+import Data.Foldable (msum)
 
 import Agda.Interaction.Options
 import Agda.Interaction.Highlighting.Generate (storeDisambiguatedName)
@@ -25,7 +25,7 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Position
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern
-import Agda.Syntax.Abstract (IsProjP(..))
+import Agda.Syntax.Abstract (IsProjP(..), MaybePostfixProjP(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views (asView)
 import qualified Agda.Syntax.Info as A
@@ -48,6 +48,7 @@ import Agda.TypeChecking.Telescope
 
 import Agda.TypeChecking.Rules.LHS.Problem
 
+import Agda.Utils.Except (catchError)
 import Agda.Utils.Functor ((<.>))
 import Agda.Utils.Lens
 import Agda.Utils.List
@@ -89,57 +90,92 @@ splitProblem mf (Problem ps qs tel pr) = do
     -- Result splitting
     splitRest :: ProblemRest -> ListT TCM SplitProblem
     splitRest (ProblemRest (p : ps) b) | Just f <- mf = do
-      let failure   = typeError $ CannotEliminateWithPattern p $ unArg b
-          notProjP  = typeError $ NotAProjectionPattern p
-          notRecord = failure -- lift $ typeError $ ShouldBeRecordType $ unArg b
-          wrongHiding :: MonadTCM tcm => QName -> tcm a
-          wrongHiding d = typeError . GenericDocError =<< do
-            liftTCM $ text "Wrong hiding used for projection " <+> prettyTCM d
       lift $ reportSDoc "tc.lhs.split" 20 $ sep
         [ text "splitting problem rest"
         , nest 2 $ text "pattern         p =" <+> prettyA p
         , nest 2 $ text "eliminates type b =" <+> prettyTCM b
         ]
+      lift $ reportSDoc "tc.lhs.split" 80 $ sep
+        [ nest 2 $ text $ "pattern (raw)   p = " ++ show p
+        ]
       -- If the pattern is not a projection pattern, that's an error.
       -- Probably then there were too many arguments.
-      caseMaybe (isProjP p) failure $ \ d -> do
+      caseMaybe (maybePostfixProjP p) failure $ \ (o, AmbQ ds) -> do
         -- So it is a projection pattern (d = projection name), is it?
-        caseMaybeM (lift $ isProjection d) notProjP $ \ proj -> case proj of
+        projs <- lift $ mapMaybeM (\ d -> fmap (d,) <$> isProjection d) ds
+        when (null projs) notProjP
+        -- If the target is not a record type, that's an error.
+        -- It could be a meta, but since we cannot postpone lhs checking, we crash here.
+        caseMaybeM (lift $ isRecordType $ unArg b) notRecord $ \(r, vs, def) -> case def of
+          Record{ recFields = fs } -> do
+            lift $ reportSDoc "tc.lhs.split" 20 $ sep
+              [ text $ "we are of record type r  = " ++ show r
+              , text   "applied to parameters vs = " <+> prettyTCM vs
+              , text $ "and have fields       fs = " ++ show fs
+              ]
+            -- The record "self" is the definition f applied to the patterns
+            let es = patternsToElims qs
+            -- Note: the module parameters are already part of qs
+            let self = defaultArg $ Def f [] `applyE` es
+                ai   = getArgInfo p
+            -- Try the projection candidates
+            msum $ map (tryProj o ai self fs vs (length projs >= 2)) projs
+
+          _ -> __IMPOSSIBLE__
+      where
+      failure   = lift $ typeError $ CannotEliminateWithPattern p $ unArg b
+      notProjP  = lift $ typeError $ NotAProjectionPattern p
+      notRecord = failure -- lift $ typeError $ ShouldBeRecordType $ unArg b
+      wrongHiding :: MonadTCM tcm => QName -> tcm a
+      wrongHiding d = typeError . GenericDocError =<< do
+        liftTCM $ text "Wrong hiding used for projection " <+> prettyTCM d
+
+      tryProj :: ProjOrigin -> ArgInfo -> Arg Term -> [Arg QName] -> Args -> Bool -> (QName, Projection) -> ListT TCM SplitProblem
+      tryProj o ai self fs vs amb (d0, proj) = do
+        -- Recoverable errors are those coming from the projection.
+        -- If we have several projections (amb) we just try the next one.
+        let ambErr err = if amb then mzero else err
+        case proj of
           -- Andreas, 2015-05-06 issue 1413 projProper=Nothing is not impossible
-          Projection{projProper = Nothing} -> notProjP
-          Projection{projProper = Just d, projIndex = n, projArgInfo = ai} -> do
+          Projection{projProper = False} -> ambErr notProjP
+          Projection{projProper = True, projOrig = d, projLams = lams} -> do
+            let ai = projArgInfo proj
             -- If projIndex==0, then the projection is already applied
             -- to the record value (like in @open R r@), and then it
             -- is no longer a projection but a record field.
-            unless (n > 0) notProjP
-            unless (getHiding p == getHiding ai) $ wrongHiding d
+            when (null lams) $ ambErr notProjP
             lift $ reportSLn "tc.lhs.split" 90 "we are a projection pattern"
             -- If the target is not a record type, that's an error.
             -- It could be a meta, but since we cannot postpone lhs checking, we crash here.
-            caseMaybeM (lift $ isRecordType $ unArg b) notRecord $ \(r, vs, def) -> case def of
-              Record{ recFields = fs } -> do
-                lift $ reportSDoc "tc.lhs.split" 20 $ sep
-                  [ text $ "we are of record type r  = " ++ show r
-                  , text   "applied to parameters vs = " <+> prettyTCM vs
-                  , text $ "and have fields       fs = " ++ show fs
-                  , text $ "original proj         d  = " ++ show d
-                  ]
-                -- Get the field decoration.
-                -- If the projection pattern @d@ is not a field name, that's an error.
-                argd <- maybe failure return $ find ((d ==) . unArg) fs
-                let es = patternsToElims qs
-                -- the record "self" is the definition f applied to the patterns
-                fvs <- lift $ freeVarsToApply f
-                let self = defaultArg $ Def f (map Apply fvs) `applyE` es
-                -- get the type of projection d applied to "self"
-                dType <- lift $ defType <$> getConstInfo d  -- full type!
+            lift $ reportSDoc "tc.lhs.split" 20 $ sep
+              [ text $ "proj                  d0 = " ++ show d0
+              , text $ "original proj         d  = " ++ show d
+              ]
+            -- Get the field decoration.
+            -- If the projection pattern name @d@ is not a field name,
+            -- we have to try the next projection name.
+            -- If this was not an ambiguous projection, that's an error.
+            argd <- maybe (ambErr failure) return $ find ((d ==) . unArg) fs
+            let ai' = setRelevance (getRelevance argd) ai
+
+            -- From here, we have the correctly disambiguated projection.
+            -- Thus, we no longer catch errors.
+            unless (getHiding p == getHiding ai) $ wrongHiding d
+
+            -- For highlighting, we remember which name we disambiguated to.
+            -- This is safe here (fingers crossed) as we won't decide on a
+            -- different projection even if we backtrack and come here again.
+            lift $ storeDisambiguatedName d0
+
+            -- Get the type of projection d applied to "self"
+            dType <- lift $ defType <$> getConstInfo d  -- full type!
+            lift $ reportSDoc "tc.lhs.split" 20 $ sep
+              [ text "we are              self = " <+> prettyTCM (unArg self)
+              , text "being projected by dType = " <+> prettyTCM dType
+              ]
+            -- This should succeed, as we have the correctly disambiguated.
+            lift $ SplitRest (Arg ai' d0) o <$> dType `piApplyM` (vs ++ [self])
 
-                lift $ reportSDoc "tc.lhs.split" 20 $ sep
-                  [ text "we are              self = " <+> prettyTCM (unArg self)
-                  , text "being projected by dType = " <+> prettyTCM dType
-                  ]
-                lift $ SplitRest argd <$> dType `piApplyM` (vs ++ [self])
-              _ -> __IMPOSSIBLE__
     -- if there are no more patterns left in the problem rest, there is nothing to split:
     splitRest _ = mzero
 
@@ -168,6 +204,9 @@ splitProblem mf (Problem ps qs tel pr) = do
         , nest 2 $ text "dom =" <+> prettyTCM dom
         ]
 
+      -- Andreas, 2016-06-30, issue #2075: need test here!
+      unless (getHiding p == getHiding ai) $ typeError WrongHidingInLHS
+
       -- Possible reinvokations:
       let -- 1. Redo this argument (after meta instantiation).
           tryAgain = splitP ps0 tel0
@@ -176,16 +215,14 @@ splitProblem mf (Problem ps qs tel pr) = do
             underAbstraction dom xtel $ \ tel -> splitP ps tel
 
       p <- lift $ expandLitPattern p
-      case asView $ namedArg p of
+      case snd $ asView $ namedArg p of
 
         -- Case: projection pattern.  That's an error.
-        (_, A.DefP _ d ps) -> typeError $
-          if null ps
-          then CannotEliminateWithPattern p (telePi tel0 $ unArg $ restType pr)
-          else IllformedProjectionPattern $ namedArg p
+        A.ProjP{} -> typeError $
+          CannotEliminateWithPattern p (telePi tel0 $ unArg $ restType pr)
 
         -- Case: literal pattern.
-        (xs, p@(A.LitP lit))  -> do
+        p@(A.LitP lit)  -> do
           -- Note that, in the presence of --without-K, this branch is
           -- based on the assumption that the types of literals are
           -- not indexed.
@@ -199,14 +236,13 @@ splitProblem mf (Problem ps qs tel pr) = do
             {- then -} keepGoing $
             {- else -} return Split
               { splitLPats   = empty
-              , splitAsNames = xs
               , splitFocus   = Arg ai $ LitFocus lit qs a
               , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
               }
               `mplus` keepGoing
 
         -- Case: record pattern
-        (xs, p@(A.RecP _patInfo fs)) -> do
+        p@(A.RecP _patInfo fs) -> do
           res <- lift $ tryRecordType a
           case res of
             -- Subcase: blocked
@@ -238,15 +274,16 @@ splitProblem mf (Problem ps qs tel pr) = do
               args <- lift $ insertMissingFields d (const $ A.WildP A.patNoRange) fs axs
               (return Split
                 { splitLPats   = empty
-                , splitAsNames = xs
-                , splitFocus   = Arg ai $ Focus c ConPRec args (getRange p) qs d pars ixs a
+                , splitFocus   = Arg ai $ Focus c ConORec args (getRange p) qs d pars ixs a
                 , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
                 }) `mplus` keepGoing
 
         -- Case: constructor pattern.
-        (xs, p@(A.ConP ci (A.AmbQ cs) args)) -> do
+        p@(A.ConP ci (A.AmbQ cs) args) -> do
           let tryInstantiate a'
                 | [c] <- cs = do
+                  lift $ reportSDoc "tc.lhs.split" 30 $
+                    text "split ConP: type is blocked"
                     -- Type is blocked by a meta and constructor is unambiguous,
                     -- in this case try to instantiate the meta.
                   ok <- lift $ do
@@ -256,9 +293,13 @@ splitProblem mf (Problem ps qs tel pr) = do
                     Sort s <- ignoreSharing . unEl <$> reduce (piApply dt vs)
                     tryConversion $ equalType a' (El s $ Def d $ map Apply vs)
                   if ok then tryAgain else keepGoing
-                | otherwise = keepGoing
+                | otherwise = do
+                  lift $ reportSDoc "tc.lhs.split" 30 $
+                    text "split ConP: type is blocked and constructor is ambiguous"
+                  keepGoing
           -- ifBlockedType reduces the type
           ifBlockedType a (const tryInstantiate) $ \ a' -> do
+            lift $ reportSDoc "tc.lhs.split" 30 $ text "split ConP: type is " <+> prettyTCM a'
             case ignoreSharing $ unEl a' of
 
               -- Subcase: split type is a Def.
@@ -326,7 +367,6 @@ splitProblem mf (Problem ps qs tel pr) = do
 
                       (return Split
                         { splitLPats   = empty
-                        , splitAsNames = xs
                         , splitFocus   = Arg ai $ Focus c (A.patOrigin ci) args (getRange p) qs d pars ixs a
                         , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
                         }) `mplus` keepGoing
