@@ -2,9 +2,14 @@
 
 module Agda.TypeChecking.Rules.Record where
 
-import Control.Applicative
+import Prelude hiding (null)
+
+import Control.Applicative hiding (empty)
 import Control.Monad
 import Data.Maybe
+import qualified Data.Set as Set
+
+import Agda.Interaction.Options
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Common
@@ -23,16 +28,14 @@ import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.CompiledClause.Compile
 
-import Agda.TypeChecking.Rules.Data ( bindParameters, fitsIn )
+import Agda.TypeChecking.Rules.Data ( bindParameters, fitsIn, forceSort)
 import Agda.TypeChecking.Rules.Term ( isType_ )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 
-import Agda.Utils.Size
-import Agda.Utils.Permutation
 import Agda.Utils.Monad
-
-import Agda.Interaction.Options
-
+import Agda.Utils.Null
+import Agda.Utils.Permutation
+import Agda.Utils.Size
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -58,7 +61,7 @@ checkRecDef
   :: Info.DefInfo              -- ^ Position and other info.
   -> QName                     -- ^ Record type identifier.
   -> Maybe (Ranged Induction)  -- ^ Optional: (co)inductive declaration.
-  -> Maybe Bool
+  -> Maybe Bool                -- ^ Optional: user specified eta/no-eta
   -> Maybe QName               -- ^ Optional: constructor name.
   -> [A.LamBinding]            -- ^ Record parameters.
   -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> Set).
@@ -83,7 +86,7 @@ checkRecDef i name ind eta con ps contel fields =
       -- Check and evaluate field types.
       reportSDoc "tc.rec" 15 $ text "checking fields"
       -- WRONG: contype <- workOnTypes $ killRange <$> (instantiateFull =<< isType_ contel)
-      contype <- killRange <$> (instantiateFull =<< isType_ contel)
+      contype <- instantiateFull =<< isType_ contel
       reportSDoc "tc.rec" 20 $ vcat
         [ text "contype = " <+> prettyTCM contype ]
 
@@ -98,16 +101,16 @@ checkRecDef i name ind eta con ps contel fields =
       -- Compute correct type of constructor
 
       -- t = tel -> t0 where t0 must be a sort s
-      t0' <- normalise t0
-      s <- case ignoreSharing $ unEl t0' of
-        Sort s  -> return s
-        _       -> typeError $ ShouldBeASort t0
-      gamma <- getContextTelescope  -- the record params (incl. module params)
-      reportSDoc "tc.rec" 20 $ vcat
-        [ text "gamma = " <+> inTopContext (prettyTCM gamma) ]
+      TelV idxTel s <- telView t0
+      unless (null idxTel) $ typeError $ ShouldBeASort t0
+      s <- forceSort s
+
+      reportSDoc "tc.rec" 20 $ do
+        gamma <- getContextTelescope  -- the record params (incl. module params)
+        text "gamma = " <+> inTopContext (prettyTCM gamma)
 
       -- record type (name applied to parameters)
-      let rect = El s $ Def name $ map Apply $ teleArgs gamma
+      rect <- El s . Def name . map Apply <$> getContextArgs
 
       -- Put in @rect@ as correct target of constructor type.
       -- Andreas, 2011-05-10 use telePi_ instead of telePi to preserve
@@ -139,37 +142,56 @@ checkRecDef i name ind eta con ps contel fields =
           indCo = rangedThing <$> ind
           -- A constructor is inductive unless declared coinductive.
           conInduction = fromMaybe Inductive indCo
-          haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
+          -- Andreas, 2016-09-20, issue #2197.
+          -- Eta is inferred by the positivity checker.
+          -- We should turn it off until it is proven to be safe.
+          haveEta      = maybe (Inferred False) Specified eta
+          -- haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
           con = ConHead conName conInduction $ map unArg fs
 
       reportSDoc "tc.rec" 30 $ text "record constructor is " <+> text (show con)
-      addConstant name $ defaultDefn defaultArgInfo name t0
-                       $ Record { recPars           = 0
-                                , recClause         = Nothing
-                                , recConHead        = con
-                                , recNamedCon       = hasNamedCon
-                                , recConType        = contype  -- addConstant adds params!
-                                , recFields         = fs
-                                , recTel            = ftel     -- addConstant adds params!
-                                , recAbstr          = Info.defAbstract i
-                                , recEtaEquality'   = haveEta
-                                , recInduction      = indCo    -- we retain the original user declaration, in case the record turns out to be recursive
-                                -- determined by positivity checker:
-                                , recRecursive      = False
-                                , recMutual         = []
-                                }
 
-      -- Add record constructor to signature
-      -- Andreas, 2011-05-19 moved this here, it was below the record module
-      --   creation
-      addConstant conName $
-        defaultDefn defaultArgInfo conName contype $
-             Constructor { conPars   = 0
-                         , conSrcCon = con
-                         , conData   = name
-                         , conAbstr  = Info.defAbstract conInfo
-                         , conInd    = conInduction
-                         }
+      -- Add the record definition.
+
+      -- Andreas, 2016-06-17, Issue #2018:
+      -- Do not rely on @addConstant@ to put in the record parameters,
+      -- as they might be renamed in the context.
+      -- By putting them ourselves (e.g. by using the original type @t@)
+      -- we make sure we get the original names!
+      let npars = size tel
+          telh  = fmap hideAndRelParams tel
+      escapeContext npars $ do
+        addConstant name $
+          defaultDefn defaultArgInfo name t $
+            Record
+              { recPars           = npars
+              , recClause         = Nothing
+              , recConHead        = con
+              , recNamedCon       = hasNamedCon
+              , recFields         = fs
+              , recTel            = telh `abstract` ftel
+              , recAbstr          = Info.defAbstract i
+              , recEtaEquality'   = haveEta
+              , recInduction      = indCo
+                  -- We retain the original user declaration [(co)inductive]
+                  -- in case the record turns out to be recursive.
+              -- Determined by positivity checker:
+              , recRecursive      = False
+              , recMutual         = []
+              }
+
+        -- Add record constructor to signature
+        addConstant conName $
+          defaultDefn defaultArgInfo conName (telh `abstract` contype) $
+            Constructor
+              { conPars   = npars
+              , conSrcCon = con
+              , conData   = name
+              , conAbstr  = Info.defAbstract conInfo
+              , conInd    = conInduction
+              , conErased = []
+              }
+
       -- Declare the constructor as eligible for instance search
       when (Info.defInstance i == InstanceDef) $ do
         addNamedInstance conName name
@@ -202,7 +224,7 @@ checkRecDef i name ind eta con ps contel fields =
         , nest 2 $ do
            tel <- getContextTelescope
            text (show tel) $+$ do
-           inContext [] $ do
+           inTopContext $ do
              prettyTCM tel $+$ do
                telA <- reify tel
                text (show telA) $+$ do
@@ -212,7 +234,7 @@ checkRecDef i name ind eta con ps contel fields =
 -}
 
       let info = setRelevance recordRelevance defaultArgInfo
-          addRecordVar = addCtxString "" $ Dom info rect
+          addRecordVar = addContext' ("", Dom info rect)
           -- the record variable has the empty name by intention, see issue 208
 
       let m = qnameToMName name  -- Name of record module.
@@ -251,7 +273,8 @@ checkRecDef i name ind eta con ps contel fields =
           -- record type.
           -- See test/Succeed/ProjectionsTakeModuleTelAsParameters.agda.
           tel' <- getContextTelescope
-          checkRecordProjections m name con tel' (raise 1 ftel) fields
+          setDefaultModuleParameters m
+          checkRecordProjections m name hasNamedCon con tel' (raise 1 ftel) fields
 
         -- Andreas 2012-02-13: postpone polarity computation until after positivity check
         -- computePolarity name
@@ -273,9 +296,9 @@ checkRecDef i name ind eta con ps contel fields =
     [@fs@   ]  the fields to be checked
 -}
 checkRecordProjections ::
-  ModuleName -> QName -> ConHead -> Telescope -> Telescope ->
+  ModuleName -> QName -> Bool -> ConHead -> Telescope -> Telescope ->
   [A.Declaration] -> TCM ()
-checkRecordProjections m r con tel ftel fs = do
+checkRecordProjections m r hasNamedCon con tel ftel fs = do
     checkProjs EmptyTel ftel fs
   where
 
@@ -299,7 +322,7 @@ checkRecordProjections m r con tel ftel fs = do
           , text "tel   =" <+> (inTopContext . prettyTCM $ tel)
           , text "ftel1 =" <+> prettyTCM ftel1
           , text "t     =" <+> prettyTCM t
-          , text "ftel2 =" <+> addCtxTel ftel1 (underAbstraction_ ftel2 prettyTCM)
+          , text "ftel2 =" <+> addContext ftel1 (underAbstraction_ ftel2 prettyTCM)
           ]
         ]
 
@@ -324,12 +347,12 @@ checkRecordProjections m r con tel ftel fs = do
       -- where Δ = Γ, tel is the current context
       let finalt   = telePi (replaceEmptyName "r" tel) t
           projname = qualify m $ qnameName x
-          projcall = Var 0 [Proj projname]
+          projcall o = Var 0 [Proj o projname]
           rel      = getRelevance ai
           -- the recursive call
           recurse  = checkProjs (abstract ftel1 $ ExtendTel (Dom ai t)
                                  $ Abs (nameToArgName $ qnameName projname) EmptyTel)
-                                (ftel2 `absApp` projcall) fs
+                                (ftel2 `absApp` projcall ProjSystem) fs
 
       reportSDoc "tc.rec.proj" 25 $ nest 2 $ text "finalt=" <+> do
         inTopContext $ prettyTCM finalt
@@ -356,47 +379,38 @@ checkRecordProjections m r con tel ftel fs = do
         -- compute body modification for irrelevant projections
         let bodyMod = case rel of
               Relevant   -> id
+              NonStrict  -> id
               Irrelevant -> DontCare
               _          -> __IMPOSSIBLE__
 
         let -- Andreas, 2010-09-09: comment for existing code
             -- split the telescope into parameters (ptel) and the type or the record
             -- (rt) which should be  R ptel
-            (ptel,[rt]) = splitAt (size tel - 1) $ telToList tel
-            projArgI    = domInfo rt
-            cpi    = ConPatternInfo (Just ConPRec) (Just $ argFromDom $ fmap snd rt)
+            telList = telToList tel
+            (_ptel,[rt]) = splitAt (size tel - 1) telList
+            cpo    = if hasNamedCon then ConOCon else ConORec
+            cpi    = ConPatternInfo (Just cpo) (Just $ argFromDom $ fmap snd rt)
             conp   = defaultArg $ ConP con cpi $
-                     [ Arg info $ unnamed $ VarP "x" | Dom info _ <- telToList ftel ]
-            nobind 0 = id
-            nobind n = Bind . Abs "_" . nobind (n - 1)
-            body   = nobind (size ftel1)
-                   $ Bind . Abs "x"
-                   $ nobind (size ftel2)
-                   $ Body $ bodyMod $ var (size ftel2)
+                     [ Arg info $ unnamed $ varP "x" | Dom info _ <- telToList ftel ]
+            body   = Just $ bodyMod $ var (size ftel2)
             cltel  = ftel
             clause = Clause { clauseRange = getRange info
                             , clauseTel       = killRange cltel
-                            , namedClausePats = [Named Nothing <$> numberPatVars (idP $ size ftel) conp]
+                            , namedClausePats = [Named Nothing <$> numberPatVars __IMPOSSIBLE__ (idP $ size ftel) conp]
                             , clauseBody      = body
                             , clauseType      = Just $ Arg ai t
                             , clauseCatchall  = False
                             }
 
-        -- Andreas, 2013-10-20
-        -- creating the projection construction function
-        let core = Lam projArgI $ Abs "r" $ bodyMod $ projcall
-            -- leading lambdas are to ignore parameter applications
-            proj = teleNoAbs ptel core
-            -- proj = foldr (\ (Dom ai (x, _)) -> Lam ai . NoAbs x) core ptel
-            projection = Projection
-              { projProper   = Just projname
+        let projection = Projection
+              { projProper   = True
+              , projOrig     = projname
               -- name of the record type:
-              , projFromType = r
+              , projFromType = defaultArg r
               -- index of the record argument (in the type),
               -- start counting with 1:
-              , projIndex    = size ptel + 1  -- which is @size tel@
-              , projDropPars = proj
-              , projArgInfo  = projArgI
+              , projIndex    = size tel -- which is @size ptel + 1@
+              , projLams     = ProjLams $ map (\ (Dom ai (x,_)) -> Arg ai x) telList
               }
 
         reportSDoc "tc.rec.proj" 80 $ sep
@@ -406,7 +420,7 @@ checkRecordProjections m r con tel ftel fs = do
         reportSDoc "tc.rec.proj" 70 $ sep
           [ text "adding projection"
           , nest 2 $ prettyTCM projname <+> text (show (clausePats clause)) <+> text "=" <+>
-                       inTopContext (addCtxTel ftel (prettyTCM (clauseBody clause)))
+                       inTopContext (addContext ftel (maybe (text "_|_") prettyTCM (clauseBody clause)))
           ]
         reportSDoc "tc.rec.proj" 10 $ sep
           [ text "adding projection"
@@ -427,22 +441,13 @@ checkRecordProjections m r con tel ftel fs = do
         escapeContext (size tel) $ do
           addConstant projname $
             (defaultDefn ai projname (killRange finalt)
-              Function { funClauses        = [clause]
-                       , funCompiled       = Just cc
-                       , funTreeless       = Nothing
-                       , funDelayed        = NotDelayed
-                       , funInv            = NotInjective
-                       , funAbstr          = ConcreteDef
-                       , funMutual         = []
-                       , funProjection     = Just projection
-                       , funSmashable      = True
-                       , funStatic         = False
-                       , funInline         = False
-                       , funTerminates     = Just True
-                       , funExtLam         = Nothing
-                       , funWith           = Nothing
-                       , funCopatternLHS   = isCopatternLHS [clause]
-                       })
+              emptyFunction
+                { funClauses        = [clause]
+                , funCompiled       = Just cc
+                , funProjection     = Just projection
+                , funTerminates     = Just True
+                , funCopatternLHS   = isCopatternLHS [clause]
+                })
               { defArgOccurrences = [StrictPos] }
           computePolarity projname
           when (Info.defInstance info == InstanceDef) $
