@@ -61,13 +61,13 @@ instance EmbPrj a => EmbPrj (Drop a) where
     valu _      = malformed
 
 instance EmbPrj a => EmbPrj (Elim' a) where
-  icod_ (Apply a) = icode1' a
-  icod_ (Proj  a) = icode1 0 a
+  icod_ (Apply a)  = icode1' a
+  icod_ (Proj a b) = icode2 0 a b
 
   value = vcase valu where
-    valu [a]    = valu1 Apply a
-    valu [0, a] = valu1 Proj a
-    valu _      = malformed
+    valu [a]       = valu1 Apply a
+    valu [0, a, b] = valu2 Proj a b
+    valu _         = malformed
 
 instance EmbPrj I.ConHead where
   icod_ (ConHead a b c) = icode3' a b c
@@ -98,7 +98,7 @@ instance EmbPrj I.Term where
   icod_ (Lam      a b) = icode2 1 a b
   icod_ (Lit      a  ) = icode1 2 a
   icod_ (Def      a b) = icode2 3 a b
-  icod_ (Con      a b) = icode2 4 a b
+  icod_ (Con    a b c) = icode3 4 a b c
   icod_ (Pi       a b) = icode2 5 a b
   icod_ (Sort     a  ) = icode1 7 a
   icod_ (MetaV    a b) = __IMPOSSIBLE__
@@ -113,7 +113,7 @@ instance EmbPrj I.Term where
     valu [1, a, b] = valu2 Lam   a b
     valu [2, a]    = valu1 Lit   a
     valu [3, a, b] = valu2 Def   a b
-    valu [4, a, b] = valu2 Con   a b
+    valu [4, a, b, c] = valu3 Con a b c
     valu [5, a, b] = valu2 Pi    a b
     valu [7, a]    = valu1 Sort  a
     valu [8, a]    = valu1 DontCare a
@@ -177,6 +177,15 @@ instance EmbPrj a => EmbPrj (Open a) where
     valu [a, b] = valu2 OpenThing a b
     valu _      = malformed
 
+instance EmbPrj a => EmbPrj (Local a) where
+  icod_ (Local a b) = icode2' a b
+  icod_ (Global a)  = icode1' a
+
+  value = vcase valu where
+    valu [a, b] = valu2 Local a b
+    valu [a]    = valu1 Global a
+    valu _      = malformed
+
 instance EmbPrj CtxId where
   icod_ (CtxId a) = icode a
   value n         = CtxId `fmap` value n
@@ -184,14 +193,14 @@ instance EmbPrj CtxId where
 instance EmbPrj DisplayTerm where
   icod_ (DTerm    a  )   = icode1' a
   icod_ (DDot     a  )   = icode1 1 a
-  icod_ (DCon     a b)   = icode2 2 a b
+  icod_ (DCon     a b c) = icode3 2 a b c
   icod_ (DDef     a b)   = icode2 3 a b
   icod_ (DWithApp a b c) = icode3 4 a b c
 
   value = vcase valu where
     valu [a]          = valu1 DTerm a
     valu [1, a]       = valu1 DDot a
-    valu [2, a, b]    = valu2 DCon a b
+    valu [2, a, b, c] = valu3 DCon a b c
     valu [3, a, b]    = valu2 DDef a b
     valu [4, a, b, c] = valu3 DWithApp a b c
     valu _            = malformed
@@ -201,14 +210,14 @@ instance EmbPrj MutualId where
   value n         = MutId `fmap` value n
 
 instance EmbPrj Definition where
-  icod_ (Defn a b c d e f g h i j k) = icode11' a b (P.killRange c) d e f g h i j k
+  icod_ (Defn a b c d e f g h i j k l) = icode12' a b (P.killRange c) d e f g h i j k l
 
   value = vcase valu where
-    valu [a, b, c, d, e, f, g, h, i, j, k] = valu11 Defn a b c d e f g h i j k
-    valu _                                 = malformed
+    valu [a, b, c, d, e, f, g, h, i, j, k, l] = valu12 Defn a b c d e f g h i j k l
+    valu _                                    = malformed
 
 instance EmbPrj NLPat where
-  icod_ (PVar a b)      = icode2 0 a b
+  icod_ (PVar a b c)    = icode3 0 a b c
   icod_ (PWild)         = icode0 1
   icod_ (PDef a b)      = icode2 2 a b
   icod_ (PLam a b)      = icode2 3 a b
@@ -217,28 +226,42 @@ instance EmbPrj NLPat where
   icod_ (PTerm a)       = icode1 6 a
 
   value = vcase valu where
-    valu [0, a, b] = valu2 PVar a b
-    valu [1]       = valu0 PWild
-    valu [2, a, b] = valu2 PDef a b
-    valu [3, a, b] = valu2 PLam a b
-    valu [4, a, b] = valu2 PPi a b
-    valu [5, a, b] = valu2 PBoundVar a b
-    valu [6, a]    = valu1 PTerm a
-    valu _         = malformed
+    valu [0, a, b, c] = valu3 PVar a b c
+    valu [1]          = valu0 PWild
+    valu [2, a, b]    = valu2 PDef a b
+    valu [3, a, b]    = valu2 PLam a b
+    valu [4, a, b]    = valu2 PPi a b
+    valu [5, a, b]    = valu2 PBoundVar a b
+    valu [6, a]       = valu1 PTerm a
+    valu _            = malformed
 
-instance EmbPrj RewriteRule where
-  icod_ (RewriteRule a b c d e) = icode5' a b c d e
+instance EmbPrj NLPType where
+  icod_ (NLPType a b) = icode2' a b
 
   value = vcase valu where
-    valu [a, b, c, d, e] = valu5 RewriteRule a b c d e
-    valu _               = malformed
+    valu [a, b] = valu2 NLPType a b
+    valu _      = malformed
+
+instance EmbPrj RewriteRule where
+  icod_ (RewriteRule a b c d e f) = icode6' a b c d e f
+
+  value = vcase valu where
+    valu [a, b, c, d, e, f] = valu6 RewriteRule a b c d e f
+    valu _                  = malformed
 
 instance EmbPrj Projection where
   icod_ (Projection a b c d e) = icode5' a b c d e
 
   value = vcase valu where
     valu [a, b, c, d, e] = valu5 Projection a b c d e
-    valu _               = malformed
+    valu _ = malformed
+
+instance EmbPrj ProjLams where
+  icod_ (ProjLams a) = icode1' a
+
+  value = vcase valu where
+    valu [a] = valu1 ProjLams a
+    valu _   = malformed
 
 instance EmbPrj ExtLamInfo where
   icod_ (ExtLamInfo a b) = icode2' a b
@@ -285,21 +308,33 @@ instance EmbPrj EtaEquality where
     valu _     = malformed
 
 instance EmbPrj Defn where
-  icod_ Axiom                                       = icode0 0
-  icod_ (Function    a b _ c d e f g h i j k l m n) = icode14 1 a b c d e f g h i j k l m n
-  icod_ (Datatype    a b c d e f g h i j)           = icode10 2 a b c d e f g h i j
-  icod_ (Record      a b c d e f g h i j k l)       = icode12 3 a b c d e f g h i j k l
-  icod_ (Constructor a b c d e)                     = icode5 4 a b c d e
-  icod_ (Primitive   a b c d)                       = icode4 5 a b c d
+  icod_ Axiom                                   = icode0 0
+  icod_ (Function    a b _ c d e f g h i j k m) = icode12 1 a b c d e f g h i j k m
+  icod_ (Datatype    a b c d e f g h i j)       = icode10 2 a b c d e f g h i j
+  icod_ (Record      a b c d e f g h i j k)     = icode11 3 a b c d e f g h i j k
+  icod_ (Constructor a b c d e f)               = icode6 4 a b c d e f
+  icod_ (Primitive   a b c d)                   = icode4 5 a b c d
+  icod_ AbstractDefn                            = __IMPOSSIBLE__
 
   value = vcase valu where
-    valu [0]                                           = valu0 Axiom
-    valu [1, a, b, c, d, e, f, g, h, i, j, k, l, m, n] = valu14 (\ a b -> Function a b Nothing) a b c d e f g h i j k l m n
-    valu [2, a, b, c, d, e, f, g, h, i, j]             = valu10 Datatype a b c d e f g h i j
-    valu [3, a, b, c, d, e, f, g, h, i, j, k, l]       = valu12 Record  a b c d e f g h i j k l
-    valu [4, a, b, c, d, e]                            = valu5 Constructor a b c d e
-    valu [5, a, b, c, d]                               = valu4 Primitive   a b c d
-    valu _                                             = malformed
+    valu [0]                                     = valu0 Axiom
+    valu [1, a, b, c, d, e, f, g, h, i, j, k, m] = valu12 (\ a b -> Function a b Nothing) a b c d e f g h i j k m
+    valu [2, a, b, c, d, e, f, g, h, i, j]       = valu10 Datatype a b c d e f g h i j
+    valu [3, a, b, c, d, e, f, g, h, i, j, k]    = valu11 Record  a b c d e f g h i j k
+    valu [4, a, b, c, d, e, f]                   = valu6 Constructor a b c d e f
+    valu [5, a, b, c, d]                         = valu4 Primitive   a b c d
+    valu _                                       = malformed
+
+instance EmbPrj FunctionFlag where
+  icod_ FunStatic       = icode0 0
+  icod_ FunInline       = icode0 1
+  icod_ FunMacro        = icode0 2
+
+  value = vcase valu where
+    valu [0] = valu0 FunStatic
+    valu [1] = valu0 FunInline
+    valu [2] = valu0 FunMacro
+    valu _   = malformed
 
 instance EmbPrj a => EmbPrj (WithArity a) where
   icod_ (WithArity a b) = icode2' a b
@@ -353,17 +388,6 @@ instance EmbPrj I.Clause where
     valu [a, b, c, d, e, f] = valu6 Clause a b c d e f
     valu _                  = malformed
 
-instance EmbPrj a => EmbPrj (I.ClauseBodyF a) where
-  icod_ (Body   a) = icode1 0 a
-  icod_ (Bind   a) = icode1' a
-  icod_ NoBody     = icode0'
-
-  value = vcase valu where
-    valu [0, a] = valu1 Body   a
-    valu [a]    = valu1 Bind   a
-    valu []     = valu0 NoBody
-    valu _      = malformed
-
 instance EmbPrj I.ConPatternInfo where
   icod_ (ConPatternInfo a b) = icode2' a b
 
@@ -371,19 +395,26 @@ instance EmbPrj I.ConPatternInfo where
     valu [a, b] = valu2 ConPatternInfo a b
     valu _      = malformed
 
+instance EmbPrj I.DBPatVar where
+  icod_ (DBPatVar a b) = icode2' a b
+
+  value = vcase valu where
+    valu [a, b] = valu2 DBPatVar a b
+    valu _      = malformed
+
 instance EmbPrj a => EmbPrj (I.Pattern' a) where
   icod_ (VarP a    ) = icode1' a
-  icod_ (ConP a b c) = icode3' a b c
+  icod_ (ConP a b c) = icode3 1 a b c
   icod_ (LitP a    ) = icode1 2 a
   icod_ (DotP a    ) = icode1 3 a
-  icod_ (ProjP a   ) = icode1 4 a
+  icod_ (ProjP a b ) = icode2 4 a b
 
   value = vcase valu where
     valu [a]       = valu1 VarP a
-    valu [a, b, c] = valu3 ConP a b c
+    valu [1, a, b, c] = valu3 ConP a b c
     valu [2, a]    = valu1 LitP a
     valu [3, a]    = valu1 DotP a
-    valu [4, a]    = valu1 ProjP a
+    valu [4, a, b] = valu2 ProjP a b
     valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Builtin a) where
