@@ -1,6 +1,5 @@
 {-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE PatternGuards            #-}
 
 module Agda.TypeChecking.SizedTypes where
 
@@ -180,7 +179,7 @@ boundedSizeMetaHook v tel0 a = do
       n <- getContextSize
       let tel | n > 0     = telFromList $ genericDrop n $ telToList tel0
               | otherwise = tel0
-      addCtxTel tel $ do
+      addContext tel $ do
         v <- sizeSuc 1 $ raise (size tel) v `apply` teleArgs tel
         -- compareSizes CmpLeq v u
         size <- sizeType
