@@ -1,10 +1,6 @@
 {-# LANGUAGE CPP                   #-}
 {-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE PatternGuards         #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TupleSections         #-}
 
 -- | Solving size constraints under hypotheses.
 --
@@ -54,13 +50,20 @@
 
 module Agda.TypeChecking.SizedTypes.Solve where
 
-import Control.Monad (unless)
+import Prelude hiding (null)
+
+import Control.Monad hiding (forM, forM_)
+import Control.Monad.Trans.Maybe
+import Control.Monad.Reader (asks)
+
 import Data.Foldable (Foldable, foldMap, forM_)
 import Data.Function
-import Data.List
-import Data.Monoid (mappend)
+import qualified Data.List as List
+import Data.Monoid
 import qualified Data.Map as Map
+import Data.Set (Set)
 import qualified Data.Set as Set
+import qualified Data.Traversable as Trav
 import Data.Traversable (Traversable, forM)
 
 import Agda.Interaction.Options
@@ -71,15 +74,13 @@ import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad as TCM hiding (Offset)
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Free
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.MetaVars
--- import {-# SOURCE #-} Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Constraints as C
--- import {-# SOURCE #-} Agda.TypeChecking.Conversion
--- import {-# SOURCE #-} Agda.TypeChecking.Constraints
 
 import qualified Agda.TypeChecking.SizedTypes as S
 import Agda.TypeChecking.SizedTypes.Syntax as Size
@@ -91,21 +92,28 @@ import Agda.Utils.Either
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Function
 import Agda.Utils.Functor
+import Agda.Utils.Lens
 
 #if MIN_VERSION_base(4,8,0)
-import Agda.Utils.List hiding ( uncons )
+import qualified Agda.Utils.List as List hiding ( uncons )
 #else
-import Agda.Utils.List
+import qualified Agda.Utils.List as List
 #endif
 
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Size
 import Agda.Utils.Tuple
+import qualified Agda.Utils.VarSet as VarSet
+
+import qualified Agda.Utils.Either as Either
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+type CC = Closure TCM.Constraint
+
 -- | Flag to control the behavior of size solver.
 data DefaultToInfty
   = DefaultToInfty      -- ^ Instantiate all unconstrained size variables to ∞.
@@ -113,65 +121,277 @@ data DefaultToInfty
   deriving (Eq, Ord, Show)
 
 -- | Solve size constraints involving hypotheses.
+
 solveSizeConstraints :: DefaultToInfty -> TCM ()
 solveSizeConstraints flag =  do
-  -- Get the constraints.
-  cs0 <- S.getSizeConstraints
+
+  -- 1. Get the constraints normalised.
+
+  cs0 <- mapM (mapClosure normalise) =<< S.getSizeConstraints
+  unless (null cs0) $
+    reportSDoc "tc.size.solve" 40 $ vcat $
+      [ text $ "Solving constraints (" ++ show flag ++ ")"
+      ] ++ map prettyTCM cs0
   let -- Error for giving up
       cannotSolve = typeError . GenericDocError =<<
-        vcat (text "Cannot solve size constraints" : map prettyTCM
-                   cs0)
-  unless (null cs0) $ solveSizeConstraints_ cs0
+        vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
+
+  -- 2. Cluster the constraints by common size metas.
+
+  -- Get all size metas.
+  sizeMetaSet <- Set.fromList . map (\ (x, _t, _tel) -> x) <$> S.getSizeMetas True
+
+  -- Pair each constraint with its list of size metas occurring in it.
+  cms <- forM cs0 $ \ cl -> enterClosure cl $ \ c -> do
+
+    -- @allMetas@ does not reduce or instantiate;
+    -- this is why we require the size constraints to be normalised.
+    return (cl, map metaId . Set.toList $
+      sizeMetaSet `Set.intersection` Set.fromList (allMetas c))
+
+  -- Now, some constraints may have no metas (clcs), the others have at least one (othercs).
+  let classify :: (a, [b]) -> Either a (a, (b,[b]))
+      classify (cl, [])     = Left  cl
+      classify (cl, (x:xs)) = Right (cl, (x,xs))
+  let (clcs, othercs) = List.mapEither classify cms
+
+  -- We cluster the constraints by their metas.
+  let ccs = cluster' othercs
+
+  -- 3. Solve each cluster
+
+  -- Solve the closed constraints, one by one.
+
+  forM_ clcs $ \ c -> () <$ solveSizeConstraints_ flag [c]
+
+  -- Solve the clusters.
+
+  constrainedMetas <- Set.unions <$> do
+    forM  (ccs) $ \ (cs :: [CC]) -> do
+      when (null cs) __IMPOSSIBLE__
+
+      -- Convert each constraint in the cluster to the largest context.
+      -- (Keep fingers crossed).
+
+      enterClosure (List.maximumBy (compare `on` (length . envContext . clEnv)) cs) $ \ _ -> do
+        -- Get all constraints that can be cast to the longest context.
+        cs' :: [TCM.Constraint] <- catMaybes <$> do
+          mapM (runMaybeT . castConstraintToCurrentContext) cs
+
+        reportSDoc "tc.size.solve" 20 $ vcat $
+          [ text "converted size constraints to context: " <+> do
+              tel <- getContextTelescope
+              inTopContext $ prettyTCM tel
+          ] ++ map (nest 2 . prettyTCM) cs'
+
+        -- Solve the converted constraints.
+        solveSizeConstraints_ flag =<<  mapM buildClosure cs'
+
+  -- 4. Possibly set remaining metas to infinity.
 
   -- Andreas, issue 1862: do not default to ∞ always, could be too early.
   when (flag == DefaultToInfty) $ do
 
-  -- Set the unconstrained size metas to ∞.
-  ms <- S.getSizeMetas False -- do not get interaction metas
-  unless (null ms) $ do
-    inf <- primSizeInf
-    forM_ ms $ \ (m, t, tel) -> unlessM (isFrozen m) $ do
-      reportSDoc "tc.size.solve" 20 $
-        text "solution " <+> prettyTCM (MetaV m []) <+>
-        text " := "      <+> prettyTCM inf
-      assignMeta 0 m t (downFrom $ size tel) inf
+    -- let constrainedMetas = Set.fromList $ concat $
+    --       for cs0 $ \ Closure{ clValue = ValueCmp _ _ u v } ->
+    --         allMetas u ++ allMetas v
 
-  -- Double check.
-  unless (null cs0 && null ms) $ do
-    flip catchError (const cannotSolve) $
-      noConstraints $
-        forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
+    -- Set the unconstrained size metas to ∞.
+    ms <- S.getSizeMetas False -- do not get interaction metas
+    unless (null ms) $ do
+      inf <- primSizeInf
+      forM_ ms $ \ (m, t, tel) -> do
+        unless (m `Set.member` constrainedMetas) $ do
+        unlessM (isFrozen m) $ do
+        reportSDoc "tc.size.solve" 20 $
+          text "solution " <+> prettyTCM (MetaV m []) <+>
+          text " := "      <+> prettyTCM inf
+        assignMeta 0 m t (List.downFrom $ size tel) inf
 
-solveSizeConstraints_ :: [Closure TCM.Constraint] -> TCM ()
-solveSizeConstraints_ cs0 = do
+  -- -- Double check.
+  -- unless (null cs0 && null ms) $ do
+  --   flip catchError (const cannotSolve) $
+  --     noConstraints $
+  --       forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
+
+  -- 5. Make sure we did not lose any constraints.
+  forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
+
+
+-- | TODO: this does not actually work!
+--
+--   We would like to use a constraint @c@ created in context @Δ@ from module @N@
+--   in the current context @Γ@ and current module @M@.
+--
+--   @Δ@ is module tel @Δ₁@ of @N@ extended by some local bindings @Δ₂@.
+--   @Γ@ is the current context.
+--   The module parameter substitution from current @M@ to @N@ be
+--   @Γ ⊢ σ : Δ₁@.
+--
+--   If @M == N@, we do not need the parameter substitution.  We try raising.
+--
+--   We first strengthen @Δ ⊢ c@ to live in @Δ₁@ and obtain @c₁ = strengthen Δ₂ c@.
+--   We then transport @c₁@ to @Γ@ and obtain @c₂ = applySubst σ c₁@.
+--
+--   This works for different modules, but if @M == N@ we should not strengthen
+--   and then weaken, because strengthening is a partial operation.
+--   We should rather lift the substitution @σ@ by @Δ₂@ and then
+--   raise by @Γ₂ - Δ₂@.
+--   This "raising" might be a strengthening if @Γ₂@ is shorter than @Δ₂@.
+--
+--   (TODO: If the module substitution does not exist, because @N@ is not
+--   a parent of @M@, we cannot use the constraint, as it has been created
+--   in an unrelated context.)
+
+castConstraintToCurrentContext' :: Closure TCM.Constraint -> MaybeT TCM TCM.Constraint
+castConstraintToCurrentContext' cl = do
+  let modN  = envCurrentModule $ clEnv cl
+      delta = envContext $ clEnv cl
+  -- The module telescope of the constraint.
+  -- The constraint could come from the module telescope of the top level module.
+  -- In this case, it does not live in any module!
+  -- Thus, getSection can return Nothing.
+  delta1 <- liftTCM $ maybe empty (^. secTelescope) <$> getSection modN
+  -- The number of locals of the constraint.
+  let delta2 = size delta - size delta1
+  unless (delta2 >= 0) __IMPOSSIBLE__
+
+  -- The current module M and context Γ.
+  modM  <- currentModule
+  gamma <- liftTCM $ getContextSize
+  -- The current module telescope.
+  -- Could also be empty, if we are in the front matter or telescope of the top-level module.
+  gamma1 <-liftTCM $ maybe empty (^. secTelescope) <$> getSection modM
+  -- The current locals.
+  let gamma2 = gamma - size gamma1
+
+  -- Γ ⊢ σ : Δ₁
+  sigma <- liftTCM $ getModuleParameterSub modN
+
+  -- Debug printing.
+  reportSDoc "tc.constr.cast" 40 $ text "casting constraint" $$ do
+    tel <- getContextTelescope
+    inTopContext $ nest 2 $ vcat $
+      [ text "current module                = " <+> prettyTCM modM
+      , text "current module telescope      = " <+> prettyTCM gamma1
+      , text "current context               = " <+> prettyTCM tel
+      , text "constraint module             = " <+> prettyTCM modN
+      , text "constraint module telescope   = " <+> prettyTCM delta1
+      , text "constraint context            = " <+> (prettyTCM =<< enterClosure cl (const $ getContextTelescope))
+      , text "constraint                    = " <+> enterClosure cl prettyTCM
+      , text "module parameter substitution = " <+> prettyTCM sigma
+      ]
+
+  -- If gamma2 < 0, we must be in the wrong context.
+  -- E.g. we could have switched to the empty context even though
+  -- we are still inside a module with parameters.
+  -- In this case, we cannot safely convert the constraint,
+  -- since the module parameter substitution may be wrong.
+  guard (gamma2 >= 0)
+
+  -- Shortcut for modN == modM:
+  -- Raise constraint from Δ to Γ, if possible.
+  -- This might save us some strengthening.
+  if modN == modM then raiseMaybe (gamma - size delta) $ clValue cl else do
+
+  -- Strengthen constraint to Δ₁ ⊢ c
+  c <- raiseMaybe (-delta2) $ clValue cl
+
+  -- Ulf, 2016-11-09: I don't understand what this function does when M and N
+  -- are not related. Certainly things can go terribly wrong (see
+  -- test/Succeed/Issue2223b.agda)
+  fv <- liftTCM $ getModuleFreeVars modN
+  guard $ fv == size delta1
+
+  -- Γ ⊢ c[σ]
+  return $ applySubst sigma c
+  where
+    raiseMaybe n c = do
+      -- Fine if we have to weaken or strengthening is safe.
+      guard $ n >= 0 || List.all (>= -n) (VarSet.toList $ allVars $ freeVars c)
+      return $ raise n c
+
+
+-- | A hazardous hack, may the Gods have mercy on us.
+--
+--   To cast to the current context, we match the context of the
+--   given constraint by 'CtxId', and as fallback, by variable name (douh!).
+--
+--   This hack lets issue 2046 go through.
+
+castConstraintToCurrentContext :: Closure TCM.Constraint -> MaybeT TCM TCM.Constraint
+castConstraintToCurrentContext cl = do
+  -- The target context
+  gamma <- asks envContext
+  -- The context where the constraint lives.
+  let delta = envContext $ clEnv cl
+  -- The constraint
+  let c = clValue cl
+  let findInGamma (Ctx cid (Dom _ (x, t))) =
+        -- try to find same CtxId (safe)
+        case List.findIndex ((cid ==) . ctxId) gamma of
+          Just i -> Just i
+          Nothing ->
+            -- match by name (hazardous)
+            -- This is one of the seven deadly sins (not respecting alpha).
+            List.findIndex ((x ==) . fst . unDom . ctxEntry) gamma
+  let cand = map findInGamma delta
+  -- The domain of our substitution
+  let coveredVars = VarSet.fromList $ catMaybes $ zipWith ($>) cand [0..]
+  -- Check that all the free variables of the constraint are contained in
+  -- coveredVars.
+  -- We ignore the free variables occurring in sorts.
+  guard $ getAll $ runFree (\ (i, _) -> All $ i `VarSet.member` coveredVars) IgnoreAll c
+  -- Turn cand into a substitution.
+  -- Since we ignored the free variables in sorts, we better patch up
+  -- the substitution with some dummy term (Sort Prop) rather than __IMPOSSIBLE__.
+  let dummy = Sort Prop
+  let sigma = parallelS $ map (maybe dummy var) cand
+  -- Apply substitution to constraint and pray that the Gods are merciful on us.
+  return $ applySubst sigma c
+  -- Note: the resulting constraint may not well-typed.
+  -- Even if it is, it may map variables to their wrong counterpart.
+
+-- | Return the size metas occurring in the simplified constraints.
+--   A constraint like @↑ _j =< ∞ : Size@ simplifies to nothing,
+--   so @_j@ would not be in this set.
+solveSizeConstraints_ :: DefaultToInfty -> [CC] -> TCM (Set MetaId)
+solveSizeConstraints_ flag cs0 = do
   -- Pair constraints with their representation as size constraints.
   -- Discard constraints that do not have such a representation.
-  ccs <- catMaybes <$> do
-    forM cs0 $ \ c -> fmap (c,) <$> computeSizeConstraint c
+  ccs :: [(CC,HypSizeConstraint)] <- catMaybes <$> do
+    forM cs0 $ \ c0 -> fmap (c0,) <$> computeSizeConstraint c0
 
   -- Simplify constraints and check for obvious inconsistencies.
-  cs <- concat <$> do
-    forM ccs $ \ (c, HypSizeConstraint cxt hids hs sc) -> do
-      case simplify1 (\ c -> return [c]) sc of
+  ccs' <- concat <$> do
+    forM ccs $ \ (c0, HypSizeConstraint cxt hids hs sc) -> do
+      case simplify1 (\ sc -> return [sc]) sc of
         Left _ -> typeError . GenericDocError =<< do
-          text "Contradictory size constraint" <+> prettyTCM c
-        Right cs -> return $ HypSizeConstraint cxt hids hs <$> cs
+          text "Contradictory size constraint" <+> prettyTCM c0
+        Right cs -> return $ (c0,) . HypSizeConstraint cxt hids hs <$> cs
 
   -- Cluster constraints according to the meta variables they mention.
   -- @csNoM@ are the constraints that do not mention any meta.
-  let (csNoM, csMs) = (`partitionMaybe` cs) $ \ c ->
-        fmap (c,) $ uncons $ map (metaId . sizeMetaId) $ Set.toList $ flexs c
+  let (csNoM, csMs) = (`List.partitionMaybe` ccs') $ \ p@(c0, c) ->
+        fmap (p,) $ List.uncons $ map (metaId . sizeMetaId) $ Set.toList $ flexs c
   -- @css@ are the clusters of constraints.
+      css :: [[(CC,HypSizeConstraint)]]
       css = cluster' csMs
-  -- There should be no constraints that do not mention a meta?
-  unless (null csNoM) __IMPOSSIBLE__
+
+  -- Check that the closed constraints are valid.
+  solveCluster flag csNoM
 
   -- Now, process the clusters.
-  forM_ css solveCluster
+  forM_ css $ solveCluster flag
 
-solveCluster :: [HypSizeConstraint] -> TCM ()
-solveCluster [] = __IMPOSSIBLE__
-solveCluster cs = do
+  return $ Set.mapMonotonic sizeMetaId $ flexs $ map (snd . fst) csMs
+
+-- | Solve a cluster of constraints sharing some metas.
+--
+solveCluster :: DefaultToInfty -> [(CC,HypSizeConstraint)] -> TCM ()
+solveCluster flag [] = return ()
+solveCluster flag ccs = do
+  let cs = map snd ccs
   let err reason = typeError . GenericDocError =<< do
         vcat $
           [ text $ "Cannot solve size constraints" ] ++ map prettyTCM cs ++
@@ -200,7 +420,7 @@ solveCluster cs = do
 -}
   -- We rely on the fact that contexts are only extended...
   -- Just take the longest context.
-  let HypSizeConstraint gamma hids hs _ = maximumBy (compare `on` (length . sizeContext)) cs
+  let HypSizeConstraint gamma hids hs _ = List.maximumBy (compare `on` (length . sizeContext)) cs
   -- Length of longest context.
   let n = size gamma
 
@@ -258,23 +478,31 @@ solveCluster cs = do
   -- There cannot be negative cycles in hypotheses graph due to scoping.
   let hg = fromRight __IMPOSSIBLE__ $ hypGraph (rigids csF) hyps
 
-  -- Construct the constraint graph.
-  --    g :: Size.Graph NamedRigid Int Label
-  g <- either err return $ constraintGraph csF hg
-  reportSDoc "tc.size.solve" 40 $ vcat $
-    [ text "Constraint graph"
-    , text (show g)
-    ]
+  -- -- Construct the constraint graph.
+  -- --    g :: Size.Graph NamedRigid Int Label
+  -- g <- either err return $ constraintGraph csF hg
+  -- reportSDoc "tc.size.solve" 40 $ vcat $
+  --   [ text "Constraint graph"
+  --   , text (show g)
+  --   ]
+
+  -- sol :: Solution NamedRigid Int <- either err return $ solveGraph Map.empty hg g
+  -- either err return $ verifySolution hg csF sol
+
+  -- Andreas, 2016-07-13, issue 2096.
+  -- Running the solver once might result in unsolvable left-over constraints.
+  -- We need to iterate the solver to detect this.
+  sol :: Solution NamedRigid Int <- either err return $
+    iterateSolver Map.empty hg csF Map.empty
 
-  sol :: Solution NamedRigid Int <- either err return $ solveGraph Map.empty hg g
-  either err return $ verifySolution hg csF sol
   -- Convert solution to meta instantiation.
   forM_ (Map.assocs sol) $ \ (m, a) -> do
     unless (validOffset a) __IMPOSSIBLE__
     -- Solution does not contain metas
     u <- unSizeExpr $ fmap __IMPOSSIBLE__ a
     let x = MetaId m
-    let SizeMeta _ xs = fromMaybe __IMPOSSIBLE__ $ find ((m==) . metaId . sizeMetaId) metas
+    let SizeMeta _ xs = fromMaybe __IMPOSSIBLE__ $
+          List.find ((m==) . metaId . sizeMetaId) metas
     -- Check that solution is well-scoped
     let ys = rigidIndex <$> Set.toList (rigids a)
         ok = all (`elem` xs) ys -- TODO: more efficient
@@ -282,10 +510,67 @@ solveCluster cs = do
     u <- if ok then return u else primSizeInf
     t <- getMetaType x
     reportSDoc "tc.size.solve" 20 $ inTopContext $ modifyContext (const gamma) $ do
-      text "solution " <+> prettyTCM (MetaV x []) <+> text " := " <+> prettyTCM u
+      let args = map (Apply . defaultArg . var) xs
+      text "solution " <+> prettyTCM (MetaV x args) <+> text " := " <+> prettyTCM u
+    reportSDoc "tc.size.solve" 60 $ vcat
+      [ text $ "  xs = " ++ show xs
+      , text $ "  u  = " ++ show u
+      ]
     assignMeta n x t xs u
+    -- WRONG:
+    -- let partialSubst = List.sort $ zip xs $ map var $ downFrom n
+    -- assignMeta' n x t (length xs) partialSubst u
+    -- WRONG: assign DirEq x (map (defaultArg . var) xs) u
 
+  -- Possibly set remaining size metas to ∞ (issue 1862)
+  -- unless we have an interaction meta in the cluster (issue 2095).
 
+  ims <- Set.fromList <$> getInteractionMetas
+
+  --  ms = unsolved size metas from cluster
+  let ms = Set.fromList (map sizeMetaId metas) Set.\\  -- Some CPP or ghc does not like trailing backslash, thus, this comment!
+             Set.mapMonotonic MetaId (Map.keysSet sol)
+  --  Make sure they do not contain an interaction point
+  let noIP = Set.null $ Set.intersection ims ms
+
+  unless (null ms) $ reportSDoc "tc.size.solve" 30 $ fsep $
+    [ text "cluster did not solve these size metas: " ] ++ map prettyTCM (Set.toList ms)
+
+  solvedAll <- do
+    -- If no metas are left, we have solved this cluster completely.
+    if Set.null ms                then return True  else do
+    -- Otherwise, we can solve it completely if we are allowed to set to ∞.
+    if flag == DontDefaultToInfty then return False else do
+    -- Which is only the case when we have no interaction points in the cluster.
+    if not noIP                   then return False else do
+    -- Try to set all unconstrained size metas to ∞.
+    inf <- primSizeInf
+    and <$> do
+      forM (Set.toList ms) $ \ m -> do
+        -- If one variable is frozen, we cannot set it (and hence not all) to ∞
+        let no = do
+              reportSDoc "tc.size.solve" 30 $
+                prettyTCM (MetaV m []) <+> text "is frozen, cannot set it to ∞"
+              return False
+        ifM (isFrozen m) no $ {-else-} do
+          reportSDoc "tc.size.solve" 20 $
+            text "solution " <+> prettyTCM (MetaV m []) <+>
+            text " := "      <+> prettyTCM inf
+          t <- jMetaType . mvJudgement <$> lookupMeta m
+          TelV tel core <- telView t
+          unlessM (isJust <$> isSizeType core) __IMPOSSIBLE__
+          assignMeta 0 m t (List.downFrom $ size tel) inf
+          return True
+
+  -- Double check.
+  when solvedAll $ do
+    let cs0 = map fst ccs
+        -- Error for giving up
+        cannotSolve = typeError . GenericDocError =<<
+          vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
+    flip catchError (const cannotSolve) $
+      noConstraints $
+        forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
 
 -- | Collect constraints from a typing context, looking for SIZELT hypotheses.
 getSizeHypotheses :: Context -> TCM [(CtxId, SizeConstraint)]
@@ -452,7 +737,9 @@ instance Flexs SizeMeta HypSizeConstraint where
 instance PrettyTCM HypSizeConstraint where
   prettyTCM (HypSizeConstraint cxt _ hs c) =
     inTopContext $ modifyContext (const cxt) $ do
+      let cxtNames = reverse $ map (fst . unDom . ctxEntry) cxt
       -- text ("[#cxt=" ++ show (size cxt) ++ "]") <+> do
+      prettyList (map prettyTCM cxtNames) <+> do
       applyUnless (null hs)
        (((hcat $ punctuate (text ", ") $ map prettyTCM hs) <+> text "|-") <+>)
        (prettyTCM c)
@@ -493,7 +780,7 @@ sizeExpr u = do
     OtherSize u -> case ignoreSharing u of
       Var i []    -> (\ x -> Just $ Rigid (NamedRigid x i) 0) . show <$> nameOfBV i
 --      MetaV m es  -> return $ Just $ Flex (SizeMeta m es) 0
-      MetaV m es | Just xs <- mapM isVar es, fastDistinct xs
+      MetaV m es | Just xs <- mapM isVar es, List.fastDistinct xs
                   -> return $ Just $ Flex (SizeMeta m xs) 0
       _           -> return Nothing
   where
