@@ -1,10 +1,4 @@
-{-# LANGUAGE DeriveFoldable             #-}
-{-# LANGUAGE DeriveFunctor              #-}
-{-# LANGUAGE DeriveTraversable          #-}
-{-# LANGUAGE FlexibleInstances          #-}
-{-# LANGUAGE FunctionalDependencies     #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
 {-# LANGUAGE UndecidableInstances       #-}
 
@@ -19,8 +13,6 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable)
 
-import Test.QuickCheck
-
 import Agda.TypeChecking.SizedTypes.Utils
 
 -- * Syntax
@@ -38,13 +30,6 @@ instance MeetSemiLattice Offset where
 instance Plus Offset Offset Offset where
   plus (O x) (O y) = O (plus x y)
 
-instance Arbitrary Offset where
-  arbitrary = do
-    NonNegative x <- arbitrary
-    return x
-
-  shrink (O x) = map O $ filter (>= 0) (shrink x)
-
 -- | Fixed size variables @i@.
 newtype Rigid  = RigidId { rigidId :: String }
   deriving (Eq, Ord)
@@ -89,9 +74,6 @@ instance MeetSemiLattice Cmp where
 instance Top Cmp where
   top = Le
 
-instance Arbitrary Cmp where
-  arbitrary = arbitraryBoundedEnum
-
 -- | Constraint: an inequation between size expressions,
 --   e.g. @X < ∞@ or @i + 3 ≤ j@.
 data Constraint' rigid flex = Constraint
@@ -150,6 +132,9 @@ instance Ord f => Substitute r f (Constraint' r f) where
 instance Substitute r f a => Substitute r f [a] where
   subst = map . subst
 
+instance Substitute r f a => Substitute r f (Map k a) where
+  subst = fmap . subst
+
 -- | Add offset to size expression.
 instance Plus (SizeExpr' r f) Offset (SizeExpr' r f) where
   plus e m =
@@ -163,7 +148,7 @@ instance Plus (SizeExpr' r f) Offset (SizeExpr' r f) where
 
 type CTrans r f = Constraint' r f -> Either String [Constraint' r f]
 
--- | Returns 'Nothing' if we have a contradictory constraint.
+-- | Returns an error message if we have a contradictory constraint.
 simplify1 :: (Show f, Show r, Eq r) => CTrans r f -> CTrans r f
 simplify1 test c = do
   let err = Left $ "size constraint " ++ show c ++ " is inconsistent"
