@@ -1,14 +1,6 @@
 {-# LANGUAGE CPP                       #-}
-{-# LANGUAGE FlexibleInstances         #-}
-{-# LANGUAGE MultiParamTypeClasses     #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
-{-# LANGUAGE PatternGuards             #-}
 {-# LANGUAGE ScopedTypeVariables       #-}
-{-# LANGUAGE TupleSections             #-}
-
-#if __GLASGOW_HASKELL__ >= 710
-{-# LANGUAGE FlexibleContexts #-}
-#endif
 
 module Agda.TypeChecking.SizedTypes.WarshallSolver where
 
@@ -26,8 +18,6 @@ import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Traversable (for)
 
-import Test.QuickCheck (Arbitrary(..), frequency, choose)
-
 import Agda.TypeChecking.SizedTypes.Syntax
 import Agda.TypeChecking.SizedTypes.Utils
 
@@ -147,12 +137,6 @@ instance Num Weight where
 instance Plus Weight Offset Weight where
   plus w k = w + (Offset k)
 
-instance Arbitrary Weight where
-  arbitrary = frequency
-    [ (1, return Infinity)
-    , (5, Offset . O <$> choose (0, 200))
-    ]
-
 -- | Test for negativity, used to detect negative cycles.
 class Negative a where
   negative :: a -> Bool
@@ -226,12 +210,6 @@ instance Top Label where
   isTop Label{}       = False
   isTop LInf          = True
 
-instance Arbitrary Label where
-  arbitrary = frequency
-    [ (1, return LInf)
-    , (5, Label <$> arbitrary <*> arbitrary)
-    ]
-
 -- * Semiring with idempotent '+' == dioid
 
 instance Dioid Weight where
@@ -581,6 +559,7 @@ data Bounds r f = Bounds
   { lowerBounds :: Bound r f
   , upperBounds :: Bound r f
   , mustBeFinite :: Set f
+    -- ^ These metas are < ∞.
   }
 
 -- | Compute a lower bound for a flexible from an edge.
@@ -869,9 +848,9 @@ solveGraph pols hg g = do
           []     -> return $ Nothing
           (a:as) -> do
             case foldM (glb hg) a as of
-              Nothing -> Left $ "inconsistent upper bound for " ++ show x
-              Just l | validOffset l -> return $ Just l
-                     | otherwise     -> return $ findRigidBelow hg l
+              Just l | validOffset l                  -> return $ Just l
+                     | Just l' <- findRigidBelow hg l -> return $ Just l'
+              _ -> Left $ "inconsistent upper bound for " ++ show x
       case (lb, ub) of
         (Just l, Nothing) -> return $ Just (x, l)  -- solve x = lower bound
         (Nothing, Just u) -> return $ Just (x, u)  -- solve x = upper bound
@@ -912,6 +891,30 @@ verifySolution hg cs sol = do
     Just cs -> Left $ "solution leaves constraints " ++ show cs
 -}
 
+-- | Iterate solver until no more metas can be solved.
+--
+--   This might trigger a (wanted) error on the second iteration (see Issue 2096)
+--   which would otherwise go unnoticed.
+
+iterateSolver
+  :: (Ord r, Ord f, Show r, Show f)
+  => Polarities f
+     -- ^ Meta variable polarities (prefer lower or upper solution?).
+  -> HypGraph r f
+     -- ^ Hypotheses (assumed to have no metas, so, fixed during iteration).
+  -> [Constraint' r f]
+     -- ^ Constraints to solve.
+  -> Solution r f
+     -- ^ Previous substitution (already applied to constraints).
+  -> Either String (Solution r f)
+     -- ^ Accumulated substition.
+
+iterateSolver pols hg cs sol0 = do
+  g <- constraintGraph cs hg
+  sol <- solveGraph pols hg g
+  if Map.null sol then return sol0 else
+    iterateSolver pols hg (subst sol cs) (Map.unionWith __IMPOSSIBLE__ sol $ subst sol sol0)
+
 -- * Tests
 
 testSuccs :: Ord f => Map (Node [Char] f) [Edge' [Char] f Label]
