@@ -1,6 +1,4 @@
 {-# LANGUAGE CPP               #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE PatternGuards     #-}
 
 module Agda.TypeChecking.Unquote where
 
@@ -38,7 +36,7 @@ import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Exception
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Reduce.Monad hiding (reportSDoc)
+import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Quote
@@ -52,13 +50,15 @@ import {-# SOURCE #-} Agda.TypeChecking.Rules.Def
 
 import Agda.Utils.Except
 import Agda.Utils.Impossible
-import Agda.Utils.Monad ( ifM )
+import Agda.Utils.Maybe
+import Agda.Utils.Monad
 import Agda.Utils.Permutation ( Permutation(Perm), compactP )
 import Agda.Utils.String ( Str(Str), unStr )
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as Set
 import Agda.Utils.Maybe.Strict (toLazy)
 import Agda.Utils.FileName
+import Agda.Utils.Lens
 
 #include "undefined.h"
 
@@ -120,12 +120,12 @@ inOriginalContext m =
 isCon :: ConHead -> TCM Term -> UnquoteM Bool
 isCon con tm = do t <- liftU tm
                   case ignoreSharing t of
-                    Con con' _ -> return (con == con')
+                    Con con' _ _ -> return (con == con')
                     _ -> return False
 
 isDef :: QName -> TCM Term -> UnquoteM Bool
 isDef f tm = do
-  t <- liftU tm
+  t <- liftU (etaContract =<< normalise =<< tm)
   case ignoreSharing t of
     Def g _ -> return (f == g)
     _       -> return False
@@ -141,11 +141,6 @@ reduceQuotedTerm t = do
 class Unquote a where
   unquote :: I.Term -> UnquoteM a
 
-unquoteH :: Unquote a => Arg Term -> UnquoteM a
-unquoteH a | isHidden a && isRelevant a =
-    unquote $ unArg a
-unquoteH a = throwException $ BadVisibility "hidden"  a
-
 unquoteN :: Unquote a => Arg Term -> UnquoteM a
 unquoteN a | notHidden a && isRelevant a =
     unquote $ unArg a
@@ -188,22 +183,22 @@ instance Unquote ArgInfo where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [h,r] -> do
+      Con c _ [h,r] -> do
         choice
-          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r)]
+          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected <*> pure False)]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "arg info" t
 
 instance Unquote a => Unquote (Arg a) where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [info,x] -> do
+      Con c _ [info,x] -> do
         choice
           [(c `isCon` primArgArg, Arg <$> unquoteN info <*> unquoteN x)]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "arg" t
 
 -- Andreas, 2013-10-20: currently, post-fix projections are not part of the
@@ -211,6 +206,16 @@ instance Unquote a => Unquote (Arg a) where
 instance Unquote R.Elim where
   unquote t = R.Apply <$> unquote t
 
+instance Unquote Bool where
+  unquote t = do
+    t <- reduceQuotedTerm t
+    case ignoreSharing t of
+      Con c _ [] ->
+        choice [ (c `isCon` primTrue,  pure True)
+               , (c `isCon` primFalse, pure False) ]
+               __IMPOSSIBLE__
+      _ -> throwException $ NonCanonical "boolean" t
+
 instance Unquote Integer where
   unquote t = do
     t <- reduceQuotedTerm t
@@ -256,7 +261,7 @@ instance Unquote ErrorPart where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [x] ->
+      Con c _ [x] ->
         choice [ (c `isCon` primAgdaErrorPartString, StrPart  <$> unquoteNString x)
                , (c `isCon` primAgdaErrorPartTerm,   TermPart <$> unquoteN x)
                , (c `isCon` primAgdaErrorPartName,   NamePart <$> unquoteN x) ]
@@ -267,40 +272,40 @@ instance Unquote a => Unquote [a] where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [x,xs] -> do
+      Con c _ [x,xs] -> do
         choice
           [(c `isCon` primCons, (:) <$> unquoteN x <*> unquoteN xs)]
           __IMPOSSIBLE__
-      Con c [] -> do
+      Con c _ [] -> do
         choice
           [(c `isCon` primNil, return [])]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "list" t
 
 instance Unquote Hiding where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [] -> do
+      Con c _ [] -> do
         choice
           [(c `isCon` primHidden,  return Hidden)
           ,(c `isCon` primInstance, return Instance)
           ,(c `isCon` primVisible, return NotHidden)]
           __IMPOSSIBLE__
-      Con c vs -> __IMPOSSIBLE__
+      Con c _ vs -> __IMPOSSIBLE__
       _        -> throwException $ NonCanonical "visibility" t
 
 instance Unquote Relevance where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [] -> do
+      Con c _ [] -> do
         choice
           [(c `isCon` primRelevant,   return Relevant)
           ,(c `isCon` primIrrelevant, return Irrelevant)]
           __IMPOSSIBLE__
-      Con c vs -> __IMPOSSIBLE__
+      Con c _ vs -> __IMPOSSIBLE__
       _        -> throwException $ NonCanonical "relevance" t
 
 instance Unquote QName where
@@ -314,11 +319,11 @@ instance Unquote a => Unquote (R.Abs a) where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [x,y] -> do
+      Con c _ [x,y] -> do
         choice
           [(c `isCon` primAbsAbs, R.Abs <$> (hint <$> unquoteNString x) <*> unquoteN y)]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "abstraction" t
 
     where hint x | not (null x) = x
@@ -331,10 +336,10 @@ instance Unquote MetaId where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Lit (LitMeta r f x) -> do
-        live <- (== f) <$> liftU getCurrentPath
-        unless live $ liftU $ do
-            m <- fromMaybe __IMPOSSIBLE__ . Map.lookup f <$> sourceToModule
+      Lit (LitMeta r f x) -> liftU $ do
+        live <- (f ==) <$> getCurrentPath
+        unless live $ do
+            m <- fromMaybe __IMPOSSIBLE__ <$> lookupModuleFromSource f
             typeError . GenericDocError =<<
               sep [ text "Can't unquote stale metavariable"
                   , pretty m <> text "." <> pretty x ]
@@ -348,16 +353,16 @@ instance Unquote R.Sort where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [] -> do
+      Con c _ [] -> do
         choice
           [(c `isCon` primAgdaSortUnsupported, return R.UnknownS)]
           __IMPOSSIBLE__
-      Con c [u] -> do
+      Con c _ [u] -> do
         choice
           [(c `isCon` primAgdaSortSet, R.SetS <$> unquoteN u)
           ,(c `isCon` primAgdaSortLit, R.LitS <$> unquoteN u)]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "sort" t
 
 instance Unquote Literal where
@@ -367,7 +372,7 @@ instance Unquote Literal where
           file <- liftU getCurrentPath
           return $ LitMeta r file x
     case ignoreSharing t of
-      Con c [x] ->
+      Con c _ [x] ->
         choice
           [ (c `isCon` primAgdaLitNat,    LitNat    noRange <$> unquoteN x)
           , (c `isCon` primAgdaLitFloat,  LitFloat  noRange <$> unquoteN x)
@@ -376,25 +381,25 @@ instance Unquote Literal where
           , (c `isCon` primAgdaLitQName,  LitQName  noRange <$> unquoteN x)
           , (c `isCon` primAgdaLitMeta,   litMeta   noRange =<< unquoteN x) ]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "literal" t
 
 instance Unquote R.Term where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [] ->
+      Con c _ [] ->
         choice
           [ (c `isCon` primAgdaTermUnsupported, return R.Unknown) ]
           __IMPOSSIBLE__
 
-      Con c [x] -> do
+      Con c _ [x] -> do
         choice
           [ (c `isCon` primAgdaTermSort,      R.Sort      <$> unquoteN x)
           , (c `isCon` primAgdaTermLit,       R.Lit       <$> unquoteN x) ]
           __IMPOSSIBLE__
 
-      Con c [x, y] ->
+      Con c _ [x, y] ->
         choice
           [ (c `isCon` primAgdaTermVar,     R.Var     <$> (fromInteger <$> unquoteN x) <*> unquoteN y)
           , (c `isCon` primAgdaTermCon,     R.Con     <$> (ensureCon =<< unquoteN x) <*> unquoteN y)
@@ -421,37 +426,37 @@ instance Unquote R.Pattern where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [] -> do
+      Con c _ [] -> do
         choice
           [ (c `isCon` primAgdaPatAbsurd, return R.AbsurdP)
           , (c `isCon` primAgdaPatDot,    return R.DotP)
           ] __IMPOSSIBLE__
-      Con c [x] -> do
+      Con c _ [x] -> do
         choice
           [ (c `isCon` primAgdaPatVar,  R.VarP  <$> unquoteNString x)
           , (c `isCon` primAgdaPatProj, R.ProjP <$> unquoteN x)
           , (c `isCon` primAgdaPatLit,  R.LitP  <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c [x, y] -> do
+      Con c _ [x, y] -> do
         choice
           [ (c `isCon` primAgdaPatCon, R.ConP <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "pattern" t
 
 instance Unquote R.Clause where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
-      Con c [x] -> do
+      Con c _ [x] -> do
         choice
           [ (c `isCon` primAgdaClauseAbsurd, R.AbsurdClause <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c [x, y] -> do
+      Con c _ [x, y] -> do
         choice
           [ (c `isCon` primAgdaClauseClause, R.Clause <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
-      Con c _ -> __IMPOSSIBLE__
+      Con c _ _ -> __IMPOSSIBLE__
       _ -> throwException $ NonCanonical "clause" t
 
 -- Unquoting TCM computations ---------------------------------------------
@@ -479,8 +484,10 @@ evalTCM v = do
     I.Def f [u] ->
       choice [ (f `isDef` primAgdaTCMInferType,          tcFun1 tcInferType          u)
              , (f `isDef` primAgdaTCMNormalise,          tcFun1 tcNormalise          u)
+             , (f `isDef` primAgdaTCMReduce,             tcFun1 tcReduce             u)
              , (f `isDef` primAgdaTCMGetType,            tcFun1 tcGetType            u)
              , (f `isDef` primAgdaTCMGetDefinition,      tcFun1 tcGetDefinition      u)
+             , (f `isDef` primAgdaTCMIsMacro,            tcFun1 tcIsMacro            u)
              , (f `isDef` primAgdaTCMFreshName,          tcFun1 tcFreshName          u) ]
              failEval
     I.Def f [u, v] ->
@@ -498,6 +505,7 @@ evalTCM v = do
              failEval
     I.Def f [_, _, u, v] ->
       choice [ (f `isDef` primAgdaTCMCatchError,    tcCatchError    (unElim u) (unElim v))
+             , (f `isDef` primAgdaTCMWithNormalisation, tcWithNormalisation (unElim u) (unElim v))
              , (f `isDef` primAgdaTCMExtendContext, tcExtendContext (unElim u) (unElim v))
              , (f `isDef` primAgdaTCMInContext,     tcInContext     (unElim u) (unElim v)) ]
              failEval
@@ -510,6 +518,11 @@ evalTCM v = do
     tcBind m k = do v <- evalTCM m
                     evalTCM (k `apply` [defaultArg v])
 
+    process :: (InstantiateFull a, Normalise a) => a -> TCM a
+    process v = do
+      norm <- view eUnquoteNormalise
+      if norm then normalise v else instantiateFull v
+
     mkT l a = El s a
       where s = Type $ Max [Plus 0 $ UnreducedLevel l]
 
@@ -518,6 +531,11 @@ evalTCM v = do
     tcCatchError m h =
       liftU2 (\ m1 m2 -> m1 `catchError` \ _ -> m2) (evalTCM m) (evalTCM h)
 
+    tcWithNormalisation :: Term -> Term -> UnquoteM Term
+    tcWithNormalisation b m = do
+      v <- unquote b
+      liftU1 (locally eUnquoteNormalise $ const v) (evalTCM m)
+
     uqFun1 :: Unquote a => (a -> UnquoteM b) -> Elim -> UnquoteM b
     uqFun1 fun a = do
       a <- unquote (unElim a)
@@ -567,17 +585,17 @@ evalTCM v = do
     tcInferType :: R.Term -> TCM Term
     tcInferType v = do
       (_, a) <- inferExpr =<< toAbstract_ v
-      quoteType =<< normalise a
+      quoteType =<< process a
 
     tcCheckType :: R.Term -> R.Type -> TCM Term
     tcCheckType v a = do
       a <- isType_ =<< toAbstract_ a
       e <- toAbstract_ v
       v <- checkExpr e a
-      quoteTerm =<< normalise v
+      quoteTerm =<< process v
 
     tcQuoteTerm :: Term -> UnquoteM Term
-    tcQuoteTerm v = liftU $ quoteTerm =<< normalise v
+    tcQuoteTerm v = liftU $ quoteTerm =<< process v
 
     tcUnquoteTerm :: Type -> R.Term -> TCM Term
     tcUnquoteTerm a v = do
@@ -590,10 +608,15 @@ evalTCM v = do
       (v, _) <- inferExpr =<< toAbstract_ v
       quoteTerm =<< normalise v
 
+    tcReduce :: R.Term -> TCM Term
+    tcReduce v = do
+      (v, _) <- inferExpr =<< toAbstract_ v
+      quoteTerm =<< reduce =<< instantiateFull v
+
     tcGetContext :: UnquoteM Term
     tcGetContext = liftU $ do
       as <- map (fmap snd) <$> getContext
-      as <- etaContract =<< normalise as
+      as <- etaContract =<< process as
       buildList <*> mapM quoteDom as
 
     extendCxt :: Arg R.Type -> UnquoteM a -> UnquoteM a
@@ -622,6 +645,14 @@ evalTCM v = do
     tcGetType :: QName -> TCM Term
     tcGetType x = quoteType . defType =<< constInfo x
 
+    tcIsMacro :: QName -> TCM Term
+    tcIsMacro x = do
+      true  <- primTrue
+      false <- primFalse
+      let qBool True  = true
+          qBool False = false
+      qBool . isMacro . theDef <$> constInfo x
+
     tcGetDefinition :: QName -> TCM Term
     tcGetDefinition x = quoteDefn =<< constInfo x
 
@@ -639,7 +670,7 @@ evalTCM v = do
         reportSDoc "tc.unquote.decl" 10 $ sep [ text "declare" <+> prettyTCM x <+> text ":"
                                               , nest 2 $ prettyTCM a ]
         a <- isType_ =<< toAbstract_ a
-        alreadyDefined <- (True <$ getConstInfo x) `catchError` \ _ -> return False
+        alreadyDefined <- isJust <$> tryMaybe (getConstInfo x)
         when alreadyDefined $ genericError $ "Multiple declarations of " ++ show x
         addConstant x $ defaultDefn i x a emptyFunction
         when (h == Instance) $ addTypedInstance x a
