@@ -1,9 +1,5 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
-{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE PatternGuards              #-}
-{-# LANGUAGE TemplateHaskell            #-}
-{-# LANGUAGE TupleSections              #-}
 
 -- | Finite bijections (implemented as a pair of maps).
 
@@ -21,8 +17,6 @@ import qualified Data.Map as Map
 import Data.Tuple
 import Data.Typeable ( Typeable )
 
-import Test.QuickCheck
-
 -- | Finite bijective map from @a@ to @b@.  There, and back again.
 data BiMap a b = BiMap
   { biMapThere :: Map a b
@@ -77,29 +71,3 @@ instance (Ord a, Ord b) => Ord (BiMap a b) where
 instance (Show a, Show b) => Show (BiMap a b) where
   show bimap = "Agda.Utils.BiMap.fromList " ++ show (toList bimap)
 
-instance (Ord a, Ord b, Arbitrary a, Arbitrary b) => Arbitrary (BiMap a b) where
-  arbitrary = fromList <$> do List.zip <$> alist <*> blist
-    where
-      alist = List.nub <$> arbitrary
-      blist = List.nub <$> arbitrary
-
-------------------------------------------------------------------------
--- * Properties
-------------------------------------------------------------------------
-
-prop_BiMap_invariant :: (Ord a, Ord b) => BiMap a b -> Bool
-prop_BiMap_invariant (BiMap t u) =
-  Map.toAscList t == List.sort (List.map swap (Map.toList u))
-
-------------------------------------------------------------------------
--- * All tests
-------------------------------------------------------------------------
-
--- Template Haskell hack to make the following $quickCheckAll work
--- under ghc-7.8.
-return [] -- KEEP!
-
-tests :: IO Bool
-tests = do
-  putStrLn "Agda.Utils.BiMap"
-  $quickCheckAll
