@@ -10,12 +10,9 @@ module Agda.Utils.Either
   , fromLeft, fromRight
   , maybeLeft, maybeRight
   , allLeft, allRight
-  , tests
+  , maybeToEither
   ) where
 
-import Agda.Utils.QuickCheck
-import Agda.Utils.TestHelpers
-
 -- | Loop while we have an exception.
 
 whileLeft :: Monad m => (a -> Either b c) -> (a -> b -> m a) -> (a -> c -> m d) -> a -> m d
@@ -105,18 +102,6 @@ allLeft = mapM maybeLeft
 allRight :: [Either a b] -> Maybe [b]
 allRight = mapM maybeRight
 
-prop_allRight :: Eq b => [Either t b] -> Bool
-prop_allRight xs =
-  allRight xs ==
-    if all isRight xs then
-      Just $ map (\ (Right x) -> x) xs
-     else
-      Nothing
-
-------------------------------------------------------------------------
--- All tests
-
-tests :: IO Bool
-tests = runTests "Agda.Utils.Either"
-  [ quickCheck' (prop_allRight :: [Either Integer Bool] -> Bool)
-  ]
+-- | Convert 'Maybe' to @'Either' ()'@
+maybeToEither :: Maybe a -> Either () a
+maybeToEither = maybe (Left ()) Right
