@@ -4,14 +4,13 @@
 
 {-| Operations on file names. -}
 module Agda.Utils.FileName
-  ( AbsolutePath
+  ( AbsolutePath(AbsolutePath)
   , filePath
-  , rootName
   , mkAbsolute
   , absolute
   , (===)
   , doesFileExistCaseSensitive
-  , tests
+  , rootPath
   ) where
 
 import Control.Applicative
@@ -29,9 +28,8 @@ import Data.Function
 import Data.Hashable
 import Data.Typeable (Typeable)
 
+import Agda.Utils.Monad
 import Agda.Utils.Pretty
-import Agda.Utils.TestHelpers
-import Agda.Utils.QuickCheck
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -58,17 +56,6 @@ instance Show AbsolutePath where
 instance Pretty AbsolutePath where
   pretty = text . filePath
 
--- | The paths have to be absolute, valid and normalised, without
--- trailing path separators.
-
-absolutePathInvariant :: AbsolutePath -> Bool
-absolutePathInvariant x =
-  isAbsolute f &&
-  isValid f &&
-  f == normalise f &&
-  f == dropTrailingPathSeparator f
-  where f = filePath x
-
 -- | Constructs 'AbsolutePath's.
 --
 -- Precondition: The path must be absolute and valid.
@@ -79,11 +66,6 @@ mkAbsolute f
       AbsolutePath $ Text.pack $ dropTrailingPathSeparator $ normalise f
   | otherwise    = __IMPOSSIBLE__
 
-prop_mkAbsolute :: FilePath -> Property
-prop_mkAbsolute f =
-  let path = rootPath ++ f
-  in  isValid path ==> absolutePathInvariant $ mkAbsolute $ path
-
 rootPath :: FilePath
 #if mingw32_HOST_OS
 rootPath = joinDrive "C:" [pathSeparator]
@@ -123,38 +105,19 @@ infix 4 ===
 (===) :: AbsolutePath -> AbsolutePath -> Bool
 (===) = equalFilePath `on` filePath
 
--- | Case-sensitive doesFileExist for Windows.
+-- | Case-sensitive 'doesFileExist' for Windows.
+--
 -- This is case-sensitive only on the file name part, not on the directory part.
 -- (Ideally, path components coming from module name components should be
 --  checked case-sensitively and the other path components should be checked
---  case insenstively.)
+--  case insensitively.)
+
 doesFileExistCaseSensitive :: FilePath -> IO Bool
 #if mingw32_HOST_OS
 doesFileExistCaseSensitive f = do
-  ex <- doesFileExist f
-  if ex then bracket (findFirstFile f) (findClose . fst) $
-               fmap (takeFileName f ==) . getFindDataFileName . snd
-        else return False
+  doesFileExist f `and2M` do
+    bracket (findFirstFile f) (findClose . fst) $
+      fmap (takeFileName f ==) . getFindDataFileName . snd
 #else
 doesFileExistCaseSensitive f = doesFileExist f
 #endif
-
-------------------------------------------------------------------------
--- Generators
-
-instance Arbitrary AbsolutePath where
-  arbitrary = mk . take 3 . map (take 2) <$>
-                listOf (listOf1 (elements "a1"))
-    where mk ps = mkAbsolute (joinPath $ rootPath : ps)
-
-instance CoArbitrary AbsolutePath where
-  coarbitrary (AbsolutePath t) = coarbitrary (Text.unpack t)
-
-------------------------------------------------------------------------
--- All tests
-
-tests :: IO Bool
-tests = runTests "Agda.Utils.FileName"
-  [ quickCheck' absolutePathInvariant
-  , quickCheck' prop_mkAbsolute
-  ]
